;*----------------------------------*
;*    >>>Principles of Object<<<    *
;*    >>> Oriented Languages <<<    *
;*                                  *
;*     Objects with inheritance     *
;*                                  *
;*           Theo D'Hondt           *
;*  VUB Programming Technology Lab  *
;*              c2001               *
;*        slipified in 2014         *

;*----------------------------------*

(define (<<TABLE>>)
  (define tab '())
  (lambda (op . rest)
    (case op
      ((instantiate)
       (let ((table (<<TABLE>>)))
         (for-each
          (lambda (elt) 
            (table 'put (car elt) (eval (cdr elt))))
          tab)
         table))
      ((copy)
       (let ((table (<<TABLE>>)))
         (for-each 
          (lambda (elt)  
            (table 'put (car elt) (cdr elt)))
          tab)
         table))
      ((get)
       (let* ((key (car rest))
              (entry (assoc key tab)))
         (if entry
             (cdr entry)
             #f)))
      ((put) 
       (let* ((key (car rest))
              (entry (assoc key tab))
              (value (cadr rest)))
         (if entry
             (error "duplicate name" key)
             (set! tab (cons (cons key value) tab)))))
      ((replace)
       (let* ((key (car rest))
              (entry (assoc key tab))
              (value (cadr rest)))
         (if entry
             (set-cdr! entry value)
             (error "undefined name" key))
         #t)))))

(define-macro VAR
  (lambda (name value)
    `(<<VARS>> 'put ',name ',value)))

(define-macro METHOD
  (lambda (msg args . body)
    `(<<METHODS>> 'put ',msg
      (lambda (<<CONTEXT>> ,@args) ,@body))))

(define-macro CLASS 
  (lambda (super . defs)
    `(letrec
         ((<<SUPER>> ,super)
          (<<METHODS>> (<<TABLE>>))
          (<<VARS>> (<<SUPER>> '<<COPY>>))
          (<<CLASS>>
           (lambda (MSG . ARGS)
             (case MSG
               ((new)
                (let* 
                    ((context (<<VARS>> 'instantiate))
                     (self 
                      (lambda (msg . args)
                        (<<CLASS>> '<<EVAL>> context msg args))))
                  (context 'replace '<<SELF>> self)
                  self))
               ((<<EVAL>>)
                (let*
                    ((context (car ARGS))
                     (msg (cadr ARGS))
                     (args (caddr ARGS))
                     (entry (<<METHODS>> 'get msg)))
                  (if entry 
                      (apply entry (cons context args))
                      (<<SUPER>> '<<EVAL>> context msg args))))
               ((<<COPY>>)
                (<<VARS>> 'copy))
               (else
                (error "invalid class message" MSG))))))
       ,@defs
       <<CLASS>>)))

(define Root
  (lambda (MSG . ARGS)
    (case MSG
      ((new)
       (error "cannot instantiate root class"))
      ((<<EVAL>>)
       (let*
           ((msg (cadr ARGS)))
         (error "method not found " (car msg))))
      ((<<COPY>>)
       (let ((tab (<<TABLE>>)))
         (tab 'put '<<SELF>> '())
         tab))
      (else
       (error "invalid class message" MSG)))))

(define-macro NEW
  (lambda (class)
    `(,class 'new)))

(define-macro SELF
  (lambda ()
    `(<<CONTEXT>> 'get '<<SELF>>)))

(define-macro SEND
  (lambda (object msg . args)
    `(,object ',msg ,@args)))

(define-macro SUPER
  (lambda (msg . args)
    `(<<SUPER>> '<<EVAL>> <<CONTEXT>> ',msg ,args)))

(define-macro ? 
  (lambda (name)    
    `(<<CONTEXT>> 'get ',name)))

(define-macro ! 
  (lambda (name value)    
    `(<<CONTEXT>> 'replace ',name ,value)))