                     /*---------------------------------------------*/
                     /*                  >>>Slip<<<                 */
                     /*                 Theo D'Hondt                */
                     /*          VUB Software Languages Lab         */
                     /*                  (c) 2016                   */
                     /*---------------------------------------------*/
                     /*                final version                */
                     /*---------------------------------------------*/
                     /*                    Slip                     */
                     /*---------------------------------------------*/

/*--------------------------------------------------------------------------------------*/
/*       AL1_error                                at least one expression required      */
/*       AM1_error                                at most one argument allowed          */
/*       AM2_error                                at most two arguments allowed         */
/*       ARI_error                                arguments restricted to integers      */
/*       ARN_error                                arguments restricted to numbers       */
/*       BFO_error                                buffer overflow                       */
/*       BND_error                                binding expected                      */
/*       BVR_error                                byte value required                   */
/*       CIP_error                                close input port error                */
/*       COP_error                                close output port error               */
/*       COR_error                                character out of range                */
/*       CRQ_error                                continuation requires single value    */
/*       DBZ_error                                division by zero                      */
/*       DLV_error                                duplicate let variable                */
/*       DPN_error                                duplicate parameter name              */
/*       DRQ_error                                digit required                        */
/*       E1X_error                                exactly one expression required       */
/*       EX0_error                                no arguments allowed                  */
/*       EX1_error                                exactly one argument required         */
/*       EX2_error                                exactly two arguments required        */
/*       EX3_error                                exactly three arguments required      */
/*       ICL_error                                invalid character literal             */
/*       IDF_error                                illegal decimal fraction              */
/*       IDS_error                                illegal decimal scale                 */
/*       ILC_error                                illegal character                     */
/*       ILA_error                                illegal use of @                      */
/*       ILB_error                                illegal use of \                      */
/*       ILH_error                                illegal use of #                      */
/*       ILS_error                                illegal session                       */
/*       IMB_error                                illegal macro body                    */
/*       IMS_error                                illegal macro shadowing               */
/*       IMV_error                                illegal macro value                   */
/*       INM_error                                Insufficient memory                   */
/*       IOR_error                                index out of range                    */
/*       IPA_error                                invalid parameter                     */
/*       IRP_error                                illegal right parenthesis             */
/*       IRS_error                                invalid response                      */
/*       ISF_error                                input stream failure                  */
/*       ITA_error                                improperly terminated argument list   */
/*       ITF_error                                improperly terminated form            */
/*       ITK_error                                improper token                        */
/*       IUK_error                                illegal use of keyword                */
/*       IUP_error                                illegal use of period                 */
/*       IVP_error                                invalid pattern in form               */
/*       IVS_error                                invalid variable shadowing            */
/*       IVV_error                                invalid variable in form              */
/*       IXT_error                                invalid expression type               */
/*       MMO_error                                memory overflow                       */
/*       MNS_error                                memory not synchronized               */
/*       MRP_error                                missing right parenthesis             */
/*       MSP_error                                missing pattern in form               */
/*       MSY_error                                macro symbol                          */
/*       NAB_error                                not a boolean                         */
/*       NAC_error                                not a character                       */
/*       NAL_error                                not a list                            */
/*       NAN_error                                not a number                          */
/*       NAP_error                                not a pair                            */
/*       NAS_error                                not a string                          */
/*       NAV_error                                not a vector                          */
/*       NIP_error                                not an input port                     */
/*       NNN_error                                non-negative number required          */
/*       NOP_error                                not an output port                    */
/*       NPR_error                                not a promise                         */
/*       NSY_error                                not a symbol                          */
/*       NTL_error                                number too large                      */
/*       NVS_error                                not a valid symbol                    */
/*       NYI_error                                not yet implemented                   */
/*       OIF_error                                open input file error                 */
/*       OOF_error                                open output file error                */
/*       OOR_error                                out of range error                    */
/*       OPR_error                                operator required                     */
/*       OSF_error                                output stream failure                 */
/*       PAC_error                                port already closed                   */
/*       PEP_error                                premature end of program              */
/*       PNR_error                                positive number required              */
/*       PRE_error                                protocol error                        */
/*       QBR_error                                quote or backslash required           */
/*       QRQ_error                                quote required                        */
/*       RDF_error                                read failure                          */
/*       SOF_error                                session output failure                */
/*       STL_error                                string too long                       */
/*       TFO_error                                too few operands in application       */
/*       TFX_error                                too few expressions in form           */
/*       TMO_error                                too many operands in application      */
/*       TMS_error                                too many sessions                     */
/*       TMX_error                                too many expressions in form          */
/*       UDF_error                                undefined variable                    */
/*       UQQ_error                                unquote(-splicing) not in quasiquote  */
/*       ULS_error                                unquote-splicing not from within list */
/*       UXT_error                                unexpected expression type            */
/*       VNF_error                                variable not found                    */
/*       VTL_error                                vector too large                      */
/*       WRF_error                                write failure                         */
/*       XIT_error                                excess input tokens                   */
/*--------------------------------------------------------------------------------------*/

typedef
  enum { Slip_read_mode,
         Slip_write_mode } Slip_mode;

typedef
  enum { Slip_end_of_stream_response,
         Slip_invalid_stream_response,
         Slip_success_response } Slip_response;

typedef     void * Slip_address;
typedef     char   Slip_char;
typedef unsigned   Slip_error;
typedef     char * Slip_text;
typedef     void   Slip_void;

/*----------------------------------- callback functions -------------------------------*/

Slip_response Slip_Close(Slip_mode , Slip_address);
Slip_void     Slip_Error(Slip_error, Slip_text);
Slip_response  Slip_Open(Slip_text , Slip_mode, Slip_address *);
Slip_response  Slip_Read(Slip_text , Slip_address);
Slip_response Slip_Write(Slip_char , Slip_address);

/*----------------------------------- exported functions -------------------------------*/

Slip_void      Slip_Eval(Slip_text);
