                     /*---------------------------------------------*/
                     /*                  >>>Slip<<<                 */
                     /*                 Theo D'Hondt                */
                     /*          VUB Software Languages Lab         */
                     /*                  (c) 2018                   */
                     /*---------------------------------------------*/
                     /*                final version                */
                     /*---------------------------------------------*/
                     /*                 dictionary                  */
                     /*---------------------------------------------*/

#include "SlipPlatform.h"

#include "SlipGrammar.h"

#include "SlipDictionary.h"

/*----------------------------------- private macros -----------------------------------*/

#define NULL_DICTIONARY           (DCT_type)Unique_Null
#define NULL_VARIABLE_SYMBOL_LINK (VSL_type)Unique_Null
#define NULL_MACRO_SYMBOL_LINK    (MSL_type)Unique_Null

/*------------------------------------ private types -----------------------------------*/

DEFINE_TYPE(MSL);
DEFINE_TYPE(VSL);

BEGIN_CHUNK(DCT);
  DEFINE_SLOT(VSL, vsl);
  DEFINE_SLOT(MSL, msl);
  DEFINE_SLOT(SMI, scp);
  DEFINE_SLOT(SMI, siz);
  DEFINE_SLOT(EXP, bln);
  DEFINE_SLOT(EXP, dct);
END_CHUNK(DCT);

BEGIN_CHUNK(MSL);
  DEFINE_SLOT(SYM, sym);
  DEFINE_SLOT(EXP, exp);
  DEFINE_SLOT(MSL, msl);
END_CHUNK(MSL);

BEGIN_CHUNK(VSL);
  DEFINE_SLOT(SYM, sym);
  DEFINE_SLOT(EXP, vsl);
END_CHUNK(VSL);

/*-------------------------------- protected variables ---------------------------------*/

static DCT_type Current_Dictionary_P;
static DCT_type Dictionary_P;
static EXP_type Expression_P;
static DCT_type Macro_Dictionary_P;
static DCT_type Saved_Current_Dictionary_P;
static DCT_type Saved_Macro_Dictionary_P;
static SYM_type Symbol_P;

/*--------------------------------- private functions ----------------------------------*/

static DCT_type make_DCT_M(VOI_type)
  { DCT_type dictionary;
    VEC_type vector;
    vector = make_VEC_M(DCT_size);
    dictionary = (DCT_type)vector;
    return dictionary; }

static MSL_type make_MSL_M(VOI_type)
  { MSL_type macro_symbol_link;
    VEC_type vector;
    vector = make_VEC_M(MSL_size);
    macro_symbol_link = (MSL_type)vector;
    return macro_symbol_link; }

static VSL_type make_VSL_M(VOI_type)
  { VEC_type vector;
    VSL_type variable_symbol_link;
    vector = make_VEC_M(VSL_size);
    variable_symbol_link = (VSL_type)vector;
    return variable_symbol_link; }

/*--------------------------------------------------------------------------------------*/

static VOI_type copy(DCT_type Source_dictionary,
                     DCT_type Destination_dictionary)
  { DCT_type dictionary;
    MSL_type macro_symbol_link;
    SMI_type frame_size_smallint,
             scope_smallint;
    VSL_type variable_symbol_link;
    variable_symbol_link = Source_dictionary->vsl;
    macro_symbol_link    = Source_dictionary->msl;
    scope_smallint       = Source_dictionary->scp;
    frame_size_smallint  = Source_dictionary->siz;
    dictionary           = Source_dictionary->dct;
    Destination_dictionary->vsl = variable_symbol_link;
    Destination_dictionary->msl = macro_symbol_link;
    Destination_dictionary->scp = scope_smallint;
    Destination_dictionary->siz = frame_size_smallint;
    Destination_dictionary->dct = dictionary; }

static DCT_type clone_M(VOI_type)
  { DCT_type cloned_dictionary;
    if (is_NUL(Current_Dictionary_P))
      return NULL_DICTIONARY;
    cloned_dictionary = make_DCT_M();
    cloned_dictionary->vsl = Current_Dictionary_P->vsl;
    cloned_dictionary->msl = Current_Dictionary_P->msl;
    cloned_dictionary->scp = Current_Dictionary_P->scp;
    cloned_dictionary->siz = Current_Dictionary_P->siz;
    cloned_dictionary->bln = Current_Dictionary_P->bln;
    cloned_dictionary->dct = Current_Dictionary_P->dct;
    return cloned_dictionary; }

static UNS_type define_variable_M(SYM_type Variable_symbol,
                                  DCT_type Dictionary)
  { SMI_type frame_size_smallint;
    UNS_type frame_size;
    VSL_type variable_symbol_link;
    Dictionary_P = Dictionary;
    Symbol_P = Variable_symbol;
    variable_symbol_link = make_VSL_M();
    variable_symbol_link->sym = Symbol_P;
    variable_symbol_link->vsl = Dictionary_P->vsl;
    frame_size_smallint       = Dictionary_P->siz;
    incr_SMU(&frame_size_smallint);
    Dictionary_P->vsl = variable_symbol_link;
    Dictionary_P->siz = frame_size_smallint;
    frame_size = get_SMU(frame_size_smallint);
    return frame_size; }

static UNS_type get_scope(DCT_type Dictionary)
  { SMI_type scope_smallint;
    UNS_type scope;
    scope_smallint = Dictionary->scp;
    scope = get_SMU(scope_smallint);
    return scope; }

static SYM_type locate_offset(UNS_type Offset,
                              DCT_type Dictionary)
  { SMI_type frame_size_smallint;
    SYM_type variable_symbol;
    UNS_type frame_size,
             offset;
    VSL_type variable_symbol_link;
    variable_symbol_link = Dictionary->vsl;
    frame_size_smallint  = Dictionary->siz;
    frame_size = get_SMU(frame_size_smallint);
    for (offset = frame_size;
         offset > Offset;
         offset -= 1)
      variable_symbol_link = variable_symbol_link->vsl;
    variable_symbol = variable_symbol_link->sym;
    return variable_symbol; }

static MSL_type locate_macro(SYM_type Macro_symbol,
                             DCT_type Dictionary)
  { MSL_type macro_symbol_link;
    for (macro_symbol_link = Dictionary->msl;
         macro_symbol_link != NULL_MACRO_SYMBOL_LINK;
         macro_symbol_link = macro_symbol_link->msl)
      if (Macro_symbol == macro_symbol_link->sym)
        return macro_symbol_link;
    return NULL_MACRO_SYMBOL_LINK; }

static UNS_type locate_variable(SYM_type Variable_symbol,
                                DCT_type Dictionary)
  { SMI_type frame_size_smallint;
    UNS_type offset;
    VSL_type variable_symbol_link;
    frame_size_smallint  = Dictionary->siz;
    offset = get_SMU(frame_size_smallint);
    for (variable_symbol_link = Dictionary->vsl;
         variable_symbol_link != NULL_VARIABLE_SYMBOL_LINK;
         variable_symbol_link = variable_symbol_link->vsl)
      { if (variable_symbol_link->sym == Variable_symbol)
          return offset;
        offset -= 1; }
    return 0; }

/*---------------------------------- public functions ----------------------------------*/

DCT_type Dictionary_Activate_Macro(VOI_type)
  { DCT_type dictionary;
    if (is_NUL(Macro_Dictionary_P))
      return NULL_DICTIONARY;
    dictionary = Current_Dictionary_P;
    Current_Dictionary_P = Macro_Dictionary_P;
    Macro_Dictionary_P = NULL_DICTIONARY;
    return dictionary; }

BLN_type Dictionary_Contains_Eval(VOI_type)
  { EXP_type boolean_expression;
    boolean_expression = Current_Dictionary_P->bln;
    if (is_TRU(boolean_expression))
      return TRUE;
    return FALSE; }

VOI_type Dictionary_Deactivate_Macro(DCT_type Dictionary)
  { if (is_NUL(Dictionary))
      return;
    Macro_Dictionary_P = Current_Dictionary_P;
    Current_Dictionary_P = Dictionary; }

BLN_type Dictionary_Define_Macro_Symbol_M(SYM_type Macro_symbol,
                                          EXP_type Macro_expression)
  { DCT_type dictionary;
    MSL_type macro_symbol_link;
    UNS_type offset;
    for (dictionary = Current_Dictionary_P;
         dictionary != NULL_DICTIONARY;
         dictionary = dictionary->dct)
     { offset = locate_variable(Macro_symbol,
                                dictionary);
       if (offset > 0)
         return TRUE; }
    macro_symbol_link = locate_macro(Macro_symbol,
                                     Current_Dictionary_P);
    if (macro_symbol_link == NULL_MACRO_SYMBOL_LINK)
      { Expression_P = Macro_expression;
        Symbol_P = Macro_symbol;
        macro_symbol_link = make_MSL_M();
        macro_symbol_link->sym = Macro_symbol;
        macro_symbol_link->exp = Macro_expression;
        macro_symbol_link->msl = Current_Dictionary_P->msl;
        Current_Dictionary_P->msl = macro_symbol_link; }
    else
      macro_symbol_link->exp = Macro_expression;
    return FALSE; }

UNS_type Dictionary_Define_Variable_M(SYM_type Variable_symbol)
  { DCT_type dictionary;
    MSL_type macro_symbol_link;
    UNS_type offset;
    for (dictionary = Current_Dictionary_P;
         dictionary != NULL_DICTIONARY;
         dictionary = dictionary->dct)
     { macro_symbol_link = locate_macro(Variable_symbol,
                                        dictionary);
       if (macro_symbol_link != NULL_MACRO_SYMBOL_LINK)
         return 0; }
    offset = locate_variable(Variable_symbol,
                             Current_Dictionary_P);
    if (offset > 0)
      return offset;
    offset = define_variable_M(Variable_symbol,
                               Current_Dictionary_P);
    return offset; }

VOI_type Dictionary_Enter_Nested_Scope_M(VOI_type)
  { DCT_type dictionary;
    SMI_type scope_smallint;
    scope_smallint = Current_Dictionary_P->scp;
    incr_SMU(&scope_smallint);
    dictionary = make_DCT_M();
    dictionary->vsl = NULL_VARIABLE_SYMBOL_LINK;
    dictionary->msl = NULL_MACRO_SYMBOL_LINK;
    dictionary->scp = scope_smallint;
    dictionary->siz = Unique_Zero_Smallint;
    dictionary->bln = Unique_False;
    dictionary->dct = Current_Dictionary_P;
    Current_Dictionary_P = dictionary; }

DCT_type Dictionary_Exit_Nested_Scope(VOI_type)
  { DCT_type dictionary;
    dictionary = Current_Dictionary_P;
    Current_Dictionary_P = dictionary->dct;
    return dictionary; }

DCT_type Dictionary_Get_Current_Dictionary(VOI_type)
  { return Current_Dictionary_P; }

UNS_type Dictionary_Get_Frame_Size(VOI_type)
  { SMI_type frame_size_smallint;
    UNS_type frame_size;
    frame_size_smallint = Current_Dictionary_P->siz;
    frame_size = get_SMU(frame_size_smallint);
    return frame_size; }

EXP_type Dictionary_Get_Macro_Expression(SYM_type Macro_symbol)
  { DCT_type dictionary;
    MSL_type macro_symbol_link;
    for (dictionary = Current_Dictionary_P;
         dictionary != NULL_DICTIONARY;
         dictionary = dictionary->dct)
     { macro_symbol_link = locate_macro(Macro_symbol,
                                        dictionary);
       if (macro_symbol_link != NULL_MACRO_SYMBOL_LINK)
         return macro_symbol_link->exp; }
    return Unique_Undefined; }

BLN_type Dictionary_Lexical_Address_M(SYM_type Variable_symbol,
                                      UNs_type Scope_reference,
                                      UNs_type Offset_reference)
  { DCT_type dictionary;
    MSL_type macro_symbol_link;
    UNS_type offset,
             scope;
    for (dictionary = Current_Dictionary_P;
         dictionary != NULL_DICTIONARY;
         dictionary = dictionary->dct)
     { macro_symbol_link = locate_macro(Variable_symbol,
                                        dictionary);
       if (macro_symbol_link != NULL_MACRO_SYMBOL_LINK)
         return TRUE; }
    offset = locate_variable(Variable_symbol,
                             Current_Dictionary_P);
    if (offset > 0)
      { *Scope_reference = 0;
        *Offset_reference = offset;
        return FALSE; }
    scope = get_scope(Current_Dictionary_P);
    if (scope == 1)
      return TRUE;
    for (dictionary = Current_Dictionary_P;
         scope > 1;
         )
      { dictionary = dictionary->dct;
        scope = get_scope(dictionary);
        offset = locate_variable(Variable_symbol,
                                 dictionary);
        if (offset > 0)
          { *Scope_reference = scope;
            *Offset_reference = offset;
            return FALSE; }}
    offset = define_variable_M(Variable_symbol,
                               dictionary);
    *Scope_reference = 1;
    *Offset_reference = offset;
    return FALSE; }

SYM_type Dictionary_Global_Variable(UNS_type Scope,
                                    UNS_type Offset)
  { DCT_type dictionary;
    SYM_type variable_symbol;
    dictionary = Current_Dictionary_P;
    while (Scope < get_scope(dictionary))
      dictionary = dictionary->dct;
    variable_symbol = locate_offset(Offset,
                                    dictionary);
    return variable_symbol; }

SYM_type Dictionary_Local_Variable(UNS_type Offset)
  { SYM_type variable_symbol;
    variable_symbol = locate_offset(Offset,
                                    Current_Dictionary_P);
    return variable_symbol; }

VOI_type Dictionary_Register_Eval(VOI_type)
  { Current_Dictionary_P->bln = Unique_True; }

DCT_type Dictionary_Replace_Dictionary(DCT_type Dictionary)
  { DCT_type dictionary;
    dictionary = Current_Dictionary_P;
    Current_Dictionary_P = Dictionary;
    return dictionary; }

VOI_type Dictionary_Restore(VOI_type)
  { copy(Saved_Current_Dictionary_P,
         Current_Dictionary_P);
    copy(Saved_Macro_Dictionary_P,
         Macro_Dictionary_P); }

VOI_type Dictionary_Save(VOI_type)
  { copy(Current_Dictionary_P,
         Saved_Current_Dictionary_P);
    copy(Macro_Dictionary_P,
         Saved_Macro_Dictionary_P); }

/*------------------------------------- initialize -------------------------------------*/

VOI_type Dictionary_Initialize_M(VOI_type)
  { Current_Dictionary_P = make_DCT_M();
    Current_Dictionary_P->vsl = NULL_VARIABLE_SYMBOL_LINK;
    Current_Dictionary_P->msl = NULL_MACRO_SYMBOL_LINK;
    Current_Dictionary_P->scp = Unique_One_Smallint;
    Current_Dictionary_P->siz = Unique_Zero_Smallint;
    Current_Dictionary_P->bln = Unique_Null;
    Current_Dictionary_P->dct = NULL_DICTIONARY;
    Saved_Current_Dictionary_P = clone_M();
    Saved_Macro_Dictionary_P = clone_M();
    REGISTER(Current_Dictionary_P);
    REGISTER(Expression_P);
    REGISTER(Dictionary_P);
    REGISTER(Macro_Dictionary_P);
    REGISTER(Saved_Current_Dictionary_P);
    REGISTER(Saved_Macro_Dictionary_P);
    REGISTER(Symbol_P); }

VOI_type Dictionary_Macro_Initialize_M(VOI_type)
  { Macro_Dictionary_P = clone_M(); }

