                     /*---------------------------------------------*/
                     /*                  >>>Slip<<<                 */
                     /*                 Theo D'Hondt                */
                     /*          VUB Software Languages Lab         */
                     /*                  (c) 2018                   */
                     /*---------------------------------------------*/
                     /*                final version                */
                     /*---------------------------------------------*/
                     /*                 environment                 */
                     /*---------------------------------------------*/

#include "SlipPlatform.h"

#include "SlipGrammar.h"

#include "SlipCache.h"

#include "SlipEnvironment.h"

/*--------------------------------------------------------------------------------------*/
/*----------------------------------- frame manipulation -------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*----------------------------------- private macros -----------------------------------*/

#define EMPTY_FRAME (FRM_type)Unique_Empty_Chunk_P
#define NULL_FRAME  (FRM_type)Unique_Null

/*---------------------------------- private constants ---------------------------------*/

enum { Initial_global_frame_size       = 256,
       Initial_global_macro_frame_size = 256,
       Frame_increment                 =   8,
       Global_frame_increment          =  32 };

/*---------------------------- private protected variables -----------------------------*/

static ENV_type Current_Environment_P;
static FRM_type Current_Frame_P;
static ENV_type Current_Macro_Environment_P;
static FRM_type Global_Frame_P;
static FRM_type Global_Macro_Frame_P;

/*---------------------------------- public functions ----------------------------------*/

FRM_type Environment_Activate_Macro(VOI_type)
  { FRM_type frame;
    if (is_NUL(Global_Macro_Frame_P))
      return NULL_FRAME;
    frame = Current_Frame_P;
    Current_Frame_P = Global_Macro_Frame_P;
    Global_Macro_Frame_P = NULL_FRAME;
    return frame; }

VOI_type Environment_Adjust_Current_Frame_Size_M(UNS_type Required_size)
  { CHK_type new_chunk;
    UNS_type index,
             new_size,
             size;
    size = size_CHK(Current_Frame_P);
    if (Required_size <= size)
      return;
    new_size = Required_size + Frame_increment;
    new_chunk = make_CHK_M(new_size);
    mark_CHK(new_chunk);
    for (index = 1;
         index <= size;
         index += 1)
      new_chunk[index] = Current_Frame_P[index];
    for (;
         index <= new_size;
         index += 1)
      new_chunk[index] = Unique_Undefined;
    Grammar_Replace_Chunk(Current_Frame_P,
                          new_chunk);
    Current_Frame_P = new_chunk; }

VOI_type Environment_Deactivate_Macro(FRM_type Frame)
  { if (is_NUL(Frame))
      return;
    Global_Macro_Frame_P = Current_Frame_P;
    Current_Frame_P = Frame; }

UNS_type Environment_Frame_Size(FRM_type Frame)
  { UNS_type size;
    size = size_CHK(Frame);
    return size; }

FRM_type Environment_Get_Current_Frame(VOI_type)
  { return Current_Frame_P; }

FRM_type Environment_Get_Macro_Frame(VOI_type)
  { return Global_Macro_Frame_P; }

EXP_type Environment_Local_Get(UNS_type Offset)
  { EXP_type value_expression;
    value_expression = Current_Frame_P[Offset];
    return value_expression; }

VOI_type Environment_Local_Set(EXP_type Value_expression,
                               UNS_type Offset)
  { Current_Frame_P[Offset] = Value_expression; }

FRM_type Environment_Make_Frame_M(UNS_type Frame_size)
  { CHK_type frame_chunk;
    FRM_type frame;
    frame_chunk = Cache_Make_Chunk_M(Frame_size);
    frame = frame_chunk;
    return frame; }

FRM_type Environment_Mark_Frame(VOI_type)
  { mark_CHK(Current_Frame_P);
    return Current_Frame_P; }

VOI_type Environment_Register_Native(EXP_type Value_expression,
                                     UNS_type Offset)
  { C_ASSERT(Offset < Initial_global_frame_size);
    C_ASSERT(Offset < Initial_global_macro_frame_size);
    Global_Frame_P[Offset] = Value_expression;
    Global_Macro_Frame_P[Offset] = Value_expression; }

VOI_type Environment_Release_Current_Frame(VOI_type)
  { Cache_Release_Chunk(Current_Frame_P);
    Current_Frame_P = EMPTY_FRAME; }

VOI_type Environment_Release_Frame(FRM_type Frame)
  { Cache_Release_Chunk(Frame); }

VOI_type Environment_Replace_Frame(FRM_type Frame)
  { Current_Frame_P = Frame; }

/*--------------------------------------------------------------------------------------*/

static VOI_type frame_initialize_M(VOI_type)
  { Current_Frame_P = make_CHK_Z(Initial_global_frame_size);
    Global_Frame_P = Current_Frame_P;
    Global_Macro_Frame_P = make_CHK_Z(Initial_global_macro_frame_size);
    mark_CHK(Global_Frame_P);
    mark_CHK(Global_Macro_Frame_P); }

/*--------------------------------------------------------------------------------------*/
/*-------------------------------- environment manipulation ----------------------------*/
/*--------------------------------------------------------------------------------------*/

/*----------------------------------- private macros -----------------------------------*/

#define EMPTY_ENVIRONMENT (ENV_type)Unique_Empty_Chunk_P

/*---------------------------------- public functions ----------------------------------*/

ENV_type Environment_Extend_M(VOI_type)
  { CHK_type environment_chunk;
    ENV_type extended_environment;
    FRM_type frame;
    UNS_type environment_size,
             scope;
    environment_size = size_CHK(Current_Environment_P);
    environment_size += 1;
    environment_chunk = Cache_Make_Chunk_M(environment_size);
    extended_environment = (ENV_type)environment_chunk;
    for (scope = 1;
         scope < environment_size;
         scope += 1)
      { frame = Current_Environment_P[scope];
        extended_environment[scope] = frame; }
    mark_CHK(Current_Frame_P);
    extended_environment[environment_size] = Current_Frame_P;
    return extended_environment; }

ENV_type Environment_Get_Current_Environment(VOI_type)
  { return Current_Environment_P; }

UNS_type Environment_Get_Current_Environment_Size(VOI_type)
  { UNS_type environment_size;
    environment_size = size_CHK(Current_Environment_P);
    return environment_size; }

EXP_type Environment_Global_Get(UNS_type Scope,
                                UNS_type Offset)
  { EXP_type value_expression;
    FRM_type frame;
    frame = Current_Environment_P[Scope];
    value_expression = frame[Offset];
    return value_expression; }

VOI_type Environment_Global_Set(EXP_type Value_expression,
                                UNS_type Scope,
                                UNS_type Offset)
  { FRM_type frame;
    frame = Current_Environment_P[Scope];
    frame[Offset] = Value_expression; }

VOI_type Environment_Replace_Environment(ENV_type Environment)
  { Current_Environment_P = Environment; }

VOI_type Environment_Reset(VOI_type)
  { Current_Environment_P = EMPTY_ENVIRONMENT;
    Current_Frame_P = Global_Frame_P; }

/*--------------------------------------------------------------------------------------*/

static VOI_type environment_initialize_M(VOI_type)
  { Current_Environment_P       = EMPTY_ENVIRONMENT;
    Current_Macro_Environment_P = EMPTY_ENVIRONMENT; }

/*------------------------------------- initialize -------------------------------------*/

VOI_type Environment_Initialize_M(VOI_type)
  { frame_initialize_M();
    environment_initialize_M();
    REGISTER(Current_Environment_P);
    REGISTER(Current_Frame_P);
    REGISTER(Current_Macro_Environment_P);
    REGISTER(Global_Frame_P);
    REGISTER(Global_Macro_Frame_P); }

