                     /*---------------------------------------------*/
                     /*                  >>>Slip<<<                 */
                     /*                 Theo D'Hondt                */
                     /*          VUB Software Languages Lab         */
                     /*                  (c) 2018                   */
                     /*---------------------------------------------*/
                     /*                final version                */
                     /*---------------------------------------------*/
                     /*                   evaluate                  */
                     /*---------------------------------------------*/

#include "SlipPlatform.h"

#include "SlipMain.h"

#include "SlipGrammar.h"

#include "SlipCompile.h"
#include "SlipDictionary.h"
#include "SlipEnvironment.h"
#include "SlipNatives.h"
#include "SlipPool.h"
#include "SlipStack.h"
#include "SlipThread.h"

#include "SlipEvaluate.h"

/*---------------------------------- private constants ---------------------------------*/

static const TXT_type Call_CC_text  = "call-with-current-continuation",
                      Evaluate_text = "eval",
                      Force_text    = "force";

/*---------------------------------- private variables ---------------------------------*/

static C_JMP_BUF Exit;

/*--------------------------------- protected variables --------------------------------*/

static EXP_type Alternate_Expression_P;
static EXP_type Body_Expression_P;
static SEQ_type Body_Sequence_P;
static EXP_type Car_Expression_P;
static EXP_type Cdr_Expression_P;
static EXP_type Consequent_Expression_P;
static CNT_type Continuation_P;
static DCT_type Dictionary_P;
static EXP_type Else_Expression_P;
static ENV_type Environment_P;
static EXP_type Expression_P;
static FRM_type Frame_P;
static NAZ_type Higher_Order_Native_P;
static EXP_type List_Expression_P;
static NAT_type Native_P;
static EXP_type Operator_Expression_P;
static PAI_type Pair_P;
static EXP_type Predicate_Expression_P;
static SEQ_type Predicate_Sequence_P;
static PRC_type Procedure_P;
static PRV_type Procedure_Vararg_P;
static PRM_type Promise_P;
static VEC_type Quasi_Vector_P;
static SEQ_type Sequence_P;
static STR_type String_P;
static SYM_type Symbol_P;
static VEC_type Vector_P;

/*--------------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------------*/
/*-------------------------------- auxiliary functions ---------------------------------*/
/*--------------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------------------*/
/*----------------------------------- error handling -----------------------------------*/
/*--------------------------------------------------------------------------------------*/

static VOI_type error(ERR_type Error,
                      TXT_type Text)
  { Main_Error(Error,
               Text); }

static VOI_type symbol_error(ERR_type Error,
                             SYM_type Symbol)
  { TXT_type text;
    text = Symbol->txt;
    error(Error,
          text); }

static VOI_type procedure_error(ERR_type Error,
                                PRC_type Procedure)
  { SYM_type symbol;
    symbol = Procedure->sym;
    symbol_error(Error,
                 symbol); }

static VOI_type procedure_vararg_error(ERR_type Error,
                                       PRV_type Procedure_vararg)
  { SYM_type symbol;
    symbol = Procedure_vararg->sym;
    symbol_error(Error,
                 symbol); }

static VOI_type keyword_error(ERR_type Error,
                              KEY_type Key)
  { SYM_type symbol;
    symbol = Pool_Keyword_Symbol(Key);
    symbol_error(Error,
                 symbol); }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------ list handling -----------------------------------*/
/*--------------------------------------------------------------------------------------*/

static UNS_type list(EXP_type List_expression,
                     TXT_type Text)
  { SGN_type size;
    size = Grammar_List_Size(List_expression);
    if (size < 0)
      error(ITA_error,
            Text);
    return (UNS_type)size; }

static VOI_type reverse(PAi_type Pair_reference)
  { EXP_type next_expression,
             previous_expression;
    PAI_type next_pair;
    previous_expression = Unique_Null;
    for (next_pair = *Pair_reference;
         is_PAI(next_pair);
         next_pair = next_expression)
      { next_expression = next_pair->cdr;
        next_pair->cdr = previous_expression;
        previous_expression = next_pair; }
    *Pair_reference = previous_expression; }

/*--------------------------------------------------------------------------------------*/

static PAI_type pair_M(EXP_type Car_expression,
                       EXP_type Cdr_expression)
  { PAI_type pair;
    Car_Expression_P = Car_expression;
    Cdr_Expression_P = Cdr_expression;
    pair = make_PAI_M();
    pair->car = Car_Expression_P;
    pair->cdr = Cdr_Expression_P;
    return pair; }

static VOI_type unpair2(EXP_type Pair_Expression,
                        EXp_type Car_reference,
                        EXp_type Cdr_reference)
  { PAI_type pair;
    pair = Pair_Expression;
    *Car_reference = pair->car;
    if (Cdr_reference)
      *Cdr_reference = pair->cdr; }

/*--------------------------------------------------------------------------------------*/

static FRM_type frame_from_list_M(EXP_type List_expression,
                                  UNS_type Size,
                                  UNS_type Count)
  { EXP_type value_expression;
    FRM_type frame;
    PAI_type list_pair;
    UNS_type position;
    List_Expression_P = List_expression;
    frame = Environment_Make_Frame_M(Size);
    for (position = 1;
         position <= Count;
         position += 1)
      { list_pair = List_Expression_P;
        value_expression = list_pair->car;
        frame[position] = value_expression;
        List_Expression_P = list_pair->cdr; }
    return frame; }

static FRM_type frame_from_vararg_list_M(EXP_type List_expression,
                                         UNS_type Size,
                                         UNS_type Count)
  { EXP_type value_expression;
    FRM_type frame;
    PAI_type list_pair;
    UNS_type position;
    List_Expression_P = List_expression;
    frame = Environment_Make_Frame_M(Size);
    for (position = 1;
         position <= Count;
         position += 1)
      { list_pair = List_Expression_P;
        value_expression = list_pair->car;
        frame[position] = value_expression;
        List_Expression_P = list_pair->cdr; }
    frame[position] = List_Expression_P;
    return frame; }

/*--------------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------------*/
/*------------------------------------ C O N T E X T -----------------------------------*/
/*--------------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(context);

THREAD_BEGIN_CHUNK(Ctx);
  THREAD_FRAME_SLOT(DCT, dct);
  THREAD_FRAME_SLOT(ENV, env);
  THREAD_FRAME_SLOT(FRM, frm);
THREAD_END_CHUNK(Ctx);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(context)
  { Ctx_type context_thread;
    DCT_type dictionary;
    ENV_type environment;
    EXP_type expression;
    FRM_type frame;
    THREAD_PEEK(context, Ctx);
    dictionary  = context_thread->dct;
    environment = context_thread->env;
    frame       = context_thread->frm;
    Dictionary_Replace_Dictionary(dictionary);
    Environment_Release_Current_Frame();
    Environment_Replace_Environment(environment);
    Environment_Replace_Frame(frame);
    THREAD_GET_RESULT(expression, EXP);
    THREAD_ZAP;
    THREAD_SET_RESULT(expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type poke_current_context_M(VOI_type)
  { Ctx_type context_thread;
    DCT_type dictionary;
    ENV_type environment;
    FRM_type frame;
    THREAD_POKE_M(context, Ctx);
    dictionary = Dictionary_Get_Current_Dictionary();
    environment = Environment_Get_Current_Environment();
    frame = Environment_Get_Current_Frame();
    context_thread->dct = dictionary;
    context_thread->env = environment;
    context_thread->frm = frame; }

static VOI_type push_current_context_M(VOI_type)
  { Ctx_type context_thread;
    DCT_type dictionary;
    ENV_type environment;
    FRM_type frame;
    THREAD_PUSH_M(context, Ctx);
    dictionary = Dictionary_Get_Current_Dictionary();
    environment = Environment_Get_Current_Environment();
    frame = Environment_Get_Current_Frame();
    context_thread->dct = dictionary;
    context_thread->env = environment;
    context_thread->frm = frame; }

static VOI_type replace_context(DCT_type Dictionary,
                                ENV_type Environment,
                                FRM_type Frame)
  { Dictionary_Replace_Dictionary(Dictionary);
    Environment_Replace_Environment(Environment);
    Environment_Replace_Frame(Frame); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_context(VOI_type)
  { THREAD_REGISTER(context); }

/*--------------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------------*/
/*------------------------------------ C L O N I N G -----------------------------------*/
/*--------------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type clone_expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/
/*------------------------------------ clone literal -----------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type clone_literal(EXP_type Expression)
  { return Expression; }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------- clone pair -------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static PAI_type clone_pair_M(PAI_type Pair)
  { EXP_type car_expression,
             cdr_expression;
    PAI_type pair;
    unpair2(Pair,
            &Car_Expression_P,
            &cdr_expression);
    Stack_Push_M(cdr_expression);
    car_expression = clone_expression_M(Car_Expression_P);
    cdr_expression = Stack_Swap(car_expression);
    cdr_expression = clone_expression_M(cdr_expression);
    car_expression = Stack_Pop();
    pair = pair_M(car_expression,
                  cdr_expression);
    return pair; }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------ clone string ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static STR_type clone_string_M(STR_type String)
  { STR_type string;
    TXT_type text;
    UNS_type length;
    length = length_STR(String);
    if (length == 0)
      return Unique_Empty_String_P;
    String_P = String;
    string = make_STR_M(length);
    text = String_P->txt;
    C_STRCPY(string->txt, text);
    return string; }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------ clone vector ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static VEC_type clone_vector_M(VEC_type Vector)
  { EXP_type expression;
    UNS_type index;
    UNS_type size;
    VEC_type vector;
    size = size_VEC(Vector);
    if (size == 0)
      return Unique_Empty_Vector_P;
    Stack_Push_M(Vector);
    Vector_P = Vector;
    vector = make_VEC_M(size);
    Stack_Push_M(vector);
    for (index = 1;
         index <= size;
         index += 1)
      { Stack_Push_M(Vector_P);
        expression = Vector_P[index];
        expression = clone_expression_M(expression);
        Vector_P = Stack_Pop();
        vector = Stack_Peek();
        vector[index] = expression; }
    vector = Stack_Pop();
    return vector; }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------- clone expression ----------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type clone_expression_M(EXP_type Expression)
  { TAG_type tag;
    tag = Grammar_Tag(Expression);
    switch (tag)
      { case PAI_tag:
          return clone_pair_M(Expression);
        case STR_tag:
          return clone_string_M(Expression);
        case VEC_tag:
          return clone_vector_M(Expression);
        case CHA_tag:
        case FLS_tag:
        case MAC_tag:
        case NUL_tag:
        case REA_tag:
        case SMI_tag:
        case SYM_tag:
        case TRU_tag:
          return clone_literal(Expression);
        default:
          keyword_error(IXT_error,
                        Pool_Quote); }
    return Unique_Undefined; }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cloning(VOI_type)
  { return; }

/*--------------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------------*/
/*-------------------------------------- A P P L Y -------------------------------------*/
/*--------------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------------------*/
/*------------------------------------ continuation ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(continuation);

THREAD_BEGIN_CHUNK(Cnt);
  THREAD_FRAME_SLOT(CNT, cnt);
THREAD_END_CHUNK(Cnt);

static Cnt_type continuation_thread;

/*--------------------------------------------------------------------------------------*/

static VOI_type    expression_M(EXP_type);
static VOI_type replace_context(DCT_type,
                                ENV_type,
                                FRM_type);

/*--------------------------------------------------------------------------------------*/

static VOI_type activate_continuation(CNT_type Continuation,
                                      EXP_type Value_expression)
  { DCT_type dictionary;
    ENV_type environment;
    FRM_type frame;
    THR_type thread;
    environment = Continuation->env;
    frame       = Continuation->frm;
    thread      = Continuation->thr;
    dictionary  = Continuation->dct;
    Environment_Release_Current_Frame();
    replace_context(dictionary,
                    environment,
                    frame);
    Thread_Replace(thread);
    THREAD_SET_RESULT(Value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type apply_direct_continuation(CNT_type Continuation,
                                          EXP_type Operand_list_expression)
  { EXP_type value_expression;
    PAI_type operand_list_pair;
    UNS_type size;
    size = list(Operand_list_expression,
                Call_CC_text);
    if (size != 1)
      error(E1X_error,
            Call_CC_text);
    operand_list_pair = Operand_list_expression;
    value_expression = operand_list_pair->car;
    activate_continuation(Continuation,
                          value_expression); }

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(continuation)
  { CNT_type continuation;
    EXP_type value_expression;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(continuation, Cnt);
    continuation = continuation_thread->cnt;
    THREAD_ZAP;
    activate_continuation(continuation,
                          value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type apply_continuation_M(CNT_type Continuation,
                                     SEQ_type Sequence)
  { UNS_type size;
    size = size_SEQ(Sequence);
    if (size != 1)
      error(CRQ_error,
            Call_CC_text);
    Continuation_P = Continuation;
    Expression_P = Sequence[1];
    THREAD_PUSH_M(continuation, Cnt);
    continuation_thread->cnt = Continuation_P;
    expression_M(Expression_P); }

static CNT_type make_continuation_M(VOI_type)
  { CNT_type continuation;
    continuation = make_CNT_M();
    continuation->env = Environment_Get_Current_Environment();
    continuation->frm = Environment_Mark_Frame();
    continuation->thr = Thread_Mark();
    continuation->dct = Dictionary_Get_Current_Dictionary();
    return continuation; }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_continuation(VOI_type)
  { THREAD_REGISTER(continuation); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- native --------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(native_M);

THREAD_BEGIN_CHUNK(Nat);
  THREAD_FRAME_SLOT(NAT, nat);
  THREAD_FRAME_SLOT(SEQ, opd);
  THREAD_FRAME_SLOT(FRM, frm);
  THREAD_FRAME_SLOT(SMI, pos);
THREAD_END_CHUNK(Nat);

/*--------------------------------------------------------------------------------------*/

static UNS_type direct_expressions(UNS_type,
                                   UNS_type,
                                   FRM_type,
                                   SEQ_type);
static VOI_type       expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

static VOI_type call_native_fuction_M(NAT_type Native,
                                      FRM_type Frame)
  { NFN_type native_fuction_M;
    native_fuction_M = Native->nfn;
    native_fuction_M(Frame); }

/*--------------------------------------------------------------------------------------*/

static VOI_type apply_direct_native_M(NAT_type Native,
                                      EXP_type Operand_list_expression)
  { FRM_type frame;
    TXT_type text;
    UNS_type size;
    text = Native->txt;
    size = list(Operand_list_expression,
                text);
    Native_P = Native;
    frame = frame_from_list_M(Operand_list_expression,
                              size,
                              size);
    call_native_fuction_M(Native_P,
                          frame); }

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(native_M)
  { Nat_type native_M_thread;
    EXP_type expression,
             value_expression;
    FRM_type frame;
    NAT_type native;
    SMI_type position_smallint;
    UNS_type position;
    UNS_type size;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(native_M, Nat);
    native            = native_M_thread->nat;
    Sequence_P        = native_M_thread->opd;
    frame             = native_M_thread->frm;
    position_smallint = native_M_thread->pos;
    position = get_SMU(position_smallint);
    frame[position] = value_expression;
    size = Environment_Frame_Size(frame);
    position += 1;
    position = direct_expressions(size,
                                  position,
                                  frame,
                                  Sequence_P);
    if (position == 0)
      { THREAD_ZAP;
        call_native_fuction_M(native,
                              frame); }
    else
      { position_smallint = make_SMU(position);
        THREAD_KEEP_M(native_M, Nat);
        native_M_thread->pos = position_smallint;
        expression = Sequence_P[position];
        expression_M(expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type apply_native_M(NAT_type Native,
                               SEQ_type Sequence)
  { Nat_type native_M_thread;
    EXP_type expression;
    SMI_type position_smallint;
    UNS_type position;
    UNS_type size;
    size = size_SEQ(Sequence);
    Native_P = Native;
    Sequence_P = Sequence;
    Frame_P = Environment_Make_Frame_M(size);
    position = direct_expressions(size,
                                  1 ,
                                  Frame_P,
                                  Sequence_P);
    if (position == 0)
      call_native_fuction_M(Native_P,
                            Frame_P);
    else
      { position_smallint = make_SMU(position);
        THREAD_PUSH_M(native_M, Nat);
        native_M_thread->nat = Native_P;
        native_M_thread->opd = Sequence_P;
        native_M_thread->frm = Frame_P;
        native_M_thread->pos = position_smallint;
        expression = Sequence_P[position];
        expression_M(expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_native(VOI_type)
  { THREAD_REGISTER(native_M); }

/*--------------------------------------------------------------------------------------*/
/*--------------------------------- higher-order native --------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(higher_order_native_M);
THREAD_ID(higher_order_native_T_M);

THREAD_BEGIN_CHUNK(Naz);
  THREAD_FRAME_SLOT(NAZ, naz);
  THREAD_FRAME_SLOT(SEQ, opd);
  THREAD_FRAME_SLOT(FRM, frm);
  THREAD_FRAME_SLOT(SMI, pos);
THREAD_END_CHUNK(Naz);

/*--------------------------------------------------------------------------------------*/

static UNS_type direct_expressions(UNS_type,
                                   UNS_type,
                                   FRM_type,
                                   SEQ_type);
static VOI_type       expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

static VOI_type call_higher_order_native_fuction_M(NAZ_type Higher_order_native,
                                                   FRM_type Frame)
  { NFN_type native_fuction_M;
    native_fuction_M = Higher_order_native->nfn;
    native_fuction_M(Frame); }

static VOI_type call_higher_order_native_fuction_T_M(NAZ_type Higher_order_native,
                                                     FRM_type Frame)
  { NFN_type native_fuction_M;
    native_fuction_M = Higher_order_native->nfz;
    native_fuction_M(Frame); }

/*--------------------------------------------------------------------------------------*/

static VOI_type apply_direct_higher_order_native_M(NAZ_type Higher_order_native,
                                                   EXP_type Operand_list_expression)
  { FRM_type frame;
    TXT_type text;
    UNS_type size;
    text = Higher_order_native->txt;
    size = list(Operand_list_expression,
                text);
    frame = frame_from_list_M(Operand_list_expression,
                              size,
                              size);
    call_higher_order_native_fuction_M(Higher_order_native,
                                       frame); }

static VOI_type apply_direct_higher_order_native_T_M(NAZ_type Higher_order_native,
                                                     EXP_type Operand_list_expression)
  { FRM_type frame;
    TXT_type text;
    UNS_type size;
    text = Higher_order_native->txt;
    size = list(Operand_list_expression,
                text);
    frame = frame_from_list_M(Operand_list_expression,
                              size,
                              size);
    call_higher_order_native_fuction_T_M(Higher_order_native,
                                         frame); }

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(higher_order_native_M)
  { Naz_type higher_order_native_M_thread;
    EXP_type expression,
             value_expression;
    FRM_type frame;
    NAZ_type higher_order_native;
    SMI_type position_smallint;
    UNS_type position;
    UNS_type size;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(higher_order_native_M, Naz);
    higher_order_native = higher_order_native_M_thread->naz;
    Sequence_P          = higher_order_native_M_thread->opd;
    frame               = higher_order_native_M_thread->frm;
    position_smallint   = higher_order_native_M_thread->pos;
    position = get_SMU(position_smallint);
    frame[position] = value_expression;
    size = Environment_Frame_Size(frame);
    position += 1;
    position = direct_expressions(size,
                                  position,
                                  frame,
                                  Sequence_P);
    if (position == 0)
      { THREAD_ZAP;
        call_higher_order_native_fuction_M(higher_order_native,
                                           frame); }
    else
      { position_smallint = make_SMU(position);
        THREAD_KEEP_M(higher_order_native_M, Naz);
        higher_order_native_M_thread->pos = position_smallint;
        expression = Sequence_P[position];
        expression_M(expression); }}

THREAD_CONTINUATION(higher_order_native_T_M)
  { Naz_type higher_order_native_T_M_thread;
    EXP_type expression,
             value_expression;
    FRM_type frame;
    NAZ_type higher_order_native;
    SMI_type position_smallint;
    UNS_type position;
    UNS_type size;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(higher_order_native_T_M, Naz);
    higher_order_native = higher_order_native_T_M_thread->naz;
    Sequence_P          = higher_order_native_T_M_thread->opd;
    frame               = higher_order_native_T_M_thread->frm;
    position_smallint   = higher_order_native_T_M_thread->pos;
    position = get_SMU(position_smallint);
    frame[position] = value_expression;
    size = Environment_Frame_Size(frame);
    position += 1;
    position = direct_expressions(size,
                                  position,
                                  frame,
                                  Sequence_P);
    if (position == 0)
      { THREAD_ZAP;
        call_higher_order_native_fuction_T_M(higher_order_native,
                                             frame); }
    else
      { position_smallint = make_SMU(position);
        THREAD_KEEP_M(higher_order_native_T_M, Naz);
        higher_order_native_T_M_thread->pos = position_smallint;
        expression = Sequence_P[position];
        expression_M(expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type apply_higher_order_native_M(NAZ_type Higher_order_native,
                                            SEQ_type Sequence)
  { Naz_type higher_order_native_M_thread;
    EXP_type expression;
    SMI_type position_smallint;
    UNS_type position;
    UNS_type size;
    size = size_SEQ(Sequence);
    Higher_Order_Native_P = Higher_order_native;
    Sequence_P = Sequence;
    Frame_P = Environment_Make_Frame_M(size);
    position = direct_expressions(size,
                                  1,
                                  Frame_P,
                                  Sequence_P);
    if (position == 0)
      call_higher_order_native_fuction_M(Higher_Order_Native_P,
                                         Frame_P);
    else
      { position_smallint = make_SMU(position);
        THREAD_PUSH_M(higher_order_native_M, Naz);
        higher_order_native_M_thread->naz = Higher_Order_Native_P;
        higher_order_native_M_thread->opd = Sequence_P;
        higher_order_native_M_thread->frm = Frame_P;
        higher_order_native_M_thread->pos = position_smallint;
        expression = Sequence_P[position];
        expression_M(expression); }}

static VOI_type apply_higher_order_native_T_M(NAZ_type Higher_order_native,
                                              SEQ_type Sequence)
  { Naz_type higher_order_native_T_M_thread;
    EXP_type expression;
    SMI_type position_smallint;
    UNS_type position;
    UNS_type size;
    size = size_SEQ(Sequence);
    Higher_Order_Native_P = Higher_order_native;
    Sequence_P = Sequence;
    Frame_P = Environment_Make_Frame_M(size);
    position = direct_expressions(size,
                                  1,
                                  Frame_P,
                                  Sequence_P);
    if (position == 0)
      call_higher_order_native_fuction_T_M(Higher_Order_Native_P,
                                           Frame_P);
    else
      { position_smallint = make_SMU(position);
        THREAD_PUSH_M(higher_order_native_T_M, Naz);
        higher_order_native_T_M_thread->naz = Higher_Order_Native_P;
        higher_order_native_T_M_thread->opd = Sequence_P;
        higher_order_native_T_M_thread->frm = Frame_P;
        higher_order_native_T_M_thread->pos = position_smallint;
        expression = Sequence_P[position];
        expression_M(expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_higher_order_native(VOI_type)
  { THREAD_REGISTER(higher_order_native_M);
    THREAD_REGISTER(higher_order_native_T_M); }

/*--------------------------------------------------------------------------------------*/
/*--------------------------------------- procedure ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(procedure_M);
THREAD_ID(procedure_T_M);

THREAD_BEGIN_CHUNK(Prc);
  THREAD_FRAME_SLOT(PRC, prc);
  THREAD_FRAME_SLOT(SEQ, opd);
  THREAD_FRAME_SLOT(FRM, frm);
  THREAD_FRAME_SLOT(SMI, pos);
THREAD_END_CHUNK(Prc);

/*--------------------------------------------------------------------------------------*/

static UNS_type     direct_expressions(UNS_type,
                                       UNS_type,
                                       FRM_type,
                                       SEQ_type);
static VOI_type           expression_M(EXP_type);
static VOI_type poke_current_context_M(VOI_type);
static VOI_type        replace_context(DCT_type,
                                       ENV_type,
                                       FRM_type);

/*--------------------------------------------------------------------------------------*/

static VOI_type procedure_body_M(PRC_type Procedure,
                                 FRM_type Frame)
  { DCT_type dictionary;
    ENV_type environment;
    EXP_type expression;
    dictionary  = Procedure->dct;
    environment = Procedure->env;
    expression  = Procedure->exp;
    replace_context(dictionary,
                    environment,
                    Frame);
    expression_M(expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type apply_direct_procedure_M(PRC_type Procedure,
                                         EXP_type Operand_list_expression)
  { SMI_type count_smallint;
    SMI_type size_smallint;
    SYM_type symbol;
    TXT_type text;
    UNS_type count,
             size;
    symbol         = Procedure->sym;
    count_smallint = Procedure->par;
    size_smallint  = Procedure->siz;
    text = symbol->txt;
    size = list(Operand_list_expression,
                text);
    count = get_SMU(count_smallint);
    if (count == size)
      { size = get_SMU(size_smallint);
        Procedure_P = Procedure;
        Frame_P = frame_from_list_M(Operand_list_expression,
                                    size,
                                    count );
        push_current_context_M();
        procedure_body_M(Procedure_P,
                         Frame_P); }
    else if (count < size)
      procedure_error(TMO_error,
                      Procedure);
    else
      procedure_error(TFO_error,
                      Procedure); }

static VOI_type apply_direct_procedure_T_M(PRC_type Procedure,
                                           EXP_type Operand_list_expression)
  { FRM_type frame;
    SMI_type count_smallint;
    SMI_type size_smallint;
    SYM_type symbol;
    TXT_type text;
    UNS_type count,
             size;
    symbol         = Procedure->sym;
    count_smallint = Procedure->par;
    size_smallint  = Procedure->siz;
    text = symbol->txt;
    size = list(Operand_list_expression,
                text);
    count = get_SMU(count_smallint);
    if (count == size)
      { size = get_SMU(size_smallint);
        Procedure_P = Procedure;
        frame = frame_from_list_M(Operand_list_expression,
                                  size,
                                  count );
        Environment_Release_Current_Frame();
        procedure_body_M(Procedure_P,
                         frame); }
    else if (count < size)
      procedure_error(TMO_error,
                      Procedure);
    else
      procedure_error(TFO_error,
                      Procedure); }

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(procedure_M)
  { Prc_type procedure_M_thread;
    EXP_type expression,
             value_expression;
    SMI_type count_smallint;
    SMI_type position_smallint;
    UNS_type position;
    UNS_type size;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(procedure_M, Prc);
    Procedure_P = procedure_M_thread->prc;
    Sequence_P  = procedure_M_thread->opd;
    Frame_P     = procedure_M_thread->frm;
    position_smallint = procedure_M_thread->pos;
    count_smallint = Procedure_P->par;
    position = get_SMU(position_smallint);
    size = get_SMU(count_smallint);
    Frame_P[position] = value_expression;
    position += 1;
    position = direct_expressions(size,
                                  position,
                                  Frame_P,
                                  Sequence_P);
    if (position == 0)
      { poke_current_context_M();
        procedure_body_M(Procedure_P,
                         Frame_P); }
    else
      { position_smallint = make_SMU(position);
        THREAD_KEEP_M(procedure_M, Prc);
        procedure_M_thread->pos = position_smallint;
        expression = Sequence_P[position];
        expression_M(expression); }}

THREAD_CONTINUATION(procedure_T_M)
  { Prc_type procedure_T_M_thread;
    EXP_type expression,
             value_expression;
    FRM_type frame;
    PRC_type procedure;
    SMI_type count_smallint;
    SMI_type position_smallint;
    UNS_type position;
    UNS_type size;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(procedure_T_M, Prc);
    procedure  = procedure_T_M_thread->prc;
    Sequence_P = procedure_T_M_thread->opd;
    frame      = procedure_T_M_thread->frm;
    position_smallint = procedure_T_M_thread->pos;
    count_smallint = procedure->par;
    position = get_SMU(position_smallint);
    size = get_SMU(count_smallint);
    frame[position] = value_expression;
    position += 1;
    position = direct_expressions(size,
                                  position,
                                  frame,
                                  Sequence_P);
    if (position == 0)
      { THREAD_ZAP;
        Environment_Release_Current_Frame();
        procedure_body_M(procedure,
                         frame); }
    else
      { position_smallint = make_SMU(position);
        THREAD_KEEP_M(procedure_T_M, Prc);
        procedure_T_M_thread->pos = position_smallint;
        expression = Sequence_P[position];
        expression_M(expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type apply_procedure_M(PRC_type Procedure,
                                  SEQ_type Sequence)
  { Prc_type procedure_M_thread;
    EXP_type expression;
    SMI_type count_smallint;
    SMI_type position_smallint;
    SMI_type size_smallint;
    UNS_type count,
             position,
             size;
    count_smallint = Procedure->par;
    size_smallint  = Procedure->siz;
    count = get_SMU(count_smallint);
    size = size_SEQ(Sequence);
    if (count > size)
      procedure_error(TFO_error,
                      Procedure);
    if (count < size)
      procedure_error(TMO_error,
                      Procedure);
    size = get_SMU(size_smallint);
    Procedure_P = Procedure;
    Sequence_P = Sequence;
    Frame_P = Environment_Make_Frame_M(size);
    position = direct_expressions(count,
                                  1,
                                  Frame_P,
                                  Sequence_P);
    if (position == 0)
      { push_current_context_M();
        procedure_body_M(Procedure_P,
                         Frame_P); }
    else
      { position_smallint = make_SMU(position);
        THREAD_PUSH_M(procedure_M, Prc);
        procedure_M_thread->prc = Procedure_P;
        procedure_M_thread->opd = Sequence_P;
        procedure_M_thread->frm = Frame_P;
        procedure_M_thread->pos = position_smallint;
        expression = Sequence_P[position];
        expression_M(expression); }}

static VOI_type apply_procedure_T_M(PRC_type Procedure,
                                    SEQ_type Sequence)
  { Prc_type procedure_T_M_thread;
    EXP_type expression;
    SMI_type count_smallint;
    SMI_type position_smallint;
    SMI_type size_smallint;
    UNS_type count,
             position,
             size;
    count_smallint = Procedure->par;
    size_smallint  = Procedure->siz;
    count = get_SMU(count_smallint);
    size = size_SEQ(Sequence);
    if (count > size)
      procedure_error(TFO_error,
                      Procedure);
    if (count < size)
      procedure_error(TMO_error,
                      Procedure);
    size = get_SMU(size_smallint);
    Procedure_P = Procedure;
    Sequence_P = Sequence;
    Frame_P = Environment_Make_Frame_M(size);
    position = direct_expressions(count,
                                  1,
                                  Frame_P,
                                  Sequence_P);
    if (position == 0)
      { Environment_Release_Current_Frame();
        procedure_body_M(Procedure_P,
                         Frame_P); }
    else
      { position_smallint = make_SMU(position);
        THREAD_PUSH_M(procedure_T_M, Prc);
        procedure_T_M_thread->prc = Procedure_P;
        procedure_T_M_thread->opd = Sequence_P;
        procedure_T_M_thread->frm = Frame_P;
        procedure_T_M_thread->pos = position_smallint;
        expression = Sequence_P[position];
        expression_M(expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_procedure(VOI_type)
  { THREAD_REGISTER(procedure_M);
    THREAD_REGISTER(procedure_T_M); }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------ procedure vararg --------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(procedure_vararg_M);
THREAD_ID(procedure_vararg_T_M);
THREAD_ID(vararg_M);
THREAD_ID(vararg_T_M);

THREAD_BEGIN_CHUNK(Prv);
  THREAD_FRAME_SLOT(PRV, prv);
  THREAD_FRAME_SLOT(SEQ, opd);
  THREAD_FRAME_SLOT(FRM, frm);
  THREAD_FRAME_SLOT(EXP, lst);
  THREAD_FRAME_SLOT(SMI, pos);
THREAD_END_CHUNK(Prv);

/*--------------------------------------------------------------------------------------*/

static VOI_type           expression_M(EXP_type);
static VOI_type poke_current_context_M(VOI_type);
static VOI_type push_current_context_M(VOI_type);
static VOI_type        replace_context(DCT_type,
                                       ENV_type,
                                       FRM_type);

/*--------------------------------------------------------------------------------------*/

static VOI_type procedure_vararg_body_M(PRV_type Procedure_vararg,
                                        FRM_type Frame)
  { DCT_type dictionary;
    ENV_type environment;
    EXP_type expression;
    dictionary  = Procedure_vararg->dct;
    environment = Procedure_vararg->env;
    expression  = Procedure_vararg->exp;
    replace_context(dictionary,
                    environment,
                    Frame);
    expression_M(expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type apply_direct_procedure_vararg_M(PRV_type Procedure_vararg,
                                                EXP_type Operand_list_expression)
  { SMI_type count_smallint;
    SMI_type size_smallint;
    SYM_type symbol;
    TXT_type text;
    UNS_type count,
             size;
    symbol         = Procedure_vararg->sym;
    count_smallint = Procedure_vararg->par;
    size_smallint  = Procedure_vararg->siz;
    text = symbol->txt;
    size = list(Operand_list_expression,
                text);
    count = get_SMU(count_smallint);
    if (count <= size)
      { size = get_SMU(size_smallint);
        Procedure_Vararg_P = Procedure_vararg;
        Frame_P = frame_from_vararg_list_M(Operand_list_expression,
                                           size,
                                           count );
        push_current_context_M();
        procedure_vararg_body_M(Procedure_Vararg_P,
                                Frame_P); }
    else
      procedure_vararg_error(TFO_error,
                             Procedure_Vararg_P); }

static VOI_type apply_direct_procedure_vararg_T_M(PRV_type Procedure_vararg,
                                                  EXP_type Operand_list_expression)
  { SMI_type count_smallint;
    SMI_type size_smallint;
    SYM_type symbol;
    TXT_type text;
    UNS_type count,
             size;
    symbol         = Procedure_vararg->sym;
    count_smallint = Procedure_vararg->par;
    size_smallint  = Procedure_vararg->siz;
    text = symbol->txt;
    size = list(Operand_list_expression,
                text);
    count = get_SMU(count_smallint);
    if (count <= size)
      { size = get_SMU(size_smallint);
        Procedure_Vararg_P = Procedure_vararg;
        Frame_P = frame_from_vararg_list_M(Operand_list_expression,
                                           size,
                                           count );
        Environment_Release_Current_Frame();
        procedure_vararg_body_M(Procedure_Vararg_P,
                                Frame_P); }
    else
      procedure_vararg_error(TFO_error,
                             Procedure_Vararg_P); }

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(procedure_vararg_M)
  { Prv_type procedure_vararg_M_thread;
    Prv_type vararg_M_thread;
    EXP_type expression,
             value_expression;
    SMI_type count_smallint;
    SMI_type position_smallint;
    UNS_type count,
             position;
    UNS_type sequence_size;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(procedure_vararg_M, Prv);
    Procedure_Vararg_P = procedure_vararg_M_thread->prv;
    Sequence_P         = procedure_vararg_M_thread->opd;
    Frame_P            = procedure_vararg_M_thread->frm;
    position_smallint  = procedure_vararg_M_thread->pos;
    position = get_SMU(position_smallint);
    Frame_P[position] = value_expression;
    sequence_size = size_SEQ(Sequence_P);
    if (position == sequence_size)
      { position += 1;
        Frame_P[position] = Unique_Null;
        poke_current_context_M();
        procedure_vararg_body_M(Procedure_Vararg_P,
                                Frame_P); }
    else
      { count_smallint = Procedure_Vararg_P->par;
        count = get_SMU(count_smallint);
        position += 1;
        position_smallint = make_SMU(position);
        if (position <= count)
          { THREAD_KEEP_M(procedure_vararg_M, Prv);
            procedure_vararg_M_thread->pos = position_smallint; }
        else
          { THREAD_PATCH_M(vararg_M, Prv);
            vararg_M_thread->pos = position_smallint; }
        expression = Sequence_P[position];
        expression_M(expression); }}

THREAD_CONTINUATION(procedure_vararg_T_M)
  { Prv_type procedure_vararg_T_M_thread;
    Prv_type vararg_T_M_thread;
    EXP_type expression,
             value_expression;
    FRM_type frame;
    SMI_type count_smallint;
    SMI_type position_smallint;
    UNS_type count,
             position;
    UNS_type sequence_size;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(procedure_vararg_T_M, Prv);
    Procedure_Vararg_P = procedure_vararg_T_M_thread->prv;
    Sequence_P         = procedure_vararg_T_M_thread->opd;
    frame              = procedure_vararg_T_M_thread->frm;
    position_smallint  = procedure_vararg_T_M_thread->pos;
    position = get_SMU(position_smallint);
    frame[position] = value_expression;
    sequence_size = size_SEQ(Sequence_P);
    if (position == sequence_size)
      { position += 1;
        frame[position] = Unique_Null;
        Environment_Release_Current_Frame();
        THREAD_ZAP;
        procedure_vararg_body_M(Procedure_Vararg_P,
                                frame); }
    else
      { count_smallint = Procedure_Vararg_P->par;
        count = get_SMU(count_smallint);
        position += 1;
        position_smallint = make_SMU(position);
        if (position <= count)
          { THREAD_KEEP_M(procedure_vararg_T_M, Prv);
            procedure_vararg_T_M_thread->pos = position_smallint; }
        else
          { THREAD_PATCH_M(vararg_T_M, Prv);
            vararg_T_M_thread->pos = position_smallint; }
        expression = Sequence_P[position];
        expression_M(expression); }}

THREAD_CONTINUATION(vararg_M)
  { Prv_type vararg_M_thread;
    EXP_type car_expression,
             cdr_expression,
             expression;
    SMI_type count_smallint;
    SMI_type position_smallint;
    UNS_type count,
             position;
    UNS_type sequence_size;
    THREAD_GET_RESULT(car_expression, EXP);
    THREAD_PEEK(vararg_M, Prv);
    Procedure_Vararg_P = vararg_M_thread->prv;
    Frame_P            = vararg_M_thread->frm;
    Sequence_P         = vararg_M_thread->opd;
    cdr_expression     = vararg_M_thread->lst;
    position_smallint  = vararg_M_thread->pos;
    Pair_P = pair_M(car_expression,
                    cdr_expression);
    position = get_SMU(position_smallint);
    sequence_size = size_SEQ(Sequence_P);
    if (position == sequence_size)
      { count_smallint = Procedure_Vararg_P->par;
        count = get_SMU(count_smallint);
        reverse(&Pair_P);
        position = count + 1;
        Frame_P[position] = Pair_P;
        poke_current_context_M();
        procedure_vararg_body_M(Procedure_Vararg_P,
                                Frame_P); }
    else
      { position += 1;
        position_smallint = make_SMU(position);
        THREAD_KEEP_M(vararg_M, Prv);
        vararg_M_thread->lst = Pair_P;
        vararg_M_thread->pos = position_smallint;
        expression = Sequence_P[position];
        expression_M(expression); }}

THREAD_CONTINUATION(vararg_T_M)
  { Prv_type vararg_T_M_thread;
    EXP_type car_expression,
             cdr_expression,
             expression;
    SMI_type count_smallint;
    SMI_type position_smallint;
    UNS_type count,
             position;
    UNS_type sequence_size;
    THREAD_GET_RESULT(car_expression, EXP);
    THREAD_PEEK(vararg_T_M, Prv);
    Procedure_Vararg_P = vararg_T_M_thread->prv;
    Frame_P            = vararg_T_M_thread->frm;
    Sequence_P         = vararg_T_M_thread->opd;
    cdr_expression     = vararg_T_M_thread->lst;
    position_smallint  = vararg_T_M_thread->pos;
    Pair_P = pair_M(car_expression,
                    cdr_expression);
    position = get_SMU(position_smallint);
    sequence_size = size_SEQ(Sequence_P);
    if (position == sequence_size)
      { count_smallint = Procedure_Vararg_P->par;
        count = get_SMU(count_smallint);
        reverse(&Pair_P);
        position = count + 1;
        Frame_P[position] = Pair_P;
        Environment_Release_Current_Frame();
        THREAD_ZAP;
        procedure_vararg_body_M(Procedure_Vararg_P,
                                Frame_P); }
    else
      { position += 1;
        position_smallint = make_SMU(position);
        THREAD_KEEP_M(vararg_T_M, Prv);
        vararg_T_M_thread->lst = Pair_P;
        vararg_T_M_thread->pos = position_smallint;
        expression = Sequence_P[position];
        expression_M(expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type apply_procedure_vararg_M(PRV_type Procedure_vararg,
                                         SEQ_type Sequence)
  { Prv_type procedure_vararg_M_thread,
             vararg_M_thread;
    EXP_type expression;
    SMI_type count_smallint;
    SMI_type position_smallint;
    SMI_type size_smallint;
    UNS_type count,
             position,
             sequence_size,
             size;
    count_smallint = Procedure_vararg->par;
    size_smallint  = Procedure_vararg->siz;
    count = get_SMU(count_smallint);
    sequence_size = size_SEQ(Sequence);
    if (count > sequence_size)
      procedure_vararg_error(TFO_error,
                             Procedure_vararg);
    size = get_SMU(size_smallint);
    Procedure_Vararg_P = Procedure_vararg;
    Sequence_P = Sequence;
    Frame_P = Environment_Make_Frame_M(size);
    position = direct_expressions(count,
                                  1,
                                  Frame_P,
                                  Sequence_P);
    if ((position == 0) && (count == sequence_size))
      { Frame_P[count + 1] = Unique_Null;
        push_current_context_M();
        procedure_vararg_body_M(Procedure_Vararg_P,
                                Frame_P); }
    else
      { if (position == 0)
          { position = count + 1;
            position_smallint = make_SMU(position);
            THREAD_PUSH_M(vararg_M, Prv);
            vararg_M_thread->prv = Procedure_Vararg_P;
            vararg_M_thread->opd = Sequence_P;
            vararg_M_thread->frm = Frame_P;
            vararg_M_thread->lst = Unique_Null;
            vararg_M_thread->pos = position_smallint; }
        else
          { position_smallint = make_SMU(position);
            THREAD_PUSH_M(procedure_vararg_M, Prv);
            procedure_vararg_M_thread->prv = Procedure_Vararg_P;
            procedure_vararg_M_thread->opd = Sequence_P;
            procedure_vararg_M_thread->frm = Frame_P;
            procedure_vararg_M_thread->lst = Unique_Null;
            procedure_vararg_M_thread->pos = position_smallint; }
        expression = Sequence_P[position];
        expression_M(expression); }}

static VOI_type apply_procedure_vararg_T_M(PRV_type Procedure_vararg,
                                           SEQ_type Sequence)
  { Prv_type procedure_vararg_T_M_thread,
             vararg_T_M_thread;
    EXP_type expression;
    SMI_type count_smallint;
    SMI_type position_smallint;
    SMI_type size_smallint;
    UNS_type count,
             position,
             sequence_size,
             size;
    count_smallint = Procedure_vararg->par;
    size_smallint  = Procedure_vararg->siz;
    count = get_SMU(count_smallint);
    sequence_size = size_SEQ(Sequence);
    if (count > sequence_size)
      procedure_vararg_error(TFO_error,
                             Procedure_vararg);
    size = get_SMU(size_smallint);
    Procedure_Vararg_P = Procedure_vararg;
    Sequence_P = Sequence;
    Frame_P = Environment_Make_Frame_M(size);
    position = direct_expressions(count,
                                  1,
                                  Frame_P,
                                  Sequence_P);
    if ((position == 0) && (count == sequence_size))
      { Frame_P[count + 1] = Unique_Null;
        push_current_context_M();
        procedure_vararg_body_M(Procedure_Vararg_P,
                                Frame_P); }
    else
      { if (position == 0)
          { position = count + 1;
            position_smallint = make_SMU(position);
            THREAD_PUSH_M(vararg_T_M, Prv);
            vararg_T_M_thread->prv = Procedure_Vararg_P;
            vararg_T_M_thread->opd = Sequence_P;
            vararg_T_M_thread->frm = Frame_P;
            vararg_T_M_thread->lst = Unique_Null;
            vararg_T_M_thread->pos = position_smallint; }
        else
          { position_smallint = make_SMU(position);
            THREAD_PUSH_M(procedure_vararg_T_M, Prv);
            procedure_vararg_T_M_thread->prv = Procedure_Vararg_P;
            procedure_vararg_T_M_thread->opd = Sequence_P;
            procedure_vararg_T_M_thread->frm = Frame_P;
            procedure_vararg_T_M_thread->lst = Unique_Null;
            procedure_vararg_T_M_thread->pos = position_smallint; }
        expression = Sequence_P[position];
        expression_M(expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_procedure_vararg(VOI_type)
  { THREAD_REGISTER(procedure_vararg_M);
    THREAD_REGISTER(procedure_vararg_T_M);
    THREAD_REGISTER(vararg_M);
    THREAD_REGISTER(vararg_T_M); }

/*--------------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------------*/
/*----------------------------------- E V A L U A T E ----------------------------------*/
/*--------------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------------------*/
/*------------------------------------------ and ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(and_M);

THREAD_BEGIN_CHUNK(And);
  THREAD_FRAME_SLOT(SEQ, prd);
  THREAD_FRAME_SLOT(SMI, pos);
THREAD_END_CHUNK(And);

/*--------------------------------------------------------------------------------------*/

static VOI_type expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(and_M)
  { And_type and_M_thread;
    EXP_type expression,
             value_expression;
    SMI_type position_smallint;
    UNS_type position;
    UNS_type size;
    THREAD_GET_RESULT(value_expression, EXP);
    if (is_FLS(value_expression))
      { THREAD_ZAP;
        THREAD_SET_RESULT(Unique_False); }
    else
      { THREAD_PEEK(and_M, And);
        Sequence_P        = and_M_thread->prd;
        position_smallint = and_M_thread->pos;
        size = size_SEQ(Sequence_P);
        position = get_SMU(position_smallint);
        if (position == size)
          { THREAD_ZAP;
            THREAD_SET_RESULT(Unique_True); }
        else
          { position += 1;
            position_smallint = make_SMU(position);
            THREAD_KEEP_M(and_M, And);
            and_M_thread->pos = position_smallint;
            expression = Sequence_P[position];
            expression_M(expression); }}}

/*--------------------------------------------------------------------------------------*/

static VOI_type and_M(AND_type And)
  { And_type and_M_thread;
    EXP_type expression;
    UNS_type size;
    Sequence_P = And->prd;
    size = size_SEQ(Sequence_P);
    if (size == 0)
      THREAD_SET_RESULT(Unique_True);
    else
      { THREAD_PUSH_M(and_M, And);
        and_M_thread->prd = Sequence_P;
        and_M_thread->pos = Unique_One_Smallint;
        expression = Sequence_P[1];
        expression_M(expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_and(VOI_type)
  { THREAD_REGISTER(and_M);  }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------- application ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(application_M);
THREAD_ID(application_T_M);

THREAD_BEGIN_CHUNK(Apl);
  THREAD_FRAME_SLOT(SEQ, seq);
THREAD_END_CHUNK(Apl);

/*--------------------------------------------------------------------------------------*/

static VOI_type            apply_direct_continuation(CNT_type,
                                                     EXP_type);
static VOI_type                apply_direct_native_M(NAT_type,
                                                     EXP_type);
static VOI_type   apply_direct_higher_order_native_M(NAZ_type,
                                                     EXP_type);
static VOI_type apply_direct_higher_order_native_T_M(NAZ_type,
                                                     EXP_type);
static VOI_type             apply_direct_procedure_M(PRC_type,
                                                     EXP_type);
static VOI_type           apply_direct_procedure_T_M(PRC_type,
                                                     EXP_type);
static VOI_type      apply_direct_procedure_vararg_M(PRV_type,
                                                     EXP_type);
static VOI_type    apply_direct_procedure_vararg_T_M(PRV_type,
                                                     EXP_type);
static VOI_type                 apply_continuation_M(CNT_type,
                                                     SEQ_type);
static VOI_type                       apply_native_M(NAT_type,
                                                     SEQ_type);
static VOI_type          apply_higher_order_native_M(NAZ_type,
                                                     SEQ_type);
static VOI_type        apply_higher_order_native_T_M(NAZ_type,
                                                     SEQ_type);
static VOI_type                    apply_procedure_M(PRC_type,
                                                     FRM_type);
static VOI_type                  apply_procedure_T_M(PRC_type,
                                                     FRM_type);
static VOI_type             apply_procedure_vararg_M(PRV_type,
                                                     FRM_type);
static VOI_type           apply_procedure_vararg_T_M(PRV_type,
                                                     FRM_type);
static VOI_type                         expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

static VOI_type direct_apply_M(EXP_type Operator_expression,
                               EXP_type Operand_expression)
  { TAG_type tag;
    tag = Grammar_Tag(Operator_expression);
    switch (tag)
      { case CNT_tag:
          apply_direct_continuation(Operator_expression,
                                    Operand_expression);
          return;
        case NAT_tag:
          apply_direct_native_M(Operator_expression,
                                Operand_expression);
          return;
        case NAZ_tag:
          apply_direct_higher_order_native_M(Operator_expression,
                                             Operand_expression);
          return;
        case PRC_tag:
          apply_direct_procedure_M(Operator_expression,
                                   Operand_expression);
          return;
        case PRV_tag:
          apply_direct_procedure_vararg_M(Operator_expression,
                                          Operand_expression);
          return;
        default:
          keyword_error(OPR_error,
                        Pool_Application); }}

static VOI_type direct_apply_T_M(EXP_type Operator_expression,
                                 EXP_type Operand_expression)
  { TAG_type tag;
    tag = Grammar_Tag(Operator_expression);
    switch (tag)
      { case CNT_tag:
          apply_direct_continuation(Operator_expression,
                                    Operand_expression);
          return;
        case NAT_tag:
          apply_direct_native_M(Operator_expression,
                                Operand_expression);
          return;
        case NAZ_tag:
          apply_direct_higher_order_native_T_M(Operator_expression,
                                               Operand_expression);
          return;
        case PRC_tag:
          apply_direct_procedure_T_M(Operator_expression,
                                     Operand_expression);
          return;
        case PRV_tag:
          apply_direct_procedure_vararg_T_M(Operator_expression,
                                            Operand_expression);
          return;
        default:
          keyword_error(OPR_error,
                        Pool_Application); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type apply_M(EXP_type Operator_expression,
                        SEQ_type Operand_sequence)
  { TAG_type tag;
    tag = Grammar_Tag(Operator_expression);
    switch (tag)
      { case CNT_tag:
          apply_continuation_M(Operator_expression,
                               Operand_sequence);
          return;
        case NAT_tag:
          apply_native_M(Operator_expression,
                         Operand_sequence);
          return;
        case NAZ_tag:
          apply_higher_order_native_M(Operator_expression,
                                      Operand_sequence);
          return;
        case PRC_tag:
          apply_procedure_M(Operator_expression,
                            Operand_sequence);
          return;
        case PRV_tag:
          apply_procedure_vararg_M(Operator_expression,
                                   Operand_sequence);
          return;
        default:
          keyword_error(OPR_error,
                        Pool_Application); }}

static VOI_type apply_T_M(EXP_type Operator_expression,
                          SEQ_type Operand_sequence)
  { TAG_type tag;
    tag = Grammar_Tag(Operator_expression);
    switch (tag)
      { case CNT_tag:
          apply_continuation_M(Operator_expression,
                               Operand_sequence);
          return;
        case NAT_tag:
          apply_native_M(Operator_expression,
                         Operand_sequence);
          return;
        case NAZ_tag:
          apply_higher_order_native_T_M(Operator_expression,
                                        Operand_sequence);
          return;
        case PRC_tag:
          apply_procedure_T_M(Operator_expression,
                              Operand_sequence);
          return;
        case PRV_tag:
          apply_procedure_vararg_T_M(Operator_expression,
                                     Operand_sequence);
          return;
        default:
          keyword_error(OPR_error,
                        Pool_Application); }}

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(application_M)
  { Apl_type application_M_thread;
    EXP_type value_expression;
    SEQ_type sequence;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(application_M, Apl);
    sequence = application_M_thread->seq;
    THREAD_ZAP;
    apply_M(value_expression,
            sequence); }

THREAD_CONTINUATION(application_T_M)
  { Apl_type application_T_M_thread;
    EXP_type value_expression;
    SEQ_type sequence;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(application_T_M, Apl);
    sequence = application_T_M_thread->seq;
    THREAD_ZAP;
    apply_T_M(value_expression,
              sequence); }

/*--------------------------------------------------------------------------------------*/

static VOI_type application_M(APL_type Application)
  { Apl_type application_M_thread;
    Expression_P = Application->exp;
    Sequence_P   = Application->arg;
    THREAD_PUSH_M(application_M, Apl);
    application_M_thread->seq = Sequence_P;
    expression_M(Expression_P); }

static VOI_type application_T_M(APL_type Application)
  { Apl_type application_T_M_thread;
    Expression_P = Application->exp;
    Sequence_P   = Application->arg;
    THREAD_PUSH_M(application_T_M, Apl);
    application_T_M_thread->seq = Sequence_P;
    expression_M(Expression_P); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_application(VOI_type)
  { THREAD_REGISTER(application_M);
    THREAD_REGISTER(application_T_M); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- begin ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static VOI_type expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

static VOI_type begin_M(BEG_type Begin)

  { EXP_type expression;
    expression = Begin->exp;
    expression_M(expression); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- cond ----------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(cond_M);
THREAD_ID(cond_else_M);

THREAD_BEGIN_CHUNK(Cnd);
  THREAD_FRAME_SLOT(SEQ, prd);
  THREAD_FRAME_SLOT(SEQ, exp);
  THREAD_FRAME_SLOT(SMI, pos);
THREAD_END_CHUNK(Cnd);

THREAD_BEGIN_CHUNK(Cne);
  THREAD_FRAME_SLOT(SEQ, prd);
  THREAD_FRAME_SLOT(SEQ, exp);
  THREAD_FRAME_SLOT(EXP, els);
  THREAD_FRAME_SLOT(SMI, pos);
THREAD_END_CHUNK(Cne);

/*--------------------------------------------------------------------------------------*/

static VOI_type expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(cond_M)
  { Cnd_type cond_M_thread;
    EXP_type expression,
             predicate_expression,
             value_expression;
    SEQ_type body_sequence;
    SMI_type position_smallint;
    UNS_type position;
    UNS_type size;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(cond_M, Cnd);
    Predicate_Sequence_P = cond_M_thread->prd;
    body_sequence        = cond_M_thread->exp;
    position_smallint    = cond_M_thread->pos;
    position = get_SMU(position_smallint);
    if (is_FLS(value_expression))
      { size = size_SEQ(Predicate_Sequence_P);
        if (position == size)
          { THREAD_ZAP;
            THREAD_SET_RESULT(Unique_Unspecified); }
        else
          { position += 1;
            position_smallint = make_SMU(position);
            THREAD_KEEP_M(cond_M, Cnd);
            cond_M_thread->pos = position_smallint;
            predicate_expression = Predicate_Sequence_P[position];
            expression_M(predicate_expression); }}
    else
      { THREAD_ZAP;
        expression = body_sequence[position];
        expression_M(expression); }}

THREAD_CONTINUATION(cond_else_M)
  { Cne_type cond_else_M_thread;
    EXP_type else_expression,
             expression,
             predicate_expression,
             value_expression;
    SEQ_type body_sequence;
    SMI_type position_smallint;
    UNS_type position;
    UNS_type size;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(cond_else_M, Cne);
    Predicate_Sequence_P = cond_else_M_thread->prd;
    body_sequence        = cond_else_M_thread->exp;
    else_expression      = cond_else_M_thread->els;
    position_smallint    = cond_else_M_thread->pos;
    position = get_SMU(position_smallint);
    if (is_FLS(value_expression))
      { size = size_SEQ(Body_Sequence_P);
        if (position == size)
          { THREAD_ZAP;
            expression_M(else_expression); }
        else
          { position += 1;
            position_smallint = make_SMU(position);
            THREAD_KEEP_M(cond_else_M, Cne);
            cond_else_M_thread->pos = position_smallint;
            predicate_expression = Predicate_Sequence_P[position];
            expression_M(predicate_expression); }}
    else
      { THREAD_ZAP;
        expression = body_sequence[position];
        expression_M(expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type cond_M(CND_type Cond_T)
  { Cnd_type cond_M_thread;
    EXP_type predicate_expression;
    Predicate_Sequence_P = Cond_T->prd;
    Body_Sequence_P      = Cond_T->bod;
    THREAD_PUSH_M(cond_M, Cnd);
    cond_M_thread->prd = Predicate_Sequence_P;
    cond_M_thread->exp = Body_Sequence_P;
    cond_M_thread->pos = Unique_One_Smallint;
    predicate_expression = Predicate_Sequence_P[1];
    expression_M(predicate_expression); }

static VOI_type cond_else_M(CNE_type Cond_else_T)
  { Cne_type cond_else_M_thread;
    EXP_type predicate_expression;
    UNS_type size;
    Predicate_Sequence_P = Cond_else_T->prd;
    Body_Sequence_P      = Cond_else_T->bod;
    Else_Expression_P    = Cond_else_T->els;
    size = size_SEQ(Predicate_Sequence_P);
    if (size == 0)
       expression_M(Else_Expression_P);
    else
      { THREAD_PUSH_M(cond_else_M, Cne);
        cond_else_M_thread->prd = Predicate_Sequence_P;
        cond_else_M_thread->exp = Body_Sequence_P;
        cond_else_M_thread->els = Else_Expression_P;
        cond_else_M_thread->pos = Unique_One_Smallint;
        predicate_expression = Predicate_Sequence_P[1];
        expression_M(predicate_expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cond(VOI_type)
  { THREAD_REGISTER(cond_M);
    THREAD_REGISTER(cond_else_M); }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------- cons-stream ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(cons_stream_M);

THREAD_BEGIN_CHUNK(Cns);
  THREAD_FRAME_SLOT(SMI, siz);
  THREAD_FRAME_SLOT(EXP, cdr);
  THREAD_FRAME_SLOT(DCT, dct);
THREAD_END_CHUNK(Cns);

/*--------------------------------------------------------------------------------------*/

static VOI_type expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(cons_stream_M)
  { Cns_type cons_stream_M_thread;
    PRM_type promise;
    SMI_type size_smallint;
    PAI_type pair;
    THREAD_GET_RESULT(Car_Expression_P, EXP);
    THREAD_PEEK(cons_stream_M, Cns);
    size_smallint    = cons_stream_M_thread->siz;
    Cdr_Expression_P = cons_stream_M_thread->cdr;
    Dictionary_P     = cons_stream_M_thread->dct;
    THREAD_ZAP;
    Environment_P = Environment_Extend_M();
    promise = make_PRM_M();
    promise->exp = Cdr_Expression_P;
    promise->siz = size_smallint;
    promise->env = Environment_P;
    promise->dct = Dictionary_P;
    pair = pair_M(Car_Expression_P,
                  promise);
    THREAD_SET_RESULT(pair); }

/*--------------------------------------------------------------------------------------*/

static VOI_type cons_stream_M(CNS_type Cons_stream_T)
  { Cns_type cons_stream_M_thread;
    SMI_type size_smallint;
    Car_Expression_P = Cons_stream_T->car;
    size_smallint    = Cons_stream_T->siz;
    Cdr_Expression_P = Cons_stream_T->cdr;
    Dictionary_P     = Cons_stream_T->dct;
    THREAD_PUSH_M(cons_stream_M, Cns);
    cons_stream_M_thread->siz = size_smallint;
    cons_stream_M_thread->cdr = Cdr_Expression_P;
    cons_stream_M_thread->dct = Dictionary_P;
    expression_M(Car_Expression_P); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cons_stream(VOI_type)
  { THREAD_REGISTER(cons_stream_M); }

/*--------------------------------------------------------------------------------------*/
/*----------------------------------- define function ----------------------------------*/
/*--------------------------------------------------------------------------------------*/

static VOI_type define_function_M(DFF_type Define_function)
  { PRC_type procedure;
    SMI_type count_smallint;
    SMI_type offset_smallint;
    SMI_type size_smallint;
    UNS_type offset;
    Symbol_P        = Define_function->sym;
    offset_smallint = Define_function->ofs;
    count_smallint  = Define_function->par;
    size_smallint   = Define_function->siz;
    Expression_P    = Define_function->exp;
    Dictionary_P    = Define_function->dct;
    Environment_P = Environment_Extend_M();
    procedure = make_PRC_M();
    procedure->sym = Symbol_P;
    procedure->par = count_smallint;
    procedure->exp = Expression_P;
    procedure->siz = size_smallint;
    procedure->env = Environment_P;
    procedure->dct = Dictionary_P;
    offset = get_SMU(offset_smallint);
    Environment_Local_Set(procedure,
                          offset);
    THREAD_SET_RESULT(procedure); }

/*--------------------------------------------------------------------------------------*/
/*-------------------------------- define function vararg ------------------------------*/
/*--------------------------------------------------------------------------------------*/

static VOI_type define_function_vararg_M(DFZ_type Define_function_vararg)
  { PRV_type procedure_vararg;
    SMI_type count_smallint;
    SMI_type offset_smallint;
    SMI_type size_smallint;
    UNS_type offset;
    Symbol_P        = Define_function_vararg->sym;
    offset_smallint = Define_function_vararg->ofs;
    count_smallint  = Define_function_vararg->par;
    size_smallint   = Define_function_vararg->siz;
    Expression_P    = Define_function_vararg->exp;
    Dictionary_P    = Define_function_vararg->dct;
    Environment_P = Environment_Extend_M();
    procedure_vararg = make_PRV_M();
    procedure_vararg->sym = Symbol_P;
    procedure_vararg->par = count_smallint;
    procedure_vararg->exp = Expression_P;
    procedure_vararg->siz = size_smallint;
    procedure_vararg->env = Environment_P;
    procedure_vararg->dct = Dictionary_P;
    offset = get_SMU(offset_smallint);
    Environment_Local_Set(procedure_vararg,
                          offset);
    THREAD_SET_RESULT(procedure_vararg); }

/*--------------------------------------------------------------------------------------*/
/*----------------------------------- define variable ----------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(define_variable);

THREAD_BEGIN_CHUNK(Dfv);
  THREAD_FRAME_SLOT(SMI, ofs);
THREAD_END_CHUNK(Dfv);

/*--------------------------------------------------------------------------------------*/

static EXP_type direct_expression(EXP_type);
static VOI_type      expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(define_variable)
  { Dfv_type define_variable_thread;
    EXP_type value_expression;
    SMI_type offset_smallint;
    UNS_type offset;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(define_variable, Dfv);
    offset_smallint = define_variable_thread->ofs;
    THREAD_ZAP;
    offset = get_SMU(offset_smallint);
    Environment_Local_Set(value_expression,
                          offset);
    THREAD_SET_RESULT(value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type define_variable_M(DFV_type Define_variable)
  { Dfv_type define_variable_thread;
    EXP_type value_expression;
    SMI_type offset_smallint;
    UNS_type offset;
    offset_smallint = Define_variable->ofs;
    Expression_P    = Define_variable->exp;
    value_expression = direct_expression(Expression_P);
    if (is_UDF(value_expression))
      { THREAD_PUSH_M(define_variable, Dfv);
        define_variable_thread->ofs = offset_smallint;
        expression_M(Expression_P); }
    else
      { offset = get_SMU(offset_smallint);
        Environment_Local_Set(value_expression,
                              offset);
        THREAD_SET_RESULT(value_expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_define_variable(VOI_type)
  { THREAD_REGISTER(define_variable); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- delay ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static VOI_type delay_M(DEL_type Delay_T)
  { PRM_type promise;
    SMI_type size_smallint;
    size_smallint = Delay_T->siz;
    Expression_P  = Delay_T->exp;
    Dictionary_P  = Delay_T->dct;
    Environment_P = Environment_Extend_M();
    promise = make_PRM_M();
    promise->exp = Expression_P;
    promise->siz = size_smallint;
    promise->env = Environment_P;
    promise->dct = Dictionary_P;
    THREAD_SET_RESULT(promise); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- eval ----------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(eval_M);
THREAD_ID(eval_T_M);

THREAD_BEGIN_CHUNK(Eva);
THREAD_END_CHUNK(Eva);

/*--------------------------------------------------------------------------------------*/

static VOI_type expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

static VOI_type evaluate_M(EXP_type Expression)
  { UNS_type size;
    Expression_P = Compile_M(Expression);
    size = Dictionary_Get_Frame_Size();
    Environment_Adjust_Current_Frame_Size_M(size);
    expression_M(Expression_P); }

static VOI_type evaluate_T_M(EXP_type Expression)
  { UNS_type size;
    Expression_P = Compile_T_M(Expression);
    size = Dictionary_Get_Frame_Size();
    Environment_Adjust_Current_Frame_Size_M(size);
    expression_M(Expression_P); }

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(eval_M)
  { THREAD_GET_RESULT(Expression_P, EXP);
    THREAD_ZAP;
    evaluate_M(Expression_P); }

THREAD_CONTINUATION(eval_T_M)
  { THREAD_GET_RESULT(Expression_P, EXP);
    THREAD_ZAP;
    evaluate_T_M(Expression_P); }

/*--------------------------------------------------------------------------------------*/

static VOI_type eval_M(EVA_type Evaluate)
  { Eva_type eval_M_thread;
    Expression_P = Evaluate->exp;
    THREAD_PUSH_M(eval_M, Eva);
    expression_M(Expression_P); }

static VOI_type eval_T_M(EVA_type Evaluate)
  { Eva_type eval_T_M_thread;
    Expression_P = Evaluate->exp;
    THREAD_PUSH_M(eval_T_M, Eva);
    expression_M(Expression_P); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_eval(VOI_type)
  { THREAD_REGISTER(eval_M);
    THREAD_REGISTER(eval_T_M); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- global --------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static VOI_type   apply_M(EXP_type,
                          SEQ_type);
static VOI_type apply_T_M(EXP_type,
                          SEQ_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type global_variable_value(GLB_type Global)
  { EXP_type value_expression;
    SYM_type symbol;
    UNS_type offset;
    UNS_type Scope;
    Scope  = Global->scp;
    offset = Global->ofs;
    value_expression = Environment_Global_Get(Scope,
                                              offset);
    if (is_UDF(value_expression))
      { symbol = Dictionary_Global_Variable(Scope,
                                            offset);
        symbol_error(UDF_error,
                     symbol); }
    return value_expression; }

/*--------------------------------------------------------------------------------------*/

static VOI_type global_application_M(GLA_type Global_application)
  { EXP_type value_expression;
    GLB_type global;
    SEQ_type sequence;
    global   = Global_application->glb;
    sequence = Global_application->arg;
    value_expression = global_variable_value(global);
    apply_M(value_expression,
            sequence); }

static VOI_type global_application_T_M(GLA_type Global_application)
  { EXP_type value_expression;
    GLB_type global;
    SEQ_type sequence;
    global   = Global_application->glb;
    sequence = Global_application->arg;
    value_expression = global_variable_value(global);
    apply_T_M(value_expression,
              sequence); }

static VOI_type global_variable(GLB_type Global)
  { EXP_type value_expression;
    value_expression = global_variable_value(Global);
    THREAD_SET_RESULT(value_expression); }

/*--------------------------------------------------------------------------------------*/
/*----------------------------------------- if -----------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(if_double_M);
THREAD_ID(if_single_M);

THREAD_BEGIN_CHUNK(Ifd);
  THREAD_FRAME_SLOT(EXP, cns);
  THREAD_FRAME_SLOT(EXP, alt);
THREAD_END_CHUNK(Ifd);

THREAD_BEGIN_CHUNK(Ifs);
  THREAD_FRAME_SLOT(EXP, cns);
THREAD_END_CHUNK(Ifs);

/*--------------------------------------------------------------------------------------*/

static VOI_type expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(if_double_M)
  { Ifd_type if_double_M_thread;
    EXP_type expression,
             value_expression;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(if_double_M, Ifd);
    if (is_FLS(value_expression))
      expression = if_double_M_thread->alt;
    else
      expression = if_double_M_thread->cns;
    THREAD_ZAP;
    expression_M(expression); }

THREAD_CONTINUATION(if_single_M)
  { Ifs_type if_single_M_thread;
    EXP_type expression,
             value_expression;
    THREAD_GET_RESULT(value_expression, EXP);
    if (is_FLS(value_expression))
      { THREAD_ZAP;
        THREAD_SET_RESULT(Unique_Unspecified); }
    else
      { THREAD_PEEK(if_single_M, Ifs);
        expression = if_single_M_thread->cns;
        THREAD_ZAP;
        expression_M(expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type if_double_M(IFD_type If_double)
  { Ifd_type if_double_M_thread;
    Predicate_Expression_P  = If_double->prd;
    Consequent_Expression_P = If_double->cns;
    Alternate_Expression_P  = If_double->alt;
    THREAD_PUSH_M(if_double_M, Ifd);
    if_double_M_thread->cns = Consequent_Expression_P;
    if_double_M_thread->alt = Alternate_Expression_P;
    expression_M(Predicate_Expression_P); }

static VOI_type if_single_M(IFS_type If_single)
  { Ifs_type if_single_M_thread;
    Predicate_Expression_P  = If_single->prd;
    Consequent_Expression_P = If_single->cns;
    THREAD_PUSH_M(if_single_M, Ifs);
    if_single_M_thread->cns = Consequent_Expression_P;
    expression_M(Predicate_Expression_P); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_if(VOI_type)
  { THREAD_REGISTER(if_double_M);
    THREAD_REGISTER(if_single_M); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- lambda --------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static VOI_type lambda_M(LMB_type Lambda)
  { PRC_type procedure;
    SMI_type count_smallint;
    SMI_type size_smallint;
    SYM_type symbol;
    count_smallint  = Lambda->par;
    size_smallint   = Lambda->siz;
    Expression_P    = Lambda->exp;
    Dictionary_P    = Lambda->dct;
    Environment_P = Environment_Extend_M();
    procedure = make_PRC_M();
    symbol = Pool_Keyword_Symbol(Pool_Anonymous);
    procedure->sym = symbol;
    procedure->par = count_smallint;
    procedure->exp = Expression_P;
    procedure->siz = size_smallint;
    procedure->env = Environment_P;
    procedure->dct = Dictionary_P;
    THREAD_SET_RESULT(procedure); }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------ lambda vararg -----------------------------------*/
/*--------------------------------------------------------------------------------------*/

static VOI_type lambda_vararg_M(LMV_type Lambda_vararg)
  { PRV_type procedure_vararg;
    SMI_type count_smallint;
    SMI_type size_smallint;
    SYM_type symbol;
    count_smallint  = Lambda_vararg->par;
    size_smallint   = Lambda_vararg->siz;
    Expression_P    = Lambda_vararg->exp;
    Dictionary_P    = Lambda_vararg->dct;
    Environment_P = Environment_Extend_M();
    procedure_vararg = make_PRV_M();
    symbol = Pool_Keyword_Symbol(Pool_Anonymous);
    procedure_vararg->sym = symbol;
    procedure_vararg->par = count_smallint;
    procedure_vararg->exp = Expression_P;
    procedure_vararg->siz = size_smallint;
    procedure_vararg->env = Environment_P;
    procedure_vararg->dct = Dictionary_P;
    THREAD_SET_RESULT(procedure_vararg); }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------------ let ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(let_M);

THREAD_BEGIN_CHUNK(Let);
  THREAD_FRAME_SLOT(SEQ, seq);
  THREAD_FRAME_SLOT(EXP, exp);
  THREAD_FRAME_SLOT(FRM, frm);
  THREAD_FRAME_SLOT(SMI, pos);
THREAD_END_CHUNK(Let);

/*--------------------------------------------------------------------------------------*/

static UNS_type     direct_expressions(UNS_type,
                                       UNS_type,
                                       FRM_type,
                                       SEQ_type);
static VOI_type           expression_M(EXP_type);
static VOI_type push_current_context_M(VOI_type);
static VOI_type        replace_context(DCT_type,
                                       ENV_type,
                                       FRM_type);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(let_M)
  { Let_type let_M_thread;
    EXP_type expression,
             value_expression;
    FRM_type frame;
    SMI_type position_smallint;
    UNS_type position;
    UNS_type size;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(let_M, Let);
    Sequence_P        = let_M_thread->seq;
    expression        = let_M_thread->exp;
    frame             = let_M_thread->frm;
    position_smallint = let_M_thread->pos;
    size = size_SEQ(Sequence_P);
    position = get_SMU(position_smallint);
    frame[position] = value_expression;
    position += 1;
    position = direct_expressions(size,
                                  position,
                                  frame,
                                  Sequence_P);
    if (position == 0)
      { THREAD_ZAP;
        expression_M(expression); }
    else
      { position_smallint = make_SMU(position);
        THREAD_KEEP_M(let_M, Let);
        let_M_thread->pos = position_smallint;
        expression = Sequence_P[position];
        expression_M(expression); }}
 
/*--------------------------------------------------------------------------------------*/

static VOI_type let_M(LET_type Let)
  { Let_type let_M_thread;
    ENV_type environment;
    EXP_type expression;
    SMI_type position_smallint;
    SMI_type size_smallint;
    UNS_type position;
    UNS_type size;
    size_smallint = Let->siz;
    Sequence_P    = Let->seq;
    Expression_P  = Let->exp;
    Dictionary_P  = Let->dct;
    push_current_context_M();
    size = get_SMU(size_smallint);
    Frame_P = Environment_Make_Frame_M(size);
    environment = Environment_Extend_M();
    replace_context(Dictionary_P,
                    environment,
                    Frame_P);
    size = size_SEQ(Sequence_P);
    position = direct_expressions(size,
                                  1,
                                  Frame_P,
                                  Sequence_P);
    if (position == 0)
      expression_M(Expression_P);
    else
      { position_smallint = make_SMU(position);
        THREAD_PUSH_M(let_M, Let);
        let_M_thread->seq = Sequence_P;
        let_M_thread->exp = Expression_P;
        let_M_thread->frm = Frame_P;
        let_M_thread->pos = position_smallint;
        expression = Sequence_P[position];
        expression_M(expression); }}

static VOI_type let_T_M(LET_type Let)
  { Let_type let_M_thread;
    ENV_type environment;
    EXP_type expression;
    SMI_type position_smallint;
    SMI_type size_smallint;
    UNS_type position;
    UNS_type size;
    size_smallint = Let->siz;
    Sequence_P    = Let->seq;
    Expression_P  = Let->exp;
    Dictionary_P  = Let->dct;
    size = get_SMU(size_smallint);
    Frame_P = Environment_Make_Frame_M(size);
    environment = Environment_Extend_M();
    replace_context(Dictionary_P,
                    environment,
                    Frame_P);
    size = size_SEQ(Sequence_P);
    position = direct_expressions(size,
                                  1,
                                  Frame_P,
                                  Sequence_P);
    if (position == 0)
      expression_M(Expression_P);
    else
      { position_smallint = make_SMU(position);
        THREAD_PUSH_M(let_M, Let);
        let_M_thread->seq = Sequence_P;
        let_M_thread->exp = Expression_P;
        let_M_thread->frm = Frame_P;
        let_M_thread->pos = position_smallint;
        expression = Sequence_P[position];
        expression_M(expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_let(VOI_type)
  { THREAD_REGISTER(let_M); }

/*------------------------------------ literal value -----------------------------------*/

static VOI_type literal(EXP_type Expression)
  { THREAD_SET_RESULT(Expression); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- local ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static VOI_type   apply_M(EXP_type,
                          SEQ_type);
static VOI_type apply_T_M(EXP_type,
                          SEQ_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type local_variable_value(LCL_type Local)
  { EXP_type value_expression;
    SYM_type symbol;
    UNS_type offset;
    offset = Local->ofs;
    value_expression = Environment_Local_Get(offset);
    if (is_UDF(value_expression))
      { symbol = Dictionary_Local_Variable(offset);
        symbol_error(UDF_error,
                     symbol); }
    return value_expression; }

/*--------------------------------------------------------------------------------------*/

static VOI_type local_application_M(LCA_type Local_application)
  { EXP_type value_expression;
    LCL_type local;
    SEQ_type sequence;
    local    = Local_application->lcl;
    sequence = Local_application->arg;
    value_expression = local_variable_value(local);
    apply_M(value_expression,
            sequence); }

static VOI_type local_application_T_M(LCA_type Local_application)
  { EXP_type value_expression;
    LCL_type local;
    SEQ_type sequence;
    local    = Local_application->lcl;
    sequence = Local_application->arg;
    value_expression = local_variable_value(local);
    apply_T_M(value_expression,
              sequence); }

static VOI_type local_variable(LCL_type Local)
  { EXP_type value_expression;
    value_expression = local_variable_value(Local);
    THREAD_SET_RESULT(value_expression); }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------------ or ----------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(or_M);

THREAD_BEGIN_CHUNK(Orr);
  THREAD_FRAME_SLOT(SEQ, prd);
  THREAD_FRAME_SLOT(SMI, pos);
THREAD_END_CHUNK(Orr);

/*--------------------------------------------------------------------------------------*/

static VOI_type expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(or_M)
  { Orr_type or_M_thread;
    EXP_type expression,
             value_expression;
    SMI_type position_smallint;
    UNS_type position;
    UNS_type size;
    THREAD_GET_RESULT(value_expression, EXP);
    if (!is_FLS(value_expression))
      { THREAD_ZAP;
        THREAD_SET_RESULT(Unique_True); }
    else
      { THREAD_PEEK(or_M, Orr);
        Sequence_P        = or_M_thread->prd;
        position_smallint = or_M_thread->pos;
        size = size_SEQ(Sequence_P);
        position = get_SMU(position_smallint);
        if (position == size)
          { THREAD_ZAP;
            THREAD_SET_RESULT(Unique_False); }
        else
          { position += 1;
            position_smallint = make_SMU(position);
            THREAD_KEEP_M(or_M, Orr);
            or_M_thread->pos = position_smallint;
            expression = Sequence_P[position];
            expression_M(expression); }}}

/*--------------------------------------------------------------------------------------*/

static VOI_type or_M(ORR_type Or)
  { Orr_type or_M_thread;
    EXP_type expression;
    UNS_type size;
    Sequence_P = Or->prd;
    size = size_SEQ(Sequence_P);
    if (size == 0)
      THREAD_SET_RESULT(Unique_False);
    else
      { THREAD_PUSH_M(or_M, Orr);
        or_M_thread->prd = Sequence_P;
        or_M_thread->pos = Unique_One_Smallint;
        expression = Sequence_P[1];
        expression_M(expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_or(VOI_type)
  { THREAD_REGISTER(or_M); }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------- quasiquote -------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(quasi_pair_car_M);
THREAD_ID(quasi_pair_cdr_M);
THREAD_ID(quasi_vector_M);
THREAD_ID(unquote_car_splice_M);
THREAD_ID(unquote_cdr_splice_M);

THREAD_BEGIN_CHUNK(Qpr);
  THREAD_FRAME_SLOT(EXP, exp);
THREAD_END_CHUNK(Qpr);

THREAD_BEGIN_CHUNK(Qvc);
  THREAD_FRAME_SLOT(VEC, vec);
  THREAD_FRAME_SLOT(VEC, qvc);
  THREAD_FRAME_SLOT(SMI, pos);
THREAD_END_CHUNK(Qvc);

/*--------------------------------------------------------------------------------------*/

static VOI_type         evaluate_M(EXP_type);
static VOI_type       evaluate_T_M(EXP_type);

static STR_type     clone_string_M(STR_type);
static VOI_type quasi_expression_M(EXP_type);
static VOI_type      quasisymbol_M(QSY_type);

/*--------------------------------------------------------------------------------------*/

static VOI_type quasi_literal(EXP_type Expression)
  { THREAD_SET_RESULT(Expression); }

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(quasi_pair_car_M)
  { Qpr_type quasi_pair_car_M_thread,
             quasi_pair_cdr_M_thread;
    THREAD_GET_RESULT(Car_Expression_P, EXP);
    THREAD_PEEK(quasi_pair_car_M, Qpr);
    Cdr_Expression_P = quasi_pair_car_M_thread->exp;
    THREAD_PATCH_M(quasi_pair_cdr_M, Qpr);
    quasi_pair_cdr_M_thread->exp = Car_Expression_P;
    quasi_expression_M(Cdr_Expression_P); }

THREAD_CONTINUATION(quasi_pair_cdr_M)
  { Qpr_type quasi_pair_cdr_M_thread;
    EXP_type car_expression,
             cdr_expression;
    PAI_type pair;
    THREAD_GET_RESULT(cdr_expression, EXP);
    THREAD_PEEK(quasi_pair_cdr_M, Qpr);
    car_expression = quasi_pair_cdr_M_thread->exp;
    THREAD_ZAP;
    pair = pair_M(car_expression,
                  cdr_expression);
    THREAD_SET_RESULT(pair); }

/*--------------------------------------------------------------------------------------*/

static VOI_type quasi_pair_M(PAI_type Quasi_pair)
  { Qpr_type quasi_pair_car_M_thread;
    Car_Expression_P = Quasi_pair->car;
    Cdr_Expression_P = Quasi_pair->cdr;
    THREAD_PUSH_M(quasi_pair_car_M, Qpr);
    quasi_pair_car_M_thread->exp = Cdr_Expression_P;
    quasi_expression_M(Car_Expression_P); }

/*--------------------------------------------------------------------------------------*/

static VOI_type quasi_quasiquote_M(EXP_type Expression)
  { quasi_expression_M(Expression_P); }

/*--------------------------------------------------------------------------------------*/

static VOI_type quasi_string_M(STR_type String)
  { STR_type string;
    string = clone_string_M(String);
    THREAD_SET_RESULT(string); }

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(quasi_vector_M)
  { Qvc_type quasi_vector_M_thread;
    EXP_type expression;
    SMI_type position_smallint;
    UNS_type position;
    UNS_type size;
    VEC_type quasi_vector;
    THREAD_GET_RESULT(expression, EXP);
    THREAD_PEEK(quasi_vector_M, Qvc);
    Vector_P          = quasi_vector_M_thread->vec;
    quasi_vector      = quasi_vector_M_thread->qvc;
    position_smallint = quasi_vector_M_thread->pos;
    position = get_SMU(position_smallint);
    quasi_vector[position] = expression;
    size = size_VEC(Vector_P);
    if (position == size)
      { THREAD_ZAP;
        THREAD_SET_RESULT(quasi_vector); }
    else
      { position += 1;
        position_smallint = make_SMU(position);
        THREAD_KEEP_M(quasi_vector_M, Qvc);
        quasi_vector_M_thread->pos = position_smallint;
        expression = Vector_P[position];
        quasi_expression_M(expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type quasi_vector_M(VEC_type Vector)
  { Qvc_type quasi_vector_M_thread;
    EXP_type expression;
    UNS_type size;
    size = size_SEQ(Vector);
    if (size == 0)
      THREAD_SET_RESULT(Unique_Empty_Vector_P);
    else
      { Vector_P = Vector;
        Quasi_Vector_P = make_VEC_M(size);
        THREAD_PUSH_M(quasi_vector_M, Qvc);
        quasi_vector_M_thread->vec = Vector_P;
        quasi_vector_M_thread->qvc = Quasi_Vector_P;
        quasi_vector_M_thread->pos = Unique_One_Smallint;
        expression = Vector_P[1];
        quasi_expression_M(expression); }}

/*--------------------------------------- unquote --------------------------------------*/

static VOI_type unquote_M(UNQ_type Unquote)
  { EXP_type expression;
    expression = Unquote->exp;
    evaluate_M(expression); }

static VOI_type unquote_T_M(UNQ_type Unquote)
  { EXP_type expression;
    expression = Unquote->exp;
    evaluate_T_M(expression); }

/*--------------------------------- unquote-list-splice --------------------------------*/

THREAD_CONTINUATION(unquote_car_splice_M)
  { Qpr_type unquote_car_splice_M_thread,
             unquote_cdr_splice_M_thread;
    THREAD_GET_RESULT(List_Expression_P, EXP);
    THREAD_PEEK(unquote_car_splice_M, Qpr);
    Expression_P = unquote_car_splice_M_thread->exp;
    THREAD_PATCH_M(unquote_cdr_splice_M, Qpr);
    unquote_cdr_splice_M_thread->exp = List_Expression_P;
    quasi_expression_M(Expression_P); }

THREAD_CONTINUATION(unquote_cdr_splice_M)
  { Qpr_type unquote_cdr_splice_M_thread;
    EXP_type expression,
             list_expression;
    PAI_type pair;
    TAG_type tag;
    THREAD_GET_RESULT(expression, EXP);
    THREAD_PEEK(unquote_cdr_splice_M, Qpr);
    list_expression = unquote_cdr_splice_M_thread->exp;
    THREAD_ZAP;
    tag = Grammar_Tag(list_expression);
    switch (tag)
      { case NUL_tag:
          THREAD_SET_RESULT(expression);
          return;
        case PAI_tag:
          THREAD_SET_RESULT(list_expression);
          do
            { pair = list_expression;
              list_expression = pair->cdr; }
          while (is_PAI(list_expression));
          if (is_NUL(list_expression))
            { pair->cdr = expression;
              return; }
        default:
          keyword_error(NAL_error,
                        Pool_Unquote_Splicing); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type unquote_list_splice_M(ULS_type Unquote_list_splice)
  { Qpr_type unquote_car_splice_M_thread;
    Expression_P = Unquote_list_splice->exp;
    THREAD_PATCH_M(unquote_car_splice_M, Qpr);
    evaluate_M(Expression_P); }

/*-------------------------------- unquote-vector-splice -------------------------------*/

static VOI_type unquote_vector_splice(UVS_type Unquote_vector_splice)
  { keyword_error(NYI_error,
                  Pool_Unquote_Splicing); }

/*------------------------------------- quasi expression -------------------------------*/

static VOI_type quasi_expression_M(EXP_type Expression)
  { TAG_type tag;
    tag = Grammar_Tag(Expression);
    switch (tag)
      { case PAI_tag:
          quasi_pair_M(Expression);
          return;
        case QQU_tag:
          quasi_quasiquote_M(Expression);
          return;
        case QSY_tag:
          quasisymbol_M(Expression);
          return;
        case STR_tag:
          quasi_string_M(Expression);
          return;
        case ULS_tag:
          unquote_list_splice_M(Expression);
          return;
        case UNQ_tag:
          unquote_M(Expression);
          return;
        case UNq_tag:
          unquote_T_M(Expression);
          return;
        case UVS_tag:
          unquote_vector_splice(Expression);
          return;
        case VEC_tag:
          quasi_vector_M(Expression);
          return;
        case CHA_tag:
        case FLS_tag:
        case NUL_tag:
        case REA_tag:
        case SMI_tag:
        case SYM_tag:
        case TRU_tag:
          quasi_literal(Expression);
          return;
        default:
          keyword_error(IXT_error,
                        Pool_Quasiquote); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type quasiquote_M(QQU_type Quasiquote)
  { EXP_type expression;
    expression = Quasiquote->exp;
    quasi_expression_M(expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_quasiquote(VOI_type)
  { THREAD_REGISTER(quasi_pair_car_M);
    THREAD_REGISTER(quasi_pair_cdr_M);
    THREAD_REGISTER(quasi_vector_M);
    THREAD_REGISTER(unquote_car_splice_M);
    THREAD_REGISTER(unquote_cdr_splice_M); }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------- quasisymbol ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static VOI_type quasisymbol_M(QSY_type Quasisymbol)
  { SMI_type count_smallint;
    SYM_type symbol;
    UNS_type count;
    symbol         = Quasisymbol->sym;
    count_smallint = Quasisymbol->ctr;
    count = get_SMU(count_smallint);
    symbol = Pool_Generate_M(symbol,
                             count);
    THREAD_SET_RESULT(symbol); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- quote ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static EXP_type clone_expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

static VOI_type quote_M(QUO_type Quote)
  { EXP_type expression;
    expression = Quote->exp;
    expression = clone_expression_M(expression);
    THREAD_SET_RESULT(expression); }

/*--------------------------------------------------------------------------------------*/
/*--------------------------------------- sequence -------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(sequence_M);

THREAD_BEGIN_CHUNK(Seq);
  THREAD_FRAME_SLOT(SEQ, seq);
  THREAD_FRAME_SLOT(SMI, pos);
THREAD_END_CHUNK(Seq);

/*--------------------------------------------------------------------------------------*/

static VOI_type expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(sequence_M)
  { Seq_type sequence_M_thread;
    EXP_type expression;
    SMI_type position_smallint;
    UNS_type position,
             size;
    THREAD_PEEK(sequence_M, Seq);
    Sequence_P        = sequence_M_thread->seq;
    position_smallint = sequence_M_thread->pos;
    position = get_SMU(position_smallint);
    size = size_SEQ(Sequence_P);
    position += 1;
    if (position < size)
      { position_smallint = make_SMU(position);
        THREAD_KEEP_M(sequence_M, Seq);
        sequence_M_thread->pos = position_smallint; }
    else
      THREAD_ZAP;
    expression = Sequence_P[position];
    expression_M(expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type sequence_M(SEQ_type Sequence)
  { Seq_type sequence_M_thread;
    EXP_type expression;
    UNS_type size;
    Sequence_P = Sequence;
    size = size_SEQ(Sequence_P);
    if (size > 1)
      { THREAD_PUSH_M(sequence_M, Seq);
        sequence_M_thread->seq = Sequence_P;
        sequence_M_thread->pos = Unique_One_Smallint; }
    expression = Sequence_P[1];
    expression_M(expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_sequence(VOI_type)
  { THREAD_REGISTER(sequence_M); }

/*--------------------------------------------------------------------------------------*/
/*-------------------------------------- set global ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(set_global);

THREAD_BEGIN_CHUNK(Stg);
  THREAD_FRAME_SLOT(SMI, scp);
  THREAD_FRAME_SLOT(SMI, ofs);
THREAD_END_CHUNK(Stg);

/*--------------------------------------------------------------------------------------*/

static EXP_type direct_expression(EXP_type);
static VOI_type      expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(set_global)
  { Stg_type set_global_thread;
    EXP_type value_expression;
    SMI_type offset_smallint;
    SMI_type scope_smallint;
    UNS_type offset;
    UNS_type scope;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(set_global, Stg);
    scope_smallint  = set_global_thread->scp;
    offset_smallint = set_global_thread->ofs;
    THREAD_ZAP;
    scope = get_SMU(scope_smallint);
    offset = get_SMU(offset_smallint);
    Environment_Global_Set(value_expression,
                           scope,
                           offset);
    THREAD_SET_RESULT(value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type set_global_M(STG_type Set_global)
  { Stg_type set_global_thread;
    EXP_type value_expression;
    SMI_type offset_smallint;
    SMI_type scope_smallint;
    UNS_type offset;
    UNS_type scope;
    scope_smallint  = Set_global->scp;
    offset_smallint = Set_global->ofs;
    Expression_P    = Set_global->exp;
    value_expression = direct_expression(Expression_P);
    if (is_UDF(value_expression))
      { THREAD_PUSH_M(set_global, Stg);
        set_global_thread->scp = scope_smallint;
        set_global_thread->ofs = offset_smallint;
        expression_M(Expression_P); }
    else
      { scope = get_SMU(scope_smallint);
        offset = get_SMU(offset_smallint);
        Environment_Global_Set(value_expression,
                               scope,
                               offset);
        THREAD_SET_RESULT(value_expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_set_global(VOI_type)
  { THREAD_REGISTER(set_global); }

/*--------------------------------------------------------------------------------------*/
/*-------------------------------------- set local -------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(set_local);

THREAD_BEGIN_CHUNK(Stl);
  THREAD_FRAME_SLOT(SMI, ofs);
THREAD_END_CHUNK(Stl);

/*--------------------------------------------------------------------------------------*/

static EXP_type direct_expression(EXP_type);
static VOI_type      expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(set_local)
  { Stl_type set_local_thread;
    EXP_type value_expression;
    SMI_type offset_smallint;
    UNS_type offset;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(set_local, Stl);
    offset_smallint = set_local_thread->ofs;
    THREAD_ZAP;
    offset = get_SMU(offset_smallint);
    Environment_Local_Set(value_expression,
                          offset);
    THREAD_SET_RESULT(value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type set_local_M(STL_type Set_local)
  { Stl_type set_local_thread;
    EXP_type value_expression;
    SMI_type offset_smallint;
    UNS_type offset;
    offset_smallint = Set_local->ofs;
    Expression_P    = Set_local->exp;
    value_expression = direct_expression(Expression_P);
    if (is_UDF(value_expression))
      { THREAD_PUSH_M(set_local, Stl);
        set_local_thread->ofs = offset_smallint;
        expression_M(Expression_P); }
    else
      { offset = get_SMU(offset_smallint);
        Environment_Local_Set(value_expression,
                              offset);
        THREAD_SET_RESULT(value_expression); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_set_local(VOI_type)
  { THREAD_REGISTER(set_local); }

/*--------------------------------------------------------------------------------------*/
/*--------------------------------------- string ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static STR_type clone_string_M(STR_type);

/*--------------------------------------------------------------------------------------*/

static VOI_type string_M(STR_type String)
  { STR_type string;
    string = clone_string_M(String);
    THREAD_SET_RESULT(string);}

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- thunk ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static VOI_type           expression_M(EXP_type);
static VOI_type push_current_context_M(VOI_type);
static VOI_type        replace_context(DCT_type,
                                       ENV_type,
                                       FRM_type);

/*--------------------------------------------------------------------------------------*/

static VOI_type thunk_M(THK_type Thunk)
  { SMI_type size_smallint;
    ENV_type environment;
    UNS_type size;
    size_smallint = Thunk->siz;
    Expression_P  = Thunk->exp;
    Dictionary_P  = Thunk->dct;
    push_current_context_M();
    size = get_SMU(size_smallint);
    Frame_P = Environment_Make_Frame_M(size);
    environment = Environment_Extend_M();
    replace_context(Dictionary_P,
                    environment,
                    Frame_P);
    expression_M(Expression_P); }

static VOI_type thunk_T_M(THK_type Thunk)
  { SMI_type size_smallint;
    ENV_type environment;
    UNS_type size;
    size_smallint = Thunk->siz;
    Expression_P  = Thunk->exp;
    Dictionary_P  = Thunk->dct;
    size = get_SMU(size_smallint);
    Frame_P = Environment_Make_Frame_M(size);
    environment = Environment_Extend_M();
    replace_context(Dictionary_P,
                    environment,
                    Frame_P);
    expression_M(Expression_P); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- vector --------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static VEC_type clone_vector_M(VEC_type);

/*--------------------------------------------------------------------------------------*/

static VOI_type vector_M(VEC_type Vector)
  { VEC_type vector;
    vector = clone_vector_M(Vector);
    THREAD_SET_RESULT(vector); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- while ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(while_body_M);
THREAD_ID(while_predicate_M);

THREAD_BEGIN_CHUNK(Whi);
  THREAD_FRAME_SLOT(EXP, prd);
  THREAD_FRAME_SLOT(EXP, exp);
  THREAD_FRAME_SLOT(EXP, res);
THREAD_END_CHUNK(Whi);

/*--------------------------------------------------------------------------------------*/

static VOI_type expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(while_body_M)
  { Whi_type while_body_M_thread,
             while_predicate_M_thread;
    THREAD_GET_RESULT(Expression_P, EXP);
    THREAD_PEEK(while_body_M, Whi);
    Predicate_Expression_P = while_body_M_thread->prd;
    THREAD_PATCH_M(while_predicate_M, Whi);
    while_predicate_M_thread->res = Expression_P;
    expression_M(Predicate_Expression_P); }

THREAD_CONTINUATION(while_predicate_M)
  { Whi_type while_body_M_thread,
             while_predicate_M_thread;
    EXP_type result_expression,
             value_expression;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(while_predicate_M, Whi);
    Body_Expression_P = while_predicate_M_thread->exp;
    result_expression = while_predicate_M_thread->res;
    if (is_FLS(value_expression))
      { THREAD_ZAP;
        THREAD_SET_RESULT(result_expression); }
    else
      { THREAD_PATCH_M(while_body_M, Whi);
        expression_M(Body_Expression_P); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type while_M(WHI_type While_T)
  { Whi_type while_predicate_M_thread;
    Predicate_Expression_P = While_T->prd;
    Body_Expression_P      = While_T->exp;
    THREAD_PUSH_M(while_predicate_M, Whi);
    while_predicate_M_thread->prd = Predicate_Expression_P;
    while_predicate_M_thread->exp = Body_Expression_P;
    while_predicate_M_thread->res = Unique_Unspecified;
    expression_M(Predicate_Expression_P); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_while(VOI_type)
  { THREAD_REGISTER(while_body_M);
    THREAD_REGISTER(while_predicate_M); }

/*--------------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------------*/
/*-------------------------------------- expression ------------------------------------*/
/*--------------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static VOI_type                    and_M(AND_type);
static VOI_type            application_M(APL_type);
static VOI_type          application_T_M(APL_type);
static VOI_type                  begin_M(BEG_type);
static VOI_type                   cond_M(CND_type);
static VOI_type              cond_else_M(CNE_type);
static VOI_type            cons_stream_M(CNS_type);
static VOI_type                  delay_M(DEL_type);
static VOI_type        define_function_M(DFF_type);
static VOI_type define_function_vararg_M(DFZ_type);
static VOI_type        define_variable_M(DFV_type);
static VOI_type                   eval_M(EVA_type);
static VOI_type                 eval_T_M(EVA_type);
static VOI_type     global_application_M(GLA_type);
static VOI_type   global_application_T_M(GLA_type);
static VOI_type          global_variable(GLB_type);
static VOI_type              if_double_M(IFD_type);
static VOI_type              if_single_M(IFS_type);
static VOI_type                  literal(EXP_type);
static VOI_type                 lambda_M(LMB_type);
static VOI_type          lambda_vararg_M(LMV_type);
static VOI_type      local_application_M(LCA_type);
static VOI_type    local_application_T_M(LCA_type);
static VOI_type           local_variable(LCL_type);
static VOI_type                    let_M(LET_type);
static VOI_type                  let_T_M(LET_type);
static VOI_type                     or_M(ORR_type);
static VOI_type             quasiquote_M(QQU_type);
static VOI_type            quasisymbol_M(QSY_type);
static VOI_type                  quote_M(QUO_type);
static VOI_type               sequence_M(SEQ_type);
static VOI_type             set_global_M(STG_type);
static VOI_type              set_local_M(STL_type);
static VOI_type                 string_M(STR_type);
static VOI_type                  thunk_M(THK_type);
static VOI_type                thunk_T_M(THK_type);
static VOI_type                 vector_M(VEC_type);
static VOI_type                  while_M(WHI_type);

static EXP_type    global_variable_value(GLB_type);
static EXP_type     local_variable_value(LCL_type);

static VOI_type   push_current_context_M(VOI_type);
static VOI_type          replace_context(DCT_type,
                                         ENV_type,
                                         FRM_type);

/*--------------------------------------------------------------------------------------*/

static EXP_type direct_expression(EXP_type Expression)
  { GLB_type global;
    LCL_type local;
    TAG_type tag;
    tag = Grammar_Tag(Expression);
    switch (tag)
      { case GLB_tag:
          global = (GLB_type)Expression;
          return global_variable_value(global);
        case LCL_tag:
          local = (LCL_type)Expression;
          return local_variable_value(local);
        case CHA_tag:
        case FLS_tag:
        case NUL_tag:
        case PAI_tag:
        case REA_tag:
        case SMI_tag:
        case STR_tag:
        case SYM_tag:
        case TRU_tag:
        case VEC_tag:
          return Expression;
        default:
          return Unique_Undefined; }}

static UNS_type direct_expressions(UNS_type Size,
                                   UNS_type Position,
                                   FRM_type Frame,
                                   SEQ_type Sequence)
  { EXP_type expression;
    UNS_type position;
    for (position = Position;
         position <= Size;
         position++)
      { expression = Sequence[position];
        expression = direct_expression(expression);
        if (is_UDF(expression))
          return position;
        Frame[position] = expression; }
    return 0; }

/*--------------------------------------------------------------------------------------*/


static VOI_type expression_M(EXP_type Expression)
  { TAG_type tag;
    tag = Grammar_Tag(Expression);
    switch (tag)
      { case AND_tag:
          and_M(Expression);
          return;
        case APL_tag:
          application_M(Expression);
          return;
        case APl_tag:
          application_T_M(Expression);
          return;
        case BEG_tag:
          begin_M(Expression);
          return;
        case CND_tag:
          cond_M(Expression);
          return;
        case CNE_tag:
          cond_else_M(Expression);
          return;
        case CNS_tag:
          cons_stream_M(Expression);
          return;
        case DEL_tag:
          delay_M(Expression);
          return;
        case DFF_tag:
          define_function_M(Expression);
          return;
        case DFV_tag:
          define_variable_M(Expression);
          return;
        case DFZ_tag:
          define_function_vararg_M(Expression);
          return;
        case EVA_tag:
          eval_M(Expression);
          return;
        case EVa_tag:
          eval_T_M(Expression);
          return;
        case GLA_tag:
          global_application_M(Expression);
          return;
        case GLa_tag:
          global_application_T_M(Expression);
          return;
        case GLB_tag:
          global_variable(Expression);
          return;
        case IFD_tag:
          if_double_M(Expression);
          return;
        case IFS_tag:
          if_single_M(Expression);
          return;
        case LCA_tag:
          local_application_M(Expression);
          return;
        case LCa_tag:
          local_application_T_M(Expression);
          return;
        case LCL_tag:
          local_variable(Expression);
          return;
        case LET_tag:
          let_M(Expression);
          return;
        case LEt_tag:
          let_T_M(Expression);
          return;
        case LMB_tag:
          lambda_M(Expression);
          return;
        case LMV_tag:
          lambda_vararg_M(Expression);
          return;
        case ORR_tag:
          or_M(Expression);
          return;
        case QQU_tag:
          quasiquote_M(Expression);
          return;
        case QSY_tag:
          quasisymbol_M(Expression);
          return;
        case QUO_tag:
          quote_M(Expression);
          return;
        case SEQ_tag:
          sequence_M(Expression);
          return;
        case STG_tag:
          set_global_M(Expression);
          return;
        case STL_tag:
          set_local_M(Expression);
          return;
        case STR_tag:
          string_M(Expression);
          return;
        case THK_tag:
          thunk_M(Expression);
          return;
        case THk_tag:
          thunk_T_M(Expression);
          return;
        case VEC_tag:
          vector_M(Expression);
          return;
        case WHI_tag:
          while_M(Expression);
          return;
        case CHA_tag:
        case FLS_tag:
        case MAC_tag:
        case NUL_tag:
        case REA_tag:
        case SMI_tag:
        case SYM_tag:
        case TRU_tag:
          literal(Expression);
          return;
        case CNT_tag:
        case END_tag:
        case FRC_tag:
        case IPT_tag:
        case NAT_tag:
        case NAZ_tag:
        case OPT_tag:
        case PRC_tag:
        case PRM_tag:
        case PRV_tag:
        case UDF_tag:
        case USP_tag:
          error(UXT_error,
                Evaluate_text);
          return;
        default:
          error(IXT_error,
                Evaluate_text); }}

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- for-each ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static const TXT_type for_each_text = "for-each";

THREAD_ID(for_each_M);
THREAD_ID(for_each_T_M);

THREAD_BEGIN_CHUNK(For);
  THREAD_FRAME_SLOT(EXP, prc);
  THREAD_FRAME_SLOT(EXP, lst);
  THREAD_FRAME_SLOT(PAI, arg);
THREAD_END_CHUNK(For);

/*--------------------------------------------------------------------------------------*/

VOI_type   direct_apply_M(EXP_type,
                          EXP_type);
VOI_type direct_apply_T_M(EXP_type,
                          EXP_type);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(for_each_M)
  { For_type for_each_M_thread;
    EXP_type list_expression,
             value_expression;
    PAI_type pair;
    TAG_type tag;
    THREAD_PEEK(for_each_M, For);
    Operator_Expression_P = for_each_M_thread->prc;
    list_expression       = for_each_M_thread->lst;
    Pair_P                = for_each_M_thread->arg;
    tag = Grammar_Tag(list_expression);
    switch (tag)
      { case PAI_tag:
          pair = list_expression;
          value_expression  = pair->car;
          List_Expression_P = pair->cdr;
          Pair_P->car = value_expression;
          if (is_NUL(List_Expression_P))
            THREAD_ZAP;
          else
            { THREAD_KEEP_M(for_each_M, For);
              for_each_M_thread->lst = List_Expression_P; }
          direct_apply_M(Operator_Expression_P,
                         Pair_P);
          return;
        default:
          error(NAL_error,
                for_each_text); }}

THREAD_CONTINUATION(for_each_T_M)
  { For_type for_each_T_M_thread;
    EXP_type list_expression,
             value_expression;
    PAI_type pair;
    TAG_type tag;
    THREAD_PEEK(for_each_T_M, For);
    Operator_Expression_P = for_each_T_M_thread->prc;
    list_expression       = for_each_T_M_thread->lst;
    Pair_P                = for_each_T_M_thread->arg;
    tag = Grammar_Tag(List_Expression_P);
    switch (tag)
      { case PAI_tag:
          pair = list_expression;
          value_expression  = pair->car;
          List_Expression_P = pair->cdr;
          Pair_P->car = value_expression;
          if (is_NUL(List_Expression_P))
            { THREAD_ZAP;
              direct_apply_T_M(Operator_Expression_P,
                               Pair_P); }
          else
            { THREAD_KEEP_M(for_each_T_M, For);
              for_each_T_M_thread->lst = List_Expression_P;
              direct_apply_M(Operator_Expression_P,
                             Pair_P); }
          return;
        default:
          error(NAL_error,
                for_each_text); }}
/*--------------------------------------------------------------------------------------*/

static VOI_type for_each_M(EXP_type Operator_expression,
                           EXP_type List_expression)
  { For_type for_each_M_thread;
    EXP_type car_expression;
    TAG_type tag;
    tag = Grammar_Tag(List_expression);
    switch (tag)
      { case NUL_tag:
          THREAD_SET_RESULT(Unique_Unspecified);
          return;
        case PAI_tag:
          unpair2(List_expression,
                  &car_expression,
                  &List_Expression_P);
          Operator_Expression_P = Operator_expression;
          Pair_P = pair_M(car_expression,
                          Unique_Null);
          if (!is_NUL(List_Expression_P))
            { THREAD_PUSH_M(for_each_M, For);
              for_each_M_thread->prc = Operator_Expression_P;
              for_each_M_thread->lst = List_Expression_P;
              for_each_M_thread->arg = Pair_P; }
          direct_apply_M(Operator_Expression_P,
                         Pair_P);
          return;
        default:
          error(NAL_error,
                for_each_text); }}

static VOI_type for_each_T_M(EXP_type Operator_expression,
                             EXP_type List_expression)
  { For_type for_each_T_M_thread;
    EXP_type car_expression;
    TAG_type tag;
    tag = Grammar_Tag(List_expression);
    switch (tag)
      { case NUL_tag:
          THREAD_SET_RESULT(Unique_Unspecified);
          return;
        case PAI_tag:
          unpair2(List_expression,
                  &car_expression,
                  &List_Expression_P);
          Operator_Expression_P = Operator_expression;
          Pair_P = pair_M(car_expression,
                          Unique_Null);
          if (is_NUL(List_Expression_P))
            direct_apply_T_M(Operator_Expression_P,
                             (EXP_type)Pair_P);
          else
            { THREAD_PUSH_M(for_each_T_M, For);
              for_each_T_M_thread->prc = Operator_Expression_P;
              for_each_T_M_thread->lst = List_Expression_P;
              for_each_T_M_thread->arg = Pair_P;
              direct_apply_M(Operator_Expression_P,
                             (EXP_type)Pair_P); }
          return;
        default:
          error(NAL_error,
                for_each_text); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_for_each(VOI_type)
  { THREAD_REGISTER(for_each_M);
    THREAD_REGISTER(for_each_T_M); }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------------ map ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static const TXT_type map_text = "map";

THREAD_ID(map_M);
THREAD_ID(map_T_M);

THREAD_BEGIN_CHUNK(Map);
  THREAD_FRAME_SLOT(EXP, prc);
  THREAD_FRAME_SLOT(EXP, lst);
  THREAD_FRAME_SLOT(PAI, arg);
  THREAD_FRAME_SLOT(EXP, res);
THREAD_END_CHUNK(Map);

/*--------------------------------------------------------------------------------------*/

VOI_type   direct_apply_M(EXP_type,
                          EXP_type);
VOI_type direct_apply_T_M(EXP_type,
                          EXP_type);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(map_M)
  { Map_type map_M_thread;
    EXP_type value_expression;
    PAI_type pair;
    TAG_type tag;
    THREAD_GET_RESULT(Car_Expression_P, EXP);
    THREAD_PEEK(map_M, Map);
    Operator_Expression_P = map_M_thread->prc;
    List_Expression_P     = map_M_thread->lst;
    Pair_P                = map_M_thread->arg;
    Cdr_Expression_P      = map_M_thread->res;
    pair = make_PAI_M();
    pair->car = Car_Expression_P;
    pair->cdr = Cdr_Expression_P;
    tag = Grammar_Tag(List_Expression_P);
    switch (tag)
      { case NUL_tag:
          THREAD_ZAP;
          reverse(&pair);
          THREAD_SET_RESULT(pair);
          return;
        case PAI_tag:
          Expression_P = pair;
          pair = List_Expression_P;
          value_expression  = pair->car;
          List_Expression_P = pair->cdr;
          Pair_P->car = value_expression;
          THREAD_KEEP_M(map_M, Map);
          map_M_thread->lst = List_Expression_P;
          map_M_thread->res = Expression_P;
          direct_apply_M(Operator_Expression_P,
                         (EXP_type)Pair_P);
          return;
        default:
          error(NAL_error,
                map_text); }}

THREAD_CONTINUATION(map_T_M)
  { Map_type map_T_M_thread;
    EXP_type value_expression;
    PAI_type pair;
    TAG_type tag;
    THREAD_GET_RESULT(Car_Expression_P, EXP);
    THREAD_PEEK(map_T_M, Map);
    Operator_Expression_P = map_T_M_thread->prc;
    List_Expression_P     = map_T_M_thread->lst;
    Pair_P                = map_T_M_thread->arg;
    Cdr_Expression_P      = map_T_M_thread->res;
    pair = make_PAI_M();
    pair->car = Car_Expression_P;
    pair->cdr = Cdr_Expression_P;
    tag = Grammar_Tag(List_Expression_P);
    switch (tag)
      { case NUL_tag:
          THREAD_ZAP;
          reverse(&pair);
          THREAD_SET_RESULT(pair);
          return;
        case PAI_tag:
          Expression_P = pair;
          pair = List_Expression_P;
          value_expression  = pair->car;
          List_Expression_P = pair->cdr;
          Pair_P->car = value_expression;
          THREAD_KEEP_M(map_T_M, Map);
          map_T_M_thread->lst = List_Expression_P;
          map_T_M_thread->res = Expression_P;
          if (is_NUL(List_Expression_P))
            direct_apply_T_M(Operator_Expression_P,
                             (EXP_type)Pair_P);
          else
            direct_apply_M(Operator_Expression_P,
                           (EXP_type)Pair_P);
          return;
        default:
          error(NAL_error,
                map_text); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type map_M(EXP_type Operator_expression,
                      EXP_type List_expression)
  { Map_type map_M_thread;
    PAI_type pair;
    TAG_type tag;
    tag = Grammar_Tag(List_expression);
    switch (tag)
      { case NUL_tag:
          THREAD_SET_RESULT(Unique_Null);
          return;
        case PAI_tag:
          pair = List_expression;
          List_Expression_P  = pair->cdr;
          Car_Expression_P = pair->car;
          Operator_Expression_P = Operator_expression;
          Pair_P = make_PAI_M();
          Pair_P->car = Car_Expression_P;
          Pair_P->cdr = Unique_Null;
          THREAD_PUSH_M(map_M, Map);
          map_M_thread->prc = Operator_Expression_P;
          map_M_thread->lst = List_Expression_P;
          map_M_thread->arg = Pair_P;
          map_M_thread->res = Unique_Null;
          direct_apply_M(Operator_Expression_P,
                         (EXP_type)Pair_P);
          return;
        default:
          error(NAL_error,
                map_text); }}

static VOI_type map_T_M(EXP_type Operator_expression,
                        EXP_type List_expression)
  { Map_type map_T_M_thread;
    PAI_type pair;
    TAG_type tag;
    tag = Grammar_Tag(List_expression);
    switch (tag)
      { case NUL_tag:
          THREAD_SET_RESULT(Unique_Null);
          return;
        case PAI_tag:
          pair = List_expression;
          List_Expression_P  = pair->cdr;
          Car_Expression_P = pair->car;
          Operator_Expression_P = Operator_expression;
          Pair_P = make_PAI_M();
          Pair_P->car = Car_Expression_P;
          Pair_P->cdr = Unique_Null;
          THREAD_PUSH_M(map_T_M, Map);
          map_T_M_thread->prc = Operator_Expression_P;
          map_T_M_thread->lst = List_Expression_P;
          map_T_M_thread->arg = Pair_P;
          map_T_M_thread->res = Unique_Null;
          if (is_NUL(List_Expression_P))
            direct_apply_T_M(Operator_Expression_P,
                             (EXP_type)Pair_P);
          else
            direct_apply_M(Operator_Expression_P,
                           (EXP_type)Pair_P);
          return;
        default:
          error(NAL_error,
                map_text); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_map(VOI_type)
  { THREAD_REGISTER(map_M);
    THREAD_REGISTER(map_T_M); }

/*--------------------------------------------------------------------------------------*/
/*----------------------------------------- force --------------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(force);

THREAD_BEGIN_CHUNK(Frc);
  PRM_type prm;
THREAD_END_CHUNK(Frc);

/*--------------------------------------------------------------------------------------*/

static EXP_type direct_expression(EXP_type);
static VOI_type      expression_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(force)
  { Frc_type force_thread;
    EXP_type value_expression;
    PRM_type promise;
    THREAD_GET_RESULT(value_expression, EXP);
    THREAD_PEEK(force, Frc);
    promise = force_thread->prm;
    THREAD_ZAP;
    convert_FRC(promise,
                value_expression);
    THREAD_SET_RESULT(value_expression); }

static VOI_type force_M(EXP_type Expression)
  { Frc_type force_thread;
    DCT_type dictionary;
    ENV_type environment;
    EXP_type expression,
             value_expression;
    FRC_type forced;
    FRM_type frame;
    SMI_type size_smallint;
    TAG_type tag;
    UNS_type size;
    tag = Grammar_Tag(Expression);
    switch (tag)
      { case FRC_tag:
          forced = (FRC_type)Expression;
          value_expression = forced->val;
          THREAD_SET_RESULT(value_expression);
          return;
        case PRM_tag:
          Promise_P = (PRM_type)Expression;
          THREAD_PUSH_M(force, Frc);
          force_thread->prm = Promise_P;
          push_current_context_M();
          size_smallint = Promise_P->siz;
          size = get_SMU(size_smallint);
          frame = Environment_Make_Frame_M(size);
          expression  = Promise_P->exp;
          environment = Promise_P->env;
          dictionary  = Promise_P->dct;
          replace_context(dictionary,
                          environment,
                          frame);
          expression_M(expression);
          return;
        default:
          error(NPR_error,
                Force_text); }}

static VOI_type force_T_M(EXP_type Expression)
  { Frc_type force_thread;
    DCT_type dictionary;
    ENV_type environment;
    EXP_type expression,
             value_expression;
    FRC_type forced;
    FRM_type frame;
    SMI_type size_smallint;
    TAG_type tag;
    UNS_type size;
    tag = Grammar_Tag(Expression);
    switch (tag)
      { case FRC_tag:
          forced = Expression;
          value_expression = forced->val;
          THREAD_SET_RESULT(value_expression);
          return;
        case PRM_tag:
          Promise_P = Expression;
          THREAD_PUSH_M(force, Frc);
          force_thread->prm = Promise_P;
          Environment_Release_Current_Frame();
          size_smallint = Promise_P->siz;
          size = get_SMU(size_smallint);
          frame = Environment_Make_Frame_M(size);
          expression  = Promise_P->exp;
          environment = Promise_P->env;
          dictionary  = Promise_P->dct;
          replace_context(dictionary,
                          environment,
                          frame);
          expression_M(expression);
          return;
        default:
          error(NPR_error,
                Force_text); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_force(VOI_type)
  { THREAD_REGISTER(force); }

/*--------------------------------- callback functions ---------------------------------*/

/*--------------------------------------------------------------------------------------*/
/*----------------------------------- read-eval-print ----------------------------------*/
/*--------------------------------------------------------------------------------------*/

THREAD_ID(exit);

THREAD_BEGIN_CHUNK(Exi);
THREAD_END_CHUNK(Exi);

/*--------------------------------------------------------------------------------------*/

THREAD_CONTINUATION(exit)
  { C_LONGJMP(Exit, 1); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_exit(VOI_type)
  { THREAD_REGISTER(exit); }

/*---------------------------------- public functions ----------------------------------*/

static VOI_type   direct_apply_M(EXP_type,
                                 EXP_type);
static VOI_type direct_apply_T_M(EXP_type,
                                 EXP_type);
static VOI_type     expression_M(EXP_type);
static VOI_type          force_M(EXP_type);
static VOI_type        force_T_M(EXP_type);

/*--------------------------------------------------------------------------------------*/

EXP_type Evaluate_M(EXP_type Expression)
  { Exi_type exit_thread;
    THREAD_PUSH_M(exit, Exi);
    if (C_SETJMP(Exit) == 0)
      { expression_M(Expression);
        Thread_Loop_M(); }
    THREAD_GET_RESULT(Expression_P, EXP);
    THREAD_ZAP;
    return Expression_P; }                                                        // !!!!!

EXP_type Evaluate_Macro_M(EXP_type Operator_expression,
                          EXP_type Operand_expression)
  { Exi_type exit_thread;
    THREAD_PUSH_M(exit, Exi);
    if (C_SETJMP(Exit) == 0)
      { direct_apply_M(Operator_expression,
                       Operand_expression);
        Thread_Loop_M(); }
    THREAD_GET_RESULT(Expression_P, EXP);
    THREAD_ZAP;
    return Expression_P; }                                                        // !!!!!

VOI_type Evaluate_Apply_M(EXP_type Operator_expression,
                          EXP_type Operand_expression)
  { direct_apply_M(Operator_expression,
                   Operand_expression);
    THREAD_GET_RESULT(Expression_P, EXP); }                                       // !!!!!

VOI_type Evaluate_Apply_T_M(EXP_type Operator_expression,
                            EXP_type Operand_expression)
  { direct_apply_T_M(Operator_expression,
                     Operand_expression); }

VOI_type Evaluate_Force_M(EXP_type Expression)
  { force_M(Expression); }

VOI_type Evaluate_Force_T_M(EXP_type Expression)
  { force_T_M(Expression); }

VOI_type Evaluate_For_Each_M(EXP_type Operator_expression,
                             EXP_type List_expression)
  { for_each_M(Operator_expression,
               List_expression); }

VOI_type Evaluate_For_Each_T_M(EXP_type Operator_expression,
                               EXP_type List_expression)
  { for_each_T_M(Operator_expression,
                 List_expression); }

VOI_type Evaluate_Map_M(EXP_type Operator_expression,
                        EXP_type List_expression)
  { map_M(Operator_expression,
          List_expression); }

VOI_type Evaluate_Map_T_M(EXP_type Operator_expression,
                          EXP_type List_expression)
  { map_T_M(Operator_expression,
             List_expression); }

CNT_type Evaluate_Make_Continuation_M(VOI_type)
  { return make_continuation_M(); }

VOI_type Evaluate_Return(EXP_type Expression)
  { THREAD_SET_RESULT(Expression); }

/*------------------------------------- initialize -------------------------------------*/

VOI_type Evaluate_Initialize_M(VOI_type)
  { Pool_Initialize_M();
    Stack_Initialize_M();
    Thread_Initialize_M();
    Environment_Initialize_M();
    Dictionary_Initialize_M();
    Native_Initialize_M();
    Compile_Initialize_M();

    initialize_and();
    initialize_application();
    initialize_cloning();
    initialize_cond();
    initialize_cons_stream();
    initialize_context();
    initialize_continuation();
    initialize_define_variable();
    initialize_eval();
    initialize_exit();
    initialize_force();
    initialize_for_each();
    initialize_higher_order_native();
    initialize_if();
    initialize_let();
    initialize_map();
    initialize_native();
    initialize_or();
    initialize_procedure();
    initialize_procedure_vararg();
    initialize_quasiquote();
    initialize_sequence();
    initialize_set_global();
    initialize_set_local();
    initialize_while();

    REGISTER(Alternate_Expression_P);
    REGISTER(Body_Expression_P);
    REGISTER(Body_Sequence_P);
    REGISTER(Car_Expression_P);
    REGISTER(Cdr_Expression_P);
    REGISTER(Consequent_Expression_P);
    REGISTER(Continuation_P);
    REGISTER(Dictionary_P);
    REGISTER(Else_Expression_P);
    REGISTER(Environment_P);
    REGISTER(Expression_P);
    REGISTER(Frame_P);
    REGISTER(Higher_Order_Native_P);
    REGISTER(List_Expression_P);
    REGISTER(Native_P);
    REGISTER(Operator_Expression_P);
    REGISTER(Pair_P);
    REGISTER(Predicate_Expression_P);
    REGISTER(Predicate_Sequence_P);
    REGISTER(Procedure_P);
    REGISTER(Procedure_Vararg_P);
    REGISTER(Promise_P);
    REGISTER(Quasi_Vector_P);
    REGISTER(Sequence_P);
    REGISTER(String_P);
    REGISTER(Symbol_P);
    REGISTER(Vector_P); }
