                     /*---------------------------------------------*/
                     /*                  >>>Slip<<<                 */
                     /*                 Theo D'Hondt                */
                     /*          VUB Software Languages Lab         */
                     /*                  (c) 2018                   */
                     /*---------------------------------------------*/
                     /*                final version                */
                     /*---------------------------------------------*/
                     /*                    grammar                  */
                     /*---------------------------------------------*/


#include "SlipPlatform.h"

#include "SlipMemory.h"

#include "SlipGrammar.h"

/*----------------------------------- private macros -----------------------------------*/

#define MAKE(SIGNATURE)                                                                  \
  { SIGNATURE##_type signature;                                                          \
    PTR_type pointer;                                                                    \
    pointer = Memory_Chunk_M(SIGNATURE##_tag,                                            \
                             SIGNATURE##_size);                                          \
    return (SIGNATURE##_type)pointer; }

#define CLEAR(VARIABLE)                                                                  \
  *(EXp_type)&VARIABLE = UNDEF(EXP)

#define UNDEF(SIGNATURE)                                                                 \
 (SIGNATURE##_type)Memory_Undefined_Value

/*---------------------------------- private constants ---------------------------------*/

enum { Minimum_memory_size = 4096 };

enum { Newline_char = '\n',
       Null_char    = '\0',
       Space_char   = ' ' };

/*------------------------------------ private types -----------------------------------*/

typedef
  enum { UDF_inline = 0,                   	                       /* inline: undefined */

         END_inline = 1,                                        /* inline: end of input */
         FLS_inline = 2,                	                             /* inline: false */
         NST_inline = 3,               	                         /* inline: null stream */
         NUL_inline = 4,               	                                /* inline: null */
         TRU_inline = 5,               	                                /* inline: true */
         USP_inline = 6,    	                                   /* inline: unspecified */

         CHA_inline = 7 } INL_type;	                               /* inline: character */

/*---------------------------------- private functions ---------------------------------*/

static UNS_type text_size_as_cells(UNS_type Text_length)
  { UNS_type chunk_size;
    chunk_size = Text_length;
    chunk_size += Memory_Cell_Width;
    chunk_size /= Memory_Cell_Width;
    return chunk_size; }

/*---------------------------------- public constants ----------------------------------*/

const UNS_type Maximum_Vector_Size   = Memory_Maximum_Chunk_Size;
const LNG_type Maximum_Smallint_Size = Memory_Maximum_Smallint;

/*---------------------------------- public variables ----------------------------------*/

CHR_type Grammar_Text_Buffer[Grammar_Text_Buffer_Length];

/*---------------------------------- private functions ---------------------------------*/

static TAG_type tag(EXP_type Expression)
  { INL_type inline_;
    PTR_type pointer;
    TAG_type tag;
    pointer = (PTR_type)Expression;
    if (Memory_Is_Smallint(pointer))
      return SMI_tag;
    if (Memory_Is_Inline(pointer))
      { inline_ = (INL_type)Memory_Get_Inline(pointer);
        switch (inline_)
          { case CHA_inline:
              return CHA_tag;
            case END_inline:
              return END_tag;
            case FLS_inline:
              return FLS_tag;
            case NST_inline:
              return NST_tag;
            case NUL_inline:
              return NUL_tag;
            case TRU_inline:
              return TRU_tag;
            case UDF_inline:
              return UDF_tag;
            case USP_inline:
              return USP_tag; }}
    if (Memory_Is_Pair(pointer))
      return PAI_tag;
    tag =  (TAG_type)Memory_Get_Tag(pointer);
    return tag; }

/*---------------------------------- hidden prototypes ---------------------------------*/

VOI_type _Grammar_Register(EXp_type Reference)                                                               \
  { Memory_Register(Reference); }

/*---------------------------------- public functions ----------------------------------*/

VOI_type Grammar_Force_Collect_M(VOI_type)
  { Memory_Collect_C(); }

SGN_type Grammar_List_Size(EXP_type List_expression)
  { EXP_type list_expression;
    PAI_type list_pair;
    SGN_type size;
    list_expression = List_expression;
    for (size = 0;
         is_PAI(list_expression);
         size += 1)
      { list_pair = (PAI_type)list_expression;
        list_expression = list_pair->cdr; }
    if (!is_NUL(list_expression))
      return -1;
    return size; }

VOI_type Grammar_Replace_Chunk(CHK_type Old_chunk,
                               CHK_type New_chunk)
  { Memory_Replace((PTR_type)Old_chunk,
                   (PTR_type)New_chunk); }

TAG_type Grammar_Tag(EXP_type Expression)
  { return tag(Expression); }

/*----------------------------------- AND declarations ---------------------------------*/

BLN_type is_AND(EXP_type Expression)
  { return (tag(Expression) == AND_tag); }

AND_type make_AND_M(VOI_type)
  { AND_type and;
    PTR_type pointer;
    pointer = Memory_Chunk_M(AND_tag,
                             AND_size);
    and = (AND_type)pointer;
    return and; }

/*----------------------------------- APL declarations ---------------------------------*/

BLN_type is_APL(EXP_type Expression)
  { return (tag(Expression) == APL_tag); }

APL_type make_APL_M(VOI_type)
  { APL_type application;
    PTR_type pointer;
    pointer = Memory_Chunk_M(APL_tag,
                             APL_size);
    application = (APL_type)pointer;
    return application; }

VOI_type tail_APL(APL_type Application)
  { Memory_Set_Tag((PTR_type)Application,
                   APl_tag); }

/*----------------------------------- BEG declarations ---------------------------------*/

BLN_type is_BEG(EXP_type Expression)
  { return (tag(Expression) == BEG_tag); }

BEG_type make_BEG_M(VOI_type)
  { BEG_type begin;
    PTR_type pointer;
    pointer = Memory_Chunk_M(BEG_tag,
                             BEG_size);
    begin = (BEG_type)pointer;
    return begin; }

/*----------------------------------- CHA declarations ---------------------------------*/

CHA_type Unique_Newline_Character;
CHA_type Unique_Null_Character;
CHA_type Unique_Space_Character;

BLN_type is_CHA(EXP_type Expression)
  { return (tag(Expression) == CHA_tag); }

CHA_type make_CHA(CHR_type Char)
  { PTR_type pointer;
    pointer = Memory_Inline((UNS_type)CHA_inline,
                            (UNS_type)Char);
    return (CHA_type)pointer; }

CHR_type get_CHA(CHA_type Character)
  { UNS_type inline_value;
    inline_value = Memory_Get_Qualifier((PTR_type)Character);
    return (CHR_type)inline_value; }

static VOI_type initialize_CHA(VOI_type)
  { Unique_Newline_Character = make_CHA(Newline_char);
    Unique_Null_Character    = make_CHA(Null_char);
    Unique_Space_Character   = make_CHA(Space_char); }

/*----------------------------------- CHK declarations ---------------------------------*/

CHK_type Unique_Empty_Chunk_P;

BLN_type is_CHK(EXP_type Expression)
  { return (tag(Expression) == CHK_tag); }

BLN_type is_CHk(EXP_type Expression)
  { return (tag(Expression) == CHk_tag); }

CHK_type make_CHK_M(UNS_type Chunk_size)
  { CHK_type chunk;
    PTR_type pointer;
    if (Chunk_size == 0)
      return Unique_Empty_Chunk_P;
    pointer = Memory_Chunk_M(CHK_tag,
                             Chunk_size);
    chunk = (CHK_type)pointer;
    return chunk; }

CHK_type make_CHK_Z(UNS_type Chunk_size)
  { CHK_type chunk;
    UNS_type index;
    if (Chunk_size == 0)
      return Unique_Empty_Chunk_P;
    chunk = make_CHK_M(Chunk_size);
    for (index = 1;
         index <= Chunk_size;
         index += 1)
      CLEAR(chunk[index]);
    return chunk; }

BLN_type marked_CHK(EXP_type Expression)
  { TAG_type tag;
    tag = (TAG_type)Memory_Get_Tag((PTR_type)Expression);
    return (tag == CHk_tag); }

VOI_type mark_CHK(CHK_type Chunk)
  { Memory_Set_Tag((PTR_type)Chunk,
                   CHk_tag); }

UNS_type size_CHK(CHK_type Chunk)
  { UNS_type chunk_size;
    chunk_size = (UNS_type)Memory_Get_Size((PTR_type)Chunk);
    return chunk_size; }

static VOI_type initialize_CHK(VOI_type)
  { PTR_type pointer;
    REGISTER(Unique_Empty_Chunk_P);
    pointer = Memory_Chunk_M(CHK_tag,
                             0);
    Unique_Empty_Chunk_P = (CHK_type)pointer; }

/*----------------------------------- CND declarations ---------------------------------*/

BLN_type is_CND(EXP_type Expression)
  { return (tag(Expression) == CND_tag); }

CND_type make_CND_M(VOI_type)
  { CND_type cond;
    PTR_type pointer;
    pointer = Memory_Chunk_M(CND_tag,
                             CND_size);
    cond = (CND_type)pointer;
    return cond; }

/*----------------------------------- CNE declarations ---------------------------------*/

BLN_type is_CNE(EXP_type Expression)
  { return (tag(Expression) == CNE_tag); }

CNE_type make_CNE_M(VOI_type)
  { CNE_type cond_else;
    PTR_type pointer;
    pointer = Memory_Chunk_M(CNE_tag,
                             CNE_size);
    cond_else = (CNE_type)pointer;
    return cond_else; }

/*----------------------------------- CNS declarations ---------------------------------*/

BLN_type is_CNS(EXP_type Expression)
  { return (tag(Expression) == CNS_tag); }

CNS_type make_CNS_M(VOI_type)
  { CNS_type cons_stream;
    PTR_type pointer;
    pointer = Memory_Chunk_M(CNS_tag,
                             CNS_size);
    cons_stream = (CNS_type)pointer;
    return cons_stream; }

/*----------------------------------- CNT declarations ---------------------------------*/

BLN_type is_CNT(EXP_type Expression)
  { return (tag(Expression) == CNT_tag); }

CNT_type make_CNT_M(VOI_type)
  { CNT_type continuation;
    PTR_type pointer;
    pointer = Memory_Chunk_M(CNT_tag,
                             CNT_size);
    continuation = (CNT_type)pointer;
    return continuation; }

/*----------------------------------- DEL declarations ---------------------------------*/

BLN_type is_DEL(EXP_type Expression)
  { return (tag(Expression) == DEL_tag); }

DEL_type make_DEL_M(VOI_type)
  { DEL_type delay;
    PTR_type pointer;
    pointer = Memory_Chunk_M(DEL_tag,
                             DEL_size);
    delay = (DEL_type)pointer;
    return delay; }

/*----------------------------------- DFF declarations ---------------------------------*/

BLN_type is_DFF(EXP_type Expression)
  { return (tag(Expression) == DFF_tag); }

DFF_type make_DFF_M(VOI_type)
  { DFF_type define_function;
    PTR_type pointer;
    pointer = Memory_Chunk_M(DFF_tag,
                             DFF_size);
    define_function = (DFF_type)pointer;
    return define_function; }

/*----------------------------------- DFV declarations ---------------------------------*/

BLN_type is_DFV(EXP_type Expression)
  { return (tag(Expression) == DFV_tag); }

DFV_type make_DFV_M(VOI_type)
  { DFV_type define_variable;
    PTR_type pointer;
    pointer = Memory_Chunk_M(DFV_tag,
                             DFV_size);
    define_variable = (DFV_type)pointer;
    return define_variable; }

/*----------------------------------- DFZ declarations ---------------------------------*/

BLN_type is_DFZ(EXP_type Expression)
  { return (tag(Expression) == DFZ_tag); }

DFZ_type make_DFZ_M(VOI_type)
  { DFZ_type define_function_vararg;
    PTR_type pointer;
    pointer = Memory_Chunk_M(DFZ_tag,
                             DFZ_size);
    define_function_vararg = (DFZ_type)pointer;
    return define_function_vararg; }

/*----------------------------------- END declarations ---------------------------------*/

END_type Unique_End_Of_Input = UNDEF(END);

BLN_type is_END(EXP_type Expression)
  { return (tag(Expression) == END_tag); }

static VOI_type initialize_END(VOI_type)
  { PTR_type ptr;
    ptr = Memory_Inline((UNS_type)END_inline,
                        0);
    Unique_End_Of_Input = (END_type)ptr; }

/*----------------------------------- EVA declarations ---------------------------------*/

BLN_type is_EVA(EXP_type Expression)
  { return (tag(Expression) == EVA_tag); }

EVA_type make_EVA_M(VOI_type)
  { EVA_type evaluate;
    PTR_type pointer;
    pointer = Memory_Chunk_M(EVA_tag,
                             EVA_size);
    evaluate = (EVA_type)pointer;
    return evaluate; }

VOI_type tail_EVA(EVA_type Evaluate)
  { Memory_Set_Tag((PTR_type)Evaluate,
                   EVa_tag); }

/*----------------------------------- FLS declarations ---------------------------------*/

FLS_type Unique_False = UNDEF(FLS);

BLN_type is_FLS(EXP_type Expression)
  { return (tag(Expression) == FLS_tag); }

static VOI_type initialize_FLS(VOI_type)
  { PTR_type ptr;
    ptr = Memory_Inline((UNS_type)FLS_inline,
                        0);
    Unique_False = (FLS_type)ptr; }

/*----------------------------------- FRC declarations ---------------------------------*/

BLN_type is_FRC(EXP_type Expression)
  { return (tag(Expression) == FRC_tag); }

FRC_type make_FRC_M(VOI_type)
  { FRC_type forced;
    PTR_type pointer;
    pointer = Memory_Chunk_M(FRC_tag,
                             FRC_size);
    forced = (FRC_type)pointer;
    return forced; }

VOI_type convert_FRC(PRM_type Promise,
                     EXP_type Value_expression)
  { FRC_type forced;
    Memory_Set_Tag((PTR_type)Promise,
                   FRC_tag);
    Promise->exp = (EXP_type)Unique_Undefined;
    Promise->siz = (SMI_type)Unique_Undefined;
    Promise->env = (ENV_type)Unique_Undefined;
    Promise->dct = (DCT_type)Unique_Undefined;
    forced = (FRC_type)Promise;
    forced->val = Value_expression; }

/*----------------------------------- GLA declarations ---------------------------------*/

BLN_type is_GLA(EXP_type Expression)
  { return (tag(Expression) == GLA_tag); }

GLA_type make_GLA_M(VOI_type)
  { GLA_type global_application;
    PTR_type pointer;
    pointer = Memory_Chunk_M(GLA_tag,
                             GLA_size);
    global_application = (GLA_type)pointer;
    return global_application; }

VOI_type tail_GLA(GLA_type Global_application)
  { Memory_Set_Tag((PTR_type)Global_application,
                   GLa_tag); }

/*----------------------------------- GLB declarations ---------------------------------*/

BLN_type is_GLB(EXP_type Expression)
  { return (tag(Expression) == GLB_tag); }

GLB_type make_GLB_M(VOI_type)
  { GLB_type global;
    PTR_type pointer;
    pointer = Memory_Raw_Chunk_M(GLB_tag,
                                 GLB_size);
    global = (GLB_type)pointer;
    return global; }

/*----------------------------------- IFD declarations ---------------------------------*/

BLN_type is_IFD(EXP_type Expression)
  { return (tag(Expression) == IFD_tag); }

IFD_type make_IFD_M(VOI_type)
  { IFD_type if_double;
    PTR_type pointer;
    pointer = Memory_Chunk_M(IFD_tag,
                             IFD_size);
    if_double = (IFD_type)pointer;
    return if_double; }

/*----------------------------------- IFS declarations ---------------------------------*/

BLN_type is_IFS(EXP_type Expression)
  { return (tag(Expression) == IFS_tag); }

IFS_type make_IFS_M(VOI_type)
  { IFS_type if_single;
    PTR_type pointer;
    pointer = Memory_Chunk_M(IFS_tag,
                             IFS_size);
    if_single = (IFS_type)pointer;
    return if_single; }

/*----------------------------------- IPT declarations ---------------------------------*/

BLN_type is_IPT(EXP_type Expression)
  { return (tag(Expression) == IPT_tag); }

IPT_type make_IPT_M(UNS_type Length)
  { IPT_type input_port;
    PTR_type pointer;
    UNS_type size;
    size = IPT_size + text_size_as_cells(Length);
    pointer = Memory_Raw_Chunk_M(IPT_tag,
                                 size);
    input_port = (IPT_type)pointer;
    return input_port; }

/*----------------------------------- LCA declarations ---------------------------------*/

BLN_type is_LCA(EXP_type Expression)
  { return (tag(Expression) == LCA_tag); }

LCA_type make_LCA_M(VOI_type)
  { LCA_type local_application;
    PTR_type pointer;
    pointer = Memory_Chunk_M(LCA_tag,
                             LCA_size);
    local_application = (LCA_type)pointer;
    return local_application; }

VOI_type tail_LCA(LCA_type Local_application)
  { Memory_Set_Tag((PTR_type)Local_application,
                   LCa_tag); }

/*----------------------------------- LCL declarations ---------------------------------*/

BLN_type is_LCL(EXP_type Expression)
  { return (tag(Expression) == LCL_tag); }

LCL_type make_LCL_M(VOI_type)
  { LCL_type local;
    PTR_type pointer;
    pointer = Memory_Raw_Chunk_M(LCL_tag,
                                 LCL_size);
    local = (LCL_type)pointer;
    return local; }

/*----------------------------------- LET declarations ---------------------------------*/

BLN_type is_LET(EXP_type Expression)
  { return (tag(Expression) == LET_tag); }

LET_type make_LET_M(VOI_type)
  { LET_type let;
    PTR_type pointer;
    pointer = Memory_Chunk_M(LET_tag,
                             LET_size);
    let = (LET_type)pointer;
    return let; }

VOI_type tail_LET(LET_type Let)
  { Memory_Set_Tag((PTR_type)Let,
                   LEt_tag); }

/*----------------------------------- LMB declarations ---------------------------------*/

BLN_type is_LMB(EXP_type Expression)
  { return (tag(Expression) == LMB_tag); }

LMB_type make_LMB_M(VOI_type)
  { LMB_type lambda;
    PTR_type pointer;
    pointer = Memory_Chunk_M(LMB_tag,
                             LMB_size);
    lambda = (LMB_type)pointer;
    return lambda; }

/*----------------------------------- LMV declarations ---------------------------------*/

BLN_type is_LMV(EXP_type Expression)
  { return (tag(Expression) == LMV_tag); }

LMV_type make_LMV_M(VOI_type)
  { LMV_type lambda_vararg;
    PTR_type pointer;
    pointer = Memory_Chunk_M(LMV_tag,
                             LMV_size);
    lambda_vararg = (LMV_type)pointer;
    return lambda_vararg; }

/*----------------------------------- MAC declarations ---------------------------------*/

BLN_type is_MAC(EXP_type Expression)
  { return (tag(Expression) == MAC_tag); }

MAC_type make_MAC_M(VOI_type)
  { MAC_type macro;
    PTR_type pointer;
    pointer = Memory_Chunk_M(MAC_tag,
                             MAC_size);
    macro = (MAC_type)pointer;
    return macro; }

/*----------------------------------- MVL declarations ---------------------------------*/

BLN_type is_MVL(EXP_type Expression)
  { return (tag(Expression) == MVL_tag); }

MVL_type make_MVL_M(VOI_type)
  { MVL_type macro_expression;
    PTR_type pointer;
    pointer = Memory_Chunk_M(MVL_tag,
                             MVL_size);
    macro_expression = (MVL_type)pointer;
    return macro_expression; }

/*----------------------------------- NAT declarations ---------------------------------*/

BLN_type is_NAT(EXP_type Expression)
  { return (tag(Expression) == NAT_tag); }

NAT_type make_NAT_M(UNS_type Length)
  { NAT_type native;
    PTR_type pointer;
    UNS_type chunk_size;
    chunk_size = NAT_size + text_size_as_cells(Length);
    pointer = Memory_Raw_Chunk_M(NAT_tag,
                                 chunk_size);
    native = (NAT_type)pointer;
    return native; }

/*----------------------------------- NAZ declarations ---------------------------------*/

BLN_type is_NAZ(EXP_type Expression)
  { return (tag(Expression) == NAZ_tag); }

NAZ_type make_NAZ_M(UNS_type Length)
  { NAZ_type higher_order_native;
    PTR_type pointer;
    UNS_type chunk_size;
    chunk_size = NAZ_size + text_size_as_cells(Length);
    pointer = Memory_Raw_Chunk_M(NAZ_tag,
                                 chunk_size);
    higher_order_native = (NAZ_type)pointer;
    return higher_order_native; }

/*----------------------------------- NST declarations ---------------------------------*/

NST_type Unique_Null_Stream;

BLN_type is_NST(EXP_type Expression)
  { return (Expression == Unique_Null_Stream); }

static VOI_type initialize_NST(VOI_type)
  { PTR_type ptr;
    ptr = Memory_Inline((UNS_type)NST_inline,
                        0);
    Unique_Null_Stream = (NST_type)ptr; }

/*----------------------------------- NUL declarations ---------------------------------*/

NUL_type Unique_Null;

BLN_type is_NUL(EXP_type Expression)
  { return (Expression == Unique_Null); }

static VOI_type initialize_NUL(VOI_type)
  { PTR_type ptr;
    ptr = Memory_Inline((UNS_type)NUL_inline,
                        0);
    Unique_Null = (NUL_type)ptr; }

/*----------------------------------- OPT declarations ---------------------------------*/

BLN_type is_OPT(EXP_type Expression)
  { return (tag(Expression) == OPT_tag); }

OPT_type make_OPT_M(UNS_type Length)
  { OPT_type output_port;
    PTR_type pointer;
    UNS_type size;
    size = OPT_size + text_size_as_cells(Length);
    pointer = Memory_Raw_Chunk_M(OPT_tag,
                                 size);
    output_port = (OPT_type)pointer;
    return output_port; }

/*----------------------------------- ORR declarations ---------------------------------*/

BLN_type is_ORR(EXP_type Expression)
  { return (tag(Expression) == ORR_tag); }

ORR_type make_ORR_M(VOI_type)
  { ORR_type or;
    PTR_type pointer;
    pointer = Memory_Chunk_M(ORR_tag,
                             ORR_size);
    or = (ORR_type)pointer;
    return or; }

/*----------------------------------- PAI declarations ---------------------------------*/

BLN_type is_PAI(EXP_type Expression)
  { return (tag(Expression) == PAI_tag); }

PAI_type make_PAI_M(VOI_type)
  { PAI_type pair;
    PTR_type pointer;
    pointer = Memory_Pair_M();
    pair = (PAI_type)pointer;
    return pair; }

/*----------------------------------- PRC declarations ---------------------------------*/

BLN_type is_PRC(EXP_type Expression)
  { return (tag(Expression) == PRC_tag); }

PRC_type make_PRC_M(VOI_type)
  { PRC_type procedure;
    PTR_type pointer;
    pointer = Memory_Chunk_M(PRC_tag,
                             PRC_size);
    procedure = (PRC_type)pointer;
    return procedure; }

/*----------------------------------- PRM declarations ---------------------------------*/

BLN_type is_PRM(EXP_type Expression)
  { return (tag(Expression) == PRM_tag); }

PRM_type make_PRM_M(VOI_type)
  { PRM_type promise;
    PTR_type pointer;
    pointer = Memory_Chunk_M(PRM_tag,
                             PRM_size);
    promise = (PRM_type)pointer;
    return promise; }

/*----------------------------------- PRV declarations ---------------------------------*/

BLN_type is_PRV(EXP_type Expression)
  { return (tag(Expression) == PRV_tag); }

PRV_type make_PRV_M(VOI_type)
  { PRV_type procedure_vararg;
    PTR_type pointer;
    pointer = Memory_Chunk_M(PRV_tag,
                             PRV_size);
    procedure_vararg = (PRV_type)pointer;
    return procedure_vararg; }

/*----------------------------------- QQU declarations ---------------------------------*/

BLN_type is_QQU(EXP_type Expression)
  { return (tag(Expression) == QQU_tag); }

QQU_type make_QQU_M(VOI_type)
  { QQU_type quasiquote;
    PTR_type pointer;
    pointer = Memory_Chunk_M(QQU_tag,
                             QQU_size);
    quasiquote = (QQU_type)pointer;
    return quasiquote; }

/*----------------------------------- QSY declarations ---------------------------------*/

BLN_type is_QSY(EXP_type Expression)
  { return (tag(Expression) == QSY_tag); }

QSY_type make_QSY_M(VOI_type)
  { QSY_type quasisymbol;
    PTR_type pointer;
    pointer = Memory_Chunk_M(QSY_tag,
                             QSY_size);
    quasisymbol = (QSY_type)pointer;
    return quasisymbol; }

/*----------------------------------- QUO declarations ---------------------------------*/

BLN_type is_QUO(EXP_type Expression)
  { return (tag(Expression) == QUO_tag); }

QUO_type make_QUO_M(VOI_type)
  { QUO_type quote;
    PTR_type pointer;
    pointer = Memory_Chunk_M(QUO_tag,
                             QUO_size);
    quote = (QUO_type)pointer;
    return quote; }

/*----------------------------------- REA declarations ---------------------------------*/

BLN_type is_REA(EXP_type Expression)
  { return (tag(Expression) == REA_tag); }

REA_type make_REA_M(VOI_type)
  { REA_type real;
    PTR_type pointer;
    pointer = Memory_Raw_Chunk_M(REA_tag,
                                 REA_size);
    real = (REA_type)pointer;
    return real; }

/*----------------------------------- SEQ declarations ---------------------------------*/

SEQ_type Unique_Empty_Sequence_P;

BLN_type is_SEQ(EXP_type Expression)
  { return (tag(Expression) == SEQ_tag); }

SEQ_type make_SEQ_M(UNS_type Sequence_length)
  { PTR_type pointer;
    SEQ_type sequence;
    UNS_type index;
    if (Sequence_length == 0)
      return Unique_Empty_Sequence_P;
    pointer = Memory_Chunk_M(SEQ_tag,
                             Sequence_length);
    sequence = (VEC_type)pointer;
    for (index = 1;
         index <= Sequence_length;
         index += 1)
      sequence[index] = UNDEF(EXP);
    return sequence; }

UNS_type size_SEQ(SEQ_type Sequence)
  { UNS_type sequence_size;
    sequence_size = Memory_Get_Size((PTR_type)Sequence);
    return sequence_size; }

static VOI_type initialize_SEQ(VOI_type)
  { PTR_type pointer;
    REGISTER(Unique_Empty_Sequence_P);
    pointer = Memory_Chunk_M(SEQ_tag,
                             0);
    Unique_Empty_Sequence_P = (SEQ_type)pointer; }

/*----------------------------------- SMI declarations ---------------------------------*/

SMI_type Unique_One_Smallint;
SMI_type Unique_Zero_Smallint;

BLN_type is_SMI(EXP_type Expression)
  { return (tag(Expression) == SMI_tag); }

SMI_type make_SMI(SGN_type Signed)
  { SMI_type smallint;
    smallint = (SMI_type)Memory_Smallint(Signed);
    return smallint; }

SMI_type make_SMU(UNS_type Unsigned)
  { SMI_type smallint;
    smallint = (SMI_type)Memory_Smallint((SGN_type)Unsigned);
    return smallint; }

SGN_type get_SMI(SMI_type Smallint)
  { return Memory_Get_Smallint((PTR_type)Smallint); }

UNS_type get_SMU(SMI_type Smallint)
  { SGN_type signed_;
    signed_ = Memory_Get_Smallint((PTR_type)Smallint);
    return (UNS_type)signed_; }

VOI_type incr_SMU(SMi_type Smallint_reference)
  { Memory_Increment((REF_type)Smallint_reference); }

static VOI_type initialize_SMI(VOI_type)
  { Unique_One_Smallint  = make_SMI(1);
    Unique_Zero_Smallint = make_SMI(0); }

/*----------------------------------- STG declarations ---------------------------------*/

BLN_type is_STG(EXP_type Expression)
  { return (tag(Expression) == STG_tag); }

STG_type make_STG_M(VOI_type)
  { STG_type set_global;
    PTR_type pointer;
    pointer = Memory_Chunk_M(STG_tag,
                             STG_size);
    set_global = (STG_type)pointer;
    return set_global; }

/*----------------------------------- STL declarations ---------------------------------*/

BLN_type is_STL(EXP_type Expression)
  { return (tag(Expression) == STL_tag); }

STL_type make_STL_M(VOI_type)
  { STL_type set_local;
    PTR_type pointer;
    pointer = Memory_Chunk_M(STL_tag,
                             STL_size);
    set_local = (STL_type)pointer;
    return set_local; }

/*----------------------------------- STR declarations ---------------------------------*/

static STR_type make_string_M(UNS_type Length)
  { STR_type string;
    PTR_type pointer;
    UNS_type size;
    size = STR_size + text_size_as_cells(Length);
    pointer = Memory_Raw_Chunk_M(STR_tag,
                                 size);
    string = (STR_type)pointer;
    return string; }

/*--------------------------------------------------------------------------------------*/

STR_type Unique_Empty_String_P;

BLN_type is_STR(EXP_type Expression)
  { return (tag(Expression) == STR_tag); }

STR_type make_STR_M(UNS_type Length)
  { STR_type string;
    PTR_type pointer;
    UNS_type size;
    size = STR_size + text_size_as_cells(Length);
    pointer = Memory_Raw_Chunk_M(STR_tag,
                                 size);
    string = (STR_type)pointer;
    return string; }

UNS_type length_STR(STR_type String)
  { TXT_type text;
    UNS_type text_length;
    text = String->txt;
    text_length = C_STRLEN(text);
    return text_length; }

static VOI_type initialize_STR(VOI_type)
  { PTR_type pointer;
    pointer = Memory_Chunk_M(STR_tag,
                             0);
    Unique_Empty_String_P = (STR_type)pointer; }

/*----------------------------------- SYM declarations ---------------------------------*/

BLN_type is_SYM(EXP_type Expression)
  { return (tag(Expression) == SYM_tag); }

SYM_type make_SYM_M(UNS_type Length)
  { PTR_type pointer;
    SYM_type symbol;
    UNS_type size;
    size = SYM_size + text_size_as_cells(Length);
    pointer = Memory_Raw_Chunk_M(SYM_tag,
                                 size);
    symbol = (SYM_type)pointer;
    return symbol; }

UNS_type length_SYM(SYM_type Symbol)
  { TXT_type text;
    UNS_type text_length;
    text = Symbol->txt;
    text_length = C_STRLEN(text);
    return text_length; }

/*----------------------------------- THK declarations ---------------------------------*/

BLN_type is_THK(EXP_type Expression)
  { return (tag(Expression) == THK_tag); }

THK_type make_THK_M(VOI_type)
  { THK_type thunk;
    PTR_type pointer;
    pointer = Memory_Chunk_M(THK_tag,
                             THK_size);
    thunk = (THK_type)pointer;
    return thunk; }

VOI_type tail_THK(THK_type Thunk)
  { Memory_Set_Tag((PTR_type)Thunk,
                   THk_tag); }

/*----------------------------------- TRU declarations ---------------------------------*/

TRU_type Unique_True;

BLN_type is_TRU(EXP_type Expression)
  { return (Expression == Unique_True); }

static VOI_type initialize_TRU(VOI_type)
  { PTR_type ptr;
    ptr = Memory_Inline((UNS_type)TRU_inline,
                        0);
    Unique_True = (TRU_type)ptr; }

/*----------------------------------- UDF declarations ---------------------------------*/

UDF_type Unique_Undefined;

BLN_type is_UDF(EXP_type Expression)
  { return (Expression == Unique_Undefined); }

static VOI_type initialize_UDF(VOI_type)
  { CLEAR(Unique_Undefined); }

/*----------------------------------- ULS declarations ---------------------------------*/

BLN_type is_ULS(EXP_type Expression)
  { return (tag(Expression) == ULS_tag); }

ULS_type make_ULS_M(VOI_type)
  { ULS_type unquote_list_splicing;
    PTR_type pointer;
    pointer = Memory_Chunk_M(ULS_tag,
                             ULS_size);
    unquote_list_splicing = (ULS_type)pointer;
    return unquote_list_splicing; }

/*----------------------------------- UNQ declarations ---------------------------------*/

BLN_type is_UNQ(EXP_type Expression)
  { return (tag(Expression) == UNQ_tag); }

UNQ_type make_UNQ_M(VOI_type)
  { UNQ_type unquote;
    PTR_type pointer;
    pointer = Memory_Chunk_M(UNQ_tag,
                             UNQ_size);
    unquote = (UNQ_type)pointer;
    return unquote; }

VOI_type tail_UNQ(UNQ_type Unquote)
  { Memory_Set_Tag((PTR_type)Unquote,
                   UNq_tag); }

/*----------------------------------- USP declarations ---------------------------------*/

USP_type Unique_Unspecified;

BLN_type is_USP(EXP_type Expression)
  { return (Expression == Unique_Unspecified); }

static VOI_type initialize_USP(VOI_type)
  { PTR_type ptr;
    ptr = Memory_Inline(USP_inline,
                        0);
    Unique_Unspecified = (USP_type)ptr; }

/*----------------------------------- UVS declarations ---------------------------------*/

BLN_type is_UVS(EXP_type Expression)
  { return (tag(Expression) == UVS_tag); }

UVS_type make_UVS_M(VOI_type)
  { UVS_type unquote_vector_splicing;
    PTR_type pointer;
    pointer = Memory_Chunk_M(UVS_tag,
                             UVS_size);
    unquote_vector_splicing = (UVS_type)pointer;
    return unquote_vector_splicing; }


/*----------------------------------- VEC declarations ---------------------------------*/

VEC_type Unique_Empty_Vector_P;

BLN_type is_VEC(EXP_type Expression)
  { return (tag(Expression) == VEC_tag); }

VEC_type make_VEC_M(UNS_type Vector_size)
  { PTR_type pointer;
    VEC_type vector;
    if (Vector_size == 0)
      return Unique_Empty_Vector_P;
    pointer = Memory_Chunk_M(VEC_tag,
                             Vector_size);
    vector = (VEC_type)pointer;
    return vector; }

VEC_type make_VEC_Z(UNS_type Vector_size)
  { UNS_type index;
    VEC_type vector;
    if (Vector_size == 0)
      return Unique_Empty_Vector_P;
    vector = make_VEC_M(Vector_size);
    for (index = 1;
         index <= Vector_size;
         index += 1)
      vector[index] = Unique_Undefined;
    return vector; }

UNS_type size_VEC(VEC_type Vector)
  { UNS_type vector_size;
    vector_size = Memory_Get_Size((PTR_type)Vector);
    return vector_size; }

static VOI_type initialize_VEC(VOI_type)
  { PTR_type pointer;
    pointer = Memory_Chunk_M(VEC_tag,
                             0);
    Unique_Empty_Vector_P = (VEC_type)pointer; }

/*----------------------------------- WHI declarations ---------------------------------*/

BLN_type is_WHI(EXP_type Expression)
  { return (tag(Expression) == WHI_tag); }

WHI_type make_WHI_M(VOI_type)
  { WHI_type while_;
    PTR_type pointer;
    pointer = Memory_Chunk_M(WHI_tag,
                             WHI_size);
    while_ = (WHI_type)pointer;
    return while_; }

/*------------------------------------- initialize -------------------------------------*/

VOI_type Grammar_Initialize_M(VOI_type)
  { Memory_Initialize();
    initialize_CHA();
    initialize_CHK();
    initialize_END();
    initialize_FLS();
    initialize_NST();
    initialize_NUL();
    initialize_SEQ();
    initialize_SMI();
    initialize_STR();
    initialize_TRU();
    initialize_UDF();
    initialize_USP();
    initialize_VEC(); }
