                     /*---------------------------------------------*/
                     /*                  >>>Slip<<<                 */
                     /*                 Theo D'Hondt                */
                     /*          VUB Software Languages Lab         */
                     /*                  (c) 2018                   */
                     /*---------------------------------------------*/
                     /*                final version                */
                     /*---------------------------------------------*/
                     /*                    grammar                  */
                     /*---------------------------------------------*/

/*------------------------------------ public macros -----------------------------------*/

#define DEFINE_TYPE(SIGNATURE)                                                           \
  typedef struct SIGNATURE * SIGNATURE##_type

#define BEGIN_CHUNK(SIGNATURE)                                                           \
  typedef struct SIGNATURE {                                                             \
  DEFINE_SLOT(USP, __h)

#define BEGIN_PAIR(SIGNATURE)                                                            \
  typedef struct SIGNATURE {

#define DEFINE_SLOT(SIGNATURE, NAME)                                                     \
  SIGNATURE##_type NAME

#define END_CHUNK(SIGNATURE)                                                             \
  } SIGNATURE;                                                                           \
static const UNS_type SIGNATURE##_size = sizeof(SIGNATURE)/sizeof(ADR_type) - 1

#define END_PAIR(SIGNATURE)                                                              \
  } SIGNATURE;                                                                           \
static const UNS_type SIGNATURE##_size = sizeof(SIGNATURE)/sizeof(ADR_type);

#define REGISTER(VARIABLE)                                                               \
  _Grammar_Register((EXp_type)&VARIABLE)

/*---------------------------------- public constants ----------------------------------*/

enum { Grammar_Text_Buffer_Length = 256 };

/*------------------------------------ public types ------------------------------------*/

typedef
  enum { // small integers
         SMI_tag = -10,                           	                /* s-expr: smallint */
         // pairs
         PAI_tag = -9,                                                  /* s-expr: pair */
         // immediate pointers
         CHA_tag = -8,	                                           /* s-expr: character */
         END_tag = -7,                             	            /* s-expr: end of input */
         FLS_tag = -6,                	                               /* s-expr: false */
         NST_tag = -5,               	                            /* value: null stream */
         NUL_tag = -4,                       	                          /* s-expr: null */
         TRU_tag = -3,                       	                          /* s-expr: true */
         UDF_tag = -2,                       	                   /* internal: undefined */
         USP_tag = -1,                       	                    /* value: unspecified */
         // regular variable size chunks
         CHK_tag =  0,                                 	             /* internal: chunk */
         CHk_tag =  1,               	                        /* internal: marked chunk */
         SEQ_tag =  2,               	                            /* internal: sequence */
         VEC_tag =  3,               	                                /* s-expr: vector */
         // raw variable size chunks
         IPT_tag =  4,               	                             /* value: input port */
         NAT_tag =  5,               	                                 /* value: native */
         NAZ_tag =  6,               	                    /* value: higher-order native */
         OPT_tag =  7,               	                            /* value: output port */
         STR_tag =  8,               	                                /* s-expr: string */
         SYM_tag =  9,                                                /* s-expr: symbol */
         X10_tag = 10,                                                        /* unused */
         X11_tag = 11,                                                        /* unused */
         X12_tag = 12,                                                        /* unused */
         X13_tag = 13,                                                        /* unused */
         X14_tag = 14,                                                        /* unused */
         X15_tag = 15,                                                        /* unused */
         // regular fixed size chunks
         AND_tag = 16,  	                                               /* syntax: and */
         APL_tag = 17,               	                           /* syntax: application */
         APl_tag = 18,               	                 /* syntax: tail call application */
         BEG_tag = 19,               	                                 /* syntax: begin */
         CND_tag = 20,               	                         /* syntax: cond w/o else */
         CNE_tag = 21,               	                        /* syntax: cond with else */
         CNS_tag = 22,               	                           /* syntax: cons-stream */
         CNT_tag = 23,               	                           /* value: continuation */
/******/ DCT_tag = 24,               	                          /* internal: dictionary */
         DEL_tag = 25,               	                                 /* syntax: delay */
         DFF_tag = 26,               	                       /* syntax: define function */
         DFV_tag = 27,               	                       /* syntax: define variable */
         DFZ_tag = 28,               	          /* syntax: define function with varargs */
         EVA_tag = 29,               	                              /* syntax: evaluate */
         EVa_tag = 30,               	                    /* syntax: tail call evaluate */
         FRC_tag = 31,               	                          /* value: force wrapper */
         GLA_tag = 32,               	                    /* syntax: global application */
         GLa_tag = 33,               	          /* syntax: global tail call application */
         IFD_tag = 34,               	                         /* syntax: two-branch if */
         IFS_tag = 35,               	                         /* syntax: one-branch if */
         LCA_tag = 36,               	                     /* syntax: local application */
         LCa_tag = 37,               	           /* syntax: local tail call application */
         LET_tag = 38,               	                                   /* syntax: let */
         LEt_tag = 39,               	                         /* syntax: tail call let */
         LMB_tag = 40,               	                                /* syntax: lambda */
         LMV_tag = 41,               	                   /* syntax: lambda with varargs */
         MAC_tag = 42,               	                                  /* value: macro */
/******/ MSL_tag = 43,               	                     /* syntax: macro symbol link */
         MVL_tag = 44,               	                          /* internal: macrovalue */
         ORR_tag = 45,               	                                    /* syntax: or */
         PRC_tag = 46,               	                              /* value: procedure */
         PRM_tag = 47,               	                                /* value: promise */
         PRV_tag = 48,               	                 /* value: procedure with varargs */
         QQU_tag = 49,               	                            /* syntax: quasiquote */
         QSY_tag = 50,               	                           /* syntax: quasisymbol */
         QUO_tag = 51,               	                                 /* syntax: quote */
         STG_tag = 52,               	                 /* syntax: store global variable */
         STL_tag = 53,               	                  /* syntax: store local variable */
         THK_tag = 54,               	                               /* internal: thunk */
         THk_tag = 55,               	                     /* internal: tail call thunk */
         ULS_tag = 56,               	                   /* syntax: unquote list splice */
         UNQ_tag = 57,               	                               /* syntax: unquote */
         UNq_tag = 58,               	                     /* syntax: tail call unquote */
         UVS_tag = 59,               	                 /* syntax: unquote vector splice */
/******/ VSL_tag = 60,               	                /* internal: variable symbol link */
         WHI_tag = 61,             	                                   /* syntax: while */
         // raw fixed size chunks
         GLB_tag = 62,               	                       /* syntax: global variable */
         LCL_tag = 63,               	                        /* syntax: local variable */
         REA_tag = 64 } TAG_type;            	                          /* s-expr: real */

DEFINE_TYPE(AND);
DEFINE_TYPE(APL);
DEFINE_TYPE(BEG);
DEFINE_TYPE(CHA);
DEFINE_TYPE(CND);
DEFINE_TYPE(CNE);
DEFINE_TYPE(CNS);
DEFINE_TYPE(CNT);
DEFINE_TYPE(DEL);
DEFINE_TYPE(DFF);
DEFINE_TYPE(DFV);
DEFINE_TYPE(DFZ);
DEFINE_TYPE(END);
DEFINE_TYPE(EVA);
DEFINE_TYPE(FLS);
DEFINE_TYPE(FRC);
DEFINE_TYPE(GLA);
DEFINE_TYPE(GLB);
DEFINE_TYPE(IFD);
DEFINE_TYPE(IFS);
DEFINE_TYPE(IPT);
DEFINE_TYPE(LCA);
DEFINE_TYPE(LCa);
DEFINE_TYPE(LCL);
DEFINE_TYPE(LET);
DEFINE_TYPE(LMB);
DEFINE_TYPE(LMV);
DEFINE_TYPE(MAC);
DEFINE_TYPE(MVL);
DEFINE_TYPE(NAT);
DEFINE_TYPE(NAZ);
DEFINE_TYPE(NST);
DEFINE_TYPE(NUL);
DEFINE_TYPE(OPT);
DEFINE_TYPE(ORR);
DEFINE_TYPE(QQU);
DEFINE_TYPE(QSY);
DEFINE_TYPE(QUO);
DEFINE_TYPE(PAI);
DEFINE_TYPE(PRC);
DEFINE_TYPE(PRM);
DEFINE_TYPE(PRV);
DEFINE_TYPE(REA);
DEFINE_TYPE(SMI);
DEFINE_TYPE(STG);
DEFINE_TYPE(STL);
DEFINE_TYPE(STR);
DEFINE_TYPE(SYM);
DEFINE_TYPE(THK);
DEFINE_TYPE(TRU);
DEFINE_TYPE(UDF);
DEFINE_TYPE(ULS);
DEFINE_TYPE(UNQ);
DEFINE_TYPE(UNq);
DEFINE_TYPE(USP);
DEFINE_TYPE(UVS);
DEFINE_TYPE(WHI);

typedef VOI_type * EXP_type;                                         /* expression type */

typedef EXP_type * CHK_type;                                              /* chunk type */
typedef EXP_type * SEQ_type;                                           /* sequence type */
typedef EXP_type * VEC_type;                                             /* vector type */

typedef EXP_type * EXp_type;                               /* expression reference type */
typedef PAI_type * PAi_type;                                     /* pair reference type */
typedef SMI_type * SMi_type;                                 /* smallint reference type */

DEFINE_TYPE(DCT);
DEFINE_TYPE(THR);

typedef CHK_type   ENV_type;                                        /* environment type */
typedef CHK_type   FRM_type;                                              /* frame type */

typedef UNS_type * UNs_type;                                 /* unsigned reference type */

typedef VOI_type (* NFN_type) (FRM_type);                           /* native prototype */
typedef VOI_type (* TFN_type) (VOI_type);                            /* thunk prototype */

/*---------------------------------- public constants ----------------------------------*/

extern const LNG_type Maximum_Smallint_Size;
extern const UNS_type Maximum_Vector_Size;

/*---------------------------------- public variables ----------------------------------*/

extern CHR_type Grammar_Text_Buffer[];

/*---------------------------------- hidden prototypes ---------------------------------*/

VOI_type       _Grammar_Register(EXp_type);
                                                             
/*---------------------------------- public prototypes ---------------------------------*/

VOI_type Grammar_Force_Collect_M(VOI_type);
SGN_type       Grammar_List_Size(EXP_type);
VOI_type   Grammar_Replace_Chunk(CHK_type,
                                 CHK_type);
TAG_type             Grammar_Tag(EXP_type);

/*----------------------------------- AND declarations ---------------------------------*/

BEGIN_CHUNK(AND);
  DEFINE_SLOT(SEQ, prd);
END_CHUNK(AND);

BLN_type      is_AND(EXP_type);

AND_type  make_AND_M(VOI_type);

/*----------------------------------- APL declarations ---------------------------------*/

BEGIN_CHUNK(APL);
  DEFINE_SLOT(EXP, exp);
  DEFINE_SLOT(SEQ, arg);
END_CHUNK(APL);

BLN_type      is_APL(EXP_type);

APL_type make_APL_M(VOI_type);

VOI_type    tail_APL(APL_type);

/*----------------------------------- BEG declarations ---------------------------------*/

BEGIN_CHUNK(BEG);
  DEFINE_SLOT(EXP, exp);
END_CHUNK(BEG);

BLN_type      is_BEG(EXP_type);

BEG_type  make_BEG_M(VOI_type);

/*----------------------------------- CHA declarations ---------------------------------*/

extern CHA_type Unique_Newline_Character;
extern CHA_type Unique_Null_Character;
extern CHA_type Unique_Space_Character;

BLN_type      is_CHA(EXP_type);

CHA_type    make_CHA(CHR_type);

CHR_type     get_CHA(CHA_type);

/*----------------------------------- CHK declarations ---------------------------------*/

extern CHK_type Unique_Empty_Chunk_P;

BLN_type      is_CHK(EXP_type);
BLN_type      is_CHk(EXP_type);

CHK_type  make_CHK_M(UNS_type);
CHK_type  make_CHK_Z(UNS_type);

BLN_type  marked_CHK(EXP_type);

VOI_type    mark_CHK(CHK_type);
UNS_type    size_CHK(CHK_type);

/*----------------------------------- CND declarations ---------------------------------*/

BEGIN_CHUNK(CND);
  DEFINE_SLOT(SEQ, prd);
  DEFINE_SLOT(SEQ, bod);
END_CHUNK(CND);

BLN_type      is_CND(EXP_type);

CND_type  make_CND_M(VOI_type);

/*----------------------------------- CNE declarations ---------------------------------*/

BEGIN_CHUNK(CNE);
  DEFINE_SLOT(SEQ, prd);
  DEFINE_SLOT(SEQ, bod);
  DEFINE_SLOT(EXP, els);
END_CHUNK(CNE);

BLN_type      is_CNE(EXP_type);

CNE_type  make_CNE_M(VOI_type);

/*----------------------------------- CNS declarations ---------------------------------*/

BEGIN_CHUNK(CNS);
  DEFINE_SLOT(EXP, car);
  DEFINE_SLOT(EXP, cdr);
  DEFINE_SLOT(SMI, siz);
  DEFINE_SLOT(DCT, dct);
END_CHUNK(CNS);

BLN_type      is_CNS(EXP_type);

CNS_type  make_CNS_M(VOI_type);

/*----------------------------------- CNT declarations ---------------------------------*/

BEGIN_CHUNK(CNT);
  DEFINE_SLOT(ENV, env);
  DEFINE_SLOT(FRM, frm);
  DEFINE_SLOT(THR, thr);
  DEFINE_SLOT(DCT, dct);
END_CHUNK(CNT);

BLN_type      is_CNT(EXP_type);

CNT_type  make_CNT_M(VOI_type);

/*----------------------------------- DEL declarations ---------------------------------*/

BEGIN_CHUNK(DEL);
  DEFINE_SLOT(EXP, exp);
  DEFINE_SLOT(SMI, siz);
  DEFINE_SLOT(DCT, dct);
END_CHUNK(DEL);

BLN_type      is_DEL(EXP_type);

DEL_type  make_DEL_M(VOI_type);

/*----------------------------------- DFF declarations ---------------------------------*/

BEGIN_CHUNK(DFF);
  DEFINE_SLOT(SYM, sym);
  DEFINE_SLOT(SMI, ofs);
  DEFINE_SLOT(SMI, par);
  DEFINE_SLOT(EXP, exp);
  DEFINE_SLOT(SMI, siz);
  DEFINE_SLOT(DCT, dct);
END_CHUNK(DFF);

BLN_type      is_DFF(EXP_type);

DFF_type  make_DFF_M(VOI_type);

/*----------------------------------- DFV declarations ---------------------------------*/

BEGIN_CHUNK(DFV);
  DEFINE_SLOT(SMI, ofs);
  DEFINE_SLOT(EXP, exp);
END_CHUNK(DFV);

BLN_type      is_DFV(EXP_type);

DFV_type  make_DFV_M(VOI_type);

/*----------------------------------- DFZ declarations ---------------------------------*/

BEGIN_CHUNK(DFZ);
  DEFINE_SLOT(SYM, sym);
  DEFINE_SLOT(SMI, ofs);
  DEFINE_SLOT(SMI, par);
  DEFINE_SLOT(EXP, exp);
  DEFINE_SLOT(SMI, siz);
  DEFINE_SLOT(DCT, dct);
END_CHUNK(DFZ);

BLN_type      is_DFZ(EXP_type);

DFZ_type  make_DFZ_M(VOI_type);

/*----------------------------------- END declarations ---------------------------------*/

extern END_type Unique_End_Of_Input;

BLN_type      is_END(EXP_type);

/*----------------------------------- EVA declarations ---------------------------------*/

BEGIN_CHUNK(EVA);
  DEFINE_SLOT(EXP, exp);
END_CHUNK(EVA);

BLN_type      is_EVA(EXP_type);

EVA_type  make_EVA_M(VOI_type);

VOI_type    tail_EVA(EVA_type);

/*----------------------------------- FLS declarations ---------------------------------*/

extern FLS_type Unique_False;

BLN_type      is_FLS(EXP_type);

/*----------------------------------- FRC declarations ---------------------------------*/

BEGIN_CHUNK(FRC);
  DEFINE_SLOT(EXP, val);
END_CHUNK(FRC);

BLN_type      is_FRC(EXP_type);

FRC_type  make_FRC_M(VOI_type);

VOI_type convert_FRC(PRM_type,
                     EXP_type);

/*----------------------------------- GLA declarations ---------------------------------*/

BEGIN_CHUNK(GLA);
  DEFINE_SLOT(GLB, glb);
  DEFINE_SLOT(SEQ, arg);
END_CHUNK(GLA);

BLN_type      is_GLA(EXP_type);

GLA_type  make_GLA_M(VOI_type);

VOI_type    tail_GLA(GLA_type);

/*----------------------------------- GLB declarations ---------------------------------*/

BEGIN_CHUNK(GLB);
  DEFINE_SLOT(UNS, scp);
  DEFINE_SLOT(UNS, ofs);
END_CHUNK(GLB);

BLN_type      is_GLB(EXP_type);

GLB_type  make_GLB_M(VOI_type);

/*----------------------------------- IFD declarations ---------------------------------*/

BEGIN_CHUNK(IFD);
  DEFINE_SLOT(EXP, prd);
  DEFINE_SLOT(EXP, cns);
  DEFINE_SLOT(EXP, alt);
END_CHUNK(IFD);

BLN_type      is_IFD(EXP_type);

IFD_type  make_IFD_M(VOI_type);

/*----------------------------------- IFS declarations ---------------------------------*/

BEGIN_CHUNK(IFS);
  DEFINE_SLOT(EXP, prd);
  DEFINE_SLOT(EXP, cns);
END_CHUNK(IFS);

BLN_type      is_IFS(EXP_type);

IFS_type  make_IFS_M(VOI_type);

/*----------------------------------- IPT declarations ---------------------------------*/

BEGIN_CHUNK(IPT);
  DEFINE_SLOT(ADR, adr);
  DEFINE_SLOT(CHR, txt[]);
END_CHUNK(IPT);

BLN_type      is_IPT(EXP_type);

IPT_type  make_IPT_M(UNS_type);

/*----------------------------------- LCA declarations ---------------------------------*/

BEGIN_CHUNK(LCA);
  DEFINE_SLOT(LCL, lcl);
  DEFINE_SLOT(SEQ, arg);
END_CHUNK(LCA);

BLN_type      is_LCA(EXP_type);

LCA_type  make_LCA_M(VOI_type);

VOI_type    tail_LCA(LCA_type);

/*----------------------------------- LCL declarations ---------------------------------*/

BEGIN_CHUNK(LCL);
  DEFINE_SLOT(UNS, ofs);
END_CHUNK(LCL);

BLN_type      is_LCL(EXP_type);

LCL_type  make_LCL_M(VOI_type);

/*----------------------------------- LET declarations ---------------------------------*/

BEGIN_CHUNK(LET);
  DEFINE_SLOT(SEQ, seq);
  DEFINE_SLOT(EXP, exp);
  DEFINE_SLOT(SMI, siz);
  DEFINE_SLOT(DCT, dct);
END_CHUNK(LET);

BLN_type      is_LET(EXP_type);

LET_type  make_LET_M(VOI_type);

VOI_type    tail_LET(LET_type);

/*----------------------------------- LMB declarations ---------------------------------*/

BEGIN_CHUNK(LMB);
  DEFINE_SLOT(SMI, par);
  DEFINE_SLOT(EXP, exp);
  DEFINE_SLOT(SMI, siz);
  DEFINE_SLOT(DCT, dct);
END_CHUNK(LMB);

BLN_type      is_LMB(EXP_type);

LMB_type  make_LMB_M(VOI_type);

/*----------------------------------- LMV declarations ---------------------------------*/

BEGIN_CHUNK(LMV);
  DEFINE_SLOT(SMI, par);
  DEFINE_SLOT(EXP, exp);
  DEFINE_SLOT(SMI, siz);
  DEFINE_SLOT(DCT, dct);
END_CHUNK(LMV);

BLN_type      is_LMV(EXP_type);

LMV_type  make_LMV_M(VOI_type);

/*----------------------------------- MAC declarations ---------------------------------*/

BEGIN_CHUNK(MAC);
  DEFINE_SLOT(SYM, sym);
END_CHUNK(MAC);

BLN_type      is_MAC(EXP_type);

MAC_type  make_MAC_M(VOI_type);

/*----------------------------------- MVL declarations ---------------------------------*/

BEGIN_CHUNK(MVL);
  DEFINE_SLOT(EXP, exp);
  DEFINE_SLOT(ENV, env);
  DEFINE_SLOT(FRM, frm);
  DEFINE_SLOT(DCT, dct);
END_CHUNK(MVL);

BLN_type      is_MVL(EXP_type);

MVL_type  make_MVL_M(VOI_type);

/*----------------------------------- NAT declarations ---------------------------------*/

BEGIN_CHUNK(NAT);
  DEFINE_SLOT(NFN, nfn);
  DEFINE_SLOT(CHR, txt[]);
END_CHUNK(NAT);

BLN_type      is_NAT(EXP_type);

NAT_type  make_NAT_M(UNS_type);

/*----------------------------------- NAZ declarations ---------------------------------*/

BEGIN_CHUNK(NAZ);
  DEFINE_SLOT(NFN, nfn);
  DEFINE_SLOT(NFN, nfz);
  DEFINE_SLOT(CHR, txt[]);
END_CHUNK(NAZ);

BLN_type      is_NAZ(EXP_type);

NAZ_type  make_NAZ_M(UNS_type);

/*----------------------------------- NST declarations ---------------------------------*/

extern NST_type Unique_Null_Stream;

BLN_type      is_NST(EXP_type);

/*----------------------------------- NUL declarations ---------------------------------*/

extern NUL_type Unique_Null;

BLN_type      is_NUL(EXP_type);

/*----------------------------------- OPT declarations ---------------------------------*/

BEGIN_CHUNK(OPT);
  DEFINE_SLOT(ADR, adr);
  DEFINE_SLOT(CHR, txt[]);
END_CHUNK(OPT);

BLN_type      is_OPT(EXP_type);

OPT_type  make_OPT_M(UNS_type);

/*----------------------------------- ORR declarations ---------------------------------*/

BEGIN_CHUNK(ORR);
  DEFINE_SLOT(SEQ, prd);
END_CHUNK(ORR);

BLN_type      is_ORR(EXP_type);

ORR_type  make_ORR_M(VOI_type);

/*----------------------------------- PAI declarations ---------------------------------*/

BEGIN_PAIR(PAI)
  DEFINE_SLOT(EXP, car);
  DEFINE_SLOT(EXP, cdr);
END_PAIR(PAI)

BLN_type      is_PAI(EXP_type);

PAI_type  make_PAI_M(VOI_type);

/*----------------------------------- PRC declarations ---------------------------------*/

BEGIN_CHUNK(PRC);
  DEFINE_SLOT(SYM, sym);
  DEFINE_SLOT(SMI, par);
  DEFINE_SLOT(EXP, exp);
  DEFINE_SLOT(SMI, siz);
  DEFINE_SLOT(ENV, env);
  DEFINE_SLOT(DCT, dct);
END_CHUNK(PRC);

BLN_type      is_PRC(EXP_type);

PRC_type  make_PRC_M(VOI_type);

/*----------------------------------- PRM declarations ---------------------------------*/

BEGIN_CHUNK(PRM);
  DEFINE_SLOT(EXP, exp);
  DEFINE_SLOT(SMI, siz);
  DEFINE_SLOT(ENV, env);
  DEFINE_SLOT(DCT, dct);
END_CHUNK(PRM);

BLN_type      is_PRM(EXP_type);

PRM_type  make_PRM_M(VOI_type);

/*----------------------------------- PRV declarations ---------------------------------*/

BEGIN_CHUNK(PRV);
  DEFINE_SLOT(SYM, sym);
  DEFINE_SLOT(SMI, par);
  DEFINE_SLOT(EXP, exp);
  DEFINE_SLOT(SMI, siz);
  DEFINE_SLOT(ENV, env);
  DEFINE_SLOT(DCT, dct);
END_CHUNK(PRV);

BLN_type      is_PRV(EXP_type);

PRV_type  make_PRV_M(VOI_type);

/*----------------------------------- QQU declarations ---------------------------------*/

BEGIN_CHUNK(QQU);
  DEFINE_SLOT(EXP, exp);
END_CHUNK(QQU);

BLN_type      is_QQU(EXP_type);

QQU_type  make_QQU_M(VOI_type);

/*----------------------------------- QSY declarations ---------------------------------*/

BEGIN_CHUNK(QSY);
  DEFINE_SLOT(SYM, sym);
  DEFINE_SLOT(SMI, ctr);
END_CHUNK(QSY);

BLN_type      is_QSY(EXP_type);

QSY_type  make_QSY_M(VOI_type);

/*----------------------------------- QUO declarations ---------------------------------*/

BEGIN_CHUNK(QUO);
  DEFINE_SLOT(EXP, exp);
END_CHUNK(QUO);

BLN_type      is_QUO(EXP_type);

QUO_type  make_QUO_M(VOI_type);

/*----------------------------------- REA declarations ---------------------------------*/

BEGIN_CHUNK(REA);
  DEFINE_SLOT(FLO, flo);
END_CHUNK(REA);

BLN_type      is_REA(EXP_type);

REA_type  make_REA_M(VOI_type);

/*----------------------------------- SEQ declarations ---------------------------------*/

extern SEQ_type Unique_Empty_Sequence_P;

BLN_type      is_SEQ(EXP_type);

SEQ_type  make_SEQ_M(UNS_type);

UNS_type    size_SEQ(SEQ_type);

/*----------------------------------- SMI declarations ---------------------------------*/

extern SMI_type Unique_One_Smallint;
extern SMI_type Unique_Zero_Smallint;

BLN_type      is_SMI(EXP_type);

SMI_type    make_SMI(SGN_type);
SMI_type    make_SMU(UNS_type);

SGN_type     get_SMI(SMI_type);
UNS_type     get_SMU(SMI_type);
VOI_type    incr_SMU(SMi_type);

/*----------------------------------- STG declarations ---------------------------------*/

BEGIN_CHUNK(STG);
  DEFINE_SLOT(SMI, scp);
  DEFINE_SLOT(SMI, ofs);
  DEFINE_SLOT(EXP, exp);
END_CHUNK(STG);

BLN_type      is_STG(EXP_type);

STG_type  make_STG_M(VOI_type);

/*----------------------------------- STL declarations ---------------------------------*/

BEGIN_CHUNK(STL);
  DEFINE_SLOT(SMI, ofs);
  DEFINE_SLOT(EXP, exp);
END_CHUNK(STL);

BLN_type      is_STL(EXP_type);
STL_type  make_STL_M(VOI_type);

/*----------------------------------- STR declarations ---------------------------------*/

extern STR_type Unique_Empty_String_P;

BEGIN_CHUNK(STR);
  DEFINE_SLOT(CHR, txt[]);
END_CHUNK(STR);

BLN_type      is_STR(EXP_type);

STR_type  make_STR_M(UNS_type);

UNS_type  length_STR(STR_type);

/*----------------------------------- SYM declarations ---------------------------------*/

BEGIN_CHUNK(SYM);
  DEFINE_SLOT(CHR, txt[]);
END_CHUNK(SYM);

BLN_type      is_SYM(EXP_type);

SYM_type  make_SYM_M(UNS_type);

UNS_type  length_SYM(SYM_type);

/*----------------------------------- THK declarations ---------------------------------*/

BEGIN_CHUNK(THK);
  DEFINE_SLOT(EXP, exp);
  DEFINE_SLOT(SMI, siz);
  DEFINE_SLOT(DCT, dct);
END_CHUNK(THK);

BLN_type      is_THK(EXP_type);

THK_type  make_THK_M(VOI_type);

VOI_type    tail_THK(THK_type);

/*----------------------------------- TRU declarations ---------------------------------*/

extern TRU_type Unique_True;

BLN_type      is_TRU(EXP_type);

/*----------------------------------- UDF declarations ---------------------------------*/

extern UDF_type Unique_Undefined;

BLN_type      is_UDF(EXP_type);

/*----------------------------------- ULS declarations ---------------------------------*/

BEGIN_CHUNK(ULS);
  DEFINE_SLOT(EXP, exp);
END_CHUNK(ULS);

BLN_type      is_ULS(EXP_type);

ULS_type  make_ULS_M(VOI_type);

/*----------------------------------- UNQ declarations ---------------------------------*/

BEGIN_CHUNK(UNQ);
  DEFINE_SLOT(EXP, exp);
END_CHUNK(UNQ);

BLN_type      is_UNQ(EXP_type);

UNQ_type  make_UNQ_M(VOI_type);

VOI_type    tail_UNQ(UNQ_type);

/*----------------------------------- USP declarations ---------------------------------*/

extern USP_type Unique_Unspecified;

BLN_type      is_USP(EXP_type);

/*----------------------------------- UVS declarations ---------------------------------*/

BEGIN_CHUNK(UVS);
  DEFINE_SLOT(EXP, exp);
END_CHUNK(UVS);

BLN_type      is_UVS(EXP_type);

UVS_type  make_UVS_M(VOI_type);

/*----------------------------------- VEC declarations ---------------------------------*/

extern VEC_type Unique_Empty_Vector_P;

BLN_type      is_VEC(EXP_type);

VEC_type  make_VEC_M(UNS_type);
VEC_type  make_VEC_Z(UNS_type);

UNS_type    size_VEC(VEC_type);

/*----------------------------------- WHI declarations ---------------------------------*/

BEGIN_CHUNK(WHI);
  DEFINE_SLOT(EXP, prd);
  DEFINE_SLOT(EXP, exp);
END_CHUNK(WHI);

BLN_type      is_WHI(EXP_type);

WHI_type  make_WHI_M(VOI_type);

/*------------------------------------- initialize -------------------------------------*/

VOI_type Grammar_Initialize_M(VOI_type);

