                     /*---------------------------------------------*/
                     /*                  >>>Slip<<<                 */
                     /*                 Theo D'Hondt                */
                     /*          VUB Software Languages Lab         */
                     /*                  (c) 2018                   */
                     /*---------------------------------------------*/
                     /*                final version                */
                     /*---------------------------------------------*/
                     /*                    memory                   */
                     /*---------------------------------------------*/

#define BITHEAP

#include "SlipPlatform.h"

#include "SlipMain.h"

#ifdef BITHEAP
#include "SlipBitheap.h"
#endif

#include "SlipMemory.h"

/*--------------------------------------------------------------------------------------*/
/*                         3         2          1                                       */
/*                       21098765432109876543 21098 765 4 3 2 1 0                       */
/*                                                                                      */
/*     inline           |00000000000000000000|qqqqq qqq|i i i|0|m|                      */
/*                                                                                      */
/*     pointer          |pppppppppppppppppppp ppppp ppp p p p|0|m|                      */
/*                                                                                      */
/*     smallint         |nnnnnnnnnnnnnnnnnnnn nnnnn nnn n n n|1|m|                      */
/*                                                                                      */
/*     short tag header |ssssssssssssssssssss sssss|ttt t|r|0|1|m|                      */
/*                                                                                      */
/*     long tag header  |tttttttttttttttttttt ttttt|sss s|r|1|1|m|                      */
/*                                                                                      */
/*       i inl-bits   (3)        inline value                                           */
/*       m mrk-bit    (1)        marked/unmarked                                        */
/*       n nbr-bits   (30)       smallint value                                         */
/*       p ptr-bit    (30)       pointer to chunk or pair                               */
/*       q qua-bit    (8)        qualifier value                                        */
/*       r raw-bit    (1)        raw/regular                                            */
/*       s siz-bits   (4/24)     size                                                   */
/*       t tag-bits   (24/4)     tag                                                    */
/*                                                                                      */
/*--------------------------------------------------------------------------------------*/


/*----------------------------------- private macros -----------------------------------*/

#define Max_inline    0x00001FFCu
#define Max_size      0x00FFFFFFu
#define Max_short_tag 0x0000000Fu

#define Inl_mask      0x00000007u
#define Ltg_mask      0x000000FFu
#define Odd_mask      0x00000004u
#define Ptr_mask      0xFFFFFFFCu
#define Raw_mask      0x00000008u
#define Sht_mask      0x0000000Fu
#define Stg_mask      0xFFFFFF0Fu

/*---------------------------------- private constants ---------------------------------*/

static const PTR_type Null = (PTR_type)Memory_Undefined_Value;

enum { Fmt_offset = 2,
       Inl_offset = 2,
       Mrk_offset = 0,
       Lng_offset = 8,
       Qua_offset = 5,
       Raw_offset = 3,
       Sht_offset = 4,
       Smi_offset = 2,
       Typ_offset = 1 };

enum { Pair_width = 2,
       Root_size  = 200 };

enum { Memory_size   = 900000000 };

/*----------------------------------- private types ------------------------------------*/

typedef enum { Format    = 1 << Fmt_offset,
               Long      = 1 << Fmt_offset,
               Short     = 0 << Fmt_offset } FMT_type;

typedef enum { Marking   = 1 << Mrk_offset,
               Marked    = 1 << Mrk_offset,
               Unmarked  = 0 << Mrk_offset } MRK_type;

typedef enum { Raw       = 1 << Raw_offset,
               Regular   = 0 << Raw_offset } RAW_type;

typedef enum { Type      = 1 << Typ_offset,
               Atom      = 1 << Typ_offset,
               Header    = 1 << Typ_offset,
               Pointer   = 0 << Typ_offset } TYP_type;

typedef union CEL { UNS_type uns;
                    PTR_type ptr;  } CEL_type;

/*---------------------------------- private variables ---------------------------------*/

#ifdef BITHEAP
static PTR_type Destination;
static PTR_type Previous;
#endif

static PTR_type First_pair;
static PTR_type Free_chunk;
static PTR_type Free_pair;
static PTR_type Last_pair;

static UNS_type Allocated_chunk_size = 0;
static UNS_type Allocated_pair_size = 0;

static REF_type Force_collect[Root_size];

static UNS_type Root_counter = 0;

static CEL_type Memory[Memory_size ];

/*---------------------------------- private functions ---------------------------------*/

static UNS_type is_long_tag(UNS_type Header)
  { return Header & Format; }

static PTR_type long_tag_header(UNS_type Tag,
                                UNS_type Size,
                                RAW_type Rawness)
  { return (PTR_type)((Tag  << Lng_offset) |
                      (Size << Sht_offset) |
                      Rawness |
                      Long); }

static PTR_type short_tag_header(UNS_type Tag,
                                 UNS_type Size,
                                 RAW_type Rawness)
  { return (PTR_type)((Size << Lng_offset) |
                      (Tag  << Sht_offset) |
                      Rawness |
                      Short); }

static UNS_type get_size(PTR_type Pointer)
  { UNS_type header;
    header = (UNS_type)Pointer;
    if (is_long_tag(header))
      return (header >> Sht_offset) & Sht_mask;
    return header >> Lng_offset; }

static UNS_type get_tag(PTR_type Pointer)
  { UNS_type header;
    header = (UNS_type)Pointer;
    if (is_long_tag(header))
      return header >> Lng_offset;
    return (header >> Sht_offset) & Sht_mask; }

static VOI_type set_tag(PTR_type Pointer,
                        UNS_type Tag)
  { UNS_type header;
    header = Pointer->uns;
    if (is_long_tag(header))
      { header &= Ltg_mask;
        Pointer->uns = header | (Tag << Lng_offset); }
    else
      { header &= Stg_mask;
        Pointer->uns = header | (Tag << Sht_offset); }}

/*--------------------------------------------------------------------------------------*/

static CEL_type cell(PTR_type Pointer,
                     TYP_type Type,
                     MRK_type Mark)
  { return (CEL_type){ (UNS_type)Pointer | Type | Mark }; }

static PTR_type hdr(UNS_type Size)
  { return short_tag_header(0,
                            Size,
                            Regular); }

static UNS_type is_chunk(PTR_type Pointer)
  { return (Pointer < Free_chunk); }

static UNS_type is_odd(PTR_type Pointer)
  { return ((UNS_type)Pointer & Odd_mask); }

static UNS_type is_raw(PTR_type Pointer)
  { return ((UNS_type)Pointer & Raw_mask); }

static UNS_type marker(PTR_type Pointer)
  { return (Pointer->uns & Marking); }

static PTR_type ptr(PTR_type Pointer)
  { return (PTR_type)(Pointer->uns & Ptr_mask); }

static VOI_type set_marked(PTR_type Pointer)
  { Pointer->uns |= Marked; }

static VOI_type set_pointer(PTR_type Reference_pointer,
                            PTR_type New_pointer)
  { *Reference_pointer = (CEL_type){ (UNS_type)New_pointer | Pointer | Unmarked }; }

static VOI_type set_unmarked(PTR_type Pointer)
  { Pointer->uns &= (UNS_type)~Marking; }

static UNS_type type(PTR_type Pointer)
  { if (Pointer->uns <= Max_inline)
      return Atom;
    return (Pointer->uns & Type); }

static UNS_type size(PTR_type Pointer)
  { return get_size(Pointer); }

/*--------------------------------------------------------------------------------------*/

static VOI_type collect(UNS_type);

static UNS_type is_inline(PTR_type Pointer)
  { if ((UNS_type)Pointer & Type)
      return 0;
    return ((UNS_type)Pointer <= Max_inline); }

static UNS_type is_pair(PTR_type Pointer)
  { return (Pointer >= First_pair); }

static UNS_type is_smallint(PTR_type Pointer)
  { return (UNS_type)Pointer & Type; }

static UNS_type get_inline(PTR_type Pointer)
  { return ((UNS_type)Pointer >> Inl_offset) & Inl_mask; }

static UNS_type get_qualifier(PTR_type Pointer)
  { return (UNS_type)Pointer >> Qua_offset; }

static SGN_type get_smallint(PTR_type Pointer)
  { return (SGN_type)Pointer >> Smi_offset; }

static PTR_type make_chunk(UNS_type Tag,
                           UNS_type Size,
                           RAW_type Rawness)
  { PTR_type header,
             pointer;
    if (First_pair <= Size + Free_chunk + 1)
      collect(Size);
    pointer = Free_chunk;
    Free_chunk += Size + 1;
    if (Tag > Max_short_tag)
      header = long_tag_header(Tag,
                               Size,
                               Rawness);
    else
      header = short_tag_header(Tag,
                                Size,
                                Rawness);
    *pointer = cell(header,
                    Header,
                    Unmarked);
    return pointer; }

static PTR_type make_inline(UNS_type Inline,
                            UNS_type Qualifier)
  { return (PTR_type)((Qualifier << Qua_offset) |
                      (Inline << Inl_offset) |
                      Pointer |
                      Unmarked); }

#ifdef BITHEAP

static PTR_type make_pair(VOI_type)
  { PTR_type free,
             pointer,
             previous;
    if (Free_pair == Null)
      if (First_pair <= Pair_width + Free_chunk + 1)
        { collect(Pair_width);
          return make_pair(); }
      else
        return First_pair = First_pair - Pair_width;
    pointer = Free_pair;
    free = ptr(pointer);
    previous = ptr(pointer + 1);
    if (pointer == previous)
      Free_pair = free;
    else
      { set_pointer(pointer - 2, free);
        set_pointer(pointer - 1, previous);
        Free_pair -= Pair_width; }
    return pointer; }

#else

static PTR_type make_pair(VOI_type)
  { PTR_type pointer;
    if (Free_pair == Null)
      if (First_pair <= Pair_width + Free_chunk + 1)
        { collect(Pair_width);
          return make_pair(); }
      else
        return First_pair = First_pair - Pair_width;
    pointer = Free_pair;
    Free_pair = ptr(Free_pair);
    return pointer; }

#endif

static PTR_type make_smallint(SGN_type Signed)
  { return (PTR_type)((Signed << Smi_offset) |
                      Atom |
                      Unmarked); }

/*-------------------------------------- collect ---------------------------------------*/

static VOI_type mark_thread(VOI_type)
  { PTR_type C, C_, C__,
             P, P_;
    P = Null;                                                // P <- null
    for (C = Memory;;)                                       // C <- memory
      { C_ = ptr(C);                                         // C^
        if (marker(C) == Marked)                             // Cv = m
          if (C_ == Memory)                                  // C^ = memory
            return;                                          // stop
          else                                               // C^ ≠ memory
            if (is_chunk(C))                                 // chunk?(C)
              if (is_chunk(C_))                              // chunk?(C^)
                C = C_ - 1;                                  // C <- C^ - 1
              else                                           // pair?(C^)
                C = C_;                                      // C <- C^
            else                                             // pair?(C)
              if (is_odd(C))                                 // odd?(C)
                { P_ = ptr(P);                               // P^
                  if (is_chunk(P))                           // chunk?(P)
                    { *P = cell(C - 1, Pointer, Unmarked);   // *P <- [C - 1, p, u]
                      C = P - 1;  }                          // C = P - 1
                  else                                       // pair?(P)
                    { *P = cell(C - 1, Pointer, Marked);     // *P <- [C - 1 ,p, m]
                      C = P; }                               // C = P
                  P = P_; }                                  // P <- P^
              else                                           // even?(C)
                C += 1;                                      // C <- C + 1
        else                                                 // Cv = u
          { if (type(C) == Atom)                             // Cw = a
              if (is_chunk(C))                               // chunk?(C)
                C -= 1;                                      // C <- C - 1
              else                                           // pair?(C)
                set_marked(C);                               // Cv <- m
            else                                             // Cw = p
              if (is_chunk(C_))                              // chunk?(C^)
                { C__ = ptr(C_);                             // C^^
                  if (marker(C_) == Marked)                  // C^v = m
                    { *C = *C__;                             // *C <- *C^^
                      *C__ = cell(C, Pointer, Marked);       // *C^^ = [C, p, m]
                      if (is_chunk(C))                       // chunk?(C)
                        C -= 1;                              // C <- C - 1
                      else; }                                // pair?(C)
                  else                                       // C^v = u
                    {
                      #ifdef BITHEAP
                      Bitheap_Set((UNS_type)C_);
                      #endif
                      *C = cell(C__, Header, Marked);        // *C <- [C^^, h, m]
                      *C_ = cell(C, Pointer, Marked);        // *C^ = [C, p, m]
                      if (is_raw(C__))                       // raw?(C^^)
                        if (is_chunk(C))                     // chunk?(C)
                          C -= 1;                            // C <- C - 1
                        else
                          continue;                          // pair?(C)
                      else                                   // regular?(C^^)
                        C = C_ + size(C__); }}               // C <- C^ + size(C^^)
              else                                           // pair?(C^)
                {
                  #ifdef BITHEAP
                  Bitheap_Set((UNS_type)C_);
                  #endif
                  *C = cell(P, Pointer, Marked);             // *C <- [P, p, m]
                  P = C;                                     // P <- C
                  C = C_; }}}}                               // C <- C^

/*--------------------------------------------------------------------------------------*/

#ifdef BITHEAP

static C_JMP_BUF Exit_unthread;

static VOI_type unthread_lambda(UNS_type Address)
  { PTR_type A, S, S_;
    A = (PTR_type)Address;
    if (A >= Free_chunk)
      C_LONGJMP(Exit_unthread, !0);
    for (S = ptr(A);                                  // S <- A^
         type(S) == Pointer;                          // Sw = p
         S = S_)                                      // S <- S^
      { S_ = ptr(S);                                  // S^
        if (is_chunk(S))                              // chunk?(S)
          *S = cell(Destination, Pointer, Unmarked);  // *S <- [dest, p, u]
        else                                          // pair?(S)
          *S = cell(Destination, Pointer, Marked); }  // *S <- [dest, p, m]
    S_ = ptr(S);                                      // S^
    *A = *S;                                          // *A <- *S
    if (is_chunk(S))                                  // chunk?(S)
      *S = cell(Destination, Pointer, Unmarked);      // *S <- [dest, p, u]
    else                                              // pair?(S)
      *S = cell(Destination, Pointer, Marked);        // *S <- [dest, p, m]
    Destination += size(S_) + 1; }                    // dest <- dest + size(S^) + 1

static VOI_type unthread(VOI_type)
  { Destination = Memory + 1;
    if (C_SETJMP(Exit_unthread) == 0)
      Bitheap_Do(unthread_lambda,
                 Leave_Bits); }

#else

static VOI_type unthread(VOI_type)
  { PTR_type A, D, S, S_;
    UNS_type C, L;
    C = 0;                                                   // C <- 0
    A = Null;                                                // A <- null
    *Free_chunk = cell(Null, Pointer, Marked);               // *free <- [null, p, m]
    for (S = D = Memory + 1;;)                               // S <- D <- memory + 1
      { S_ = ptr(S);                                         // S^
        if (A == Null)                                       // A = null
          if (marker(S) == Marked)                           // Sv = m
            if (S < Free_chunk)                              // S < free
              { A = S;                                       // A <- S
                S = S_;                                      // S <- S^
                C = 0; }                                     // C <- 0
            else                                             // S = free
              return;                                        // stop
          else                                               // Sv = u
            { L = size(S_) + 1;                              // size(S^) + 1
              A = S;                                         // A <- S
              S += L;                                        // S <- S + size(S^) + 1
              C = L; }                                       // C <- size(S^) + 1
        else                                                 // A ≠ null
          if (C == 0)                                        // C = 0
            if (type(S) == Header)                           // Sw = h
              { L = size(S_) + 1;                            // size(S^) + 1
                *A = *S;                                     // *A <- *S
                if (is_chunk(S))                             // chunk?(S)
                  *S = cell(D, Pointer, Unmarked);           // *S <- [D, p, u]
                else                                         // pair?(S)
                  *S = cell(D, Pointer, Marked);             // *S <- [D, p, m]
                S = A + L;                                   // S <- A + size(S^) + 1
                D += L;                                      // D <- D + size(S^) + 1
                A = Null; }                                  // A <- null
            else                                             // Sw = p
              { if (is_chunk(S))                             // chunk?(S)
                  *S = cell(D, Pointer, Unmarked);           // *S <- [D, p, u]
                else                                         // pair?(S)
                  *S = cell(D, Pointer, Marked);             // *S <- [D, p, m]
                S = S_; }                                    // S <- S^
          else                                               // C > 0
            if (marker(S) == Marked)                         // Sv = m
              { *A = cell(hdr(C - 1), Header, Unmarked);     // *A <- [hdr(C-1), h, u]
                A = Null; }                                  // A <- null
            else                                             // Sv = u
              { L = size(S_) + 1;                            // size(S^)
                if (C + L > Max_size)                        // C + size(S^) ≥ max
                  { *A = cell(hdr(C - 1), Header, Unmarked); // *A <- [hdr(C-1), h, u]
                    A = Null; }                              // A <- null
                else                                         // C + size(S^) < max
                  { S += L;                                  // S <- S + size(S^) + 1
                    C += L; }}}}                             // C <- C + size(S^) + 1
  #endif

/*--------------------------------------------------------------------------------------*/

#ifdef BITHEAP

static VOI_type compact_link_lambda(UNS_type Address)
  { PTR_type S, S_;
    UNS_type L;
    S = (PTR_type)Address;
    if (S < Free_chunk)
      { S_ = ptr(S);                                             // S^
        Allocated_chunk_size += size(S_) + 1;                    // REPORTING
        *Destination = cell(S_, Header, Unmarked);               // *dest <- [S^, h, u]
        for (L = size(S_);                                       // L <- size(S^)
             L > 0;                                              // L > 0
             L -= 1)                                             // L <- L - 1
          { S += 1;                                              // S <- S + 1
            Destination += 1;                                    // dest <- dest + 1
            *Destination = *S; }                                 // *dest <- S
        Destination += 1; }                                      // dest <- dest + 1
    else
      { Allocated_pair_size += Pair_width;                       // REPORTING
        if (Previous == Null)                                    // previous = null
          First_pair = S;                                        // first <- S
        else                                                     // previous ≠ null
          if (S > Previous + Pair_width)                         // S > previous + 2
            { set_pointer(S - 2, Free_pair);                     // (S - 2)^ <- free
              set_pointer(S - 1, Previous + Pair_width);         // (S - 1)^ <- previous + 2
              Free_pair = S - Pair_width; }                      // free <- S - 2
          else;                                                  // S = previous + 2
        set_unmarked(S);                                         // Sv <- u
        set_unmarked(S + 1);                                     // (S + 1)v <- u
        Previous = S; }}                                         // previous <- S

static VOI_type compact_link(VOI_type)
  { Allocated_chunk_size = 0;
    Destination = Memory + 1;
    Previous = Free_pair = Null;
    Allocated_pair_size = 0;
    Bitheap_Do(compact_link_lambda,
               Clear_Bits);
    Free_chunk = Destination; }

#else

static VOI_type compact_link(VOI_type)
  { PTR_type D, S, S_;
    UNS_type C;
    Allocated_chunk_size = 0;                                // REPORTING
    C = 0;                                                   // C <- 0
    *Free_chunk = cell(Null, Pointer, Marked);               // *free = [null, p, m]
    for (S = D = Memory + 1;;                                // S <- D <- memory + 1
         S += 1)                                             // S <- S + 1
      { S_ = ptr(S);                                         // S^
        if (C == 0)                                          // C = 0
          if (marker(S) == Marked)                           // Sv = m
            if (S < Free_chunk)                              // S < free
              { *D = cell(S_, Header, Unmarked);             // *D <- [S^, h, u]
                C = size(S_);                                // C <- size(S^)
                Allocated_chunk_size += C + 1;               // REPORTING
                D += 1; }                                    // D <- D + 1
            else                                             // S = free
              { Free_chunk = D;                              // free <- D
                break; }                                     // exit
          else                                               // Sv = u
            S += size(S_);                                   // S <- S + size(S^)
         else                                                // C > 0
            { *D = *S;                                       // *D <- *S
              C -= 1;                                        // C <- C - 1
              D += 1; }}                                     // D <- D + 1
    Allocated_pair_size = 0;                                 // REPORTING
    S = First_pair;                                          // S <- first
    *Last_pair = cell(Null, Pointer, Marked);                // *last = [null, p, m]
    First_pair = Null;                                       // first <- null
    for (Free_pair = Null;;                                  // free <- null
         S += Pair_width)                                    // S <- S + 2
      if (marker(S) == Marked)                               // Sv = m
        { Allocated_pair_size += Pair_width;                 // REPORTING
          if (First_pair == Null)                            // first = null
            First_pair = S;                                  // first <- S
          else;                                              // first ≠ null
          if (S < Last_pair)                                 // S < last
            { set_unmarked(S);                               // Sv <- u
              set_unmarked(S + 1); }                         // (S + 1)v <- u
          else                                               // S = last
            return; }                                        // stop
      else                                                   // Sv = u
        if (First_pair != Null)                              // first ≠ null
          { *S = cell(Free_pair, Pointer, Unmarked);         // *S = [free, p, u]
            Free_pair = S; }                                 // free <- S
        else; }                                              // first = null

  #endif

/*--------------------------------------------------------------------------------------*/

static VOI_type collect(UNS_type Size)
  { FLO_type start_time,
             stop_time;
    PTR_type root_pointer;
    REF_type reference;
    UNS_type index;
    UNS_type allocated_chunk_size = Allocated_chunk_size;
    UNS_type allocated_pair_size = Allocated_pair_size;
    root_pointer = Memory->ptr;
    for (index = 0;
         index < Root_counter;
         index++)
      { reference = Force_collect[index];
        root_pointer += 1;
        root_pointer->ptr = *reference; }
    printf("\nCollecting garbage ...\n");
    start_time = (FLO_type)C_CLOCK;
    mark_thread();
    stop_time = (FLO_type)C_CLOCK;
    printf("   mark-thread:    %#.6f\n", stop_time - start_time);
    start_time = stop_time;
    unthread();
    stop_time = (FLO_type)C_CLOCK;
    printf("   unthread:       %#.6f\n", stop_time - start_time);
    start_time = stop_time;
    compact_link();
    stop_time = (FLO_type)C_CLOCK;
    printf("   compact-link:   %#.6f\n", stop_time - start_time);
    root_pointer = Memory->ptr;
    for (index = 0;
         index < Root_counter;
         index++)
      { reference = Force_collect[index];
        root_pointer += 1;
        *reference = root_pointer->ptr; }
    printf("used chunk space before GC: %10u(%05.2f%%)\n",
           allocated_chunk_size, (FLO_type)allocated_chunk_size/Memory_size*100.0);
    printf("used chunk space after  GC: %10u(%05.2f%%)\n",
           Allocated_chunk_size, (FLO_type)Allocated_chunk_size/Memory_size*100.0);
    printf("used  pair space before GC: %10u(%05.2f%%)\n",
           allocated_pair_size,  (FLO_type)allocated_pair_size /Memory_size*100.0);
    printf("used  pair space after  GC: %10u(%05.2f%%)\n",
           Allocated_pair_size,  (FLO_type)Allocated_pair_size /Memory_size*100.0);
    
    if (First_pair <= Size + Free_chunk + 1)
      Main_Error(MMO_error,
                 ""); }

/*-------------------------------------- replace ---------------------------------------*/

static void substitute_mark(PTR_type Match,
                            PTR_type Replacement)
  { PTR_type C, C_, C__, P, P_;
    P = Null;                                                // P <- null
    for (C = Memory;;)                                       // C <- memory
      { C_ = ptr(C);                                         // C^
        if (marker(C) == Marked)                             // Cv = m
          if (is_chunk(C))                                   // chunk?(C)
            { *C = *C_;                                      // *C <- *C^
              *C_ = cell(C, Pointer, Marked);                // *C^ <- [C, p, m]
              if (is_chunk(C_))                              // chunk?(C^)
                if (C_ == Memory)                            // C^ = memory
                  break;                                     // stop
                else                                         // C^ ≠ memory
                  C = C_ - 1;                                // C <- C^ - 1
                else                                         // pair?(C^)
                  C = C_; }                                  // C <- C^
          else                                               // pair?(C)
            if (is_odd(C))                                   // odd?(C)
              { P_ = ptr(P);                                 // P^
                *P = cell(C - 1, Pointer, Unmarked);         // *P <- [C - 1, p, m]
                if (is_chunk(P))                             // chunk?(P)
                  C = P - 1;                                 // C <- P - 1
                else                                         // pair?(P)
                  C = P;                                     // C <- P
                P = P_; }                                    // P <- P^
            else                                             // even?(C)
              C += 1;                                        // C <- C + 1
            else                                             // Cv = u
              if (type(C) == Atom)                           // Cw = a
                { set_marked(C);                             // Cv <- m
                  if (is_chunk(C))                           // chunk?(C)
                    C -= 1;                                  // C <- C - 1
                  else; }                                    // pair?(C)
              else                                           // Cw = p
                if (C_ == Match)                             // C^ = M
                  *C = cell(Replacement, Pointer, Unmarked); // C <- R;
                else                                         // C^ ≠ M
                  if (is_chunk(C_))                          // chunk?(C^)
                    { set_marked(C);                         // Cv <- m
                      if (marker(C_) == Marked)              // C^v = m
                        if (is_chunk(C))                     // chunk?(C)_
                          C -= 1;                            // C <- C - 1
                        else;                                // pair?(C)
                      else                                   // C^v = u
                        { C__ = ptr(C_);                     // C^^
                          set_marked(C_);                    // C^v <- m
                          if (is_raw(C__))                   // raw?(C^^)
                            if (is_chunk(C))                 // chunk?(C)
                              C -= 1;                        // C <- C - 1
                            else;                            // pair?(C)
                          else                               // regular?(C^)
                            { *C = cell(C__, Header, Marked);// *C <- [C^^, h, m]
                              *C_ = cell(C, Pointer, Marked);// *C^ <- [C, p, m]
                              C = C_ + size(C__); }}}        // C <- C^ + size(C^^)
                  else                                       // pair?(C^)
                    { *C = cell(P, Pointer, Marked);         // *C <- [P, p, m]
                      P = C;                                 // P <- C
                      C = C_; }}}                            // C <- C^

/*--------------------------------------------------------------------------------------*/

static void substitute_unmark(void)
  { PTR_type C, C_, C__, P, P_;
    P = Null;                                                 // P <- null
    for (C = Memory;;)                                        // C <- memory
      { C_ = ptr(C);                                          // C^
        if (marker(C) == Unmarked)                            // Cv = u
          if (is_chunk(C))                                    // chunk?(C)
            { *C = *C_;                                       // *C <- *C^
              *C_ = cell(C, Pointer, Unmarked);               // *C^ <- [C, p, u]
              if (is_chunk(C_))                               // chunk?(C^)
                if (C_ == Memory)                             // C^ = memory
                  break;                                      // stop
                else                                          // C^ ≠ memory
                  C = C_ - 1;                                 // C <- C^ - 1
                else                                          // pair?(C^)
                  C = C_; }                                   // C <- C^
          else                                                // pair?(C)
            if (is_odd(C))                                    // odd?(C)
              { P_ = ptr(P);                                  // P^
                *P = cell(C - 1, Pointer, Unmarked);          // *P <- [C - 1, p, u]
                if (is_chunk(P))                              // chunk?(P)
                  C = P - 1;                                  // C <- P - 1
                else                                          // pair?(P)
                  C = P;                                      // C <- P
                P = P_; }                                     // P <- P^
            else                                              // even?(C)
              C += 1;                                         // C <- C + 1
            else                                              // Cv = m
              if (type(C) == Atom)                            // Cw = a
                { set_unmarked(C);                            // Cv <- m
                  if (is_chunk(C))                            // chunk?(C)
                    C -= 1;                                   // C <- C - 1
                  else; }                                     // pair?(C)
              else                                            // Cw = p
                if (is_chunk(C_))                             // chunk?(C^)
                  { set_unmarked(C);                          // Cv <- u
                    if (marker(C_) == Unmarked)               // C^v = u
                      if (is_chunk(C))                        // chunk?(C)_
                        C -= 1;                               // C <- C - 1
                      else;                                   // pair?(C)
                    else                                      // C^v = u
                      { C__ = ptr(C_);                        // C^^
                        set_unmarked(C_);                     // Cv <- u
                        if (is_raw(C__))                      // raw?(C^^)
                          if (is_chunk(C))                    // chunk?(C)
                            C -= 1;                           // C <- C - 1
                          else;                               // pair?(C)
                        else                                  // regular?(C^)
                          { *C = cell(C__, Header, Unmarked); // *C <- [C^^, h, u]
                            *C_ = cell(C, Pointer, Unmarked); // *C^ <- [C, p, u]
                            C = C_ + size(C__); }}}           // C <- C^ + size(C^^)
                else                                          // pair?(C^)
                  { *C = cell(P, Pointer, Unmarked);          // *C <- [P, p, u]
                    P = C;                                    // P <- C
                    C = C_; }}}                               // C <- C^

/*--------------------------------------------------------------------------------------*/

static VOI_type replace(PTR_type Match,
                        PTR_type Replacement)
  { FLO_type start_time,
             stop_time;
    if (Match != Replacement)
      { printf("\nReplacing pointer ...\n");
        start_time = (FLO_type)C_CLOCK;
        substitute_mark(Match,
                        Replacement);
        stop_time = (FLO_type)C_CLOCK;
        printf("   mark-replace:   %#.6f\n", stop_time - start_time);
        start_time = stop_time;
        substitute_unmark();
        stop_time = (FLO_type)C_CLOCK;
        printf("   unmark-restore: %#.6f\n", stop_time - start_time); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type register_(ADR_type Address)
  { C_ASSERT(Root_counter < Root_size);
    Force_collect[Root_counter++] = Address; }

/*---------------------------------- public functions ----------------------------------*/

UNS_type Memory_Available(VOI_type)
  { return (UNS_type)(First_pair - Free_chunk - 1); }

PTR_type Memory_Chunk_M(UNS_type Tag,
                        UNS_type Size)
  { Allocated_chunk_size += Size + 1;
    return make_chunk(Tag,
                      Size,
                      Regular); }

VOI_type Memory_Collect_C(VOI_type)
  { collect(0); }

UNS_type Memory_Get_Inline(PTR_type Pointer)
  { return get_inline(Pointer); }

UNS_type Memory_Get_Qualifier(PTR_type Pointer)
  { return get_qualifier(Pointer); }

UNS_type Memory_Get_Size(PTR_type Pointer)
  { return get_size(Pointer->ptr); }

SGN_type Memory_Get_Smallint(PTR_type Pointer)
  { return get_smallint(Pointer); }

UNS_type Memory_Get_Tag(PTR_type Pointer)
  { return get_tag(Pointer->ptr); }

VOI_type Memory_Increment(REF_type Reference)
  { *Reference += 1; }

PTR_type Memory_Inline(UNS_type Inline,
                       UNS_type Qualifier)
  { return make_inline(Inline,
                       Qualifier); }

UNS_type Memory_Is_Inline(PTR_type Pointer)
  { return is_inline(Pointer); }

UNS_type Memory_Is_Pair(PTR_type Pointer)
  { return is_pair(Pointer); }

UNS_type Memory_Is_Smallint(PTR_type Pointer)
  { return is_smallint(Pointer); }

PTR_type Memory_Pair_M(VOI_type)
  { Allocated_pair_size += Pair_width;
    return make_pair(); }

PTR_type Memory_Raw_Chunk_M(UNS_type Tag,
                            UNS_type Size)
  { if (First_pair <= Size + Free_chunk + 1)
      collect(Size);
    Allocated_chunk_size += Size + 1;
    return make_chunk(Tag,
                      Size,
                      Raw); }

VOI_type Memory_Register(ADR_type Address)
  { register_(Address); }

VOI_type Memory_Replace(PTR_type Match,
                        PTR_type Replacement)
  { replace(Match,
            Replacement); }

VOI_type Memory_Set_Tag(PTR_type Pointer,
                        UNS_type Tag)
  { set_tag(Pointer,
            Tag); }

PTR_type Memory_Smallint(SGN_type Signed)
  { return make_smallint(Signed); }

/*----------------------------------- initialization -----------------------------------*/

VOI_type Memory_Initialize(VOI_type)
  { C_ASSERT(sizeof(UNS_type) == Memory_Cell_Width);
    C_ASSERT(sizeof(LNG_type) == 2*Memory_Cell_Width);
    C_ASSERT(sizeof(FLO_type) == 2*Memory_Cell_Width);
    C_ASSERT(sizeof(SGN_type) == Memory_Cell_Width);
    C_ASSERT(sizeof(CEL_type) == Memory_Cell_Width);
    C_ASSERT(Memory > (PTR_type)Max_inline);
    C_ASSERT(-1>>1 < 0);
    #ifdef BITHEAP
    Bitheap_Initialize();
    #endif
    Free_chunk = Memory + 1;
    Last_pair  = &Memory[Memory_size - 1];
    if (is_odd(Last_pair))
      Last_pair -= 1;
    First_pair = Last_pair - Pair_width;
    Free_pair  = Null;
    Memory->ptr = make_chunk(0,
                             Root_size,
                             Regular); }

