                     /*---------------------------------------------*/
                     /*                  >>>Slip<<<                 */
                     /*                 Theo D'Hondt                */
                     /*          VUB Software Languages Lab         */
                     /*                  (c) 2018                   */
                     /*---------------------------------------------*/
                     /*                final version                */
                     /*---------------------------------------------*/
                     /*                   natives                   */
                     /*---------------------------------------------*/

#include "SlipPlatform.h"

#include "SlipMain.h"

#include "SlipGrammar.h"

#include "SlipCompile.h"
#include "SlipDictionary.h"
#include "SlipEnvironment.h"
#include "SlipEvaluate.h"
#include "SlipPool.h"
#include "SlipPrint.h"
#include "SlipRead.h"
#include "SlipThread.h"

#include "SlipNatives.h"

/* Scheme compatibility

STANDARD
YES -
YES *
YES /
YES +
YES <
YES <=
YES =
YES >
YES >=
YES abs
YES acos
NO  angle
YES append → binary version only
YES apply → single argument list only
YES asin
YES assoc
YES assq
NO  assv
YES atan → single argument form only
YES boolean?
YES caaaar
YES caaadr
YES caaar
YES caadar
YES caaddr
YES caadr
YES caar
YES cadaar
YES cadadr
YES cadar
YES caddar
YES cadddr
YES caddr
YES cadr
YES call-with-current-continuation
NO  call-with-input-file
NO  call-with-output-file
YES car
YES cdaaar
YES cdaadr
YES cdaar
YES cdadar
YES cdaddr
YES cdadr
YES cdar
YES cddaar
YES cddadr
YES cddar
YES cdddar
YES cddddr
YES cdddr
YES cddr
YES cdr
NO  ceiling
YES char->integer
NO  char-alphabetic?
NO  char-ci<?
NO  char-ci<=?
NO  char-ci=?
NO  char-ci>?
NO  char-ci>=?
NO  char-downcase
NO  char-lower-case?
NO  char-numeric?
NO  char-ready?
NO  char-upcase
NO  char-upper-case?
NO  char-whitespace?
YES char?
YES char<?
YES char<=?
YES char=?
YES char>?
YES char>=?
YES close-input-port
YES close-output-port
NO  complex?
YES cons
YES cos
NO  current-input-port
NO  current-output-port
NO  denominator
YES display
YES eof-object?
YES eq?
YES equal?
NO  eqv?
NO  eval → special form <evaluate>
YES even?
NO  exact->inexact
NO  exact?
YES exp
YES expt
NO  floor
YES for-each
YES force
NO  gcd
NO  imag-part
NO  inexact->exact
NO  inexact?
YES input-port?
NO  integer->char
YES integer?
YES length
YES list
NO  list->string
YES list->vector
NO  list-ref
NO  list-tail
NO  load → special form
YES log
NO  magnitude
NO  make-polar
NO  make-rectangular
YES make-string
YES make-vector
YES map
YES max
YES member
YES memq
NO  memv
YES min
YES modulo
YES negative?
YES newline
YES not
YES null?
YES number->string
YES number?
NO  numerator
YES odd?
YES open-input-file
YES open-output-file
YES output-port?
YES pair?
NO  peek-char
YES positive?
YES procedure?
YES quotient
NO  rational?
NO  rationalize
YES read
YES read-char
NO  real-part
YES real?
YES remainder
YES reverse
YES round
NO  scheme-report-environment
YES set-car!
YES set-cdr!
YES sin
YES sqrt
NO  string
NO  string->list
YES string->number
YES string->symbol
YES string-append
NO  string-ci<?
NO  string-ci<=?
NO  string-ci=?
NO  string-ci>?
NO  string-ci>=?
YES string-copy
NO  string-fill!
YES string-length
YES string-ref
YES string-set!
YES string?
YES string<?
YES string<=?
YES string=?
YES string>?
YES string>=?
YES substring
YES symbol->string
YES symbol?
YES tan
NO  transcript-off
NO  transcript-on
YES truncate
YES vector
YES vector->list
NO  vector-fill!
YES vector-length
YES vector-ref
YES vector-set!
YES vector?
NO  with-input-from-file
NO  with-output-to-file
YES write
YES write-char
YES zero?

STREAMS
stream-null?
the-empty-stream
cons-stream
stream-car
stream-cdr

EXTENSIONS
YES circularity-level
YES clock
YES collect
YES exit
YES false
YES format
YES gensym
YES random
YES true

*/

/*----------------------------------- private macros -----------------------------------*/

#define NULL_PAIR        (PAI_type)Unique_Null

#define FALSE_EXPRESSION (EXP_type)Unique_False
#define TRUE_EXPRESSION  (EXP_type)Unique_True

#define RETURN_BOOLEAN(TYP, EXP)                                                         \
  Evaluate_Return((is_##TYP(value_expression)? TRUE_EXPRESSION : FALSE_EXPRESSION))

#define RETURN_2_BOOLEAN(TP1, TP2, EXP)                                                  \
  Evaluate_Return((is_##TP1(value_expression) ||                                         \
                   is_##TP2(value_expression)? TRUE_EXPRESSION : FALSE_EXPRESSION))

#define RETURN_4_BOOLEAN(TP1, TP2, TP3, TP4, EXP)                                        \
  Evaluate_Return((is_##TP1(value_expression) ||                                         \
                   is_##TP2(value_expression) ||                                       \
                   is_##TP3(value_expression) ||                                       \
                   is_##TP4(value_expression)? TRUE_EXPRESSION : FALSE_EXPRESSION))

/*--------------------------------- private constants ----------------------------------*/

enum { Initial_jump     = 0,
       Terminate_jump   = 1 };

enum { Space_char       = ' ' };

static const TXT_type Default_input_port_text  = "default-input";
static const TXT_type Default_output_port_text = "default-output";

/*--------------------------------- private variables ----------------------------------*/

static C_JMP_BUF Exit_point;

/*-------------------------------- protected variables ---------------------------------*/

static EXP_type Car_Expression_P;
static EXP_type Cdr_Expression_P;
static SYM_type Symbol_P;
static IPT_type Default_Input_Port_P;
static OPT_type Default_Output_Port_P;
static EXP_type Expression_P;
static FRM_type Frame_P;
static IPT_type Input_Port_P;
static EXP_type List_Expression_P;
static OPT_type Output_Port_P;
static PAI_type Pair_P;
static STR_type String_P;
static VEC_type Vector_P;

/*--------------------------------------------------------------------------------------*/
/*------------------------------------- auxiliary --------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*----------------------------------- error handling -----------------------------------*/

static VOI_type error(UNS_type Error,
                      TXT_type Message_text)
  { Main_Error(Error,
               Message_text); }

/*------------------------------------ name binding ------------------------------------*/

static VOI_type enter_native_name(UNS_type Offset,
                                  EXP_type Value_expression)
  { Environment_Register_Native(Value_expression,
                                Offset); }

static VOI_type native_define_M(TXT_type Text,
                                NFN_type Native_thunk_M)
  { NAT_type native;
    SYM_type symbol;
    UNS_type length,
             offset;
    symbol = Pool_Enter_Native_M(Text);
    offset = Dictionary_Define_Variable_M(symbol);
    C_ASSERT(offset > 0);
    length = C_STRLEN(Text);
    native = make_NAT_M(length);
    native->nfn = Native_thunk_M;
    C_STRCPY(native->txt, Text);
    enter_native_name(offset,
                      native); }

static VOI_type higher_order_native_define_M(TXT_type Text,
                                             NFN_type Native_thunk_M,
                                             NFN_type Native_thunk_T_M)
  { NAZ_type higher_order_native;
    SYM_type symbol;
    UNS_type length,
             offset;
    symbol = Pool_Enter_Native_M(Text);
    offset = Dictionary_Define_Variable_M(symbol);
    C_ASSERT(offset > 0);
    length = C_STRLEN(Text);
    higher_order_native = make_NAZ_M(length);
    higher_order_native->nfn = Native_thunk_M;
    higher_order_native->nfz = Native_thunk_T_M;
    C_STRCPY(higher_order_native->txt, Text);
    enter_native_name(offset,
                      higher_order_native); }

/*------------------------------------ constructors ------------------------------------*/

static SMI_type make_smallint(LNG_type Long,
                              TXT_type Message_text)
  { SGN_type int_;
    SMI_type smallint;
    if (C_LLABS(Long) > Maximum_Smallint_Size)
      error(NTL_error,
            Message_text);
    int_ = (SGN_type)Long;
    smallint = make_SMI(int_);
    return smallint; }

static REA_type validate_float_M(FLO_type Float,
                                 TXT_type Message_text)
  { REA_type real;
    if (isnan(Float))
      error(OOR_error,
            Message_text);
    if (isinf(Float))
      error(OOR_error,
            Message_text);
    real = make_REA_M();
    real->flo = Float;
    return real; }

/*------------------------------- argument list parsing --------------------------------*/

static VOI_type require_0_arguments(FRM_type Frame,
                                    TXT_type Message_text)
  { UNS_type number_of_arguments;
    number_of_arguments = Environment_Frame_Size(Frame);
    if (number_of_arguments != 0)
      error(EX0_error,
            Message_text); }

static VOI_type require_0_or_1_arguments(FRM_type Frame,
                                         EXp_type Reference,
                                         TXT_type Message_text)
  { UNS_type number_of_arguments;
    number_of_arguments = Environment_Frame_Size(Frame);
    if (number_of_arguments > 1)
      error(AM1_error,
            Message_text);
    if (number_of_arguments == 0)
      *Reference = Unique_Undefined;
    else
      *Reference = Frame[1];
    Environment_Release_Frame(Frame); }

static VOI_type require_1_argument(FRM_type Frame,
                                   EXp_type Reference,
                                   TXT_type Message_text)
  { UNS_type number_of_arguments;
    number_of_arguments = Environment_Frame_Size(Frame);
    if (number_of_arguments != 1)
      error(EX1_error,
            Message_text);
    *Reference = Frame[1];
    Environment_Release_Frame(Frame); }

static VOI_type require_1_or_2_arguments(FRM_type Frame,
                                         EXp_type First_reference,
                                         EXp_type Second_reference,
                                         TXT_type Message_text)
  { UNS_type number_of_arguments;
    number_of_arguments = Environment_Frame_Size(Frame);
    switch (number_of_arguments)
      { case 0:
          error(AL1_error,
                Message_text);
        case 1:
          *First_reference = Frame[1];
          *Second_reference = Unique_Undefined;
          break;
        case 2:
          *First_reference = Frame[1];
          *Second_reference = Frame[2];
          break;
        default:
          error(AM2_error,
                Message_text); }
    Environment_Release_Frame(Frame); }

static VOI_type require_2_arguments(FRM_type Frame,
                                    EXp_type First_reference,
                                    EXp_type Second_reference,
                                    TXT_type Message_text)
  { UNS_type number_of_arguments;
    number_of_arguments = Environment_Frame_Size(Frame);
    if (number_of_arguments != 2)
      error(EX2_error,
            Message_text);
    *First_reference  = Frame[1];
    *Second_reference = Frame[2];
    Environment_Release_Frame(Frame); }

static VOI_type require_3_arguments(FRM_type Frame,
                                    EXp_type First_reference,
                                    EXp_type Second_reference,
                                    EXp_type Third_reference,
                                    TXT_type Message_text)
  { UNS_type number_of_arguments;
    number_of_arguments = Environment_Frame_Size(Frame);
    if (number_of_arguments != 3)
      error(EX3_error,
            Message_text);
    *First_reference  = Frame[1];
    *Second_reference = Frame[2];
    *Third_reference  = Frame[3];
    Environment_Release_Frame(Frame); }

/*-------------------------------------- casting ---------------------------------------*/

static EXP_type cast_to_car(EXP_type Expression,
                            TXT_type Message_text)
  { EXP_type value_expression;
    PAI_type pair;
    if (!is_PAI(Expression))
      error(NAP_error,
            Message_text);
    pair = (PAI_type)Expression;
    value_expression = pair->car;
    return value_expression; }

static EXP_type cast_to_cdr(EXP_type Expression,
                            TXT_type Message_text)
  { EXP_type value_expression;
    PAI_type pair;
    if (!is_PAI(Expression))
      error(NAP_error,
            Message_text);
    pair = (PAI_type)Expression;
    value_expression = pair->cdr;
    return value_expression; }

static CHA_type cast_to_character(EXP_type Expression,
                                  TXT_type Message_text)
  { CHA_type character;
    if (!is_CHA(Expression))
      error(NAC_error,
            Message_text);
    character = (CHA_type)Expression;
    return character; }

static IPT_type cast_to_input_port(EXP_type Expression,
                                   TXT_type Message_text)
  { IPT_type input_port;
    if (!is_IPT(Expression))
      error(NIP_error,
            Message_text);
    input_port = (IPT_type)Expression;
    return input_port; }

static OPT_type cast_to_output_port(EXP_type Expression,
                                    TXT_type Message_text)
  { OPT_type output_port;
    if (!is_OPT(Expression))
      error(NOP_error,
            Message_text);
    output_port = (OPT_type)Expression;
    return output_port; }

static PAI_type cast_to_pair(EXP_type Expression,
                             TXT_type Message_text)
  { PAI_type pair;
    if (!is_PAI(Expression))
      error(NAP_error,
            Message_text);
    pair = (PAI_type)Expression;
    return pair; }

static SMI_type cast_to_smallint(EXP_type Expression,
                                 TXT_type Message_text)
  { SMI_type smallint;
    if (!is_SMI(Expression))
      error(NAN_error,
            Message_text);
    smallint = (SMI_type)Expression;
    return smallint; }

static STR_type cast_to_string(EXP_type Expression,
                               TXT_type Message_text)
  { STR_type string;
    if (!is_STR(Expression))
      error(NAS_error,
            Message_text);
    string = (STR_type)Expression;
    return string; }

static SYM_type cast_to_symbol(EXP_type Expression,
                               TXT_type Message_text)
  { SYM_type symbol;
    if (!is_SYM(Expression))
      error(NSY_error,
            Message_text);
    symbol = (SYM_type)Expression;
    return symbol; }

static VEC_type cast_to_vector(EXP_type Expression,
                               TXT_type Message_text)
  { VEC_type vector;
    if (!is_VEC(Expression))
      error(NAV_error,
            Message_text);
    vector = (VEC_type)Expression;
    return vector; }

/*------------------------------ reification/deification -------------------------------*/

static CHR_type reify_character(EXP_type Expression,
                                TXT_type Message_text)
  { CHA_type character;
    CHR_type char_;
    character = cast_to_character(Expression,
                                  Message_text);
    char_ = get_CHA(character);
    return char_; }

static TXT_type reify_string(EXP_type Expression,
                             TXT_type Message_text)
  { STR_type string;
    TXT_type text;
    string = cast_to_string(Expression,
                            Message_text);
    text = string->txt;
    return text; }

static UNS_type reify_unsigned(EXP_type Expression,
                               TXT_type Message_text)
  { SGN_type int_;
    SMI_type smallint;
    UNS_type unsigned_;
    smallint = cast_to_smallint(Expression,
                                Message_text);
    int_ = get_SMI(smallint);
    if (int_ < 0)
      error (NNN_error,
             Message_text);
    unsigned_ = (UNS_type)int_;
    return unsigned_; }

/*--------------------------------------------------------------------------------------*/
/*-------------------------------------- booleans --------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*-------------------------------------- boolean? --------------------------------------*/

static const TXT_type is_boolean_text = "boolean?";

static VOI_type is_boolean_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_boolean_text);
    RETURN_2_BOOLEAN(TRU, FLS, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_boolean_native_M(VOI_type)
  { native_define_M(is_boolean_text,
                    is_boolean_native); }

/*----------------------------------------- not ----------------------------------------*/

static const TXT_type not_text = "not";

static VOI_type not_native(FRM_type Frame)
  { EXP_type boolean_expression;
    require_1_argument(Frame,
                       &boolean_expression,
                       not_text);
    if (is_FLS(boolean_expression))
      Evaluate_Return(TRUE_EXPRESSION);
    else if (is_TRU(boolean_expression))
      Evaluate_Return(FALSE_EXPRESSION);
    else
      error(NAB_error,
            not_text); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_not_native_M(VOI_type)
  { native_define_M(not_text,
                    not_native); }

/*--------------------------------------------------------------------------------------*/
/*-------------------------------- equivalence predicates ------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*----------------------------------------- eq? ----------------------------------------*/

static const TXT_type is_eq_text = "eq?";

static VOI_type is_eq_native(FRM_type Frame)
  { EXP_type left_expression,
             right_expression;
    require_2_arguments(Frame,
                        &left_expression,
                        &right_expression,
                        is_eq_text);
    if (left_expression == right_expression)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_eq_native_M(VOI_type)
  { native_define_M(is_eq_text,
                    is_eq_native); }

/*--------------------------------------- equal? ---------------------------------------*/

static const TXT_type is_equal_text = "equal?";

static BLN_type is_equal(EXP_type,
                         EXP_type,
                         TXT_type);

/*--------------------------------------------------------------------------------------*/

static BLN_type is_equal_pairs(PAI_type Left_pair,
                               PAI_type Right_pair,
                               TXT_type Message_text)
  { EXP_type left_car_expression,
             left_cdr_expression,
             right_car_expression,
             right_cdr_expression;
    left_car_expression = Left_pair->car;
    right_car_expression = Right_pair->car;
    if (is_equal(left_car_expression,
                 right_car_expression,
                 Message_text) == FALSE)
      return FALSE;
    left_cdr_expression = Left_pair->cdr;
    right_cdr_expression = Right_pair->cdr;
    if (is_equal(left_cdr_expression,
                 right_cdr_expression,
                 Message_text) == FALSE)
      return FALSE;
    return TRUE; }

static BLN_type is_equal_reals(REA_type Left_real,
                               REA_type Right_real)
  { FLO_type left_float,
             right_float;
    left_float = Left_real->flo;
    right_float = Left_real->flo;
    if (left_float == right_float)
      return TRUE;
    return FALSE; }

static BLN_type is_equal_strings(STR_type Left_string,
                                 STR_type Right_string)
  { TXT_type left_text,
             right_text;
    left_text = Left_string->txt;
    right_text = Right_string->txt;
    if (C_STRCMP(left_text, right_text) == 0)
      return TRUE;
    return FALSE; }

static BLN_type is_equal_vectors(VEC_type Left_vector,
                                 VEC_type Right_vector,
                                 TXT_type Message_text)
  { EXP_type left_value_expression,
             right_value_expression;
    UNS_type index,
             left_size,
             right_size;
    left_size = size_VEC(Left_vector);
    right_size = size_VEC(Right_vector);
    if (left_size != right_size)
      return FALSE;
    for (index = 1;
         index <= left_size;
         index += 1)
      { left_value_expression = Left_vector[index];
        right_value_expression = Right_vector[index];
        if (is_equal(left_value_expression,
                     right_value_expression,
                     Message_text) == FALSE)
          return FALSE; }
    return TRUE; }

static BLN_type is_equal(EXP_type Left_expression,
                         EXP_type Right_expression,
                         TXT_type Message_text)
  { TAG_type tag;
    tag = Grammar_Tag(Left_expression);
      switch (tag)
      { case PAI_tag:
          if (is_PAI(Right_expression))
            return is_equal_pairs(Left_expression,
                                  Right_expression,
                                  Message_text);
          break;
        case REA_tag:
          if (is_REA(Right_expression))
            return is_equal_reals(Left_expression,
                                  Right_expression);
          break;
        case STR_tag:
          if (is_STR(Right_expression))
            return is_equal_strings(Left_expression,
                                    Right_expression);
          break;
        case VEC_tag:
          if (is_VEC(Right_expression))
            return is_equal_vectors(Left_expression,
                                    Right_expression,
                                    Message_text);
          break;
        case CHA_tag:
        case CNT_tag:
        case END_tag:
        case FLS_tag:
        case FRC_tag:
        case IPT_tag:
        case MAC_tag:    //   ????
        case NAT_tag:
        case NAZ_tag:
        case NST_tag:
        case NUL_tag:
        case OPT_tag:
        case PRC_tag:
        case PRM_tag:
        case PRV_tag:
        case SMI_tag:
        case SYM_tag:
        case TRU_tag:
        case USP_tag:
          if (Left_expression == Right_expression)
            return TRUE;
          break;
        default:
          error(IXT_error,
                Message_text); }
    return FALSE; }

/*--------------------------------------------------------------------------------------*/

static VOI_type is_equal_native(FRM_type Frame)
  { EXP_type left_expression,
             right_expression;
    require_2_arguments(Frame,
                        &left_expression,
                        &right_expression,
                        is_equal_text);
    if (is_equal(left_expression,
                 right_expression,
                 is_equal_text))
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_equal_native_M(VOI_type)
  { native_define_M(is_equal_text,
                    is_equal_native); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- pairs ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

#define CXR(P)                                                                           \
  { EXP_type pair_expression,                                                            \
             value_expression;                                                           \
    require_1_argument(Frame,                                                            \
                       &pair_expression,                                                 \
                       c##P##r_text);                                                    \
    value_expression = cast_to_c##P##r(pair_expression,                                  \
                                       c##P##r_text);                                    \
    Evaluate_Return(value_expression); }

#define CXXR(P, Q)                                                                       \
  { EXP_type pair_expression,                                                            \
             value_expression;                                                           \
    require_1_argument(Frame,                                                            \
                       &pair_expression,                                                 \
                       c##P##Q##r_text);                                                 \
    pair_expression = cast_to_c##Q##r(pair_expression,                                   \
                                      c##P##Q##r_text);                                  \
    value_expression = cast_to_c##P##r(pair_expression,                                  \
                                       c##P##Q##r_text);                                 \
    Evaluate_Return(value_expression); }

#define CXXXR(P, Q, R)                                                                   \
  { EXP_type pair_expression,                                                            \
             value_expression;                                                           \
    require_1_argument(Frame,                                                            \
                       &pair_expression,                                                 \
                       c##P##Q##R##r_text);                                              \
    pair_expression = cast_to_c##R##r(pair_expression,                                   \
                                      c##P##Q##R##r_text);                               \
    pair_expression = cast_to_c##Q##r(pair_expression,                                   \
                                      c##P##Q##R##r_text);                               \
    value_expression = cast_to_c##P##r(pair_expression,                                  \
                                       c##P##Q##R##r_text);                              \
    Evaluate_Return(value_expression); }

#define CXXXXR(P, Q, R, S)                                                               \
  { EXP_type pair_expression,                                                            \
             value_expression;                                                           \
    require_1_argument(Frame,                                                            \
                       &pair_expression,                                                 \
                       c##P##Q##R##S##r_text);                                           \
    pair_expression = cast_to_c##S##r(pair_expression,                                   \
                                      c##P##Q##R##S##r_text);                            \
    pair_expression = cast_to_c##R##r(pair_expression,                                   \
                                      c##P##Q##R##S##r_text);                            \
    pair_expression = cast_to_c##Q##r(pair_expression,                                   \
                                      c##P##Q##R##S##r_text);                            \
    value_expression = cast_to_c##P##r(pair_expression,                                  \
                                       c##P##Q##R##S##r_text);                           \
    Evaluate_Return(value_expression); }

/*---------------------------------------- pair? ---------------------------------------*/

static const TXT_type is_pair_text = "pair?";

static VOI_type is_pair_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_pair_text);
    RETURN_BOOLEAN(PAI, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_pair_native_M(VOI_type)
  { native_define_M(is_pair_text,
                    is_pair_native); }

/*---------------------------------------- cons ----------------------------------------*/

static const TXT_type cons_text = "cons";

static VOI_type cons_native_M(FRM_type Frame)
  { PAI_type pair;
    require_2_arguments(Frame,
                        &Car_Expression_P,
                        &Cdr_Expression_P,
                        cons_text);
    pair = make_PAI_M();
    pair->car = Car_Expression_P;
    pair->cdr = Cdr_Expression_P;
    Evaluate_Return(pair); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cons_native_M(VOI_type)
  { native_define_M(cons_text,
                    cons_native_M); }

/*----------------------------------------- car ----------------------------------------*/

static const TXT_type car_text = "car";

static VOI_type car_native(FRM_type Frame)
  CXR(a)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_car_native_M(VOI_type)
  { native_define_M(car_text,
                    car_native); }

/*----------------------------------------- cdr ----------------------------------------*/

static const TXT_type cdr_text = "cdr";

static VOI_type cdr_native(FRM_type Frame)
  CXR(d)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cdr_native_M(VOI_type)
  { native_define_M(cdr_text,
                    cdr_native); }

/*----------------------------------------- caar ---------------------------------------*/

static const TXT_type caar_text = "caar";

static VOI_type caar_native(FRM_type Frame)
  CXXR(a, a)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_caar_native_M(VOI_type)
  { native_define_M(caar_text,
                    caar_native); }

/*----------------------------------------- cadr ---------------------------------------*/

static const TXT_type cadr_text = "cadr";

static VOI_type cadr_native(FRM_type Frame)
  CXXR(a, d)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cadr_native_M(VOI_type)
  { native_define_M(cadr_text,
                    cadr_native); }

/*----------------------------------------- cdar ---------------------------------------*/

static const TXT_type cdar_text = "cdar";

static VOI_type cdar_native(FRM_type Frame)
  CXXR(d, a)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cdar_native_M(VOI_type)
  { native_define_M(cdar_text,
                    cdar_native); }

/*----------------------------------------- cddr ---------------------------------------*/

static const TXT_type cddr_text = "cddr";

static VOI_type cddr_native(FRM_type Frame)
  CXXR(d, d)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cddr_native_M(VOI_type)
  { native_define_M(cddr_text,
                    cddr_native); }

/*----------------------------------------- caaar --------------------------------------*/

static const TXT_type caaar_text = "caaar";

static VOI_type caaar_native(FRM_type Frame)
  CXXXR(a, a, a)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_caaar_native_M(VOI_type)
  { native_define_M(caaar_text,
                    caaar_native); }

/*----------------------------------------- caadr --------------------------------------*/

static const TXT_type caadr_text = "caadr";

static VOI_type caadr_native(FRM_type Frame)
  CXXXR(a, a, d)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_caadr_native_M(VOI_type)
  { native_define_M(caadr_text,
                    caadr_native); }

/*----------------------------------------- cadar --------------------------------------*/

static const TXT_type cadar_text = "cadar";

static VOI_type cadar_native(FRM_type Frame)
  CXXXR(a, d, a)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cadar_native_M(VOI_type)
  { native_define_M(cadar_text,
                    cadar_native); }

/*----------------------------------------- caddr --------------------------------------*/

static const TXT_type caddr_text = "caddr";

static VOI_type caddr_native(FRM_type Frame)
  CXXXR(a, d, d)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_caddr_native_M(VOI_type)
  { native_define_M(caddr_text,
                    caddr_native); }

/*----------------------------------------- cdaar --------------------------------------*/

static const TXT_type cdaar_text = "cdaar";

static VOI_type cdaar_native(FRM_type Frame)
  CXXXR(d, a, a)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cdaar_native_M(VOI_type)
  { native_define_M(cdaar_text,
                    cdaar_native); }

/*----------------------------------------- cdadr --------------------------------------*/

static const TXT_type cdadr_text = "cdadr";

static VOI_type cdadr_native(FRM_type Frame)
  CXXXR(d, a, d)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cdadr_native_M(VOI_type)
  { native_define_M(cdadr_text,
                    cdadr_native); }

/*----------------------------------------- cddar --------------------------------------*/

static const TXT_type cddar_text = "cddar";

static VOI_type cddar_native(FRM_type Frame)
  CXXXR(d, d, a)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cddar_native_M(VOI_type)
  { native_define_M(cddar_text,
                    cddar_native); }

/*----------------------------------------- cdddr --------------------------------------*/

static const TXT_type cdddr_text = "cdddr";

static VOI_type cdddr_native(FRM_type Frame)
  CXXXR(d, d, d)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cdddr_native_M(VOI_type)
  { native_define_M(cdddr_text,
                    cdddr_native); }

/*----------------------------------------- caaaar -------------------------------------*/

static const TXT_type caaaar_text = "caaaar";

static VOI_type caaaar_native(FRM_type Frame)
  CXXXXR(a, a, a, a)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_caaaar_native_M(VOI_type)
  { native_define_M(caaaar_text,
                    caaaar_native); }

/*----------------------------------------- caaadr -------------------------------------*/

static const TXT_type caaadr_text = "caaadr";

static VOI_type caaadr_native(FRM_type Frame)
  CXXXXR(a, a, a, d)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_caaadr_native_M(VOI_type)
  { native_define_M(caaadr_text,
                    caaadr_native); }

/*----------------------------------------- caadar -------------------------------------*/

static const TXT_type caadar_text = "caadar";

static VOI_type caadar_native(FRM_type Frame)
  CXXXXR(a, a, d, a)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_caadar_native_M(VOI_type)
  { native_define_M(caadar_text,
                    caadar_native); }

/*----------------------------------------- caaddr -------------------------------------*/

static const TXT_type caaddr_text = "caaddr";

static VOI_type caaddr_native(FRM_type Frame)
  CXXXXR(a, a, d, d)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_caaddr_native_M(VOI_type)
  { native_define_M(caaddr_text,
                    caaddr_native); }

/*----------------------------------------- cadaar -------------------------------------*/

static const TXT_type cadaar_text = "cadaar";

static VOI_type cadaar_native(FRM_type Frame)
  CXXXXR(a, d, a, a)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cadaar_native_M(VOI_type)
  { native_define_M(cadaar_text,
                    cadaar_native); }

/*----------------------------------------- cadadr -------------------------------------*/

static const TXT_type cadadr_text = "cadadr";

static VOI_type cadadr_native(FRM_type Frame)
  CXXXXR(a, d, a, d)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cadadr_native_M(VOI_type)
  { native_define_M(cadadr_text,
                    cadadr_native); }

/*----------------------------------------- caddar -------------------------------------*/

static const TXT_type caddar_text = "caddar";

static VOI_type caddar_native(FRM_type Frame)
  CXXXXR(a, d, d, a)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_caddar_native_M(VOI_type)
  { native_define_M(caddar_text,
                    caddar_native); }

/*----------------------------------------- cadddr -------------------------------------*/

static const TXT_type cadddr_text = "cadddr";

static VOI_type cadddr_native(FRM_type Frame)
  CXXXXR(a, d, d, d)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cadddr_native_M(VOI_type)
  { native_define_M(cadddr_text,
                    cadddr_native); }

/*----------------------------------------- cdaaar -------------------------------------*/

static const TXT_type cdaaar_text = "cdaaar";

static VOI_type cdaaar_native(FRM_type Frame)
  CXXXXR(d, a, a, a)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cdaaar_native_M(VOI_type)
  { native_define_M(cdaaar_text,
                    cdaaar_native); }

/*----------------------------------------- cdaadr -------------------------------------*/

static const TXT_type cdaadr_text = "cdaadr";

static VOI_type cdaadr_native(FRM_type Frame)
  CXXXXR(d, a, a, d)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cdaadr_native_M(VOI_type)
  { native_define_M(cdaadr_text,
                    cdaadr_native); }

/*----------------------------------------- cdadar -------------------------------------*/

static const TXT_type cdadar_text = "cdadar";

static VOI_type cdadar_native(FRM_type Frame)
  CXXXXR(d, a, d, a)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cdadar_native_M(VOI_type)
  { native_define_M(cdadar_text,
                    cdadar_native); }

/*----------------------------------------- cdaddr -------------------------------------*/

static const TXT_type cdaddr_text = "cdaddr";

static VOI_type cdaddr_native(FRM_type Frame)
  CXXXXR(d, a, d, d)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cdaddr_native_M(VOI_type)
  { native_define_M(cdaddr_text,
                    cdaddr_native); }

/*----------------------------------------- cddaar -------------------------------------*/

static const TXT_type cddaar_text = "cddaar";

static VOI_type cddaar_native(FRM_type Frame)
   CXXXXR(d, d, a, a)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cddaar_native_M(VOI_type)
  { native_define_M(cddaar_text,
                    cddaar_native); }

/*----------------------------------------- cddadr -------------------------------------*/

static const TXT_type cddadr_text = "cddadr";

static VOI_type cddadr_native(FRM_type Frame)
   CXXXXR(d, d, a, d)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cddadr_native_M(VOI_type)
  { native_define_M(cddadr_text,
                    cddadr_native); }

/*----------------------------------------- cdddar -------------------------------------*/

static const TXT_type cdddar_text = "cdddar";

static VOI_type cdddar_native(FRM_type Frame)
   CXXXXR(d, d, d, a)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cdddar_native_M(VOI_type)
  { native_define_M(cdddar_text,
                    cdddar_native); }

/*----------------------------------------- cddddr -------------------------------------*/

static const TXT_type cddddr_text = "cddddr";

static VOI_type cddddr_native(FRM_type Frame)
   CXXXXR(d, d, d, d)

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cddddr_native_M(VOI_type)
  { native_define_M(cddddr_text,
                    cddddr_native); }

/*-------------------------------------- set-car! --------------------------------------*/

static const TXT_type set_car_text = "set-car!";

static VOI_type set_car_native(FRM_type Frame)
  { EXP_type pair_expression,
             value_expression;
    PAI_type pair;
    require_2_arguments(Frame,
                        &pair_expression,
                        &value_expression,
                        set_car_text);
    pair = cast_to_pair(pair_expression,
                        set_car_text);
    pair->car = value_expression;
    Evaluate_Return(value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_set_car_native_M(VOI_type)
  { native_define_M(set_car_text,
                    set_car_native); }

/*-------------------------------------- set-cdr! --------------------------------------*/

static const TXT_type set_cdr_text = "set-cdr!";

static VOI_type set_cdr_native(FRM_type Frame)
  { EXP_type pair_expression,
             value_expression;
    PAI_type pair;
    require_2_arguments(Frame,
                        &pair_expression,
                        &value_expression,
                        set_cdr_text);
    pair = cast_to_pair(pair_expression,
                        set_cdr_text);
    pair->cdr = value_expression;
    Evaluate_Return(value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_set_cdr_native_M(VOI_type)
  { native_define_M(set_cdr_text,
                    set_cdr_native); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- lists ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*---------------------------------------- null? ---------------------------------------*/

static const TXT_type is_null_text = "null?";

static VOI_type is_null_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_null_text);
    RETURN_BOOLEAN(NUL, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_null_native_M(VOI_type)
  { native_define_M(is_null_text,
                    is_null_native); }

/*---------------------------------------- list ----------------------------------------*/

static const TXT_type list_text = "list";

static VOI_type list_native_M(FRM_type Frame)
  { PAI_type pair;
    UNS_type index,
             number_of_arguments;
    number_of_arguments = Environment_Frame_Size(Frame);
    if (number_of_arguments == 0)
      { Evaluate_Return(Unique_Null);
        return; }
    pair = NULL_PAIR;
    Frame_P = Frame;
    for (index = number_of_arguments;
         index > 0 ;
         index -= 1)
      { Car_Expression_P = Frame_P[index];
        Cdr_Expression_P = (EXP_type)pair;
        pair = make_PAI_M();
        pair->car = Car_Expression_P;
        pair->cdr = Cdr_Expression_P; }
    Environment_Release_Frame(Frame_P);
    Evaluate_Return(pair); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_list_native_M(VOI_type)
  { native_define_M(list_text,
                    list_native_M); }

/*---------------------------------------- length --------------------------------------*/

static const TXT_type length_text = "length";

static UNS_type list_length(EXP_type List_expression,
                            TXT_type message_text)
  { EXP_type expression;
    PAI_type pair;
    UNS_type length;
    length = 0;
    for (expression = List_expression;
         is_PAI(expression);
         length += 1)
      { pair = (PAI_type)expression;
        expression = pair->cdr; }
    if (!is_NUL(expression))
      error(NAL_error,
            message_text);
    return length; }

/*--------------------------------------------------------------------------------------*/

static VOI_type length_native(FRM_type Frame)
  { EXP_type list_expression;
    SMI_type smallint;
    UNS_type length;
    require_1_argument(Frame,
                       &list_expression,
                       length_text);
    length = list_length(list_expression,
                         length_text);
    smallint = make_SMU(length);
    Evaluate_Return(smallint); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_length_native_M(VOI_type)
  { native_define_M(length_text,
                    length_native); }

/*--------------------------------------- append ---------------------------------------*/

static const TXT_type append_text = "append";

static VOI_type append_native_M(FRM_type Frame)
  { EXP_type expression;
    PAI_type new_pair,
             pair;
    TAG_type tag;
    require_2_arguments(Frame,
                        &expression,
                        &List_Expression_P,
                        append_text);
    tag = Grammar_Tag(expression);
    switch (tag)
      { case NUL_tag:
          if (is_PAI(List_Expression_P))
            { pair = List_Expression_P;
              Car_Expression_P = pair->car;
              Cdr_Expression_P = pair->cdr;
              new_pair = make_PAI_M();
              new_pair->car = Car_Expression_P;
              new_pair->cdr = Unique_Null;
              Evaluate_Return(new_pair);
              while (is_PAI(Cdr_Expression_P))
                { pair = Cdr_Expression_P;
                  Car_Expression_P = pair->car;
                  Cdr_Expression_P = pair->cdr;
                  pair = make_PAI_M();
                  new_pair->cdr = pair;
                  new_pair = pair;
                  new_pair->car = Car_Expression_P;
                  new_pair->cdr = Unique_Null; }}
          else
            Evaluate_Return(List_Expression_P);
          return;
        case PAI_tag:
          pair = (PAI_type)expression;
          Expression_P     = pair->cdr;
          Car_Expression_P = pair->car;
          Pair_P = make_PAI_M();
          Pair_P->car = Car_Expression_P;
          Pair_P->cdr = Unique_Null;
          Evaluate_Return(Pair_P);
          while (is_PAI(Expression_P))
            { pair = (PAI_type)Expression_P;
              Expression_P     = pair->cdr;
              Car_Expression_P = pair->car;
              pair = make_PAI_M();
              pair->car = Car_Expression_P;
              pair->cdr = Unique_Null;
              Pair_P->cdr = pair;
              Pair_P = pair; }
          if (!is_NUL(Expression_P))
            error(NAL_error,
                  append_text);
          while (is_PAI(List_Expression_P))
            { pair = (PAI_type)List_Expression_P;
              List_Expression_P = pair->cdr;
              Car_Expression_P  = pair->car;
              pair = make_PAI_M();
              pair->car = Car_Expression_P;
              pair->cdr = Unique_Null;
              Pair_P->cdr = pair;
              Pair_P = pair; }
          Pair_P->cdr = List_Expression_P;
          return;
        default:
          error(NAL_error,
                append_text); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_append_native_M(VOI_type)
  { native_define_M(append_text,
                    append_native_M); }

/*---------------------------------------- memq ----------------------------------------*/

static const TXT_type memq_text = "memq";

static VOI_type memq_native(FRM_type Frame)
  { EXP_type list_expression,
             member_expression,
             value_expression;
    PAI_type pair;
    TAG_type tag;
    require_2_arguments(Frame,
                        &value_expression,
                        &list_expression,
                        memq_text);
    for (;;)
      { tag = Grammar_Tag(list_expression);
        switch (tag)
          { case NUL_tag:
              Evaluate_Return(FALSE_EXPRESSION);
              return;
            case PAI_tag:
              pair = (PAI_type)list_expression;
              member_expression = pair->car;
              if (member_expression == value_expression)
                Evaluate_Return(list_expression);
              else
                list_expression = pair->cdr;
              break;
            default:
              error(NAL_error,
                    memq_text); }}}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_memq_native_M(VOI_type)
  { native_define_M(memq_text,
                    memq_native); }

/*--------------------------------------- member ---------------------------------------*/

static const TXT_type member_text = "member";

static VOI_type member_native(FRM_type Frame)
  { EXP_type list_expression,
             member_expression,
             value_expression;
    PAI_type pair;
    TAG_type tag;
    require_2_arguments(Frame,
                        &value_expression,
                        &list_expression,
                        member_text);
    for (;;)
      { tag = Grammar_Tag(list_expression);
        switch (tag)
          { case NUL_tag:
              Evaluate_Return(FALSE_EXPRESSION);
              return;
            case PAI_tag:
              pair = (PAI_type)list_expression;
              member_expression = pair->car;
              if (is_equal(member_expression,
                           value_expression,
                           member_text))
                Evaluate_Return(list_expression);
              else
                list_expression = pair->cdr;
              break;
            default:
              error(NAL_error,
                    member_text); }}}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_member_native_M(VOI_type)
  { native_define_M(member_text,
                    member_native); }

/*---------------------------------------- assq ----------------------------------------*/

static const TXT_type assq_text = "assq";

static VOI_type assq_native(FRM_type Frame)
  { EXP_type expression,
             list_expression,
             value_expression;
    PAI_type assq_pair,
             pair;
    TAG_type tag;
    require_2_arguments(Frame,
                        &value_expression,
                        &list_expression,
                        assq_text);
    for (;;)
      { tag = Grammar_Tag(list_expression);
        switch (tag)
          { case NUL_tag:
              Evaluate_Return(FALSE_EXPRESSION);
              return;
            case PAI_tag:
              pair = (PAI_type)list_expression;
              expression = pair->car;
              assq_pair = cast_to_pair(expression,
                                       assq_text);
              expression = assq_pair->car;
              if (expression == value_expression)
                Evaluate_Return(assq_pair);
              else
                list_expression = pair->cdr;
              break;
            default:
              error(NAL_error,
                    assq_text); }}}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_assq_native_M(VOI_type)
  { native_define_M(assq_text,
                    assq_native); }

/*---------------------------------------- assoc ---------------------------------------*/

static const TXT_type assoc_text = "assoc";

static VOI_type assoc_native(FRM_type Frame)
  { EXP_type expression,
             list_expression,
             value_expression;
    PAI_type assoc_pair,
             pair;
    TAG_type tag;
    require_2_arguments(Frame,
                        &value_expression,
                        &list_expression,
                        assoc_text);
    for (;;)
      { tag = Grammar_Tag(list_expression);
        switch (tag)
          { case NUL_tag:
              Evaluate_Return(FALSE_EXPRESSION);
              return;
            case PAI_tag:
              pair = (PAI_type)list_expression;
              expression = pair->car;
              assoc_pair = cast_to_pair(expression,
                                        assoc_text);
              expression = assoc_pair->car;
              if (is_equal(expression,
                           value_expression,
                           assoc_text))
                { Evaluate_Return(assoc_pair);
                  return; }
              list_expression = pair->cdr;
              break;
            default:
              error(NAL_error,
                    assoc_text); }}}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_assoc_native_M(VOI_type)
  { native_define_M(assoc_text,
                    assoc_native); }

/*-------------------------------------- reverse ---------------------------------------*/

static const TXT_type reverse_text = "reverse";

static VOI_type reverse_native_M(FRM_type Frame)
  { PAI_type pair;
    require_1_argument(Frame,
                       &Expression_P,
                       reverse_text);
    pair = NULL_PAIR;
    while (is_PAI(Expression_P))
      { Cdr_Expression_P = pair;
        pair = Expression_P;
        Expression_P     = pair->cdr;
        Car_Expression_P = pair->car;
        pair = make_PAI_M();
        pair->car = Car_Expression_P;
        pair->car = Cdr_Expression_P; }
    if (!is_NUL(Expression_P))
      error(NAL_error,
            reverse_text);
    Evaluate_Return(pair); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_reverse_native_M(VOI_type)
  { native_define_M(reverse_text,
                    reverse_native_M); }

/*--------------------------------------------------------------------------------------*/
/*--------------------------------------- symbols --------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*--------------------------------------- symbol? --------------------------------------*/

static const TXT_type is_symbol_text = "symbol?";

static VOI_type is_symbol_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_symbol_text);
    RETURN_BOOLEAN(SYM, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_symbol_native_M(VOI_type)
  { native_define_M(is_symbol_text,
                    is_symbol_native); }

/*----------------------------------- symbol->string -----------------------------------*/

static const TXT_type symbol_to_string_text = "symbol->string";

static VOI_type symbol_to_string_native(FRM_type Frame)
  { EXP_type symbol_expression;
    STR_type string;
    TXT_type text;
    UNS_type length;
    require_1_argument(Frame,
                       &symbol_expression,
                       symbol_to_string_text);
    Symbol_P = cast_to_symbol(symbol_expression,
                              symbol_to_string_text);
    length = length_SYM(Symbol_P);
    if (length == 0)
      Evaluate_Return(Unique_Empty_String_P);
    else
      { text = Symbol_P->txt;
        string = make_STR_M(length);
        C_STRCPY(string->txt, text);
        Evaluate_Return(string); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_symbol_to_string_native_M(VOI_type)
  { native_define_M(symbol_to_string_text,
                    symbol_to_string_native); }

/*----------------------------------- string->symbol -----------------------------------*/

static const TXT_type string_to_symbol_text = "string->symbol";

static VOI_type string_to_symbol_native(FRM_type Frame)
  { EXP_type string_expression;
    STR_type string;
    SYM_type symbol;
    TXT_type text;
    UNS_type text_length;
    require_1_argument(Frame,
                       &string_expression,
                       string_to_symbol_text);
    string = cast_to_string(string_expression,
                            string_to_symbol_text);
    text = string->txt;
    text_length = C_STRLEN(text);
    if (text_length >= Grammar_Text_Buffer_Length)
      error(STL_error,
            string_to_symbol_text);
    C_STRCPY(Grammar_Text_Buffer, text);
    symbol = Pool_Enter_Text_M(Grammar_Text_Buffer);
    if (!Pool_Validate_Symbol(symbol))
      error(NVS_error,
            string_to_symbol_text);
    Evaluate_Return(symbol); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_string_to_symbol_native_M(VOI_type)
  { native_define_M(string_to_symbol_text,
                    string_to_symbol_native); }

/*--------------------------------------------------------------------------------------*/
/*--------------------------------------- numbers --------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*--------------------------------------- integer? -------------------------------------*/

static const TXT_type is_integer_text = "integer?";

static VOI_type is_integer_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_integer_text);
    RETURN_BOOLEAN(SMI, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_integer_native_M(VOI_type)
  { native_define_M(is_integer_text,
                    is_integer_native); }

/*--------------------------------------- number? --------------------------------------*/

static const TXT_type is_number_text = "number?";

static VOI_type is_number_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_number_text);
    RETURN_2_BOOLEAN(SMI, REA, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_number_native_M(VOI_type)
  { native_define_M(is_number_text,
                    is_number_native); }

/*---------------------------------------- real? ---------------------------------------*/

static const TXT_type is_real_text = "real?";

static VOI_type is_real_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_real_text);
    RETURN_BOOLEAN(REA, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_real_native_M(VOI_type)
  { native_define_M(is_real_text,
                    is_real_native); }

/*--------------------------------------- zero? ----------------------------------------*/

static const TXT_type is_zero_text = "zero?";

static VOI_type is_zero_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    require_1_argument(Frame,
                       &number_expression,
                       is_zero_text);
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(number_expression);
          if (int_ == 0)
            Evaluate_Return(TRUE_EXPRESSION);
          else
            Evaluate_Return(FALSE_EXPRESSION);
          return;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          if (float_ == 0)
            Evaluate_Return(TRUE_EXPRESSION);
          else
            Evaluate_Return(FALSE_EXPRESSION);
          return;
       default:
          error(ARN_error,
                is_zero_text); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_zero_native_M(VOI_type)
  { native_define_M(is_zero_text,
                    is_zero_native_M); }

/*------------------------------------- positive? --------------------------------------*/

static const TXT_type is_positive_text = "positive?";

static VOI_type is_positive_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    require_1_argument(Frame,
                       &number_expression,
                       is_positive_text);
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(number_expression);
          if (int_ > 0)
            Evaluate_Return(TRUE_EXPRESSION);
          else
            Evaluate_Return(FALSE_EXPRESSION);
          return;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          if (float_ > 0)
            Evaluate_Return(TRUE_EXPRESSION);
          else
            Evaluate_Return(FALSE_EXPRESSION);
          return;
       default:
          error(ARN_error,
                is_zero_text); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_positive_native_M(VOI_type)
  { native_define_M(is_positive_text,
                    is_positive_native_M); }

/*------------------------------------- negative? --------------------------------------*/

static const TXT_type is_negative_text = "negative?";

static VOI_type is_negative_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    require_1_argument(Frame,
                       &number_expression,
                       is_negative_text);
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(number_expression);
          if (int_ < 0)
            Evaluate_Return(TRUE_EXPRESSION);
          else
            Evaluate_Return(FALSE_EXPRESSION);
          return;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          if (float_ < 0)
            Evaluate_Return(TRUE_EXPRESSION);
          else
            Evaluate_Return(FALSE_EXPRESSION);
          return;
       default:
          error(ARN_error,
                is_zero_text); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_negative_native_M(VOI_type)
  { native_define_M(is_negative_text,
                    is_negative_native_M); }

/*--------------------------------------- odd? ----------------------------------------*/

static const TXT_type is_odd_text = "odd?";

static VOI_type is_odd_native_M(FRM_type Frame)
  { EXP_type number_expression;
    SGN_type int_;
    require_1_argument(Frame,
                       &number_expression,
                       is_odd_text);
    if (is_SMI(number_expression))
      { int_ = get_SMI(number_expression);
        if (int_ % 2 != 0)
          Evaluate_Return(TRUE_EXPRESSION);
        else
          Evaluate_Return(FALSE_EXPRESSION); }
    else
      error(ARI_error,
            is_odd_text); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_odd_native_M(VOI_type)
  { native_define_M(is_odd_text,
                    is_odd_native_M); }

/*--------------------------------------- even? ----------------------------------------*/

static const TXT_type is_even_text = "even?";

static VOI_type is_even_native_M(FRM_type Frame)
  { EXP_type number_expression;
    SGN_type int_;
    require_1_argument(Frame,
                       &number_expression,
                       is_even_text);
    if (is_SMI(number_expression))
      { int_ = get_SMI(number_expression);
        if (int_ % 2 == 0)
          Evaluate_Return(TRUE_EXPRESSION);
        else
          Evaluate_Return(FALSE_EXPRESSION); }
    else
      error(ARI_error,
            is_even_text); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_even_native_M(VOI_type)
  { native_define_M(is_even_text,
                    is_even_native_M); }

/*----------------------------------------- = ------------------------------------------*/

static const TXT_type is_number_equal_text = "=";

static BLN_type is_smallint_equal(SMI_type Left_smallint,
                                  EXP_type Right_number_expression)
  { BLN_type boolean;
    FLO_type right_float;
    SGN_type left_int,
             right_int;
    REA_type right_real;
    TAG_type tag;
    left_int = get_SMI(Left_smallint);
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          right_int = get_SMI(Right_number_expression);
          boolean = (left_int == right_int);
          break;
        case REA_tag:
          right_real = (REA_type)Right_number_expression;
          right_float = right_real->flo;
          boolean = (left_int == right_float);
          break;
        default:
          boolean = FALSE;
          error(ARN_error,
                is_number_equal_text); }
    return boolean; }

static BLN_type is_real_equal(REA_type Left_real,
                              EXP_type Right_number_expression)
  { BLN_type boolean;
    FLO_type left_float,
             right_float;
    SGN_type right_int;
    REA_type right_real;
    TAG_type tag;
    left_float = Left_real->flo;
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          right_int = get_SMI(Right_number_expression);
          boolean = (left_float == right_int);
          break;
        case REA_tag:
          right_real = (REA_type)Right_number_expression;
          right_float = right_real->flo;
          boolean = (left_float == right_float);
          break;
        default:
          boolean = FALSE;
          error(ARN_error,
                is_number_equal_text); }
    return boolean; }

static VOI_type is_number_equal_native(FRM_type Frame)
  { BLN_type boolean;
    EXP_type left_number_expression,
             right_number_expression;
    TAG_type tag;
    require_2_arguments(Frame,
                        &left_number_expression,
                        &right_number_expression,
                        is_number_equal_text);
    tag = Grammar_Tag(left_number_expression);
    switch (tag)
      { case SMI_tag:
          boolean = is_smallint_equal(left_number_expression,
                                      right_number_expression);
          break;
        case REA_tag:
          boolean = is_real_equal(left_number_expression,
                                  right_number_expression);
          break;
        default:
          boolean = FALSE;
          error(ARN_error,
                is_number_equal_text); }
    if (boolean)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_number_equal_native_M(VOI_type)
  { native_define_M(is_number_equal_text,
                    is_number_equal_native); }

/*----------------------------------------- < ------------------------------------------*/

static const TXT_type is_number_less_text = "<";

static BLN_type is_smallint_less(SMI_type Left_smallint,
                                 EXP_type Right_number_expression)
  { BLN_type boolean;
    FLO_type right_float;
    SGN_type left_int,
             right_int;
    REA_type right_real;
    TAG_type tag;
    left_int = get_SMI(Left_smallint);
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          right_int = get_SMI(Right_number_expression);
          boolean = (left_int < right_int);
          break;
        case REA_tag:
          right_real = (REA_type)Right_number_expression;
          right_float = right_real->flo;
          boolean = (left_int < right_float);
          break;
        default:
          boolean = FALSE;
          error(ARN_error,
                is_number_less_text); }
    return boolean; }

static BLN_type is_real_less(REA_type Left_real,
                             EXP_type Right_number_expression)
  { BLN_type boolean;
    FLO_type left_float,
             right_float;
    SGN_type right_int;
    REA_type right_real;
    TAG_type tag;
    left_float = Left_real->flo;
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          right_int = get_SMI(Right_number_expression);
          boolean = (left_float < right_int);
          break;
        case REA_tag:
          right_real = (REA_type)Right_number_expression;
          right_float = right_real->flo;
          boolean = (left_float < right_float);
          break;
        default:
          boolean = FALSE;
          error(ARN_error,
                is_number_less_text); }
    return boolean; }

static VOI_type is_number_less_native(FRM_type Frame)
  { BLN_type boolean;
    EXP_type left_number_expression,
             right_number_expression;
    TAG_type tag;
    require_2_arguments(Frame,
                        &left_number_expression,
                        &right_number_expression,
                        is_number_less_text);
    tag = Grammar_Tag(left_number_expression);
    switch (tag)
      { case SMI_tag:
          boolean = is_smallint_less(left_number_expression,
                                     right_number_expression);
          break;
        case REA_tag:
          boolean = is_real_less(left_number_expression,
                                 right_number_expression);
          break;
        default:
          boolean = FALSE;
          error(ARN_error,
                is_number_less_text); }
    if (boolean)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_number_less_native_M(VOI_type)
  { native_define_M(is_number_less_text,
                    is_number_less_native); }

/*----------------------------------------- > ------------------------------------------*/

static const TXT_type is_number_greater_text = ">";

static BLN_type is_smallint_greater(SMI_type Left_smallint,
                                    EXP_type Right_number_expression)
  { BLN_type boolean;
    FLO_type right_float;
    SGN_type left_int,
             right_int;
    REA_type right_real;
    TAG_type tag;
    left_int = get_SMI(Left_smallint);
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          right_int = get_SMI(Right_number_expression);
          boolean = (left_int > right_int);
          break;
        case REA_tag:
          right_real = (REA_type)Right_number_expression;
          right_float = right_real->flo;
          boolean = (left_int > right_float);
          break;
        default:
          boolean = FALSE;
          error(ARN_error,
                is_number_greater_text); }
    return boolean; }

static BLN_type is_real_greater(REA_type Left_real,
                                EXP_type Right_number_expression)
  { BLN_type boolean;
    FLO_type left_float,
             right_float;
    SGN_type right_int;
    REA_type right_real;
    TAG_type tag;
    left_float = Left_real->flo;
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          right_int = get_SMI(Right_number_expression);
          boolean = (left_float > right_int);
          break;
        case REA_tag:
          right_real = (REA_type)Right_number_expression;
          right_float = right_real->flo;
          boolean = (left_float > right_float);
          break;
        default:
          boolean = FALSE;
          error(ARN_error,
                is_number_greater_text); }
    return boolean; }

static VOI_type is_number_greater_native(FRM_type Frame)
  { BLN_type boolean;
    EXP_type left_number_expression,
             right_number_expression;
    TAG_type tag;
    require_2_arguments(Frame,
                        &left_number_expression,
                        &right_number_expression,
                        is_number_greater_text);
    tag = Grammar_Tag(left_number_expression);
    switch (tag)
      { case SMI_tag:
          boolean = is_smallint_greater(left_number_expression,
                                        right_number_expression);
          break;
        case REA_tag:
          boolean = is_real_greater(left_number_expression,
                                    right_number_expression);
          break;
        default:
          boolean = FALSE;
          error(ARN_error,
                is_number_greater_text); }
    if (boolean)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_number_greater_native_M(VOI_type)
  { native_define_M(is_number_greater_text,
                    is_number_greater_native); }

/*----------------------------------------- <= ------------------------------------------*/

static const TXT_type is_number_less_or_equal_text = "<=";

static BLN_type is_smallint_less_or_equal(SMI_type Left_smallint,
                                          EXP_type Right_number_expression)
  { BLN_type boolean;
    FLO_type right_float;
    SGN_type left_int,
             right_int;
    REA_type right_real;
    TAG_type tag;
    left_int = get_SMI(Left_smallint);
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          right_int = get_SMI(Right_number_expression);
          boolean = (left_int <= right_int);
          break;
        case REA_tag:
          right_real = (REA_type)Right_number_expression;
          right_float = right_real->flo;
          boolean = (left_int <= right_float);
          break;
        default:
          boolean = FALSE;
          error(ARN_error,
                is_number_less_or_equal_text); }
    return boolean; }

static BLN_type is_real_less_or_equal(REA_type Left_real,
                                      EXP_type Right_number_expression)
  { BLN_type boolean;
    FLO_type left_float,
             right_float;
    SGN_type right_int;
    REA_type right_real;
    TAG_type tag;
    left_float = Left_real->flo;
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          right_int = get_SMI(Right_number_expression);
          boolean = (left_float <= right_int);
          break;
        case REA_tag:
          right_real = (REA_type)Right_number_expression;
          right_float = right_real->flo;
          boolean = (left_float <= right_float);
          break;
        default:
          boolean = FALSE;
          error(ARN_error,
                is_number_less_or_equal_text); }
    return boolean; }

static VOI_type is_number_less_or_equal_native(FRM_type Frame)
  { BLN_type boolean;
    EXP_type left_number_expression,
             right_number_expression;
    TAG_type tag;
    require_2_arguments(Frame,
                        &left_number_expression,
                        &right_number_expression,
                        is_number_less_or_equal_text);
    tag = Grammar_Tag(left_number_expression);
    switch (tag)
      { case SMI_tag:
          boolean = is_smallint_less_or_equal(left_number_expression,
                                              right_number_expression);
          break;
        case REA_tag:
          boolean = is_real_less_or_equal(left_number_expression,
                                          right_number_expression);
          break;
        default:
          boolean = FALSE;
          error(ARN_error,
                is_number_less_or_equal_text); }
    if (boolean)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_number_less_or_equal_native_M(VOI_type)
  { native_define_M(is_number_less_or_equal_text,
                    is_number_less_or_equal_native); }

/*----------------------------------------- >= -----------------------------------------*/

static const TXT_type is_number_greater_or_equal_text = ">=";

static BLN_type is_smallint_greater_or_equal(SMI_type Left_smallint,
                                             EXP_type Right_number_expression)
  { BLN_type boolean;
    FLO_type right_float;
    SGN_type left_int,
             right_int;
    REA_type right_real;
    TAG_type tag;
    left_int = get_SMI(Left_smallint);
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          right_int = get_SMI(Right_number_expression);
          boolean = (left_int >= right_int);
          break;
        case REA_tag:
          right_real = (REA_type)Right_number_expression;
          right_float = right_real->flo;
          boolean = (left_int >= right_float);
          break;
        default:
          boolean = FALSE;
          error(ARN_error,
                is_number_greater_or_equal_text); }
    return boolean; }

static BLN_type is_real_greater_or_equal(REA_type Left_real,
                                         EXP_type Right_number_expression)
  { BLN_type boolean;
    FLO_type left_float,
             right_float;
    SGN_type right_int;
    REA_type right_real;
    TAG_type tag;
    left_float = Left_real->flo;
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          right_int = get_SMI(Right_number_expression);
          boolean = (left_float >= right_int);
          break;
        case REA_tag:
          right_real = (REA_type)Right_number_expression;
          right_float = right_real->flo;
          boolean = (left_float >= right_float);
          break;
        default:
          boolean = FALSE;
          error(ARN_error,
                is_number_greater_or_equal_text); }
    return boolean; }

static VOI_type is_number_greater_or_equal_native(FRM_type Frame)
  { BLN_type boolean;
    EXP_type left_number_expression,
             right_number_expression;
    TAG_type tag;
    require_2_arguments(Frame,
                        &left_number_expression,
                        &right_number_expression,
                        is_number_greater_or_equal_text);
    tag = Grammar_Tag(left_number_expression);
    switch (tag)
      { case SMI_tag:
          boolean = is_smallint_greater_or_equal(left_number_expression,
                                                 right_number_expression);
          break;
        case REA_tag:
          boolean = is_real_greater_or_equal(left_number_expression,
                                             right_number_expression);
          break;
        default:
          boolean = FALSE;
          error(ARN_error,
                is_number_greater_or_equal_text); }
    if (boolean)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_number_greater_or_equal_native_M(VOI_type)
  { native_define_M(is_number_greater_or_equal_text,
                    is_number_greater_or_equal_native); }

/*---------------------------------------- max -----------------------------------------*/

static const TXT_type max_text = "max";

static EXP_type smallint_max(SMI_type Left_smallint,
                             EXP_type Right_number_expression)
  { EXP_type expression;
    FLO_type right_float;
    SGN_type left_int,
             right_int;
    REA_type right_real;
    TAG_type tag;
    left_int = get_SMI(Left_smallint);
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          right_int = get_SMI(Right_number_expression);
          if (left_int > right_int)
            { expression = (EXP_type)Left_smallint;
              return expression; }
          return Right_number_expression;
        case REA_tag:
          right_real = (REA_type)Right_number_expression;
          right_float = right_real->flo;
          if (left_int > right_float)
            { expression = (EXP_type)Left_smallint;
              return expression; }
          return Right_number_expression;
        default:
          error(ARN_error,
                max_text);
          return Unique_Zero_Smallint; }}

static EXP_type real_max(REA_type Left_real,
                         EXP_type Right_number_expression)
  { EXP_type expression;
    FLO_type left_float,
             right_float;
    SGN_type right_int;
    REA_type right_real;
    TAG_type tag;
    left_float = Left_real->flo;
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          right_int = get_SMI(Right_number_expression);
          if (left_float > right_int)
            { expression = (EXP_type)Left_real;
              return expression; }
          return Right_number_expression;
        case REA_tag:
          right_real = (REA_type)Right_number_expression;
          right_float = right_real->flo;
          if (left_float > right_float)
            { expression = (EXP_type)Left_real;
              return expression; }
          return Right_number_expression;
        default:
          error(ARN_error,
                max_text);
          return Unique_Zero_Smallint; }}

static VOI_type max_native(FRM_type Frame)
  { EXP_type left_number_expression,
             number_expression,
             right_number_expression;
    REA_type left_real;
    SMI_type left_smallint;
    TAG_type tag;
    require_2_arguments(Frame,
                        &left_number_expression,
                        &right_number_expression,
                        max_text);
    tag = Grammar_Tag(left_number_expression);
    switch (tag)
      { case SMI_tag:
          left_smallint = (SMI_type)left_number_expression;
          number_expression = smallint_max(left_smallint,
                                           right_number_expression);
          break;
        case REA_tag:
          left_real = (REA_type)left_number_expression;
          number_expression = real_max(left_real,
                                       right_number_expression);
          break;
        default:
          error(ARN_error,
                max_text);
          number_expression = Unique_Zero_Smallint; }
    Evaluate_Return(number_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_max_native_M(VOI_type)
  { native_define_M(max_text,
                    max_native); }

/*---------------------------------------- min -----------------------------------------*/

static const TXT_type min_text = "min";

static EXP_type smallint_min(SMI_type Left_smallint,
                             EXP_type Right_number_expression)
  { EXP_type expression;
    FLO_type right_float;
    SGN_type left_int,
             right_int;
    REA_type right_real;
    TAG_type tag;
    left_int = get_SMI(Left_smallint);
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          right_int = get_SMI(Right_number_expression);
          if (left_int < right_int)
            { expression = (EXP_type)Left_smallint;
              return expression; }
          return Right_number_expression;
        case REA_tag:
          right_real = (REA_type)Right_number_expression;
          right_float = right_real->flo;
          if (left_int < right_float)
            { expression = (EXP_type)Left_smallint;
              return expression; }
          return Right_number_expression;
        default:
          error(ARN_error,
                min_text);
          return Unique_Zero_Smallint; }}

static EXP_type real_min(REA_type Left_real,
                         EXP_type Right_number_expression)
  { EXP_type expression;
    FLO_type left_float,
             right_float;
    SGN_type right_int;
    REA_type right_real;
    TAG_type tag;
    left_float = Left_real->flo;
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          right_int = get_SMI(Right_number_expression);
          if (left_float < right_int)
            { expression = (EXP_type)Left_real;
              return expression; }
          return Right_number_expression;
        case REA_tag:
          right_real = (REA_type)Right_number_expression;
          right_float = right_real->flo;
          if (left_float < right_float)
            { expression = (EXP_type)Left_real;
              return expression; }
          return Right_number_expression;
        default:
          error(ARN_error,
                min_text);
          return Unique_Zero_Smallint; }}

static VOI_type min_native(FRM_type Frame)
  { EXP_type left_number_expression,
             number_expression,
             right_number_expression;
    REA_type left_real;
    SMI_type left_smallint;
    TAG_type tag;
    require_2_arguments(Frame,
                        &left_number_expression,
                        &right_number_expression,
                        min_text);
    tag = Grammar_Tag(left_number_expression);
    switch (tag)
      { case SMI_tag:
          left_smallint = (SMI_type)left_number_expression;
          number_expression = smallint_min(left_smallint,
                                           right_number_expression);
          break;
        case REA_tag:
          left_real = (REA_type)left_number_expression;
          number_expression = real_min(left_real,
                                       right_number_expression);
          break;
        default:
          error(ARN_error,
                min_text);
          number_expression = Unique_Zero_Smallint; }
    Evaluate_Return(number_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_min_native_M(VOI_type)
  { native_define_M(min_text,
                    min_native); }

/*----------------------------------------- + ------------------------------------------*/

static const TXT_type add_text = "+";

static EXP_type real_add_M(FRM_type Frame,
                           FLO_type Float,
                           UNS_type Index)
  { EXP_type value_expression;
    FLO_type accumulated_float,
             float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             number_of_arguments;
    accumulated_float = Float;
    number_of_arguments = Environment_Frame_Size(Frame);
    for (index = Index;
         index <= number_of_arguments;
         index += 1)
      { value_expression = Frame[index];
        tag = Grammar_Tag(value_expression);
        switch (tag)
          { case SMI_tag:
              int_ = get_SMI(value_expression);
              accumulated_float += (FLO_type)int_;
              break;
            case REA_tag:
              real = (REA_type)value_expression;
              float_ = real->flo;
              accumulated_float += float_;
              break;
            default:
              error(ARN_error,
                    add_text); }}
    real = make_REA_M();
    real->flo = accumulated_float;
    return real; }

static EXP_type smallint_add_M(FRM_type Frame,
                               SGN_type Int)
  { EXP_type number_expression;
    FLO_type accumulated_float,
             float_;
    SGN_type int_;
    LNG_type accumulated_long;
    REA_type real;
    SMI_type smallint;
    TAG_type tag;
    UNS_type index,
             number_of_arguments;
    accumulated_long = (LNG_type)Int;
    number_of_arguments = Environment_Frame_Size(Frame);
    for (index = 2;
         index <= number_of_arguments;
         index += 1)
      { number_expression = Frame[index];
        tag = Grammar_Tag(number_expression);
        switch (tag)
          { case SMI_tag:
              int_ = get_SMI(number_expression);
              accumulated_long += (LNG_type)int_;
              break;
            case REA_tag:
              real = (REA_type)number_expression;
              float_ = real->flo;
              accumulated_float = (FLO_type)accumulated_long + float_;
              real = real_add_M(Frame,
                                accumulated_float,
                                index + 1);
              return real;
            default:
              error(ARN_error,
                    add_text); }}
    smallint = make_smallint(accumulated_long,
                             add_text);
    return smallint; }

static VOI_type add_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    UNS_type number_of_arguments;
    number_of_arguments = Environment_Frame_Size(Frame);
    if (number_of_arguments == 0)
      number_expression = Unique_Zero_Smallint;
    else
      { number_expression = Frame[1];
        tag = Grammar_Tag(number_expression);
        switch (tag)
          { case SMI_tag:
              if (number_of_arguments > 1)
                { int_ = get_SMI(number_expression);
                  number_expression = smallint_add_M(Frame,
                                                     int_); }
              break;
            case REA_tag:
              if (number_of_arguments > 1)
                { real = (REA_type)number_expression;
                  float_ = real->flo;
                  number_expression = real_add_M(Frame,
                                                 float_,
                                                 2); }
              break;
           default:
             error(ARN_error,
                   add_text); }
        Environment_Release_Frame(Frame); }
    Evaluate_Return(number_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_add_native_M(VOI_type)
  { native_define_M(add_text,
                    add_native_M); }

/*----------------------------------------- * ------------------------------------------*/

static const TXT_type multiply_text = "*";

static EXP_type real_multiply_M(FRM_type Frame,
                                FLO_type Float,
                                UNS_type Index)
  { EXP_type value_expression;
    FLO_type accumulated_float,
             float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             number_of_arguments;
    accumulated_float = Float;
    number_of_arguments = Environment_Frame_Size(Frame);
    for (index = Index;
         index <= number_of_arguments;
         index += 1)
      { value_expression = Frame[index];
        tag = Grammar_Tag(value_expression);
        switch (tag)
          { case SMI_tag:
              int_ = get_SMI(value_expression);
              accumulated_float *= (FLO_type)int_;
              break;
            case REA_tag:
              real = (REA_type)value_expression;
              float_ = real->flo;
              accumulated_float *= float_;
              break;
            default:
              error(ARN_error,
                    multiply_text); }}
    real = make_REA_M();
    real->flo = accumulated_float;
    return real; }

static EXP_type smallint_multiply_M(FRM_type Frame,
                                    SGN_type Int)
  { EXP_type number_expression;
    FLO_type accumulated_float,
             float_;
    SGN_type int_;
    LNG_type accumulated_long;
    REA_type real;
    SMI_type smallint;
    TAG_type tag;
    UNS_type index,
             number_of_arguments;
    accumulated_long = (LNG_type)Int;
    number_of_arguments = Environment_Frame_Size(Frame);
    for (index = 2;
         index <= number_of_arguments;
         index += 1)
      { number_expression = Frame[index];
        tag = Grammar_Tag(number_expression);
        switch (tag)
          { case SMI_tag:
              int_ = get_SMI(number_expression);
              accumulated_long *= (LNG_type)int_;
              break;
            case REA_tag:
              real = (REA_type)number_expression;
              float_ = real->flo;
              accumulated_float = (FLO_type)accumulated_long * float_;
              real = real_multiply_M(Frame,
                                     accumulated_float,
                                     index + 1);
              return real;
            default:
              error(ARN_error,
                    multiply_text); }}
    smallint = make_smallint(accumulated_long,
                             multiply_text);
    return smallint; }

static VOI_type multiply_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    UNS_type number_of_arguments;
    number_of_arguments = Environment_Frame_Size(Frame);
    if (number_of_arguments == 0)
      number_expression = Unique_One_Smallint;
    else
      { number_expression = Frame[1];
        tag = Grammar_Tag(number_expression);
        switch (tag)
          { case SMI_tag:
              if (number_of_arguments > 1)
                { int_ = get_SMI(number_expression);
                  number_expression = smallint_multiply_M(Frame,
                                                          int_); }
              break;
            case REA_tag:
              if (number_of_arguments > 1)
                { real = (REA_type)number_expression;
                  float_ = real->flo;
                  number_expression = real_multiply_M(Frame,
                                                      float_,
                                                      2); }
              break;
           default:
             error(ARN_error,
                   multiply_text); }
        Environment_Release_Frame(Frame); }
    Evaluate_Return(number_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_multiply_native_M(VOI_type)
  { native_define_M(multiply_text,
                    multiply_native_M); }

/*----------------------------------------- - ------------------------------------------*/

static const TXT_type subtract_text = "-";

static EXP_type real_subtract_M(FRM_type Frame,
                           FLO_type Float,
                                UNS_type Index)
  { EXP_type value_expression;
    FLO_type accumulated_float,
             float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             number_of_arguments;
    accumulated_float = Float;
    number_of_arguments = Environment_Frame_Size(Frame);
    for (index = Index;
         index <= number_of_arguments;
         index += 1)
      { value_expression = Frame[index];
        tag = Grammar_Tag(value_expression);
        switch (tag)
          { case SMI_tag:
              int_ = get_SMI(value_expression);
              accumulated_float -= (FLO_type)int_;
              break;
            case REA_tag:
              real = (REA_type)value_expression;
              float_ = real->flo;
              accumulated_float -= float_;
              break;
            default:
              error(ARN_error,
                    subtract_text); }}
    real = make_REA_M();
    real->flo = accumulated_float;
    return real; }

static EXP_type smallint_subtract_M(FRM_type Frame,
                                    SGN_type Int)
  { EXP_type number_expression;
    FLO_type accumulated_float,
             float_;
    SGN_type int_;
    LNG_type accumulated_long;
    REA_type real;
    SMI_type smallint;
    TAG_type tag;
    UNS_type index,
             number_of_arguments;
    accumulated_long = (LNG_type)Int;
    number_of_arguments = Environment_Frame_Size(Frame);
    for (index = 2;
         index <= number_of_arguments;
         index += 1)
      { number_expression = Frame[index];
        tag = Grammar_Tag(number_expression);
        switch (tag)
          { case SMI_tag:
              int_ = get_SMI(number_expression);
              accumulated_long -= (LNG_type)int_;
              break;
            case REA_tag:
              real = (REA_type)number_expression;
              float_ = real->flo;
              accumulated_float = (FLO_type)accumulated_long - float_;
              real = real_subtract_M(Frame,
                                     accumulated_float,
                                     index + 1);
              return real;
            default:
              error(ARN_error,
                    subtract_text); }}
    smallint = make_smallint(accumulated_long,
                             subtract_text);
    return smallint; }

static VOI_type subtract_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    SMI_type smallint;
    TAG_type tag;
    UNS_type number_of_arguments;
    number_of_arguments = Environment_Frame_Size(Frame);
    if (number_of_arguments == 0)
      error(AL1_error,
            subtract_text);
    number_expression = Frame[1];
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(number_expression);
          if (number_of_arguments > 1)
            { int_ = get_SMI(number_expression);
              number_expression = smallint_subtract_M(Frame,
                                                      int_); }
          else
            { int_ = - int_;
              smallint = make_smallint(int_,
                                       subtract_text);
              number_expression = (EXP_type)smallint; }
          break;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          if (number_of_arguments > 1)
            number_expression = real_subtract_M(Frame,
                                                float_,
                                                2);
          else
            { float_ = - float_;
              real = make_REA_M();
              real->flo = float_;
              number_expression = (EXP_type)real; }
          break;
       default:
         error(ARN_error,
               subtract_text); }
    Environment_Release_Frame(Frame);
    Evaluate_Return(number_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_subtract_native_M(VOI_type)
  { native_define_M(subtract_text,
                    subtract_native_M); }

/*----------------------------------------- / ------------------------------------------*/

static const TXT_type divide_text = "/";

static EXP_type real_divide_M(FRM_type Frame,
                           FLO_type Float,
                              UNS_type Index)
  { EXP_type value_expression;
    FLO_type accumulated_float,
             float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             number_of_arguments;
    accumulated_float = Float;
    number_of_arguments = Environment_Frame_Size(Frame);
    for (index = Index;
         index <= number_of_arguments;
         index += 1)
      { value_expression = Frame[index];
        tag = Grammar_Tag(value_expression);
        switch (tag)
          { case SMI_tag:
              int_ = get_SMI(value_expression);
              if (int_ == 0)
                error(DBZ_error,
                      divide_text);
              accumulated_float /= (FLO_type)int_;
              break;
            case REA_tag:
              real = (REA_type)value_expression;
              float_ = real->flo;
              if (float_ == 0)
                error(DBZ_error,
                      divide_text);
              accumulated_float /= float_;
              break;
            default:
              error(ARN_error,
                    divide_text); }}
    real = make_REA_M();
    real->flo = accumulated_float;
    return real; }

static EXP_type smallint_divide_M(FRM_type Frame,
                                  SGN_type Int)
  { EXP_type number_expression;
    FLO_type accumulated_float,
             float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             number_of_arguments;
    accumulated_float = (FLO_type)Int;
    number_of_arguments = Environment_Frame_Size(Frame);
    for (index = 2;
         index <= number_of_arguments;
         index += 1)
      { number_expression = Frame[index];
        tag = Grammar_Tag(number_expression);
        switch (tag)
          { case SMI_tag:
              int_ = get_SMI(number_expression);
              if (int_ == 0)
                error(DBZ_error,
                      divide_text);
              else
                accumulated_float /= (FLO_type)int_;
              break;
            case REA_tag:
              real = (REA_type)number_expression;
              float_ = real->flo;
              if (float_ == 0)
                error(DBZ_error,
                      divide_text);
              accumulated_float = accumulated_float / float_;
              real = real_divide_M(Frame,
                                   accumulated_float,
                                   index + 1);
              return real;
            default:
              error(ARN_error,
                    divide_text); }}
    real = make_REA_M();
    real->flo = accumulated_float;
    return real; }

static VOI_type divide_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    UNS_type number_of_arguments;
    number_of_arguments = Environment_Frame_Size(Frame);
    if (number_of_arguments == 0)
      error(AL1_error,
            divide_text);
    number_expression = Frame[1];
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(number_expression);
          if (number_of_arguments > 1)
            { int_ = get_SMI(number_expression);
              number_expression = smallint_divide_M(Frame,
                                                    int_); }
          else if (int_ == 0)
            error(DBZ_error,
                  divide_text);
          else
            { float_ = 1.0 / int_;
              real = make_REA_M();
              real->flo = float_;
              number_expression = (EXP_type)real; }
          break;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          if (number_of_arguments > 1)
            number_expression = real_divide_M(Frame,
                                              float_,
                                              2);
          else if (float_ == 0)
            error(DBZ_error,
                  divide_text);
          else
            { float_ = 1.0 / float_;
              real = make_REA_M();
              real->flo = float_;
              number_expression = (EXP_type)real; }
          break;
       default:
         error(ARN_error,
               divide_text); }
    Environment_Release_Frame(Frame);
    Evaluate_Return(number_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_divide_native_M(VOI_type)
  { native_define_M(divide_text,
                    divide_native_M); }

/*---------------------------------------- abs -----------------------------------------*/

static const TXT_type abs_text = "abs";

static VOI_type abs_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    SMI_type smallint;
    TAG_type tag;
    require_1_argument(Frame,
                       &number_expression,
                       abs_text);
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(number_expression);
          int_ = C_ABS(int_);
          smallint = make_smallint(int_,
                                   abs_text);
          Evaluate_Return(smallint);
          return;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          float_ = C_FABS(float_);
              real = make_REA_M();
              real->flo = float_;
          Evaluate_Return(real);
          return;
       default:
          error(ARN_error,
                abs_text); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_abs_native_M(VOI_type)
  { native_define_M(abs_text,
                    abs_native_M); }

/*--------------------------------------- round ----------------------------------------*/

static const TXT_type round_text = "round";

static VOI_type round_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    REA_type real;
    TAG_type tag;
    require_1_argument(Frame,
                       &number_expression,
                       round_text);
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          Evaluate_Return(number_expression);
          return;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          float_ = C_ROUND(float_);
              real = make_REA_M();
              real->flo = float_;
          Evaluate_Return(real);
          return;
       default:
          error(ARN_error,
                round_text); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_round_native_M(VOI_type)
  { native_define_M(round_text,
                    round_native_M); }

/*------------------------------------- truncate ---------------------------------------*/

static const TXT_type truncate_text = "truncate";

static VOI_type truncate_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    TAG_type tag;
    REA_type real;
    require_1_argument(Frame,
                       &number_expression,
                       truncate_text);
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          Evaluate_Return(number_expression);
          return;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          float_ = C_TRUNC(float_);
          real = make_REA_M();
          real->flo = float_;
          Evaluate_Return(real);
          return;
       default:
          error(ARN_error,
                truncate_text); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_truncate_native_M(VOI_type)
  { native_define_M(truncate_text,
                    truncate_native_M); }

/*-------------------------------------- quotient --------------------------------------*/

static const TXT_type quotient_text = "quotient";

static EXP_type smallint_quotient_M(SMI_type Left_smallint,
                                    EXP_type Right_number_expression)
  { SGN_type left_int,
             right_int;
    LNG_type long_;
    SMI_type right_smallint,
             smallint;
    TAG_type tag;
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          left_int  = get_SMI(Left_smallint);
          right_smallint = (SMI_type)Right_number_expression;
          right_int = get_SMI(right_smallint);
          if (right_int == 0)
            error(DBZ_error,
                  quotient_text);
          else
            { long_ = (LNG_type)left_int / (LNG_type)right_int;
              smallint = make_smallint(long_,
                                       quotient_text);
              return smallint; }
        default:
          error(ARI_error,
                quotient_text); }
    return Unique_Zero_Smallint; }

static VOI_type quotient_native(FRM_type Frame)
  { EXP_type left_number_expression,
             number_expression,
             right_number_expression;
    SMI_type left_smallint;
    TAG_type tag;
    require_2_arguments(Frame,
                        &left_number_expression,
                        &right_number_expression,
                        quotient_text);
    tag = Grammar_Tag(left_number_expression);
    switch (tag)
      { case SMI_tag:
          left_smallint = (SMI_type)left_number_expression;
          number_expression = smallint_quotient_M(left_smallint,
                                                  right_number_expression);
          Evaluate_Return(number_expression);
          return;
        default:
          error(ARI_error,
                quotient_text); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_quotient_native_M(VOI_type)
  { native_define_M(quotient_text,
                    quotient_native); }

/*------------------------------------- remainder --------------------------------------*/

static const TXT_type remainder_text = "remainder";

static EXP_type smallint_remainder_M(SMI_type Left_smallint,
                                     EXP_type Right_number_expression)
  { SGN_type left_int,
             right_int;
    LNG_type long_;
    SMI_type right_smallint,
             smallint;
    TAG_type tag;
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          left_int  = get_SMI(Left_smallint);
          right_smallint = (SMI_type)Right_number_expression;
          right_int = get_SMI(right_smallint);
          if (right_int == 0)
            error(DBZ_error,
                  remainder_text);
          else
            { long_ = (LNG_type)left_int % (LNG_type)right_int;
              smallint = make_smallint(long_,
                                       remainder_text);
              return smallint; }
        default:
          error(ARI_error,
                remainder_text); }
    return Unique_Zero_Smallint; }

static VOI_type remainder_native(FRM_type Frame)
  { EXP_type left_number_expression,
             number_expression,
             right_number_expression;
    SMI_type left_smallint;
    TAG_type tag;
    require_2_arguments(Frame,
                        &left_number_expression,
                        &right_number_expression,
                        remainder_text);
    tag = Grammar_Tag(left_number_expression);
    switch (tag)
      { case SMI_tag:
          left_smallint = (SMI_type)left_number_expression;
          number_expression = smallint_remainder_M(left_smallint,
                                                   right_number_expression);
          Evaluate_Return(number_expression);
          return;
        default:
          error(ARI_error,
                remainder_text); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_remainder_native_M(VOI_type)
  { native_define_M(remainder_text,
                    remainder_native); }

/*-------------------------------------- modulo ----------------------------------------*/

static const TXT_type modulo_text = "modulo";

static EXP_type smallint_modulo_M(SMI_type Left_smallint,
                                  EXP_type Right_number_expression)
  { SGN_type left_int,
             right_int;
    LNG_type long_;
    SMI_type right_smallint,
             smallint;
    TAG_type tag;
    tag = Grammar_Tag(Right_number_expression);
    switch (tag)
      { case SMI_tag:
          left_int  = get_SMI(Left_smallint);
          right_smallint = (SMI_type)Right_number_expression;
          right_int = get_SMI(right_smallint);
          if (right_int == 0)
            error(DBZ_error,
                  modulo_text);
          else
            { if (right_int > 0)
                long_ = C_LLABS(left_int) % (LNG_type)right_int;
              else
                long_ = -C_LLABS(left_int) % (LNG_type)right_int;
              smallint = make_smallint(long_,
                                       modulo_text);
              return smallint; }
        default:
          error(ARI_error,
                modulo_text); }
    return Unique_Zero_Smallint; }

static VOI_type modulo_native(FRM_type Frame)
  { EXP_type left_number_expression,
             number_expression,
             right_number_expression;
    SMI_type left_smallint;
    TAG_type tag;
    require_2_arguments(Frame,
                        &left_number_expression,
                        &right_number_expression,
                        modulo_text);
    tag = Grammar_Tag(left_number_expression);
    switch (tag)
      { case SMI_tag:
          left_smallint = (SMI_type)left_number_expression;
          number_expression = smallint_modulo_M(left_smallint,
                                                right_number_expression);
          Evaluate_Return(number_expression);
          return;
        default:
          error(ARI_error,
                modulo_text); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_modulo_native_M(VOI_type)
  { native_define_M(modulo_text,
                    modulo_native); }

/*---------------------------------------- exp -----------------------------------------*/

static const TXT_type exp_text = "exp";

static VOI_type exp_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    require_1_argument(Frame,
                       &number_expression,
                       exp_text);
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(number_expression);
          float_ = C_EXP(int_);
          break;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          float_ = C_EXP(float_);
          break;
        default:
          float_ = 0;
          error(ARN_error,
                exp_text); }
    real = validate_float_M(float_,
                            exp_text);
    Evaluate_Return(real); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_exp_native_M(VOI_type)
  { native_define_M(exp_text,
                    exp_native_M); }

/*---------------------------------------- log -----------------------------------------*/

static const TXT_type log_text = "log";

static VOI_type log_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    require_1_argument(Frame,
                       &number_expression,
                       log_text);
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(number_expression);
          float_ = C_LOG(int_);
          break;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          float_ = C_LOG(float_);
          break;
        default:
          float_ = 0;
          error(ARN_error,
                log_text); }
    real = validate_float_M(float_,
                            log_text);
    Evaluate_Return(real); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_log_native_M(VOI_type)
  { native_define_M(log_text,
                    log_native_M); }

/*---------------------------------------- sin -----------------------------------------*/

static const TXT_type sin_text = "sin";

static VOI_type sin_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    require_1_argument(Frame,
                       &number_expression,
                       sin_text);
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(number_expression);
          float_ = C_SIN(int_);
          break;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          float_ = C_SIN(float_);
          break;
        default:
          float_ = 0;
          error(ARN_error,
                sin_text); }
    real = validate_float_M(float_,
                            sin_text);
    Evaluate_Return(real); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_sin_native_M(VOI_type)
  { native_define_M(sin_text,
                    sin_native_M); }

/*---------------------------------------- cos -----------------------------------------*/

static const TXT_type cos_text = "cos";

static VOI_type cos_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    require_1_argument(Frame,
                       &number_expression,
                       cos_text);
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(number_expression);
          float_ = C_COS(int_);
          break;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          float_ = C_COS(float_);
          break;
        default:
          float_ = 0;
          error(ARN_error,
                cos_text); }
    real = validate_float_M(float_,
                            cos_text);
    Evaluate_Return(real); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_cos_native_M(VOI_type)
  { native_define_M(cos_text,
                    cos_native_M); }

/*---------------------------------------- tan -----------------------------------------*/

static const TXT_type tan_text = "tan";

static VOI_type tan_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    require_1_argument(Frame,
                       &number_expression,
                       tan_text);
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(number_expression);
          float_ = C_TAN(int_);
          break;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          float_ = C_TAN(float_);
          break;
        default:
          float_ = 0;
          error(ARN_error,
                tan_text); }
    real = validate_float_M(float_,
                            tan_text);
    Evaluate_Return(real); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_tan_native_M(VOI_type)
  { native_define_M(tan_text,
                    tan_native_M); }

/*---------------------------------------- asin ----------------------------------------*/

static const TXT_type asin_text = "asin";

static VOI_type asin_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    require_1_argument(Frame,
                       &number_expression,
                       asin_text);
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(number_expression);
          float_ = C_ASIN(int_);
          break;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          float_ = C_ASIN(float_);
          break;
        default:
          float_ = 0;
          error(ARN_error,
                asin_text); }
    real = validate_float_M(float_,
                            asin_text);
    Evaluate_Return(real); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_asin_native_M(VOI_type)
  { native_define_M(asin_text,
                    asin_native_M); }

/*---------------------------------------- acos ----------------------------------------*/

static const TXT_type acos_text = "acos";

static VOI_type acos_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    require_1_argument(Frame,
                       &number_expression,
                       acos_text);
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(number_expression);
          float_ = C_ACOS(int_);
          break;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          float_ = C_ACOS(float_);
          break;
        default:
          float_ = 0;
          error(ARN_error,
                acos_text); }
    real = validate_float_M(float_,
                            acos_text);
    Evaluate_Return(real); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_acos_native_M(VOI_type)
  { native_define_M(acos_text,
                    acos_native_M); }

/*---------------------------------------- atan ----------------------------------------*/

static const TXT_type atan_text = "atan";

static VOI_type atan_native_M(FRM_type Frame)
  { EXP_type denominator_expression,
             numerator_expression;
    FLO_type denominator_float,
             float_;
    SGN_type denominator_int,
             int_;
    REA_type real;
    TAG_type tag;
    require_1_or_2_arguments(Frame,
                             &numerator_expression,
                             &denominator_expression,
                             atan_text);
    tag = Grammar_Tag(numerator_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(numerator_expression);
          tag = Grammar_Tag(denominator_expression);
          switch (tag)
            { case SMI_tag:
                denominator_int = get_SMI(denominator_expression);
                float_ = C_ATAN2(int_, denominator_int);
                break;
              case REA_tag:
                real = (REA_type)denominator_expression;
                denominator_float = real->flo;
                float_ = C_ATAN2(int_, denominator_float);
                break;
              case USP_tag:
                float_ = C_ATAN(int_);
                break;
              default:
                float_ = 0;
                error(ARN_error,
                      atan_text); }
          break;
        case REA_tag:
          real = (REA_type)numerator_expression;
          float_ = real->flo;
          tag = Grammar_Tag(denominator_expression);
          switch (tag)
            { case SMI_tag:
                denominator_int = get_SMI(denominator_expression);
                float_ = C_ATAN2(float_, denominator_int);
                break;
              case REA_tag:
                real = (REA_type)denominator_expression;
                denominator_float = real->flo;
                float_ = C_ATAN2(float_, denominator_float);
                break;
              case USP_tag:
                float_ = C_ATAN(float_);
                break;
              default:
                float_ = 0;
                error(ARN_error,
                      atan_text); }
          break;
        default:
          float_ = 0;
          error(ARN_error,
                atan_text); }
    real = validate_float_M(float_,
                            atan_text);
    Evaluate_Return(real); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_atan_native_M(VOI_type)
  { native_define_M(atan_text,
                    atan_native_M); }

/*---------------------------------------- sqrt ----------------------------------------*/

static const TXT_type sqrt_text = "sqrt";

static VOI_type sqrt_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    TAG_type tag;
    require_1_argument(Frame,
                       &number_expression,
                       sqrt_text);
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(number_expression);
          float_ = C_SQRT(int_);
          break;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          float_ = C_SQRT(float_);
          break;
        default:
          float_ = 0;
          error(ARN_error,
                sqrt_text); }
    real = validate_float_M(float_,
                            sqrt_text);
    Evaluate_Return(real); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_sqrt_native_M(VOI_type)
  { native_define_M(sqrt_text,
                    sqrt_native_M); }

/*---------------------------------------- expt ----------------------------------------*/

static const TXT_type expt_text = "expt";

static VOI_type expt_native_M(FRM_type Frame)
  { EXP_type base_expression,
             exponent_expression;
    FLO_type base_float,
             exponent_float,
             float_;
    SGN_type base_int,
             exponent_int;
    REA_type real;
    TAG_type tag;
    require_2_arguments(Frame,
                        &base_expression,
                        &exponent_expression,
                        expt_text);
    tag = Grammar_Tag(base_expression);
    switch (tag)
      { case SMI_tag:
          base_int = get_SMI(base_expression);
          tag = Grammar_Tag(exponent_expression);
          switch (tag)
            { case SMI_tag:
                exponent_int = get_SMI(exponent_expression);
                float_ = C_POW(base_int, exponent_int);
                break;
              case REA_tag:
                real = (REA_type)exponent_expression;
                exponent_float = real->flo;
                float_ = C_POW(base_int, exponent_float);
                break;
              default:
                float_ = 0;
                error(ARN_error,
                      expt_text); }
          break;
        case REA_tag:
          real = (REA_type)base_expression;
          base_float = real->flo;
          tag = Grammar_Tag(exponent_expression);
          switch (tag)
            { case SMI_tag:
                exponent_int = get_SMI(exponent_expression);
                float_ = C_POW(base_float, exponent_int);
                break;
              case REA_tag:
                real = (REA_type)exponent_expression;
                exponent_float = real->flo;
                float_ = C_POW(base_float,  exponent_float);
                break;
              default:
                float_ = 0;
                error(ARN_error,
                      expt_text); }
          break;
        default:
          float_ = 0;
          error(ARN_error,
                expt_text); }
    real = validate_float_M(float_,
                            expt_text);
    Evaluate_Return(real); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_expt_native_M(VOI_type)
  { native_define_M(expt_text,
                    expt_native_M); }

/*--------------------------------------------------------------------------------------*/
/*------------------------------ numerical input and output ----------------------------*/
/*--------------------------------------------------------------------------------------*/

/*----------------------------------- string->number -----------------------------------*/

static const TXT_type string_to_number_text = "string->number";

static VOI_type string_to_number_native_M(FRM_type Frame)
  { EXP_type string_expression;
    LNG_type long_;
    SMI_type smallint;
    STR_type string;
    TXT_type text;
    require_1_argument(Frame,
                       &string_expression,
                       string_to_number_text);
    string = cast_to_string(string_expression,
                            string_to_number_text);
    text = string->txt;
    long_ = C_STRTOL(text, VOID, 10);
    if (C_ERANGE)
      error(NTL_error,
            string_to_number_text);
    smallint = make_smallint(long_,
               string_to_number_text);
    Evaluate_Return(smallint); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_string_to_number_native_M(VOI_type)
  { native_define_M(string_to_number_text,
                    string_to_number_native_M); }

/*----------------------------------- number->string -----------------------------------*/

static const TXT_type number_to_string_text = "number->string";

static VOI_type number_to_string_native_M(FRM_type Frame)
  { EXP_type number_expression;
    FLO_type float_;
    SGN_type int_;
    REA_type real;
    STR_type string;
    TAG_type tag;
    UNS_type length;
    require_1_argument(Frame,
                       &number_expression,
                       number_to_string_text);
    tag = Grammar_Tag(number_expression);
    switch (tag)
      { case SMI_tag:
          int_ = get_SMI(number_expression);
          C_SNPRINTF(Grammar_Text_Buffer, Grammar_Text_Buffer_Length, "%d", int_);
          break;
        case REA_tag:
          real = (REA_type)number_expression;
          float_ = real->flo;
          C_SNPRINTF(Grammar_Text_Buffer, Grammar_Text_Buffer_Length, "%g", float_);
          break;
        default:
          error(ARN_error,
                number_to_string_text); }
    length = C_STRLEN(Grammar_Text_Buffer);
    if (length == 0)
      Evaluate_Return(Unique_Empty_String_P);
    else
      { string = make_STR_M(length);
        C_STRCPY(string->txt, Grammar_Text_Buffer);
        Evaluate_Return(string); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_number_to_string_native_M(VOI_type)
  { native_define_M(number_to_string_text,
                    number_to_string_native_M); }

/*--------------------------------------------------------------------------------------*/
/*-------------------------------------- characters ------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*----------------------------------------- char? --------------------------------------*/

static const TXT_type is_char_text = "char?";

static VOI_type is_char_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_char_text);
    RETURN_BOOLEAN(CHA, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_char_native_M(VOI_type)
  { native_define_M(is_char_text,
                    is_char_native); }

/*---------------------------------------- char=? --------------------------------------*/

static const TXT_type is_char_equal_text = "char=?";

static VOI_type is_char_equal_native(FRM_type Frame)
  { CHR_type left_char,
             right_char;
    EXP_type left_character_expression,
             right_character_expression;
    require_2_arguments(Frame,
                        &left_character_expression,
                        &right_character_expression,
                        is_char_equal_text);
    left_char = reify_character(left_character_expression,
                                is_char_equal_text);
    right_char = reify_character(left_character_expression,
                                 is_char_equal_text);
    if (left_char == right_char)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_char_equal_native_M(VOI_type)
  { native_define_M(is_char_equal_text,
                    is_char_equal_native); }

/*---------------------------------------- char<? --------------------------------------*/

static const TXT_type is_char_less_text = "char<?";

static VOI_type is_char_less_native(FRM_type Frame)
  { CHR_type left_char,
             right_char;
    EXP_type left_character_expression,
             right_character_expression;
    require_2_arguments(Frame,
                        &left_character_expression,
                        &right_character_expression,
                        is_char_less_text);
    left_char = reify_character(left_character_expression,
                                is_char_less_text);
    right_char = reify_character(left_character_expression,
                                 is_char_less_text);
    if (left_char < right_char)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_char_less_native_M(VOI_type)
  { native_define_M(is_char_less_text,
                    is_char_less_native); }

/*---------------------------------------- char>? --------------------------------------*/

static const TXT_type is_char_greater_text = "char>?";

static VOI_type is_char_greater_native(FRM_type Frame)
  { CHR_type left_char,
             right_char;
    EXP_type left_character_expression,
             right_character_expression;
    require_2_arguments(Frame,
                        &left_character_expression,
                        &right_character_expression,
                        is_char_greater_text);
    left_char = reify_character(left_character_expression,
                                is_char_greater_text);
    right_char = reify_character(left_character_expression,
                                 is_char_greater_text);
    if (left_char > right_char)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_char_greater_native_M(VOI_type)
  { native_define_M(is_char_greater_text,
                    is_char_greater_native); }

/*---------------------------------------- char<=? --------------------------------------*/

static const TXT_type is_char_less_or_equal_text = "char<=?";

static VOI_type is_char_less_or_equal_native(FRM_type Frame)
  { CHR_type left_char,
             right_char;
    EXP_type left_character_expression,
             right_character_expression;
    require_2_arguments(Frame,
                        &left_character_expression,
                        &right_character_expression,
                        is_char_less_or_equal_text);
    left_char = reify_character(left_character_expression,
                                is_char_less_or_equal_text);
    right_char = reify_character(left_character_expression,
                                 is_char_less_or_equal_text);
    if (left_char <= right_char)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_char_less_or_equal_native_M(VOI_type)
  { native_define_M(is_char_less_or_equal_text,
                    is_char_less_or_equal_native); }

/*---------------------------------------- char>=? --------------------------------------*/

static const TXT_type is_char_greater_or_equal_text = "char>=?";

static VOI_type is_char_greater_or_equal_native(FRM_type Frame)
  { CHR_type left_char,
             right_char;
    EXP_type left_character_expression,
             right_character_expression;
    require_2_arguments(Frame,
                        &left_character_expression,
                        &right_character_expression,
                        is_char_greater_or_equal_text);
    left_char = reify_character(left_character_expression,
                                is_char_greater_or_equal_text);
    right_char = reify_character(left_character_expression,
                                is_char_greater_or_equal_text);
    if (left_char >= right_char)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_char_greater_or_equal_native_M(VOI_type)
  { native_define_M(is_char_greater_or_equal_text,
                    is_char_greater_or_equal_native); }

/*------------------------------------ char->integer -----------------------------------*/

static const TXT_type char_to_integer_text = "char->integer";

static VOI_type char_to_integer_native_M(FRM_type Frame)
  { CHR_type char_;
    EXP_type character_expression;
    SGN_type int_;
    SMI_type smallint;
    require_1_argument(Frame,
                       &character_expression,
                       char_to_integer_text);
    char_ = reify_character(character_expression,
                            char_to_integer_text);
    int_ = (SGN_type)char_;
    smallint = make_SMI(int_);
    Evaluate_Return(smallint); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_char_to_integer_native_M(VOI_type)
  { native_define_M(char_to_integer_text,
                    char_to_integer_native_M); }

/*------------------------------------ integer->char -----------------------------------*/

static const TXT_type integer_to_char_text = "integer->char";

static VOI_type integer_to_char_native(FRM_type Frame)
  { CHA_type character;
    CHR_type char_;
    EXP_type smallint_expression;
    UNS_type unsigned_;
    require_1_argument(Frame,
                       &smallint_expression,
                       integer_to_char_text);
    unsigned_ = reify_unsigned(smallint_expression,
                               integer_to_char_text);
    if (unsigned_ > 255)
      error(COR_error,
            integer_to_char_text);
    char_ = (CHR_type)unsigned_;
    character = make_CHA(char_);
    Evaluate_Return(character); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_integer_to_char_native_M(VOI_type)
  { native_define_M(integer_to_char_text,
                    integer_to_char_native); }

/*--------------------------------------------------------------------------------------*/
/*-------------------------------------- strings ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

static UNS_type string_length(EXP_type String_expression,
                              TXT_type Message_text)
  { STR_type string;
    UNS_type length;
    string = cast_to_string(String_expression,
                            Message_text);
    length = length_STR(string);
    return length; }

/*--------------------------------------- string? --------------------------------------*/

static const TXT_type is_string_text = "string?";

static VOI_type is_string_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_string_text);
    RETURN_BOOLEAN(STR, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_string_native_M(VOI_type)
  { native_define_M(is_string_text,
                    is_string_native); }

/*------------------------------------- make-string ------------------------------------*/

static const TXT_type make_string_text = "make-string";

static VOI_type make_string_native_M(FRM_type Frame)
  { CHR_type char_;
    EXP_type character_expression,
             string_size_expression;
    STR_type string;
    TXT_type text;
    UNS_type length;
    require_1_or_2_arguments(Frame,
                             &string_size_expression,
                             &character_expression,
                             make_string_text);
    length = reify_unsigned(string_size_expression,
                            make_string_text);
    if (length == 0)
      { Evaluate_Return(Unique_Empty_String_P);
        return; }
    if (is_UDF(character_expression))
      char_ = Space_char;
    else
      char_ = reify_character(character_expression,
                              make_string_text);
    string = make_STR_M(length);
    text = string->txt;
    C_MEMSET(text, char_, length);
    text[length] = 0;
    Evaluate_Return(string); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_make_string_native_M(VOI_type)
  { native_define_M(make_string_text,
                    make_string_native_M); }

/*------------------------------------- string-copy ------------------------------------*/

static const TXT_type string_copy_text = "string-copy";

static VOI_type string_copy_native_M(FRM_type Frame)
  { EXP_type string_expression;
    STR_type copied_string;
    TXT_type copied_text,
             text;
    UNS_type length;
    require_1_argument(Frame,
                       &string_expression,
                       string_copy_text);
    String_P = cast_to_string(string_expression,
                              string_copy_text);
    length = length_STR(String_P);
    if (length == 0)
      Evaluate_Return(Unique_Empty_String_P);
    else
      { copied_string = make_STR_M(length);
        text = String_P->txt;
        copied_text = copied_string->txt;
        C_STRCPY(copied_text, text);
        Evaluate_Return(copied_string); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_string_copy_native_M(VOI_type)
  { native_define_M(string_copy_text,
                    string_copy_native_M); }

/*------------------------------------ string-length -----------------------------------*/

static const TXT_type string_length_text = "string-length";

/*--------------------------------------------------------------------------------------*/

static VOI_type string_length_native(FRM_type Frame)
  { EXP_type string_expression;
    SMI_type length_smallint;
    UNS_type length;
    require_1_argument(Frame,
                       &string_expression,
                       string_length_text);
    length = string_length(string_expression,
                           string_length_text);
    length_smallint = make_SMU(length);
    Evaluate_Return(length_smallint); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_string_length_M(VOI_type)
  { native_define_M(string_length_text,
                    string_length_native); }

/*------------------------------------ string->list ------------------------------------*/

static const TXT_type string_to_list_text = "string->list";

static VOI_type string_to_list_native_M(FRM_type Frame)
  { CHA_type character;
    CHR_type char_;
    EXP_type list_expression,
             string_expression;
    PAI_type pair;
    TXT_type text;
    UNS_type length,
             index;
    require_1_argument(Frame,
                       &string_expression,
                       string_to_list_text);
    String_P = cast_to_string(string_expression,
                              string_to_list_text);
    length = length_STR(String_P);
    if (length == 0)
      { Evaluate_Return(Unique_Null);
        return; }
    list_expression = Unique_Null;
    for (index = length;
         index > 0;)
      { index -= 1;
        text = String_P->txt;
        char_ = text[index];
        character = make_CHA(char_);
        Car_Expression_P = character;
        Cdr_Expression_P = list_expression;
        pair = make_PAI_M();
        pair->car = Car_Expression_P;
        pair->cdr = Cdr_Expression_P;
        list_expression = pair; }
    Evaluate_Return(list_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_string_to_list_native_M(VOI_type)
  { native_define_M(string_to_list_text,
                    string_to_list_native_M); }

/*------------------------------------ list->string ------------------------------------*/

static const TXT_type list_to_string_text = "list->string";

static UNS_type list_length(EXP_type,
                            TXT_type);

/*--------------------------------------------------------------------------------------*/

static VOI_type list_to_string_native_M(FRM_type Frame)
  { CHR_type char_;
    EXP_type value_expression;
    PAI_type pair;
    STR_type string;
    TXT_type text;
    UNS_type length,
             index;
    require_1_argument(Frame,
                       &List_Expression_P,
                       list_to_string_text);
    length = list_length(List_Expression_P,
                         list_to_string_text);
    if (length == 0)
      { Evaluate_Return(Unique_Empty_String_P);
        return; }
    string = make_STR_M(length);
    text = string->txt;
    for (index = 0;
         index < length;
         index += 1)
      { pair = (PAI_type)List_Expression_P;
        value_expression  = pair->car;
        List_Expression_P = pair->cdr;
        char_ = reify_character(value_expression,
                                list_to_string_text);
        text[index] = char_; }
    Evaluate_Return(string); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_list_to_string_native_M(VOI_type)
  { native_define_M(list_to_string_text,
                    list_to_string_native_M); }

/*------------------------------------- string-ref -------------------------------------*/

static const TXT_type string_ref_text = "string-ref";

static VOI_type string_ref_native(FRM_type Frame)
  { CHA_type character;
    CHR_type char_;
    EXP_type index_expression,
             string_expression;
    TXT_type text;
    UNS_type index,
             length;
    STR_type string;
    require_2_arguments(Frame,
                        &string_expression,
                        &index_expression,
                        string_ref_text);
    length = string_length(string_expression,
                           string_ref_text);
    index = reify_unsigned(index_expression,
                           string_ref_text);
    if (index >= length)
      error(IOR_error,
            string_ref_text);
    string = (STR_type)string_expression;
    text = string->txt;
    char_ = text[index];
    character = make_CHA(char_);
    Evaluate_Return(character); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_string_ref_native_M(VOI_type)
  { native_define_M(string_ref_text,
                    string_ref_native); }

/*------------------------------------- string-set! ------------------------------------*/

static const TXT_type string_set_text = "string-set!";

static VOI_type string_set_native(FRM_type Frame)
  { CHR_type char_;
    EXP_type character_expression,
             index_expression,
             string_expression;
    TXT_type text;
    UNS_type index,
             length;
    STR_type string;
    require_3_arguments(Frame,
                        &string_expression,
                        &index_expression,
                        &character_expression,
                        string_set_text);
    length = string_length(string_expression,
                           string_set_text);
    index = reify_unsigned(index_expression,
                           string_set_text);
    char_ = reify_character(character_expression,
                            string_set_text);
    if (index >= length)
      error(IOR_error,
            string_set_text);
    string = (STR_type)string_expression;
    text = string->txt;
    text[index] = char_;
    Evaluate_Return(string); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_string_set_native_M(VOI_type)
  { native_define_M(string_set_text,
                    string_set_native); }

/*-------------------------------------- string=? -------------------------------------*/

static const TXT_type is_string_equal_text = "string=?";

static VOI_type is_string_equal_native(FRM_type Frame)
  { EXP_type left_string_expression,
             right_string_expression;
    TXT_type left_text,
             right_text;
    require_2_arguments(Frame,
                        &left_string_expression,
                        &right_string_expression,
                        is_string_equal_text);
    left_text = reify_string(left_string_expression,
                             is_string_equal_text);
    right_text = reify_string(right_string_expression,
                              is_string_equal_text);
    if (C_STRCMP(left_text, right_text) == 0)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_string_equal_native_M(VOI_type)
  { native_define_M(is_string_equal_text,
                    is_string_equal_native); }

/*-------------------------------------- string<? -------------------------------------*/

static const TXT_type is_string_less_text = "string<?";

static VOI_type is_string_less_native(FRM_type Frame)
  { EXP_type left_string_expression,
             right_string_expression;
    TXT_type left_text,
             right_text;
    require_2_arguments(Frame,
                        &left_string_expression,
                        &right_string_expression,
                        is_string_less_text);
    left_text = reify_string(left_string_expression,
                             is_string_less_text);
    right_text = reify_string(right_string_expression,
                              is_string_less_text);
    if (C_STRCMP(left_text, right_text) < 0)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_string_less_native_M(VOI_type)
  { native_define_M(is_string_less_text,
                    is_string_less_native); }

/*-------------------------------------- string>? -------------------------------------*/

static const TXT_type is_string_greater_text = "string>?";

static VOI_type is_string_greater_native(FRM_type Frame)
  { EXP_type left_string_expression,
             right_string_expression;
    TXT_type left_text,
             right_text;
    require_2_arguments(Frame,
                        &left_string_expression,
                        &right_string_expression,
                        is_string_greater_text);
    left_text = reify_string(left_string_expression,
                             is_string_greater_text);
    right_text = reify_string(right_string_expression,
                              is_string_greater_text);
    if (C_STRCMP(left_text, right_text) > 0)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_string_greater_native_M(VOI_type)
  { native_define_M(is_string_greater_text,
                    is_string_greater_native); }

/*-------------------------------------- string<=? -------------------------------------*/

static const TXT_type is_string_less_or_equal_text = "string<=?";

static VOI_type is_string_less_or_equal_native(FRM_type Frame)
  { EXP_type left_string_expression,
             right_string_expression;
    TXT_type left_text,
             right_text;
    require_2_arguments(Frame,
                        &left_string_expression,
                        &right_string_expression,
                        is_string_less_or_equal_text);
    left_text = reify_string(left_string_expression,
                             is_string_less_or_equal_text);
    right_text = reify_string(right_string_expression,
                              is_string_less_or_equal_text);
    if (C_STRCMP(left_text, right_text) <= 0)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_string_less_or_equal_native_M(VOI_type)
  { native_define_M(is_string_less_or_equal_text,
                    is_string_less_or_equal_native); }

/*-------------------------------------- string>=? -------------------------------------*/

static const TXT_type is_string_greater_or_equal_text = "string>=?";

static VOI_type is_string_greater_or_equal_native(FRM_type Frame)
  { EXP_type left_string_expression,
             right_string_expression;
    TXT_type left_text,
             right_text;
    require_2_arguments(Frame,
                        &left_string_expression,
                        &right_string_expression,
                        is_string_greater_or_equal_text);
    left_text = reify_string(left_string_expression,
                             is_string_greater_or_equal_text);
    right_text = reify_string(right_string_expression,
                              is_string_greater_or_equal_text);
    if (C_STRCMP(left_text, right_text) >= 0)
      Evaluate_Return(TRUE_EXPRESSION);
    else
      Evaluate_Return(FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_string_greater_or_equal_native_M(VOI_type)
  { native_define_M(is_string_greater_or_equal_text,
                    is_string_greater_or_equal_native); }

/*-------------------------------------- substring -------------------------------------*/

static const TXT_type substring_text = "substring";

static VOI_type substring_native_M(FRM_type Frame)
  { EXP_type start_smallint_expression,
             string_expression,
             stop_smallint_expression;
    TXT_type text,
             sub_text;
    STR_type sub_string;
    UNS_type length,
             start_unsigned,
             stop_unsigned,
             sub_length;
    require_3_arguments(Frame,
                        &string_expression,
                        &start_smallint_expression,
                        &stop_smallint_expression,
                        substring_text);
    length = string_length(string_expression,
                           substring_text);
    start_unsigned = reify_unsigned(start_smallint_expression,
                                    substring_text);
    stop_unsigned = reify_unsigned(stop_smallint_expression,
                                   substring_text);
    if (start_unsigned >= length)
      error(IOR_error,
            string_ref_text);
    if (stop_unsigned >= length)
      error(IOR_error,
            string_ref_text);
    if (start_unsigned > stop_unsigned)
      error(IOR_error,
            string_ref_text);
    sub_length = stop_unsigned - start_unsigned;
    if (sub_length == 0)
      { Evaluate_Return(Unique_Empty_String_P);
        return; }
    String_P = (STR_type)string_expression;
    sub_string = make_STR_M(sub_length);
    text = String_P->txt;
    sub_text = sub_string->txt;
    C_STRNCPY(sub_text, text + start_unsigned, sub_length);
    sub_text[stop_unsigned + 1] = 0;
    Evaluate_Return(sub_string); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_substring_M(VOI_type)
  { native_define_M(substring_text,
                    substring_native_M); }

/*------------------------------------ string-append -----------------------------------*/

static const TXT_type string_append_text = "string-append";

static VOI_type string_append_native_M(FRM_type Frame)
  { EXP_type string_expression;
    STR_type appended_string,
             string;
    TXT_type appended_text,
             text;
    UNS_type accumulated_length,
             index,
             length,
             number_of_arguments;
    number_of_arguments = Environment_Frame_Size(Frame);
    if (number_of_arguments == 0)
      { Evaluate_Return(Unique_Empty_String_P);
        return; }
    accumulated_length = 0;
    for (index = 1;
         index <= number_of_arguments ;
         index += 1)
      { string_expression = Frame[index];
        length = string_length(string_expression,
                               string_append_text);
        accumulated_length += length; }
    if (accumulated_length == 0)
      { Evaluate_Return(Unique_Empty_String_P);
        return; }
    Frame_P = Frame;
    appended_string = make_STR_M(accumulated_length);
    appended_text = appended_string->txt;
    string = (STR_type)Frame_P[1];
    text = string->txt;
    C_STRCPY(appended_text, text);
    for (index = 2;
         index <= number_of_arguments ;
         index += 1)
      { string = (STR_type)Frame_P[index];
        text = string->txt;
        C_STRCAT(appended_text, text); }
    Evaluate_Return(appended_string); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_string_append_M(VOI_type)
  { native_define_M(string_append_text,
                    string_append_native_M); }

/*--------------------------------------------------------------------------------------*/
/*-------------------------------------- vectors ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*--------------------------------------- vector? --------------------------------------*/

static const TXT_type is_vector_text = "vector?";

static VOI_type is_vector_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_vector_text);
    RETURN_BOOLEAN(VEC, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_vector_native_M(VOI_type)
  { native_define_M(is_vector_text,
                    is_vector_native); }

/*------------------------------------- make-vector ------------------------------------*/

static const TXT_type make_vector_text = "make-vector";

static VOI_type make_vector_native_M(FRM_type Frame)
  { EXP_type vector_size_expression;
    UNS_type index,
             size;
    VEC_type vector;
    require_1_or_2_arguments(Frame,
                             &vector_size_expression,
                             &Expression_P,
                             make_vector_text);
    size = reify_unsigned(vector_size_expression,
                          make_vector_text);
    if (size == 0)
      { Evaluate_Return(Unique_Empty_Vector_P);
        return; }
    if (size > Maximum_Vector_Size)
      error(VTL_error,
            make_vector_text);
    vector = make_VEC_Z(size);
    for (index = 1;
         index <= size;
         index += 1)
      vector[index] = Expression_P;
    Evaluate_Return(vector); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_make_vector_native_M(VOI_type)
  { native_define_M(make_vector_text,
                    make_vector_native_M); }

/*--------------------------------------- vector ---------------------------------------*/

static const TXT_type vector_text = "vector";

static VOI_type vector_native_M(FRM_type Frame)
  { EXP_type expression;
    UNS_type index,
             number_of_arguments;
    VEC_type vector;
    number_of_arguments = Environment_Frame_Size(Frame);
    if (number_of_arguments == 0)
      { Evaluate_Return(Unique_Empty_Vector_P);
        return; }
    vector = make_VEC_Z(number_of_arguments);
    for (index = 1;
         index <= number_of_arguments;
         index += 1)
      { expression = Frame[index];
        vector[index] = expression; }
    Evaluate_Return(vector); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_vector_native_M(VOI_type)
  { native_define_M(vector_text,
                    vector_native_M); }

/*------------------------------------ vector-length -----------------------------------*/

static const TXT_type vector_length_text = "vector-length";

static VOI_type vector_length_native(FRM_type Frame)
  { EXP_type vector_expression;
    SMI_type vector_size_smallint;
    UNS_type vector_size;
    VEC_type vector;
    require_1_argument(Frame,
                       &vector_expression,
                       vector_length_text);
    vector = cast_to_vector(vector_expression,
                            vector_length_text);
    vector_size = size_VEC(vector);
    vector_size_smallint = make_SMU(vector_size);
    Evaluate_Return(vector_size_smallint); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_vector_length_native_M(VOI_type)
  { native_define_M(vector_length_text,
                    vector_length_native); }

/*------------------------------------- vector-ref -------------------------------------*/

static const TXT_type vector_ref_text = "vector-ref";

static VOI_type vector_ref_native(FRM_type Frame)
  { EXP_type index_expression,
             value_expression,
             vector_expression;
    UNS_type index,
             vector_size;
    VEC_type vector;
    require_2_arguments(Frame,
                        &vector_expression,
                        &index_expression,
                        vector_ref_text);
    vector = cast_to_vector(vector_expression,
                            vector_ref_text);
    index = reify_unsigned(index_expression,
                           vector_ref_text);
    vector_size = size_VEC(vector);
    if (index >= vector_size)
      error(IOR_error,
            vector_ref_text);
    value_expression = vector[index + 1];
    Evaluate_Return(value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_vector_ref_native_M(VOI_type)
  { native_define_M(vector_ref_text,
                    vector_ref_native); }

/*------------------------------------- vector-set! ------------------------------------*/

static const TXT_type vector_set_text = "vector-set!";

static VOI_type vector_set_native(FRM_type Frame)
  { EXP_type index_expression,
             value_expression,
             vector_expression;
    UNS_type index,
             vector_size;
    VEC_type vector;
    require_3_arguments(Frame,
                        &vector_expression,
                        &index_expression,
                        &value_expression,
                        vector_set_text);
    vector = cast_to_vector(vector_expression,
                            vector_set_text);
    index = reify_unsigned(index_expression,
                           vector_set_text);
    vector_size = size_VEC(vector);
    if (index >= vector_size)
      error(IOR_error,
            vector_set_text);
    vector[index + 1] = value_expression;
    Evaluate_Return(vector); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_vector_set_native_M(VOI_type)
  { native_define_M(vector_set_text,
                    vector_set_native); }

/*------------------------------------ vector->list ------------------------------------*/

static const TXT_type vector_to_list_text = "vector->list";

static VOI_type vector_to_list_native_M(FRM_type Frame)
  { EXP_type list_expression,
             vector_expression;
    PAI_type pair;
    UNS_type index,
             vector_size;
    require_1_argument(Frame,
                       &vector_expression,
                       vector_length_text);
    Vector_P = cast_to_vector(vector_expression,
                              vector_length_text);
    vector_size = size_VEC(Vector_P);
    list_expression = Unique_Null;
    for (index = vector_size;
         index >= 1;
         index -= 1)
      { Car_Expression_P = Vector_P[index];
        Cdr_Expression_P = list_expression;
        pair = make_PAI_M();
        pair->car = Car_Expression_P;
        pair->cdr = Cdr_Expression_P;
        list_expression = (EXP_type)pair; }
    Evaluate_Return(list_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_vector_to_list_native_M(VOI_type)
  { native_define_M(vector_to_list_text,
                    vector_to_list_native_M); }

/*------------------------------------ list->vector ------------------------------------*/

static const TXT_type list_to_vector_text = "list->vector";

static UNS_type list_length(EXP_type,
                       TXT_type);

/*--------------------------------------------------------------------------------------*/

static VOI_type list_to_vector_native_M(FRM_type Frame)
  { EXP_type value_expression;
    PAI_type pair;
    UNS_type index,
             length;
    VEC_type vector;
    require_1_argument(Frame,
                       &List_Expression_P,
                       list_to_vector_text);
    length = list_length(List_Expression_P,
                         list_to_vector_text);
    if (length == 0)
      { Evaluate_Return(Unique_Empty_Vector_P);
        return; }
    vector = make_VEC_Z(length);
    for (index = 1;
         index <= length;
         index += 1)
      { pair = List_Expression_P;
        value_expression = pair->car;
        vector[index] = value_expression;
        List_Expression_P = pair->cdr; }
    Evaluate_Return(vector); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_list_to_vector_native_M(VOI_type)
  { native_define_M(list_to_vector_text,
                    list_to_vector_native_M); }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------- procedures -------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*------------------------------------- procedure? -------------------------------------*/

static const TXT_type is_procedure_text = "procedure?";

static VOI_type is_procedure_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_procedure_text);
    RETURN_4_BOOLEAN(PRC, PRV, NAT, NAZ, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_procedure_native_M(VOI_type)
  { native_define_M(is_procedure_text,
                    is_procedure_native); }

/*--------------------------------------- apply ----------------------------------------*/

static const TXT_type apply_text = "apply";

static VOI_type apply_native_M(FRM_type Frame)
  { EXP_type operator_expression,
             operand_expression;
    require_2_arguments(Frame,
                        &operator_expression,
                        &operand_expression,
                        apply_text);
    Evaluate_Apply_M(operator_expression,
                     operand_expression); }

static VOI_type apply_native_T_M(FRM_type Frame)
  { EXP_type operator_expression,
             operand_expression;
    require_2_arguments(Frame,
                        &operator_expression,
                        &operand_expression,
                        apply_text);
    Evaluate_Apply_T_M(operator_expression,
                       operand_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_apply_native_M(VOI_type)
  { higher_order_native_define_M(apply_text,
                                 apply_native_M,
                                 apply_native_T_M); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------- control features ----------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*---------------------------- call-with-current-continuation --------------------------*/

static const TXT_type call_with_current_continuation_text
                                                       = "call-with-current-continuation";

/*--------------------------------------------------------------------------------------*/

static VOI_type call_with_current_continuation_native_M(FRM_type Frame)
  { PAI_type pair;
    require_1_argument(Frame,
                       &Expression_P,
                       call_with_current_continuation_text);
    Car_Expression_P = Evaluate_Make_Continuation_M();
    pair = make_PAI_M();
    pair->car = Car_Expression_P;
    pair->cdr = Unique_Null;
    Evaluate_Apply_M(Expression_P,
                     pair); }

static VOI_type call_with_current_continuation_native_T_M(FRM_type Frame)
  { PAI_type pair;
    require_1_argument(Frame,
                       &Expression_P,
                       call_with_current_continuation_text);
    Car_Expression_P = Evaluate_Make_Continuation_M();
    pair = make_PAI_M();
    pair->car = Car_Expression_P;
    pair->cdr = Unique_Null;
    Evaluate_Apply_T_M(Expression_P,
                       pair); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_call_with_current_continuation_native_M(VOI_type)
  { higher_order_native_define_M(call_with_current_continuation_text,
                                 call_with_current_continuation_native_M,
                                 call_with_current_continuation_native_T_M); }

/*-------------------------------------- for-each --------------------------------------*/

static const TXT_type for_each_text = "for-each";

static VOI_type for_each_native_M(FRM_type Frame)
  { EXP_type list_expression,
             operator_expression;
    require_2_arguments(Frame,
                        &operator_expression,
                        &list_expression,
                        for_each_text);
    Evaluate_For_Each_M(operator_expression,
                        list_expression); }

 static VOI_type for_each_native_T_M(FRM_type Frame)
  { EXP_type list_expression,
             operator_expression;
    require_2_arguments(Frame,
                        &operator_expression,
                        &list_expression,
                        for_each_text);
    Evaluate_For_Each_T_M(operator_expression,
                          list_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_for_each_native_M(VOI_type)
  { higher_order_native_define_M(for_each_text,
                                 for_each_native_M,
                                 for_each_native_T_M); }

/*--------------------------------------------------------------------------------------*/

static const TXT_type force_text = "force";

static VOI_type force_native_M(FRM_type Frame)
  { EXP_type expression;
    require_1_argument(Frame,
                       &expression,
                       force_text);
    Evaluate_Force_M(expression); }

static VOI_type force_native_T_M(FRM_type Frame)
  { EXP_type expression;
    require_1_argument(Frame,
                       &expression,
                       force_text);
    Evaluate_Force_T_M(expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_force_native_M(VOI_type)
  { higher_order_native_define_M(force_text,
                                 force_native_M,
                                 force_native_T_M); }

/*---------------------------------------- map -----------------------------------------*/

static const TXT_type map_text = "map";

static VOI_type map_native_M(FRM_type Frame)
  { EXP_type list_expression,
             operator_expression;
    require_2_arguments(Frame,
                        &operator_expression,
                        &list_expression,
                        map_text);
    Evaluate_Map_M(operator_expression,
                   list_expression); }

static VOI_type map_native_T_M(FRM_type Frame)
  { EXP_type list_expression,
             operator_expression;
    require_2_arguments(Frame,
                        &operator_expression,
                        &list_expression,
                        map_text);
    Evaluate_Map_T_M(operator_expression,
                     list_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_map_native_M(VOI_type)
  { higher_order_native_define_M(map_text,
                                 map_native_M,
                                 map_native_T_M); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------- input and output ----------------------------------*/
/*--------------------------------------------------------------------------------------*/

static const TXT_type is_input_port_text = "input-port?";

static VOI_type is_input_port_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_input_port_text);
    RETURN_BOOLEAN(IPT, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_input_port_native_M(VOI_type)
  { native_define_M(is_input_port_text,
                    is_input_port_native); }

/*------------------------------------- output-port? -----------------------------------*/

static const TXT_type is_output_port_text = "output-port?";

static VOI_type is_output_port_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_output_port_text);
    RETURN_BOOLEAN(OPT, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_output_port_native_M(VOI_type)
  { native_define_M(is_output_port_text,
                    is_output_port_native); }

/*----------------------------------- open-input-file ----------------------------------*/

static const TXT_type open_input_file_text = "open-input-file";

static VOI_type open_input_file_native_M(FRM_type Frame)
  { ADR_type address;
    EXP_type string_expression;
    IPT_type input_port;
    TXT_type text;
    UNS_type length;
    require_1_argument(Frame,
                       &string_expression,
                       open_input_file_text);
    String_P = cast_to_string(string_expression,
                              open_input_file_text);
    text = String_P->txt;
    address = Main_Open_Input_Port_M(text);
    if (!address)
      error(OIF_error,
            open_input_file_text);
    length = length_STR(String_P);
    input_port = make_IPT_M(length);
    text = String_P->txt;
    input_port->adr = address;
    C_STRCPY(input_port->txt, text);
    Evaluate_Return(input_port); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_open_input_file_native_M(VOI_type)
  { native_define_M(open_input_file_text,
                    open_input_file_native_M); }

/*----------------------------------- open-output-file ---------------------------------*/

static const TXT_type open_output_file_text = "open-output-file";

static VOI_type open_output_file_native_M(FRM_type Frame)
  { ADR_type address;
    EXP_type string_expression;
    OPT_type output_port;
    TXT_type text;
    UNS_type length;
    require_1_argument(Frame,
                       &string_expression,
                       open_output_file_text);
    String_P = cast_to_string(string_expression,
                              open_output_file_text);
    text = String_P->txt;
    address = Main_Open_Output_Port_M(text);
    if (!address)
      error(OOF_error,
            open_output_file_text);
    length = length_STR(String_P);
    output_port = make_OPT_M(length);
    text = String_P->txt;
    output_port->adr = address;
    C_STRCPY(output_port->txt, text);
    Evaluate_Return(output_port); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_open_output_file_native_M(VOI_type)
  { native_define_M(open_output_file_text,
                    open_output_file_native_M); }

/*----------------------------------- close-input-port ---------------------------------*/

static const TXT_type close_input_port_text = "close-input-port";

static VOI_type close_input_port_native(FRM_type Frame)
  { ADR_type address;
    EXP_type input_port_expression;
    IPT_type input_port;
    require_1_argument(Frame,
                       &input_port_expression,
                       close_input_port_text);
    input_port = cast_to_input_port(input_port_expression,
                                    close_input_port_text);
    address = input_port->adr;
    if (address == VOID)
      error(PAC_error,
            close_input_port_text);
    if (!Main_Close_Input_Port(address))
      error(CIP_error,
            close_input_port_text);
    input_port->adr = VOID;
    Evaluate_Return(input_port); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_close_input_port_native_M(VOI_type)
  { native_define_M(close_input_port_text,
                    close_input_port_native); }

/*---------------------------------- close-output-port ---------------------------------*/

static const TXT_type close_output_port_text = "close-output-port";

static VOI_type close_output_port_native(FRM_type Frame)
  { ADR_type output_address;
    EXP_type output_port_expression;
    OPT_type output_port;
    require_1_argument(Frame,
                       &output_port_expression,
                       close_output_port_text);
    output_port = cast_to_output_port(output_port_expression,
                                      close_output_port_text);
    output_address = output_port->adr;
    if (output_address == VOID)
      error(PAC_error,
            close_output_port_text);
    if (Main_Close_Output_Port(output_address))
      error(COP_error,
            close_output_port_text);
    output_port->adr = VOID;
    Evaluate_Return(output_port); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_close_output_port_native_M(VOI_type)
  { native_define_M(close_output_port_text,
                    close_output_port_native); }

/*--------------------------------------------------------------------------------------*/
/*---------------------------------------- input ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*----------------------------------------- read ---------------------------------------*/

static const TXT_type read_text = "read";

static IPT_type get_input_port(EXP_type Input_port_expression,
                               TXT_type Message_text)
  { IPT_type input_port;
    if (is_UDF(Input_port_expression))
      input_port = Default_Input_Port_P;
    else
      input_port = cast_to_input_port(Input_port_expression,
                                      Message_text);
    return input_port; }

/*--------------------------------------------------------------------------------------*/

static VOI_type read_native_M(FRM_type Frame)
  { EXP_type expression,
             input_port_expression;
    IPT_type input_port;
    require_0_or_1_arguments(Frame,
                             &input_port_expression,
                             read_text);
    input_port = get_input_port(input_port_expression,
                                read_text);
    expression = Read_Port_M(input_port);
    Evaluate_Return(expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_read_native_M(VOI_type)
  { UNS_type length;
    length = C_STRLEN(Default_input_port_text);
    Default_Input_Port_P = make_IPT_M(length);
    Default_Input_Port_P->adr = VOID;
    C_STRCPY(Default_Input_Port_P->txt, Default_input_port_text);
    native_define_M(read_text,
                    read_native_M); }

/*-------------------------------------- read-char -------------------------------------*/

static const TXT_type read_char_text = "read-char";

static VOI_type read_char_native(FRM_type Frame)
  { ADR_type address;
    CHA_type character;
    CHR_type char_;
    EXP_type input_port_expression;
    IPT_type input_port;
    STA_type status;
    require_0_or_1_arguments(Frame,
                             &input_port_expression,
                             read_char_text);
    input_port = get_input_port(input_port_expression,
                                read_char_text);
    address = input_port->adr;
    status = Main_Read_Char(&char_,
                            address);
    switch (status)
      { case Main_Aborted_Status:
          error(RDF_error,
                read_char_text);
          return;
        case Main_Terminated_Status:
          Evaluate_Return(Unique_Undefined);
          return;
        case Main_Successful_Status:
          break; }
    character = make_CHA(char_);
    Evaluate_Return(character); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_read_char_native_M(VOI_type)
  { native_define_M(read_char_text,
                    read_char_native); }

/*--------------------------------------- eof-object? -----------------------------------*/

static const TXT_type is_end_of_file_object_text = "eof-object?";

static VOI_type is_end_of_file_object_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_end_of_file_object_text);
    RETURN_BOOLEAN(END, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_end_of_file_object_M(VOI_type)
  { native_define_M(is_end_of_file_object_text,
                    is_end_of_file_object_native); }

/*--------------------------------------------------------------------------------------*/
/*--------------------------------------- output ---------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*--------------------------------------- display --------------------------------------*/

static const TXT_type display_text = "display";

static OPT_type get_output_port(EXP_type Output_port_expression,
                                TXT_type Message_text)
  { ADR_type output_address;
    OPT_type output_port;
    if (is_UDF(Output_port_expression))
      output_port = Default_Output_Port_P;
    else
      { output_port = cast_to_output_port(Output_port_expression,
                                        Message_text);
        output_address = output_port->adr;
        if (output_address == VOID)
          error(NOP_error,
                Message_text); }
    return output_port; }

/*--------------------------------------------------------------------------------------*/

static VOI_type display_native(FRM_type Frame)
  { EXP_type output_port_expression,
             value_expression;
    OPT_type output_port;
    require_1_or_2_arguments(Frame,
                             &value_expression,
                             &output_port_expression,
                             display_text);
    output_port = get_output_port(output_port_expression,
                                  display_text);
    Print_Display(value_expression,
                  output_port);
    Evaluate_Return(Unique_Unspecified);  }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_display_native_M(VOI_type)
  { UNS_type length;
    length = C_STRLEN(Default_output_port_text);
    Default_Output_Port_P = make_OPT_M(length);
    Default_Output_Port_P->adr = VOID;
    C_STRCPY(Default_Output_Port_P->txt, Default_output_port_text);
    native_define_M(display_text,
                    display_native); }

/*--------------------------------------- newline --------------------------------------*/

static const TXT_type newline_text = "newline";

static VOI_type newline_native(FRM_type Frame)
  { EXP_type output_port_expression;
    OPT_type output_port;
    require_0_or_1_arguments(Frame,
                             &output_port_expression,
                             newline_text);
    output_port = get_output_port(output_port_expression,
                                  newline_text);
    Print_Display(Unique_Newline_Character,
                  output_port);
    Evaluate_Return(Unique_Unspecified);  }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_newline_native_M(VOI_type)
  { native_define_M(newline_text,
                    newline_native); }

/*--------------------------------------- write ----------------------------------------*/

static const TXT_type write_text = "write";

static VOI_type write_native_M(FRM_type Frame)
  { EXP_type output_port_expression,
             value_expression;
    OPT_type output_port;
    require_1_or_2_arguments(Frame,
                             &value_expression,
                             &output_port_expression,
                             write_text);
    output_port = get_output_port(output_port_expression,
                                  write_text);
    Print_Write(value_expression,
                output_port);
    Evaluate_Return(Unique_Unspecified);  }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_write_native_M(VOI_type)
  { native_define_M(write_text,
                    write_native_M); }

/*------------------------------------ write-char --------------------------------------*/

static const TXT_type write_char_text = "write-char";

static VOI_type write_char_native(FRM_type Frame)
  { ADR_type address;
    CHR_type char_;
    EXP_type character_expression,
             output_port_expression;
    OPT_type output_port;
    STA_type status;
    require_1_or_2_arguments(Frame,
                             &character_expression,
                             &output_port_expression,
                             write_char_text);
    char_ = reify_character(character_expression,
                            write_char_text);
    output_port = get_output_port(output_port_expression,
                                  write_char_text);
    address = output_port->adr;
    status = Main_Write_Char(char_,
                             address);
    switch (status)
      { case Main_Aborted_Status:
          error(WRF_error,
                write_char_text);
          return;
        default:
          error(IRS_error,
                write_char_text); }
    Evaluate_Return(Unique_Unspecified); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_write_char_native_M(VOI_type)
  { native_define_M(write_char_text,
                    write_char_native); }

/*--------------------------------------------------------------------------------------*/
/*--------------------------------------- streams --------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*----------------------------------- the-empty-stream ---------------------------------*/

static const TXT_type the_empty_stream_text = "the-empty-stream";

static UNS_type the_empty_stream_offset;

static VOI_type reset_the_empty_stream(VOI_type)
  { enter_native_name(the_empty_stream_offset,
                      Unique_Null_Stream); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_the_empty_stream_M(VOI_type)
  { SYM_type symbol;
    symbol = Pool_Enter_Native_M(the_empty_stream_text);
    the_empty_stream_offset = Dictionary_Define_Variable_M(symbol);
    C_ASSERT(the_empty_stream_offset > 0);
    reset_the_empty_stream(); }

/*-------------------------------------- promise? --------------------------------------*/

static const TXT_type is_promise_text = "promise?";

static VOI_type is_promise_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_promise_text);
    RETURN_2_BOOLEAN(FRC, PRM, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_promise_native_M(VOI_type)
  { native_define_M(is_promise_text,
                    is_promise_native); }

/*------------------------------------ stream-null? ------------------------------------*/

static const TXT_type is_stream_null_text = "stream-null?";

static VOI_type is_stream_null_native(FRM_type Frame)
  { EXP_type value_expression;
    require_1_argument(Frame,
                       &value_expression,
                       is_stream_null_text);
    RETURN_BOOLEAN(NST, value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_is_stream_null_native_M(VOI_type)
  { native_define_M(is_stream_null_text,
                    is_stream_null_native); }

/*------------------------------------- stream-car -------------------------------------*/

static const TXT_type stream_car_text = "stream-car";

static VOI_type stream_car_native(FRM_type Frame)
  { EXP_type delayed_expression,
             pair_expression,
             value_expression;
    require_1_argument(Frame,
                       &pair_expression,
                       stream_car_text);
    value_expression = cast_to_car(pair_expression,
                                   stream_car_text);
    delayed_expression = cast_to_cdr(pair_expression,
                                     stream_car_text);
    if (!(is_FRC(delayed_expression) ||
          is_PRM(delayed_expression)))
      error(NPR_error,
            stream_car_text);
    Evaluate_Return(value_expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_stream_car_native_M(VOI_type)
  { native_define_M(stream_car_text,
                    stream_car_native); }

/*------------------------------------- stream-cdr -------------------------------------*/

static const TXT_type stream_cdr_text = "stream-cdr";

static VOI_type stream_cdr_native_M(FRM_type Frame)
  { EXP_type expression;
    PAI_type pair;
    require_1_argument(Frame,
                       &expression,
                       stream_cdr_text);
    if (!is_PAI(expression))
      error(NAP_error,
            stream_cdr_text);
    pair = (PAI_type)expression;
    expression = pair->cdr;
    Evaluate_Force_M(expression); }

static VOI_type stream_cdr_native_T_M(FRM_type Frame)
  { EXP_type expression;
    PAI_type pair;
    require_1_argument(Frame,
                       &expression,
                       stream_cdr_text);
    if (!is_PAI(expression))
      error(NAP_error,
            stream_cdr_text);
    pair = (PAI_type)expression;
    expression = pair->cdr;
    Evaluate_Force_T_M(expression); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_stream_cdr_native_M(VOI_type)
  { higher_order_native_define_M(stream_cdr_text,
                                 stream_cdr_native_M,
                                 stream_cdr_native_T_M); }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------- extensions -------------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*----------------------------------- circularity-level --------------------------------*/

static const TXT_type circularity_level_text = "circularity-level";

static UNS_type circularity_level_offset;

static VOI_type reset_circularity_level(VOI_type)
  { enter_native_name(circularity_level_offset,
                      Unique_Zero_Smallint); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_circularity_level_M(VOI_type)
  { SYM_type symbol;
    symbol = Pool_Enter_Native_M(circularity_level_text);
    circularity_level_offset = Dictionary_Define_Variable_M(symbol);
    C_ASSERT(circularity_level_offset > 0);
    reset_circularity_level(); }

/*------------------------------------------ false -------------------------------------*/

static const TXT_type false_text = "false";

static UNS_type false_offset;

static VOI_type reset_false(VOI_type)
  { enter_native_name(false_offset,
                      FALSE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_false_M(VOI_type)
  { SYM_type symbol;
    symbol = Pool_Enter_Native_M(false_text);
    false_offset = Dictionary_Define_Variable_M(symbol);
    C_ASSERT(false_offset > 0);
    reset_false(); }

/*------------------------------------------ true --------------------------------------*/

static const TXT_type true_text = "true";

static UNS_type true_offset;

static VOI_type reset_true(VOI_type)
  { enter_native_name(true_offset,
                      TRUE_EXPRESSION); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_true_M(VOI_type)
  { SYM_type symbol;
    symbol = Pool_Enter_Native_M(true_text);
    true_offset = Dictionary_Define_Variable_M(symbol);
    C_ASSERT(true_offset > 0);
    reset_true(); }

/*--------------------------------------- clock ----------------------------------------*/

static const TXT_type clock_text = "clock";

static FLO_type start_time;

static VOI_type clock_native(FRM_type Frame)
  { FLO_type float_;
    REA_type real;
    require_0_arguments(Frame,
                        clock_text);
    float_ = (FLO_type)C_CLOCK;
    float_ -= start_time;
    real = make_REA_M();
    real->flo = float_;
    Evaluate_Return(real); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_clock_native_M(VOI_type)
  { start_time = (FLO_type)C_CLOCK;
    native_define_M(clock_text,
                    clock_native); }

/*-------------------------------------- collect ---------------------------------------*/

static const TXT_type collect_text = "collect";

static VOI_type collect_native_M(FRM_type Frame)
  { require_0_arguments(Frame,
                        collect_text);
    Grammar_Force_Collect_M();
    Evaluate_Return(Unique_Unspecified); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_collect_native_M(VOI_type)
  { native_define_M(collect_text,
                    collect_native_M); }

/*---------------------------------------- exit ----------------------------------------*/

static const TXT_type exit_text = "exit";

static VOI_type escape(VOI_type)
  { Environment_Release_Current_Frame();
    Environment_Reset();
    Thread_Reset();
    C_LONGJMP(Exit_point, Terminate_jump); }

static VOI_type exit_native(FRM_type Frame)
  { EXP_type value_expression;
    UNS_type index,
             size;
    size = Environment_Frame_Size(Frame);
    for (index = 1;
         index <= size;
         index += 1)
      { value_expression = Frame[index];
        Print_Print(value_expression,
                    Default_Output_Port_P); }
    escape(); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_exit_native_M(VOI_type)
  { native_define_M(exit_text,
                    exit_native); }

/*--------------------------------------- format ---------------------------------------*/

static const TXT_type format_text = "format";

/*--------------------------------------------------------------------------------------*/

static VOI_type format_native_M(FRM_type Frame)
  { ADR_type address;
    EXP_type expression,
             output_port_expression,
             string_expression;
    TXT_type text;
    UNS_type length;
    require_1_or_2_arguments(Frame,
                             &string_expression,
                             &output_port_expression,
                             write_text);
    Output_Port_P = get_output_port(output_port_expression,
                                    format_text);
    String_P = cast_to_string(string_expression,
                              format_text);
    text = String_P->txt;
    address = Main_Open_Input_Port_M(text);
    if (!address)
      error(OIF_error,
            format_text);
    length = length_STR(String_P);
    Input_Port_P = make_IPT_M(length);
    text = String_P->txt;
    Input_Port_P->adr = address;
    C_STRCPY(Input_Port_P->txt, text);
    expression = Read_Port_M(Input_Port_P);
    if (!Main_Close_Input_Port(address))
      error(CIP_error,
            format_text);
    Input_Port_P->adr = VOID;
    expression = Compile_M(expression);
    Print_Format(expression,
                 Output_Port_P);
    Evaluate_Return(Unique_Unspecified); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_format_native_M(VOI_type)
  { native_define_M(format_text,
                    format_native_M); }

/*--------------------------------------- gensym ---------------------------------------*/

static const TXT_type gensym_text = "gensym";

static VOI_type gensym_native_M(FRM_type Frame)
  { EXP_type expression;
    SYM_type symbol;
    UNS_type counter;
    require_1_argument(Frame,
                       &expression,
                       gensym_text);
    if (!is_SYM(expression))
      { error(NSY_error,
              string_to_symbol_text);
        return; }
    symbol = (SYM_type)expression;
    counter = Main_Count();
    symbol = Pool_Generate_M(symbol,
                             counter);
    Evaluate_Return(symbol); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_gensym_native_M(VOI_type)
  { native_define_M(gensym_text,
                    gensym_native_M); }

/*--------------------------------------- random ---------------------------------------*/

static const TXT_type random_text = "random";

static VOI_type random_native(FRM_type Frame)
  { SGN_type random;
    SMI_type smallint;
    require_0_arguments(Frame,
                        random_text);
    random = C_RAND();
    random >>= 2;
    smallint = make_SMI(random);
    Evaluate_Return(smallint); }

/*--------------------------------------------------------------------------------------*/

static VOI_type initialize_random_native_M(VOI_type)
  { native_define_M(random_text,
                    random_native); }

/*----------------------------------- public functions ---------------------------------*/

VOI_type Native_Exit(VOI_type)
  { Dictionary_Restore();
    escape(); }

VOI_type Native_REP(TXT_type Text,
                    BLN_type Boolean)
  { EXP_type expression;
    UNS_type frame_size;
    if (Boolean)
      { Grammar_Initialize_M();
        Evaluate_Initialize_M(); }
    Thread_Reset();
    Dictionary_Save();
    reset_the_empty_stream();
    reset_circularity_level();
    reset_false();
    reset_true();
    if (C_SETJMP(Exit_point) == Initial_jump)
      { expression = Read_String_M(Text);
        Expression_P = Compile_M(expression);
        frame_size = Dictionary_Get_Frame_Size();
        Environment_Adjust_Current_Frame_Size_M(frame_size);
        expression = Evaluate_M(Expression_P);
        Print_Print(expression,
                    Default_Output_Port_P); }}

/*------------------------------------- initialize -------------------------------------*/

VOI_type Native_Initialize_M(VOI_type)
  { Read_Initialize_M();
    Print_Initialize();

    initialize_abs_native_M();
    initialize_acos_native_M();
    initialize_add_native_M();
    initialize_append_native_M();
    initialize_apply_native_M();
    initialize_asin_native_M();
    initialize_assoc_native_M();
    initialize_assq_native_M();
    initialize_atan_native_M();
    initialize_call_with_current_continuation_native_M();
    initialize_car_native_M();
    initialize_caar_native_M();
    initialize_caaar_native_M();
    initialize_caaaar_native_M();
    initialize_caaadr_native_M();
    initialize_caadr_native_M();
    initialize_caadar_native_M();
    initialize_caaddr_native_M();
    initialize_cadr_native_M();
    initialize_cadar_native_M();
    initialize_cadaar_native_M();
    initialize_cadadr_native_M();
    initialize_caddr_native_M();
    initialize_caddar_native_M();
    initialize_cadddr_native_M();
    initialize_cdr_native_M();
    initialize_cdar_native_M();
    initialize_cdaar_native_M();
    initialize_cdaaar_native_M();
    initialize_cdaadr_native_M();
    initialize_cdadr_native_M();
    initialize_cdadar_native_M();
    initialize_cdaddr_native_M();
    initialize_cddr_native_M();
    initialize_cddar_native_M();
    initialize_cddaar_native_M();
    initialize_cddadr_native_M();
    initialize_cdddr_native_M();
    initialize_cdddar_native_M();
    initialize_cddddr_native_M();
    initialize_circularity_level_M();
    initialize_cos_native_M();
    initialize_char_to_integer_native_M();
    initialize_clock_native_M();
    initialize_close_input_port_native_M();
    initialize_close_output_port_native_M();
    initialize_collect_native_M();
    initialize_cons_native_M();
    initialize_display_native_M();
    initialize_divide_native_M();
    initialize_exit_native_M();
    initialize_exp_native_M();
    initialize_expt_native_M();
    initialize_false_M();
    initialize_for_each_native_M();
    initialize_force_native_M();
    initialize_format_native_M();
    initialize_gensym_native_M();
    initialize_integer_to_char_native_M();
    initialize_is_boolean_native_M();
    initialize_is_char_equal_native_M();
    initialize_is_char_greater_native_M();
    initialize_is_char_greater_or_equal_native_M();
    initialize_is_char_less_native_M();
    initialize_is_char_less_or_equal_native_M();
    initialize_is_char_native_M();
    initialize_is_end_of_file_object_M();
    initialize_is_eq_native_M();
    initialize_is_equal_native_M();
    initialize_is_even_native_M();
    initialize_is_integer_native_M();
    initialize_is_input_port_native_M();
    initialize_is_number_equal_native_M();
    initialize_is_number_greater_native_M();
    initialize_is_number_greater_or_equal_native_M();
    initialize_is_number_less_native_M();
    initialize_is_number_less_or_equal_native_M();
    initialize_is_negative_native_M();
    initialize_is_null_native_M();
    initialize_is_number_native_M();
    initialize_is_odd_native_M();
    initialize_is_output_port_native_M();
    initialize_is_positive_native_M();
    initialize_is_pair_native_M();
    initialize_is_procedure_native_M();
    initialize_is_promise_native_M();
    initialize_is_real_native_M();
    initialize_is_stream_null_native_M();
    initialize_is_string_equal_native_M();
    initialize_is_string_greater_native_M();
    initialize_is_string_greater_or_equal_native_M();
    initialize_is_string_less_native_M();
    initialize_is_string_less_or_equal_native_M();
    initialize_is_string_native_M();
    initialize_is_symbol_native_M();
    initialize_is_vector_native_M();
    initialize_is_zero_native_M();
    initialize_length_native_M();
    initialize_list_native_M();
    initialize_list_to_string_native_M();
    initialize_list_to_vector_native_M();
    initialize_log_native_M();
    initialize_make_string_native_M();
    initialize_make_vector_native_M();
    initialize_map_native_M();
    initialize_max_native_M();
    initialize_member_native_M();
    initialize_memq_native_M();
    initialize_min_native_M();
    initialize_modulo_native_M();
    initialize_multiply_native_M();
    initialize_newline_native_M();
    initialize_number_to_string_native_M();
    initialize_not_native_M();
    initialize_open_input_file_native_M();
    initialize_open_output_file_native_M();
    initialize_quotient_native_M();
    initialize_random_native_M();
    initialize_read_native_M();
    initialize_read_char_native_M();
    initialize_remainder_native_M();
    initialize_reverse_native_M();
    initialize_round_native_M();
    initialize_set_car_native_M();
    initialize_set_cdr_native_M();
    initialize_sin_native_M();
    initialize_sqrt_native_M();
    initialize_stream_car_native_M();
    initialize_stream_cdr_native_M();
    initialize_string_append_M();
    initialize_string_copy_native_M();
    initialize_string_length_M();
    initialize_string_ref_native_M();
    initialize_string_set_native_M();
    initialize_string_to_list_native_M();
    initialize_string_to_number_native_M();
    initialize_string_to_symbol_native_M();
    initialize_substring_M();
    initialize_subtract_native_M();
    initialize_symbol_to_string_native_M();
    initialize_tan_native_M();
    initialize_the_empty_stream_M();
    initialize_true_M();
    initialize_truncate_native_M();
    initialize_vector_length_native_M();
    initialize_vector_native_M();
    initialize_vector_ref_native_M();
    initialize_vector_set_native_M();
    initialize_vector_to_list_native_M();
    initialize_write_native_M();
    initialize_write_char_native_M();
    
    REGISTER(Car_Expression_P);
    REGISTER(Cdr_Expression_P);
    REGISTER(Symbol_P);
    REGISTER(Default_Input_Port_P);
    REGISTER(Default_Output_Port_P);
    REGISTER(Expression_P);
    REGISTER(Frame_P);
    REGISTER(Input_Port_P);
    REGISTER(List_Expression_P);
    REGISTER(Output_Port_P);
    REGISTER(Pair_P);
    REGISTER(String_P);
    REGISTER(Vector_P); }

