                     /*---------------------------------------------*/
                     /*                  >>>Slip<<<                 */
                     /*                 Theo D'Hondt                */
                     /*          VUB Software Languages Lab         */
                     /*                  (c) 2018                   */
                     /*---------------------------------------------*/
                     /*                final version                */
                     /*---------------------------------------------*/
                     /*                    pool                     */
                     /*---------------------------------------------*/

#include "SlipPlatform.h"

#include "SlipGrammar.h"

#include "SlipPool.h"

/*----------------------------------- private macros -----------------------------------*/

#define UNDEFINED_SYMBOL (SYM_type)Unique_Undefined
#define UNDEFINED_CHUNK  (CHK_type)Unique_Undefined

/*---------------------------------- private constants ---------------------------------*/

enum { Initial_pool_size   = 512,
       Keyword_table_size  =  25,
       Pool_size_increment =  32 };

static const TXT_type Quasisymbol_format = "##%s%d";

static const TXT_type Keyword_table[Keyword_table_size] =
  { "and",
    "begin",
    "cond",
    "cons-stream",
    "define",
    "delay",
    "else",
    "eval",
    "if",
    "lambda",
    "let",
    "load",
    "macro",
    "or",
    "quasiquote",
    "quasisymbol",
    "quote",
    "set!",
    "unquote",
    "unquote-splicing",
    "while",
    "anonymous",
    "application",
    "expression",
    "sequence" };

/*---------------------------------- private variables ---------------------------------*/

static UNS_type Pool_size;
static UNS_type Top_of_pool;

/*--------------------------------- protected variables --------------------------------*/

static EXP_type Expression_P;
static VEC_type Keyword_vector_P;
static PAI_type Pair_P;
static VEC_type Pool_vector_P;

/*--------------------------------------------------------------------------------------*/

static SYM_type enter_symbol_M(TXT_type text)
  { SYM_type symbol;
    UNS_type length;
    length = C_STRLEN(text);
    symbol = make_SYM_M(length);
    C_STRCPY(symbol->txt, text);
    Pool_vector_P[Top_of_pool] = symbol;
    Top_of_pool += 1;
    return symbol; }

static SYM_type search_in_pool(TXT_type Text)
  { TXT_type text;
    SYM_type symbol;
    UNS_type index;
    for (index = 1;
         index < Top_of_pool;
         index += 1)
      { symbol = Pool_vector_P[index];
        text = symbol->txt;
        if (C_STRCMP(text, Text) == 0)
          return symbol; }
    return UNDEFINED_SYMBOL; }

static VOI_type extend_pool_if_necessary_M(VOI_type)
  { VEC_type pool_vector;
    UNS_type index;
    if (Top_of_pool == Pool_size)
      { Pool_size += Pool_size_increment;
        pool_vector = make_VEC_Z(Pool_size);
        for (index = 1;
             index <= Top_of_pool;
             index += 1)
          pool_vector[index] = Pool_vector_P[index];
        Pool_vector_P = pool_vector; }}

/*---------------------------------- public functions ----------------------------------*/

SYM_type Pool_Enter_Native_M(TXT_type Text)
  { SYM_type symbol;
    C_ASSERT(Top_of_pool < Pool_size);
    symbol = enter_symbol_M(Text);
    return symbol; }

SYM_type Pool_Enter_Text_M(TXT_type Symbol_text)
  { SYM_type symbol;
    symbol = search_in_pool(Symbol_text);
    if (!is_UDF(symbol))
      return symbol;
    extend_pool_if_necessary_M();
    symbol = enter_symbol_M(Symbol_text);
    return symbol; }

SYM_type Pool_Generate_M(SYM_type Symbol,
                         UNS_type Counter)
  { SYM_type symbol;
    TXT_type text;
    text = Symbol->txt;
    C_SNPRINTF(Grammar_Text_Buffer, Grammar_Text_Buffer_Length, Quasisymbol_format, text, Counter);
    symbol = search_in_pool(Grammar_Text_Buffer);
    if (!is_UDF(symbol))
      return symbol;
    extend_pool_if_necessary_M();
    symbol = enter_symbol_M(Grammar_Text_Buffer);
    return symbol; }

SYM_type Pool_Keyword_Symbol(KEY_type Key)
  { SYM_type symbol;
    symbol = Keyword_vector_P[Key + 1];
    return symbol; }

PAI_type Pool_Quotation_M(KEY_type Key,
                          EXP_type Expression)
  { PAI_type pair;
    Expression_P = Expression;
    Pair_P = make_PAI_M();
    Pair_P->car = Expression_P;
    Pair_P->cdr = Unique_Null;
    pair = make_PAI_M();
    pair->car = Pool_Keyword_Symbol(Key);
    pair->cdr = Pair_P;
    return pair; }

BLN_type Pool_Validate_Symbol(SYM_type Symbol)
  { UNS_type keyword_index;
    for (keyword_index = Pool_And;
         keyword_index <= Pool_While;
         keyword_index += 1)
      if (Symbol == Keyword_vector_P[keyword_index])
        return TRUE;
    return FALSE; }

/*------------------------------------- initialize -------------------------------------*/

VOI_type Pool_Initialize_M(VOI_type)
  { TXT_type keyword_text;
    UNS_type index;
    Top_of_pool = 1;
    Keyword_vector_P = make_VEC_Z(Keyword_table_size);
    Pool_vector_P = make_VEC_Z(Initial_pool_size);
    Pool_size = Initial_pool_size;
    for (index = 1;
         index <= Keyword_table_size;
         index += 1)
      { keyword_text = Keyword_table[index - 1];
        Keyword_vector_P[index] = Pool_Enter_Native_M(keyword_text); }
    REGISTER(Expression_P);
    REGISTER(Keyword_vector_P);
    REGISTER(Pair_P);
    REGISTER(Pool_vector_P); }
