                     /*---------------------------------------------*/
                     /*                  >>>Slip<<<                 */
                     /*                 Theo D'Hondt                */
                     /*          VUB Software Languages Lab         */
                     /*                  (c) 2018                   */
                     /*---------------------------------------------*/
                     /*                final version                */
                     /*---------------------------------------------*/
                     /*                    print                    */
                     /*---------------------------------------------*/

#include "SlipPlatform.h"

#include "SlipMain.h"

#include "SlipGrammar.h"

#include "SlipDictionary.h"
#include "SlipPool.h"

#include "SlipPrint.h"


/*----------------------------------- private macros -----------------------------------*/

#define FORMATTED_ITEM(FORMAT, ITEM, PORT)                                               \
  { snprintf(Text_buffer,                                                                \
             Text_buffer_size,                                                           \
             FORMAT,                                                                     \
             ITEM);                                                                      \
    text_to_port(Text_buffer,                                                            \
                 PORT); }

/*---------------------------------- private constants ---------------------------------*/

enum { Indentation_step = 2,
       Print_level_max  = 8,
       Print_size_max   = 128,
       Text_buffer_size = 256 };

enum { Backslash_char      = '\\',
       Doublequote_char    = '"',
       Illegal_char        = '?',
       Leftpar_char        = '(',
       Newline_char        = '\n',
       Overflow_char       = '~',
       Quasiquote_char     = '`',
       Quote_char          = '\'',
       Rightpar_char       = ')',
       Space_char          = ' ',
       Unprintable_char    = '?',
       Unquote_char        = ',' };

static const TXT_type Display_text             = "display",
                      Format_text              = "format",
                      Print_text               = "print",
                      Write_text               = "write";

static const TXT_type Char_text                = "#\\%c",
                      Continuation_text        = "#<continuation>",
                      Ellipsis_text            = " ... ",
                      Empty_list_text          = "()",
                      Empty_stream_text        = "#<empty-stream>",
                      Empty_text_text          = "\"\"",
                      Eof_object_text          = "#<eof-object>",
                      False_text               = "#f",
                      Floating_text            = "%#.10f",
                      Forced_text              = "#<forced>",
                      Higher_order_native_text = "#<higher order native %s>",
                      Input_port_text          = "#<input port %s>",
                      Integer_text             = "%ld",
                      Macro_text               = "#<macro %s>",
                      Native_text              = "#<native %s>",
                      Output_port_text         = "#<output port %s>",
                      Pair_text                = "( ... )",
                      Period_text              = " . ",
                      Procedure_text           = "#<procedure %s>",
                      Promise_text             = "#<promise>",
                      Quasisymbol_text         = "`@",
                      True_text                = "#t",
                      Unquote_splice_text      = ",@",
                      Unspecified_text         = "<unspecified>",
                      Vector_text              = "#(",
                      Vector0_text             = "#()";

/*---------------------------------- private variables ---------------------------------*/

static UNS_type Indentation;
static UNS_type Print_level;
static UNS_type Print_size;
static CHR_type Text_buffer[Text_buffer_size];

/*---------------------------------- private functions ---------------------------------*/

static VOI_type error(UNS_type Error,
                      TXT_type Message_text)
  { Main_Error(Error,
               Message_text); }

static VOI_type char_to_port(CHR_type Char,
                             OPT_type Output_port)
  { ADR_type address;
    address = Output_port->adr;
    Main_Write_Char(Char,
                    address); }

static VOI_type text_to_port(TXT_type Text,
                             OPT_type Output_port)
  { ADR_type address;
    CHR_type char_;
    UNS_type length,
             position;
    length = C_STRLEN(Text);
    for (position = 0;
         position < length;
         position += 1)
    { char_ = Text[position];
      address = Output_port->adr;
      Main_Write_Char(char_,
                      address); }}

/*--------------------------------------- display --------------------------------------*/

static VOI_type display_expression(EXP_type,
                                   OPT_type);
static VOI_type       display_pair(PAI_type,
                                   OPT_type);

static VOI_type display_character(CHA_type Character,
                                  OPT_type Output_port)
  { CHR_type char_;
    char_ = get_CHA(Character);
    char_to_port(char_,
                 Output_port); }

static VOI_type display_continuation(CNT_type Continuation,
                                     OPT_type Output_port)
  { text_to_port(Continuation_text,
                 Output_port); }

static VOI_type display_eof_object(OPT_type Output_port)
  { text_to_port(Eof_object_text,
                 Output_port); }

static VOI_type display_false(OPT_type Output_port)
  { text_to_port(False_text,
                 Output_port); }

static VOI_type display_force(FRC_type Force,
                              OPT_type Output_port)
  { text_to_port(Forced_text,
                 Output_port); }

static VOI_type display_higher_order_native(NAZ_type Higher_order_native,
                                            OPT_type Output_port)
  { TXT_type text;
    text = Higher_order_native->txt;
    FORMATTED_ITEM(Higher_order_native_text, text, Output_port); }

static VOI_type display_pair(PAI_type Pair,
                             OPT_type Output_port)
  { EXP_type car_expression,
             cdr_expression;
    PAI_type pair;
    TAG_type tag;
    char_to_port(Leftpar_char,
                 Output_port);
    for (pair = Pair;
         ;
         )
      { car_expression = pair->car;
        cdr_expression = pair->cdr;
        display_expression(car_expression,
                           Output_port);
        tag = Grammar_Tag(cdr_expression);
        switch (tag)
          { case NUL_tag:
              char_to_port(Rightpar_char,
                           Output_port);
              return;
            case PAI_tag:
              char_to_port(Space_char,
                           Output_port);
              pair = cdr_expression;
              break;
            default:
              text_to_port(Period_text,
                           Output_port);
              display_expression(cdr_expression,
                                 Output_port);
              char_to_port(Rightpar_char,
                           Output_port);
              return; }}}

static VOI_type display_native(NAT_type Native,
                               OPT_type Output_port)
  { TXT_type text;
    text = Native->txt;
    FORMATTED_ITEM(Native_text, text, Output_port); }

static VOI_type display_null(OPT_type Output_port)
  { text_to_port(Empty_list_text,
                 Output_port); }

static VOI_type display_null_stream(OPT_type Output_port)
  { text_to_port(Empty_stream_text,
                 Output_port); }

static VOI_type display_procedure(PRC_type Procedure,
                                  OPT_type Output_port)
  { SYM_type name_symbol;
    TXT_type text;
    name_symbol = Procedure->sym;
    text = name_symbol->txt;
    FORMATTED_ITEM(Procedure_text, text, Output_port); }

static VOI_type display_procedure_vararg(PRV_type Procedure_vararg,
                                         OPT_type Output_port)
  { SYM_type name_symbol;
    TXT_type text;
    name_symbol = Procedure_vararg->sym;
    text = name_symbol->txt;
    FORMATTED_ITEM(Procedure_text, text, Output_port); }

static VOI_type display_promise(OPT_type Output_port)
  { text_to_port(Promise_text,
                 Output_port); }

static VOI_type display_real(REA_type Real,
                             OPT_type Output_port)
  { FLO_type floating;
    floating = Real->flo;
    FORMATTED_ITEM(Floating_text,
                   floating,
                   Output_port); }

static VOI_type display_smallint(SMI_type Smi,
                                 OPT_type Output_port)
  { SGN_type integer;
    integer = get_SMI(Smi);
    FORMATTED_ITEM(Integer_text,
                   integer,
                   Output_port); }

static VOI_type display_string(STR_type String,
                               OPT_type Output_port)
  { CHR_type char_;
    TXT_type text;
    UNS_type index,
             length;
    length = length_STR(String);
    if (length == 0)
      return;
    text = String->txt;
    for (index = 0;
         index < length;
         index += 1)
      { char_ = text[index];
        if (!isprint(char_))
          if (char_ != Newline_char)
            char_ = Space_char;
        char_to_port(char_,
                     Output_port); }}

static VOI_type display_symbol(SYM_type Symbol,
                               OPT_type Output_port)
  { TXT_type text;
    text = Symbol->txt;
    text_to_port(text,
                 Output_port); }

static VOI_type display_true(TRU_type True,
                             OPT_type Output_port)
  { text_to_port(True_text,
                 Output_port); }

static VOI_type display_Unspecified(USP_type Unspecified,
                                    OPT_type Output_port)
  { text_to_port(Unspecified_text,
                 Output_port); }

static VOI_type display_vector(VEC_type Vector,
                               OPT_type Output_port)
  { EXP_type value_expression;
    UNS_type index,
             size;
    size = size_VEC(Vector);
    if (size == 0)
      { text_to_port(Vector0_text,
                     Output_port);
        return; }
    text_to_port(Vector_text,
                 Output_port);
    for (index = 1;
         index <= size;
         index += 1)
      { value_expression = Vector[index];
        display_expression(value_expression,
                           Output_port);
        if (index < size)
          char_to_port(Space_char,
                       Output_port); }
    char_to_port(Rightpar_char,
                 Output_port); }

/*--------------------------------------------------------------------------------------*/

static VOI_type display_expression(EXP_type Value_expression,              /* IPT, OPT? */
                                   OPT_type Output_port)
  { TAG_type tag;
    tag = Grammar_Tag(Value_expression);
    switch (tag)
      { case CHA_tag:
          display_character(Value_expression,
                            Output_port);
          break;
        case CNT_tag:
          display_continuation(Value_expression,
                               Output_port);
          break;
        case END_tag:
          display_eof_object(Output_port);
          break;
        case FLS_tag:
          display_false(Output_port);
          break;
        case FRC_tag:
          display_force(Value_expression,
                        Output_port);
          break;
        case NAT_tag:
          display_native(Value_expression,
                         Output_port);
          break;
        case NAZ_tag:
          display_higher_order_native(Value_expression,
                                      Output_port);
          break;
        case NST_tag:
          display_null_stream(Output_port);
          break;
        case NUL_tag:
          display_null(Output_port);
          break;
        case PAI_tag:
          display_pair(Value_expression,
                       Output_port);
          break;
        case PRC_tag:
          display_procedure(Value_expression,
                            Output_port);
          break;
        case PRM_tag:
          display_promise(Output_port);
          break;
        case PRV_tag:
          display_procedure_vararg(Value_expression,
                                   Output_port);
          break;
        case REA_tag:
          display_real(Value_expression,
                       Output_port);
          break;
        case SMI_tag:
          display_smallint(Value_expression,
                                Output_port);
          break;
        case STR_tag:
          display_string(Value_expression,
                         Output_port);
          break;
        case SYM_tag:
          display_symbol(Value_expression,
                         Output_port);
          break;
        case TRU_tag:
          display_true(Value_expression,
                       Output_port);
          break;
        case USP_tag:
          display_Unspecified(Value_expression,
                              Output_port);
          break;
        case VEC_tag:
          display_vector(Value_expression,
                         Output_port);
          break;
        default:
          error(IXT_error,
                Display_text); }}

/*---------------------------------------- format --------------------------------------*/

static VOI_type format_expression(EXP_type,
                                  BLN_type,
                                  OPT_type);
static VOI_type     format_symbol(SYM_type,
                                  BLN_type,
                                  OPT_type);
static VOI_type           keyword(KEY_type,
                                  OPT_type);
static VOI_type    local_variable(UNS_type,
                                  OPT_type);

static VOI_type begin_indent(VOI_type)
  { Indentation += Indentation_step; }

static VOI_type end_indent(VOI_type)
  { Indentation -= Indentation_step; }

static VOI_type enter_scope(DCT_type   Child_dictionary,
                            DCT_type * Parent_dictionary)
  { *Parent_dictionary =  Dictionary_Replace_Dictionary(Child_dictionary); }

static VOI_type exit_scope(DCT_type Dictionary)
  { Dictionary_Replace_Dictionary(Dictionary); }

static VOI_type newline(OPT_type Output_port)
  { UNS_type index;
    char_to_port(Newline_char,
                 Output_port);
    for (index = 0;
         index < Indentation;
         index += 1)
      char_to_port(Space_char,
                   Output_port); }

/*--------------------------------------------------------------------------------------*/

static VOI_type bindings(SEQ_type Binding_sequence,
                         BLN_type Quoted,
                         OPT_type Output_port)
  { EXP_type expression;
    UNS_type index,
             size;
    size = size_VEC(Binding_sequence);
    begin_indent();
      begin_indent();
        newline(Output_port);
        char_to_port(Leftpar_char,
                     Output_port);
        if (size > 0)
          for (index = 1;
               index <= size;
               index += 1)
            { expression = Binding_sequence[index];
              char_to_port(Leftpar_char,
                           Output_port);
              local_variable(index,
                             Output_port);
              char_to_port(Space_char,
                           Output_port);
              format_expression(expression,
                                Quoted,
                                Output_port);
              char_to_port(Rightpar_char,
                           Output_port);
              if (index < size)
                { newline(Output_port);
                  char_to_port(Space_char,
                             Output_port); }}
        char_to_port(Rightpar_char,
                     Output_port);
      end_indent();
    end_indent(); }

static VOI_type cond_clauses(SEQ_type Predicate_sequence,
                             SEQ_type Body_sequence,
                             BLN_type Quoted,
                             OPT_type Output_port)
  { EXP_type body_expression,
             predicate_expression;
    UNS_type size,
             index;
 	  size = size_VEC(Predicate_sequence);
    begin_indent();
      for (index = 1;
           index <= size;
           index += 1)
        { newline(Output_port);
          char_to_port(Leftpar_char,
                     Output_port);
          predicate_expression = Predicate_sequence[index];
          body_expression = Body_sequence[index];
          begin_indent();
            format_expression(predicate_expression,
                              Quoted,
                Output_port);
            newline(Output_port);
            format_expression(body_expression,
                              Quoted,
                Output_port);
            char_to_port(Rightpar_char,
                     Output_port);
          end_indent(); }
    end_indent(); }

static VOI_type else_clause(EXP_type Else_expression,
                            BLN_type Quoted,
                            OPT_type Output_port)
  { begin_indent();
      newline(Output_port);
      char_to_port(Leftpar_char,
                     Output_port);
      keyword(Pool_Else,
              Output_port);
      begin_indent();
        newline(Output_port);
        format_expression(Else_expression,
                          Quoted,
                Output_port);
        char_to_port(Rightpar_char,
                     Output_port);
      end_indent();
    end_indent(); }

static VOI_type global_variable(UNS_type Scope,
                                UNS_type Offset,
                                OPT_type Output_port)
  { TXT_type text;
    SYM_type variable_symbol;
    variable_symbol = Dictionary_Global_Variable(Scope,
                                                 Offset);
    text = variable_symbol->txt;
    text_to_port(text,
                 Output_port); }

static VOI_type keyword(KEY_type Key,
                        OPT_type Output_port)
  { SYM_type symbol;
    TXT_type text;
    symbol = Pool_Keyword_Symbol(Key);
    text = symbol->txt;
    text_to_port(text,
                 Output_port); }

static VOI_type local_variable(UNS_type Offset,
                               OPT_type Output_port)
  { TXT_type text;
    SYM_type variable_symbol;
    variable_symbol = Dictionary_Local_Variable(Offset);
    text = variable_symbol->txt;
    text_to_port(text,
                 Output_port); }

static VOI_type operands(SEQ_type Operand_sequence,
                         BLN_type Quoted,
                         OPT_type Output_port)
  { EXP_type expression;
    UNS_type size,
             index;
 	  size = size_VEC(Operand_sequence);
    for (index = 1;
         index <= size;
         index += 1)
      { char_to_port(Space_char,
                     Output_port);
        expression = Operand_sequence[index];
        format_expression(expression,
                          Quoted,
                          Output_port); }}

static VOI_type pattern(UNS_type Parameter_count,
                        OPT_type Output_port)
  { UNS_type index;
    if (Parameter_count > 0)
      for (index = 1;
           index <= Parameter_count;
           index += 1)
        { if (index > 1)
            char_to_port(Space_char,
                         Output_port);
          local_variable(index,
                         Output_port); }}

/*--------------------------------------------------------------------------------------*/

static VOI_type format_and(AND_type And,
                           BLN_type Quoted,
                           OPT_type Output_port)
  { SEQ_type predicate_sequence;
    predicate_sequence = And->prd;
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_And,
            Output_port);
    operands(predicate_sequence,
             Quoted,
             Output_port);
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_application(APL_type Application,
                                   BLN_type Quoted,
                                   OPT_type Output_port)
  { EXP_type operator;
    SEQ_type operand_sequence;
    operator = Application->exp;
    operand_sequence = Application->arg;
    char_to_port(Leftpar_char,
                 Output_port);
    format_expression(operator,
                      Quoted,
                      Output_port);
    operands(operand_sequence,
             Quoted,
             Output_port);
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_begin(BEG_type Begin,
                             BLN_type Quoted,
                             OPT_type Output_port)
  { EXP_type body_expression;
    body_expression = Begin->exp;
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_Begin,
            Output_port);
    begin_indent();
      newline(Output_port);
      format_expression(body_expression,
                        Quoted,
                        Output_port);
    end_indent();
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_character(CHA_type Character,
                                 BLN_type Quoted,
                                 OPT_type Output_port)
  { CHR_type char_;
    char_ = get_CHA(Character);
    FORMATTED_ITEM(Char_text,
                   char_,
                   Output_port); }

static VOI_type format_cond_with_else(CNE_type Cond_else,
                                      BLN_type Quoted,
                                      OPT_type Output_port)
  { EXP_type else_expression;
    SEQ_type body_sequence,
             predicate_sequence;
    predicate_sequence = Cond_else->prd;
    body_sequence      = Cond_else->bod;
    else_expression    = Cond_else->els;
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_Cond,
                 Output_port);
    cond_clauses(predicate_sequence,
                 body_sequence,
                 Quoted,
                 Output_port);
    else_clause(else_expression,
                Quoted,
                Output_port);
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_cond_w_o_else(CND_type Cond,
                                     BLN_type Quoted,
                                     OPT_type Output_port)
  { SEQ_type body_sequence,
             predicate_sequence;
    predicate_sequence = Cond->prd;
    body_sequence      = Cond->bod;
    char_to_port(Leftpar_char,
                     Output_port);
    keyword(Pool_Cond,
                 Output_port);
    cond_clauses(predicate_sequence,
                 body_sequence,
                 Quoted,
                 Output_port);
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_cons_stream(CNS_type Cons_stream,
                                   BLN_type Quoted,
                                   OPT_type Output_port)
  { EXP_type car_expression,
             cdr_expression;
    car_expression = Cons_stream->car;
    cdr_expression = Cons_stream->cdr;
    char_to_port(Leftpar_char,
                     Output_port);
    keyword(Pool_Cons_Stream,
            Output_port);
    char_to_port(Space_char,
                 Output_port);
    format_expression(car_expression,
                      Quoted,
                      Output_port);
    char_to_port(Space_char,
                     Output_port);
    format_expression(cdr_expression,
                      Quoted,
                      Output_port);
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_define_function(DFF_type Define_function_expression,
                                       BLN_type Quoted,
                                       OPT_type Output_port)
  { DCT_type dictionary,
             enclosing_dictionary;
    EXP_type body_expression;
    SMI_type offset_smallint,
             parameter_count_smallint;
    UNS_type offset,
             parameter_count;
    offset_smallint          = Define_function_expression->ofs;
    parameter_count_smallint = Define_function_expression->par;
    body_expression          = Define_function_expression->exp;
    dictionary               = Define_function_expression->dct;
    offset = get_SMU(offset_smallint);
    parameter_count = get_SMU(parameter_count_smallint);
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_Define,
            Output_port);
    char_to_port(Space_char,
                 Output_port);
    char_to_port(Leftpar_char,
                 Output_port);
    local_variable(offset,
                   Output_port);
    enter_scope(dictionary,
                &enclosing_dictionary);
      if (parameter_count > 0)
        { char_to_port(Space_char,
                       Output_port);
          pattern(parameter_count,
                  Output_port); }
      char_to_port(Rightpar_char,
                   Output_port);
      begin_indent();
        newline(Output_port);
        format_expression(body_expression,
                          Quoted,
                          Output_port);
      end_indent();
    exit_scope(enclosing_dictionary);;
    char_to_port(Rightpar_char,
                 Output_port);  }

static VOI_type format_define_function_vararg(DFZ_type Define_function_vararg,
                                              BLN_type Quoted,
                                              OPT_type Output_port)
  { DCT_type dictionary,
             enclosing_dictionary;
    EXP_type body_expression;
    UNS_type offset,
             parameter_count;
    SMI_type offset_smallint,
             parameter_count_smallint;
    offset_smallint          = Define_function_vararg->ofs;
    parameter_count_smallint = Define_function_vararg->par;
    body_expression          = Define_function_vararg->exp;
    dictionary               = Define_function_vararg->dct;
    offset = get_SMU(offset_smallint);
    parameter_count = get_SMU(parameter_count_smallint);
    char_to_port(Leftpar_char,
                     Output_port);
    keyword(Pool_Define,
            Output_port);
    char_to_port(Space_char,
                 Output_port);
    char_to_port(Leftpar_char,
                 Output_port);
    local_variable(offset,
                   Output_port);
    enter_scope(dictionary,
                &enclosing_dictionary);
      if (parameter_count > 0)
        { char_to_port(Space_char,
                       Output_port);
          pattern(parameter_count,
                  Output_port); }
      text_to_port(Period_text,
                   Output_port);
      local_variable(parameter_count + 1,
                     Output_port);
      char_to_port(Rightpar_char,
                   Output_port);
      begin_indent();
        newline(Output_port);
        format_expression(body_expression,
                          Quoted,
                          Output_port);
      end_indent();
    exit_scope(enclosing_dictionary);;
    char_to_port(Rightpar_char,
                 Output_port);  }

static VOI_type format_define_variable(DFV_type Define_variable,
                                       BLN_type Quoted,
                                       OPT_type Output_port)
  { EXP_type expression;
    UNS_type offset;
    SMI_type offset_smallint;
    offset_smallint = Define_variable->ofs;
    expression      = Define_variable->exp;
    offset = get_SMU(offset_smallint);
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_Define,
            Output_port);
    char_to_port(Space_char,
                 Output_port);
    local_variable(offset,
                   Output_port);
    char_to_port(Space_char,
                 Output_port);
    format_expression(expression,
                      Quoted,
                      Output_port);
    char_to_port(Rightpar_char,
                     Output_port); }

static VOI_type format_delay(DEL_type Delay,
                             BLN_type Quoted,
                             OPT_type Output_port)
  { EXP_type expression;
    expression = Delay->exp;
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_Delay,
            Output_port);
    char_to_port(Space_char,
                 Output_port);
    format_expression(expression,
                      Quoted,
                      Output_port);
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_evaluate(EVA_type Eval,
                                BLN_type Quoted,
                                OPT_type Output_port)
  { EXP_type expression;
    expression = Eval->exp;
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_Eval,
            Output_port);
    char_to_port(Space_char,
                 Output_port);
    format_expression(expression,
                      Quoted,
                      Output_port);
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_false(FLS_type False,
                             BLN_type Quoted,
                             OPT_type Output_port)
  { text_to_port(False_text,
                 Output_port); }

static VOI_type format_global(GLB_type Global,
                              BLN_type Quoted,
                              OPT_type Output_port)
  { UNS_type offset,
             scope;
    scope  = Global->scp;
    offset = Global->ofs;
    global_variable(scope,
                    offset,
                    Output_port); }

static VOI_type format_global_application(GLA_type Global_application,
                                          BLN_type Quoted,
                                          OPT_type Output_port)
  { GLB_type global;
    SEQ_type operand_sequence;
    global           = Global_application->glb;
    operand_sequence = Global_application->arg;
    char_to_port(Leftpar_char,
                Output_port);
    format_global(global,
                  Quoted,
                  Output_port);
    operands(operand_sequence,
             Quoted,
             Output_port);
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_if_double(IFD_type If_double,
                                 BLN_type Quoted,
                                 OPT_type Output_port)
  { EXP_type alternative_expression,
             consequent_expression,
             predicate_expression;
    predicate_expression   = If_double->prd;
    consequent_expression  = If_double->cns;
    alternative_expression = If_double->alt;
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_If,
            Output_port);
    char_to_port(Space_char,
                 Output_port);
    format_expression(predicate_expression,
                      Quoted,
                      Output_port);
    begin_indent();
      newline(Output_port);
      format_expression(consequent_expression,
                        Quoted,
                        Output_port);
      newline(Output_port);
      format_expression(alternative_expression,
                        Quoted,
                        Output_port);
    end_indent();
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_if_single(IFS_type If_single,
                                 BLN_type Quoted,
                                 OPT_type Output_port)
  { EXP_type consequent_expression,
             predicate_expression;
    predicate_expression  = If_single->prd;
    consequent_expression = If_single->cns;
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_If,
            Output_port);
    char_to_port(Space_char,
                 Output_port);
    format_expression(predicate_expression,
                      Quoted,
                      Output_port);
    begin_indent();
      newline(Output_port);
      format_expression(consequent_expression,
                        Quoted,
                        Output_port);
    end_indent();
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_lambda(LMB_type Lambda,
                              BLN_type Quoted,
                              OPT_type Output_port)
  { DCT_type dictionary,
             enclosing_dictionary;
    EXP_type body_expression;
    SMI_type parameter_count_smallint;
    UNS_type parameter_count;
    parameter_count_smallint = Lambda->par;
    body_expression          = Lambda->exp;
    dictionary               = Lambda->dct;
    parameter_count = get_SMU(parameter_count_smallint);
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_Lambda,
            Output_port);
    char_to_port(Space_char,
                 Output_port);
    char_to_port(Leftpar_char,
                 Output_port);
    enter_scope(dictionary,
                &enclosing_dictionary);
      pattern(parameter_count,
                 Output_port);
      char_to_port(Rightpar_char,
                   Output_port);
        begin_indent();
          newline(Output_port);
          format_expression(body_expression,
                            Quoted,
                            Output_port);
        end_indent();
    exit_scope(enclosing_dictionary);;
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_lambda_vararg(LMV_type Lambda_vararg,
                                     BLN_type Quoted,
                                     OPT_type Output_port)
  { DCT_type dictionary,
             enclosing_dictionary;
    EXP_type body_expression;
    SMI_type parameter_count_smallint;
    UNS_type parameter_count;
    parameter_count_smallint = Lambda_vararg->par;
    body_expression          = Lambda_vararg->exp;
    dictionary               = Lambda_vararg->dct;
    parameter_count = get_SMU(parameter_count_smallint);
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_Lambda,
            Output_port);
    char_to_port(Space_char,
                 Output_port);
    enter_scope(dictionary,
                &enclosing_dictionary);
      if (parameter_count == 0)
        local_variable(1,
                       Output_port);
      else
        { char_to_port(Leftpar_char,
                       Output_port);
          pattern(parameter_count,
                  Output_port);
          text_to_port(Period_text,
                       Output_port);
          local_variable(parameter_count + 1,
                         Output_port);
          char_to_port(Rightpar_char,
                       Output_port); }
      begin_indent();
        newline(Output_port);
        format_expression(body_expression,
                          Quoted,
                          Output_port);
      end_indent();
    exit_scope(enclosing_dictionary);;
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_let(LET_type Let,
                           BLN_type Quoted,
                           OPT_type Output_port)
  { DCT_type dictionary,
             enclosing_dictionary;
    EXP_type body_expression;
    SEQ_type binding_sequence;
    binding_sequence = Let->seq;
    body_expression  = Let->exp;
    dictionary       = Let->dct;
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_Let,
            Output_port);
    char_to_port(Space_char,
                 Output_port);
    enter_scope(dictionary,
                &enclosing_dictionary);
    bindings(binding_sequence,
             Quoted,
             Output_port);
    begin_indent();
      newline(Output_port);
      format_expression(body_expression,
                        Quoted,
                        Output_port);
    end_indent();
    exit_scope(enclosing_dictionary);;
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_local(LCL_type Local,
                             BLN_type Quoted,
                             OPT_type Output_port)
  { UNS_type offset;
    offset = Local->ofs;
    local_variable(offset,
                   Output_port); }

static VOI_type format_local_application(LCA_type Local_application,
                                         BLN_type Quoted,
                                         OPT_type Output_port)
  { LCL_type local;
    SEQ_type operand_sequence;
    local            = Local_application->lcl;
    operand_sequence = Local_application->arg;
    char_to_port(Leftpar_char,
                 Output_port);
    format_local(local,
                 Quoted,
                 Output_port);
    operands(operand_sequence,
             Quoted,
             Output_port);
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_macro(MAC_type Macro,
                             BLN_type Quoted,
                             OPT_type Output_port)
  { SYM_type symbol;
    symbol = Macro->sym;
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_Macro,
            Output_port);
    char_to_port(Space_char,
                 Output_port);
    format_symbol(symbol,
                  TRUE,
                  Output_port);
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_null(NUL_type Null,
                            BLN_type Quoted,
                            OPT_type Output_port)
  { text_to_port(Empty_list_text,
                 Output_port); }

static VOI_type format_smallint(SMI_type Number,
                                BLN_type Quoted,
                                OPT_type Output_port)
  { SGN_type int_;
    int_ = get_SMI(Number);
    FORMATTED_ITEM(Integer_text,
                   int_,
                   Output_port); }

static VOI_type format_or(AND_type Or,
                          BLN_type Quoted,
                          OPT_type Output_port)
  { SEQ_type predicate_sequence;
    predicate_sequence = Or->prd;
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_Or,
            Output_port);
    operands(predicate_sequence,
             Quoted,
             Output_port);
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_pair(PAI_type Pair,
                            BLN_type Quoted,
                            OPT_type Output_port)
  { EXP_type residue_expression,
             value_expression;
    PAI_type pair;
    TAG_type tag;
    if (!Quoted)
      char_to_port(Quote_char,
                   Output_port);
    char_to_port(Leftpar_char,
                 Output_port);
    for (pair = Pair;
         ;
         )
      { value_expression   = pair->car;
        residue_expression = pair->cdr;
        format_expression(value_expression,
                          TRUE,
                          Output_port);
        tag = Grammar_Tag(residue_expression);
        switch (tag)
          { case NUL_tag:
              char_to_port(Rightpar_char,
                           Output_port);
              return;
            case PAI_tag:
              char_to_port(Space_char,
                           Output_port);
              pair = residue_expression;
              break;
            default:
              text_to_port(Period_text,
                           Output_port);
              format_expression(residue_expression,
                                TRUE,
                                Output_port);
              char_to_port(Rightpar_char,
                          Output_port);
              return; }}}

static VOI_type format_quasiquote(QQU_type Quasiquote,
                                  BLN_type Quoted,
                                  OPT_type Output_port)
  { EXP_type expression;
    expression = Quasiquote->exp;
    char_to_port(Quasiquote_char,
                 Output_port);
    format_expression(expression,
                      TRUE,
                      Output_port); }

static VOI_type format_quasisymbol(QSY_type Quasisymbol,
                                   BLN_type Quoted,
                                   OPT_type Output_port)
  { SYM_type symbol;
    TXT_type text;
    symbol = Quasisymbol->sym;
    text_to_port(Quasisymbol_text,
                 Output_port);
    text = symbol->txt;
    text_to_port(text,
                 Output_port); }

static VOI_type format_quote(QUO_type Quote,
                             BLN_type Quoted,
                             OPT_type Output_port)
  { EXP_type expression;
    expression = Quote->exp;
    format_expression(expression,
                      Quoted,
                      Output_port); }

static VOI_type format_unquote(UNQ_type Unquote,
                               BLN_type Quoted,
                                  OPT_type Output_port)
  { EXP_type expression;
    expression = Unquote->exp;
    if (!Quoted)
      char_to_port(Quote_char,
                   Output_port);
    char_to_port(Unquote_char,
                 Output_port);
    format_expression(expression,
                      Quoted,
                      Output_port); }

static VOI_type format_unquote_list_splice(ULS_type Unquote_list_splice,
                                           BLN_type Quoted,
                                           OPT_type Output_port)
  { EXP_type expression;
    expression = Unquote_list_splice->exp;
    text_to_port(Unquote_splice_text,
                 Output_port);
    format_expression(expression,
                      Quoted,
                      Output_port); }

static VOI_type format_unquote_vector_splice(ULS_type Unquote_vector_splice,
                                             BLN_type Quoted,
                                             OPT_type Output_port)
  { EXP_type expression;
    expression = Unquote_vector_splice->exp;
    text_to_port(Unquote_splice_text,
                 Output_port);
    format_expression(expression,
                      Quoted,
                      Output_port); }

static VOI_type format_real(REA_type Real,
                            BLN_type Quoted,
                            OPT_type Output_port)
  { FLO_type float_;
    float_ = Real->flo;
    FORMATTED_ITEM(Floating_text,
                   float_,
                   Output_port); }

static VOI_type format_sequence(SEQ_type Sequence,
                                BLN_type Quoted,
                                OPT_type Output_port)
  { EXP_type expression;
    UNS_type index,
             size;
    size = size_SEQ(Sequence);
    expression = Sequence[1];
    format_expression(expression,
                      Quoted,
                      Output_port);
    for (index = 2;
         index <= size;
         index += 1)
      { newline(Output_port);
        expression = Sequence[index];
        format_expression(expression,
                          Quoted,
                          Output_port); }}

static VOI_type format_set_global(STG_type Set_global,
                                  BLN_type Quoted,
                                  OPT_type Output_port)
  { EXP_type expression;
    SMI_type offset_smallint,
             scope_smallint;
    UNS_type offset,
             scope;
    scope_smallint  = Set_global->scp;
    offset_smallint = Set_global->ofs;
    expression = Set_global->exp;
    scope = get_SMU(scope_smallint);
    offset = get_SMU(offset_smallint);
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_Set,
            Output_port);
    char_to_port(Space_char,
                 Output_port);
    global_variable(scope,
                    offset,
                    Output_port);
    char_to_port(Space_char,
                 Output_port);
    format_expression(expression,
                      Quoted,
                      Output_port);
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_set_local(STL_type Set_local,
                                 BLN_type Quoted,
                                 OPT_type Output_port)
  { EXP_type expression;
    SMI_type offset_smallint;
    UNS_type offset;
    offset_smallint = Set_local->ofs;
    expression      = Set_local->exp;
    offset = get_SMU(offset_smallint);
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_Set,
            Output_port);
    char_to_port(Space_char,
                 Output_port);
    local_variable(offset,
                   Output_port);
    char_to_port(Space_char,
                 Output_port);
    format_expression(expression,
                      Quoted,
                      Output_port);
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_string(STR_type String,
                              BLN_type Quoted,
                              OPT_type Output_port)
  { CHR_type char_;
    TXT_type text;
    UNS_type index,
             length;
    length = length_STR(String);
    if (length == 0)
      { text_to_port(Empty_text_text,
                     Output_port);
        return; }
    text = String->txt;
    char_to_port(Doublequote_char,
                 Output_port);
    for (index = 0;
         index < length;
         index += 1)
      { char_ = text[index];
        switch (char_)
          { case Backslash_char:
            case Doublequote_char:
              char_to_port(Backslash_char,
                           Output_port);
              break; }
        if (!isprint(char_))
          if (char_ != Newline_char)
            char_ = Space_char;
        char_to_port(char_,
                     Output_port); }
    char_to_port(Doublequote_char,
                 Output_port); }

static VOI_type format_symbol(SYM_type Symbol,
                              BLN_type Quoted,
                              OPT_type Output_port)
  { TXT_type text;
    text = Symbol->txt;
    if (!Quoted)
      char_to_port(Quote_char,
                   Output_port);
    text_to_port(text,
                 Output_port); }

static VOI_type format_thunk(THK_type Thunk,
                             BLN_type Quoted,
                             OPT_type Output_port)
  { DCT_type dictionary,
             enclosing_dictionary;
    EXP_type body_expression;
    body_expression = Thunk->exp;
    dictionary      = Thunk->dct;
    enter_scope(dictionary,
                &enclosing_dictionary);
      format_expression(body_expression,
                        Quoted,
                        Output_port);
    exit_scope(enclosing_dictionary); }

static VOI_type format_true(TRU_type True,
                            BLN_type Quoted,
                            OPT_type Output_port)
  { text_to_port(True_text,
                 Output_port); }

static VOI_type format_vector(VEC_type Vector,
                              BLN_type Quoted,
                              OPT_type Output_port)
  { EXP_type value_expression;
    UNS_type index,
             size;
    size = size_VEC(Vector);
    text_to_port(Vector_text,
                 Output_port);
    for (index = 1;
         index <= size;
         index += 1)
      { value_expression = Vector[index];
        format_expression(value_expression,
                          Quoted,
                          Output_port);
        if (index < size)
          char_to_port(Space_char,
                       Output_port); }
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_while(WHI_type While_,
                             BLN_type Quoted,
                             OPT_type Output_port)
  { EXP_type body_expression,
             predicate_expression;
    predicate_expression = While_->prd;
    body_expression      = While_->exp;
    char_to_port(Leftpar_char,
                 Output_port);
    keyword(Pool_While,
            Output_port);
    char_to_port(Space_char,
                 Output_port);
    format_expression(predicate_expression,
                      Quoted,
                      Output_port);
    begin_indent();
      newline(Output_port);
      format_expression(body_expression,
                        Quoted,
                        Output_port);
    end_indent();
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type format_expression(EXP_type Expression,
                                  BLN_type Quoted,
                                  OPT_type Output_port)
  { TAG_type tag;
    tag = Grammar_Tag(Expression);
    switch (tag)
      { case AND_tag:
          format_and(Expression,
                     Quoted,
                     Output_port);
          return;
        case APL_tag:
        case APl_tag:
          format_application(Expression,
                             Quoted,
                             Output_port);
          return;
        case BEG_tag:
          format_begin(Expression,
                       Quoted,
                       Output_port);
          return;
        case CHA_tag:
          format_character(Expression,
                           Quoted,
                           Output_port);
          return;
        case CND_tag:
          format_cond_w_o_else(Expression,
                                Quoted,
                                Output_port);
          return;
        case CNE_tag:
          format_cond_with_else(Expression,
                               Quoted,
                               Output_port);
          return;
        case CNS_tag:
          format_cons_stream(Expression,
                             Quoted,
                             Output_port);
          return;
        case DEL_tag:
          format_delay(Expression,
                       Quoted,
                       Output_port);
          return;
        case DFF_tag:
          format_define_function(Expression,
                                 Quoted,
                                 Output_port);
          return;
        case DFV_tag:
          format_define_variable(Expression,
                                 Quoted,
                                 Output_port);
          return;
        case DFZ_tag:
          format_define_function_vararg(Expression,
                                        Quoted,
                                        Output_port);
          return;
        case EVA_tag:
        case EVa_tag:
          format_evaluate(Expression,
                          Quoted,
                          Output_port);
          return;
        case FLS_tag:
          format_false(Expression,
                       Quoted,
                       Output_port);
          return;
        case GLA_tag:
        case GLa_tag:
          format_global_application(Expression,
                                    Quoted,
                                    Output_port);
          return;
        case GLB_tag:
          format_global(Expression,
                        Quoted,
                        Output_port);
          return;
        case IFD_tag:
          format_if_double(Expression,
                           Quoted,
                           Output_port);
          return;
        case IFS_tag:
          format_if_single(Expression,
                           Quoted,
                           Output_port);
          return;
        case LCA_tag:
        case LCa_tag:
          format_local_application(Expression,
                                   Quoted,
                                   Output_port);
          return;
        case LCL_tag:
          format_local(Expression,
                       Quoted,
                       Output_port);
          return;
        case LET_tag:
        case LEt_tag:
          format_let(Expression,
                     Quoted,
                     Output_port);
          return;
        case LMB_tag:
          format_lambda(Expression,
                        Quoted,
                        Output_port);
          return;
        case LMV_tag:
          format_lambda_vararg(Expression,
                               Quoted,
                               Output_port);
          return;
        case MAC_tag:
          format_macro(Expression,
                       Quoted,
                       Output_port);
          return;
        case NUL_tag:
          format_null(Expression,
                      Quoted,
                      Output_port);
          return;
        case ORR_tag:
          format_or(Expression,
                    Quoted,
                    Output_port);
          return;
        case PAI_tag:
          format_pair(Expression,
                      Quoted,
                      Output_port);
          return;
        case QQU_tag:
          format_quasiquote(Expression,
                            Quoted,
                            Output_port);
          return;
        case QSY_tag:
          format_quasisymbol(Expression,
                             Quoted,
                             Output_port);
          return;
        case QUO_tag:
          format_quote(Expression,
                       Quoted,
                       Output_port);
          return;
        case REA_tag:
          format_real(Expression,
                      Quoted,
                      Output_port);
          return;
        case SEQ_tag:
          format_sequence(Expression,
                          Quoted,
                          Output_port);
          return;
        case SMI_tag:
          format_smallint(Expression,
                          Quoted,
                          Output_port);
          return;
        case STG_tag:
          format_set_global(Expression,
                            Quoted,
                            Output_port);
          return;
        case STL_tag:
          format_set_local(Expression,
                           Quoted,
                           Output_port);
          return;
        case STR_tag:
          format_string(Expression,
                        Quoted,
                        Output_port);
          return;
        case SYM_tag:
          format_symbol(Expression,
                        Quoted,
                        Output_port);
          return;
        case THK_tag:
        case THk_tag:
          format_thunk(Expression,
                       Quoted,
                       Output_port);
          return;
        case TRU_tag:
          format_true(Expression,
                      Quoted,
                      Output_port);
          return;
        case ULS_tag:
          format_unquote_list_splice(Expression,
                                     Quoted,
                                     Output_port);
          return;
        case UNQ_tag:
          format_unquote(Expression,
                         Quoted,
                         Output_port);
          return;
        case UVS_tag:
          format_unquote_vector_splice(Expression,
                                       Quoted,
                                       Output_port);
          return;
        case VEC_tag:
          format_vector(Expression,
                        Quoted,
                        Output_port);
          return;
        case WHI_tag:
          format_while(Expression,
                       Quoted,
                       Output_port);
          return;
        default:
          error(IXT_error,
                Format_text); }}

/*----------------------------------- print expression ---------------------------------*/

static VOI_type print_expression(EXP_type,
                                 OPT_type);

static VOI_type print_character(CHA_type Character,
                                OPT_type Output_port)
  { CHR_type char_;
    char_ = get_CHA(Character);
    if (!isprint(char_))
      if (char_ != NEWLINE_CHAR)
        char_ = SPACE_CHAR;
    char_to_port(char_,
                 Output_port); }

static VOI_type print_continuation(CNT_type Continuation,
                                   OPT_type Output_port)
  { text_to_port(Continuation_text,
                 Output_port); }

static VOI_type print_eof_object(END_type Eof_object,
                                 OPT_type Output_port)
  { text_to_port(Eof_object_text,
                 Output_port); }

static VOI_type print_false(FLS_type False,
                            OPT_type Output_port)
  { text_to_port(False_text,
                 Output_port); }

static VOI_type print_forced(FRC_type Forced,
                             OPT_type Output_port)
  { text_to_port(Forced_text,
                 Output_port); }

static VOI_type print_higher_order_native(NAZ_type Higher_order_native,
                                          OPT_type Output_port)
  { TXT_type text;
    text = Higher_order_native->txt;
    FORMATTED_ITEM(Higher_order_native_text, text, Output_port); }

static VOI_type print_input_port(IPT_type Input_port,
                                 OPT_type Output_port)
  { TXT_type text;
    text = Input_port->txt;
    FORMATTED_ITEM(Input_port_text, text, Output_port); }

static VOI_type print_macro(MAC_type Macro,
                            OPT_type Output_port)
  { SYM_type name_symbol;
    TXT_type text;
    name_symbol = Macro->sym;
    text = name_symbol->txt;
    FORMATTED_ITEM(Macro_text, text, Output_port); }

static VOI_type print_native(NAT_type Native,
                             OPT_type Output_port)
  { TXT_type text;
    text = Native->txt;
    FORMATTED_ITEM(Native_text, text,
                   Output_port); }

static VOI_type print_null(NUL_type Null,
                           OPT_type Output_port)
  { text_to_port(Empty_list_text,
                 Output_port); }

static VOI_type print_null_stream(NST_type Null_stream,
                                  OPT_type Output_port)
  { text_to_port(Empty_stream_text,
                 Output_port); }

static VOI_type print_output_port(OPT_type Port,
                                  OPT_type Output_port)
  { TXT_type text;
    text = Port->txt;
    FORMATTED_ITEM(Output_port_text, text, Output_port); }

static VOI_type print_pair(PAI_type Pair,
                           OPT_type Output_port)
  { EXP_type residue_expression,
             value_expression;
    UNS_type index;
    if (Print_level == 0)
      { text_to_port(Pair_text,
                     Output_port);
        return; }
    Print_level -= 1;
    Print_size /= 2;
    char_to_port(Leftpar_char,
                 Output_port);
    index = 1;
    do
      { value_expression   = Pair->car;
        residue_expression = Pair->cdr;
        print_expression(value_expression,
                         Output_port);
        if (index == Print_size)
          { text_to_port(Ellipsis_text,
                         Output_port);
            residue_expression = Unique_Null;
            break; }
        if (is_PAI(residue_expression))
          { char_to_port(Space_char,
                         Output_port);
            Pair = residue_expression; }
        index += 1; }
    while (is_PAI(residue_expression));
    if (!is_NUL(residue_expression))
      { text_to_port(Period_text,
                     Output_port);
        print_expression(residue_expression,
                         Output_port); }
    char_to_port(Rightpar_char,
                 Output_port);
    Print_level += 1;
    Print_size *= 2; }

static VOI_type print_procedure(PRC_type Procedure,
                                OPT_type Output_port)
  { SYM_type name_symbol;
    TXT_type text;
    name_symbol = Procedure->sym;
    text = name_symbol->txt;
    FORMATTED_ITEM(Procedure_text, text, Output_port); }

static VOI_type print_procedure_vararg(PRV_type Procedure_vararg,
                                       OPT_type Output_port)
  { SYM_type name_symbol;
    TXT_type text;
    name_symbol = Procedure_vararg->sym;
    text = name_symbol->txt;
    FORMATTED_ITEM(Procedure_text, text, Output_port); }

static VOI_type print_promise(PRM_type Promise,
                              OPT_type Output_port)
  { text_to_port(Promise_text,
                 Output_port); }

static VOI_type print_real(REA_type Real,
                           OPT_type Output_port)
  { FLO_type float_;
    float_ = Real->flo;
    FORMATTED_ITEM(Floating_text, float_, Output_port); }

static VOI_type print_smallint(SMI_type Number,
                               OPT_type Output_port)
  { SGN_type int_;
    int_ = get_SMI(Number);
    FORMATTED_ITEM(Integer_text, int_, Output_port); }

static VOI_type print_string(STR_type String,
                             OPT_type Output_port)
  { CHR_type char_;
    TXT_type text;
    UNS_type index,
             length;
    length = length_STR(String);
    if (length == 0)
      { text_to_port(Empty_text_text,
                   Output_port);
        return; }
    text = String->txt;
    char_to_port(Doublequote_char,
                 Output_port);
    for (index = 0;
         index < length;
         index += 1)
      { char_ = text[index];
        switch (char_)
          { case Backslash_char:
            case Doublequote_char:
              char_to_port(Backslash_char,
                           Output_port);
              break; }
        if (!isprint(char_))
          if (char_ != Newline_char)
            char_ = Space_char;
        char_to_port(char_,
                     Output_port); }
    char_to_port(Doublequote_char,
                 Output_port); }

static VOI_type print_symbol(SYM_type Symbol,
                             OPT_type Output_port)
  { TXT_type text;
    text = Symbol->txt;
    text_to_port(text,
                 Output_port); }

static VOI_type print_true(TRU_type True,
                           OPT_type Output_port)
  { text_to_port(True_text,
                 Output_port); }

static VOI_type print_unspecified(EXP_type Value_expression,
                                  OPT_type Output_port)
  { return; }

static VOI_type print_vector(VEC_type Vector,
                             OPT_type Output_port)
  { EXP_type value_expression;
    UNS_type index,
             size;
    size = size_VEC(Vector);
    if (size == 0)
      { text_to_port(Vector0_text,
                     Output_port);
        return; }
    if (Print_level == 0)
      { text_to_port(Vector_text,
                     Output_port);
        return; }
    Print_level -= 1;
    Print_size /= 2;
    text_to_port(Vector_text,
                 Output_port);
    for (index = 1;
         index <= size;
         index += 1)
      { value_expression = Vector[index];
        print_expression(value_expression,
                         Output_port);
        if (index == Print_size)
          { text_to_port(Ellipsis_text,
                         Output_port);
            break; }
        if (index < size)
          char_to_port(Space_char,
                       Output_port); }
    char_to_port(Rightpar_char,
                 Output_port);
    Print_level += 1;
    Print_size *= 2; }

static VOI_type print_expression(EXP_type Value_expression,
                                 OPT_type Output_port)
  { TAG_type tag;
    tag = Grammar_Tag(Value_expression);
    switch (tag)
      { case CHA_tag:
          print_character(Value_expression,
                          Output_port);
          return;
        case CNT_tag:
          print_continuation(Value_expression,
                           Output_port);
          return;
        case END_tag:
          print_eof_object(Value_expression,
                           Output_port);
          return;
        case FLS_tag:
          print_false(Value_expression,
                           Output_port);
          return;
        case FRC_tag:
          print_forced(Value_expression,
                           Output_port);
          return;
        case IPT_tag:
          print_input_port(Value_expression,
                           Output_port);
          return;
        case MAC_tag:
          print_macro(Value_expression,
                      Output_port);
          return;
        case NAT_tag:
          print_native(Value_expression,
                       Output_port);
          return;
        case NAZ_tag:
          print_higher_order_native(Value_expression,
                                    Output_port);
          return;
        case NST_tag:
          print_null_stream(Value_expression,
                            Output_port);
          return;
        case NUL_tag:
          print_null(Value_expression,
                     Output_port);
          return;
        case OPT_tag:
          print_output_port(Value_expression,
                            Output_port);
          return;
        case PAI_tag:
          print_pair(Value_expression,
                     Output_port);
          return;
        case PRC_tag:
          print_procedure(Value_expression,
                          Output_port);
          return;
        case PRM_tag:
          print_promise(Value_expression,
                        Output_port);
          return;
        case PRV_tag:
          print_procedure_vararg(Value_expression,
                                 Output_port);
          return;
        case REA_tag:
          print_real(Value_expression,
                     Output_port);
          return;
        case SMI_tag:
          print_smallint(Value_expression,
                         Output_port);
          return;
        case STR_tag:
          print_string(Value_expression,
                       Output_port);
          return;
        case SYM_tag:
          print_symbol(Value_expression,
                       Output_port);
          return;
        case TRU_tag:
          print_true(Value_expression,
                     Output_port);
          return;
        case USP_tag:
          print_unspecified(Value_expression,
                            Output_port);
          return;
        case VEC_tag:
          print_vector(Value_expression,
                       Output_port);
          return;
        default:
          error(IXT_error,
                Print_text); }}

/*----------------------------------------- write --------------------------------------*/

static VOI_type write_expression(EXP_type,
                                 OPT_type);

static VOI_type write_character(CHA_type Character,
                                OPT_type Output_port)
  { CHR_type char_;
    char_ = get_CHA(Character);
    switch (char_)
      { case Space_char:
          char_to_port(Space_char,
                       Output_port);
          return;
        case Newline_char:
          char_to_port(Newline_char,
                       Output_port);
          return; }
    if (!isprint(char_))
      { char_to_port(Newline_char,
                     Output_port);
        return; }
    char_to_port(char_,
                 Output_port); }

static VOI_type write_false(OPT_type Output_port)
  { text_to_port(False_text,
                 Output_port); }

static VOI_type write_null(OPT_type Output_port)
  { text_to_port(Empty_list_text,
                 Output_port); }

static VOI_type write_smallint(SMI_type Smallint,
                                    OPT_type Output_port)
  { SGN_type int_;
    int_ = get_SMI(Smallint);
    FORMATTED_ITEM(Integer_text,
                   int_,
                   Output_port); }

static VOI_type write_pair(PAI_type Pair,
                           OPT_type Output_port)
  { EXP_type residue_expression,
             value_expression;
    char_to_port(Leftpar_char,
                 Output_port);
    do
      { value_expression   = Pair->car;
        residue_expression = Pair->cdr;
        write_expression(value_expression,
                         Output_port);
        if (is_PAI(residue_expression))
          { char_to_port(Space_char,
                         Output_port);
            Pair = residue_expression; }}
    while (is_PAI(residue_expression));
    if (!is_NUL(residue_expression))
      { text_to_port(Period_text,
                   Output_port);
        write_expression(residue_expression,
                         Output_port); }
    char_to_port(Rightpar_char,
                 Output_port); }

static VOI_type write_real(REA_type Real,
                           OPT_type Output_port)
  { FLO_type floating;
    floating = Real->flo;
    FORMATTED_ITEM(Floating_text,
                         floating,
                         Output_port); }

static VOI_type write_string(STR_type String,
                             OPT_type Output_port)
  { CHR_type char_;
    TXT_type text;
    UNS_type index,
             length;
    length = length_STR(String);
    if (length == 0)
      { text_to_port(Empty_text_text,
                   Output_port);
        return; }
    text = String->txt;
    char_to_port(Doublequote_char,
                 Output_port);
    for (index = 0;
         index < length;
         index += 1)
      { char_ = text[index];
        switch (char_)
          { case Backslash_char:
            case Doublequote_char:
              char_to_port(Backslash_char,
                           Output_port);
              break; }
        if (!isprint(char_))
          if (char_ != Newline_char)
            char_ = Space_char;
        char_to_port(char_,
                     Output_port); }
    char_to_port(Doublequote_char,
                 Output_port); }

static VOI_type write_symbol(SYM_type Symbol,
                             OPT_type Output_port)
  { TXT_type text;
    text = Symbol->txt;
    text_to_port(text,
               Output_port); }

static VOI_type write_true(OPT_type Output_port)
  { text_to_port(True_text,
               Output_port); }

static VOI_type write_vector(VEC_type Vector,
                             OPT_type Output_port)
  { EXP_type value_expression;
    UNS_type index,
             size;
    size = size_VEC(Vector);
    if (size == 0)
      { text_to_port(Vector0_text,
                     Output_port);
        return; }
    text_to_port(Vector_text,
                 Output_port);
    for (index = 1;
         index <= size;
         index += 1)
      { value_expression = Vector[index];
        write_expression(value_expression,
                         Output_port);
        if (index < size)
          char_to_port(Space_char,
                       Output_port); }
    char_to_port(Rightpar_char,
                 Output_port); }

/*----------------------------------- write expression ---------------------------------*/

static VOI_type write_expression(EXP_type Value_expression,
                                 OPT_type Output_port)
  { TAG_type tag;
    tag = Grammar_Tag(Value_expression);
    switch (tag)
      { case CHA_tag:
          write_character(Value_expression,
                          Output_port);
          return;
        case FLS_tag:
          write_false(Output_port);
          return;
        case NUL_tag:
          write_null(Output_port);
          return;
        case PAI_tag:
          write_pair(Value_expression,
                     Output_port);
          return;
        case REA_tag:
          write_real(Value_expression,
                     Output_port);
          return;
        case SMI_tag:
          write_smallint(Value_expression,
                         Output_port);
          return;
        case STR_tag:
          write_string(Value_expression,
                       Output_port);
          return;
        case SYM_tag:
          write_symbol(Value_expression,
                       Output_port);
          return;
        case TRU_tag:
          write_true(Output_port);
          return;
        case VEC_tag:
          write_vector(Value_expression,
                       Output_port);
          return;
        default:
          error(IXT_error,
                Write_text); }}

/*---------------------------------- public functions ----------------------------------*/

VOI_type Print_Display(EXP_type Value_expression,
                       OPT_type Output_port)
  { display_expression(Value_expression,
                       Output_port); }

VOI_type Print_Format(EXP_type Expression,
                      OPT_type Output_port)
  { Indentation = 0;
    format_expression(Expression,
                      FALSE,
                      Output_port);
    char_to_port(Newline_char,
                 Output_port); }

VOI_type Print_Print(EXP_type Value_expression,
                     OPT_type Output_port)
  { Print_level = Print_level_max;
    Print_size = Print_size_max;
    print_expression(Value_expression,
                     Output_port); }

VOI_type Print_Write(EXP_type Value_expression,
                     OPT_type Output_port)
  { write_expression(Value_expression,
                     Output_port); }

/*--------------------------------------------------------------------------------------*/
/*------------------------------------- initialize -------------------------------------*/
/*--------------------------------------------------------------------------------------*/

VOI_type Print_Initialize(VOI_type)
  { return; }

