                     /*---------------------------------------------*/
                     /*                  >>>Slip<<<                 */
                     /*                 Theo D'Hondt                */
                     /*          VUB Software Languages Lab         */
                     /*                  (c) 2016                   */
                     /*---------------------------------------------*/
                     /*                final version                */
                     /*---------------------------------------------*/
                     /*                     scan                    */
                     /*---------------------------------------------*/

#include "SlipPlatform.h"

#include "SlipMain.h"

#include "SlipScan.h"

/*--------------------------------------- tokens ---------------------------------------*/
/*                                                                                      */
/* {token}              ::= {character} |                                               */
/*                          {end of information} |                                      */
/*                          {false} |                                                   */
/*                          {integer} |                                                 */
/*                          {left parenthesis} |                                        */
/*                          {period} |                                                  */
/*                          {quasiquote} |                                              */
/*                          {quasisymbol} |                                             */
/*                          {quote} |                                                   */
/*                          {real} |                                                    */
/*                          {right parenthesis} |                                       */
/*                          {string} |                                                  */
/*                          {symbol} |                                                  */
/*                          {true} |                                                    */
/*                          {unquote} |                                                 */
/*                          {unquote splicing} |                                        */
/*                          {vector}                                                    */
/*                                                                                      */
/*--------------------------------------------------------------------------------------*/
/*                                                                                      */
/* {character}          ::= #\ {any character}+                                         */
/*                                                                                      */
/* {end of information} ::=                                                             */
/*                                                                                      */
/* {false}              ::= #f                                                          */
/*                                                                                      */
/* {integer}            ::= {digit}+ | + {digit}+ | - {digit}+                          */
/*                                                                                      */
/* {left parenthesis}   ::= (                                                           */
/*                                                                                      */
/* {period}             ::= .                                                           */
/*                                                                                      */
/* {quasiquote}         ::= `                                                           */
/*                                                                                      */
/* {quasisymbol}        ::= `@                                                          */
/*                                                                                      */
/* {quote}              ::= '                                                           */
/*                                                                                      */
/* {real}               ::= {fraction} | {float}                                        */
/*                                                                                      */
/* {right parenthesis}  ::= )                                                           */
/*                                                                                      */
/* {string}             ::= " {non-" character}* "                                      */
/*                                                                                      */
/* {symbol}             ::= {first} {next}*                                             */
/*                                                                                      */
/* {true}               ::= #t                                                          */
/*                                                                                      */
/* {unquote}            ::= ,                                                           */
/*                                                                                      */
/* {unquote splicing}   ::= ,@                                                          */
/*                                                                                      */
/* {vector}             ::= #(                                                          */
/*                                                                                      */
/*--------------------------------------------------------------------------------------*/
/*                                                                                      */
/* {digit}              ::= 0 | 1 | ... | 9                                             */
/*                                                                                      */
/* {first}              ::= ! | $ | % | & | * | / | : | < | = | > | ? | ^ | _ | ~       */
/*                                                                                      */
/* {float}              ::= {integer} {exponent} | {fraction} {exponent}                */
/*                                                                                      */
/* {fraction}           ::= {integer} . {digit}+                                        */
/*                                                                                      */
/* {letter}             ::= a | A | ... | z | Z                                         */
/*                                                                                      */
/* {next}               ::= {digit} | {letter} | {first} | @ | # | . | + | -            */
/*                                                                                      */
/*--------------------------------------------------------------------------------------*/
/*                                                                                      */
/* {exponent}           ::= e {integer} | E {integer}                                   */
/*                                                                                      */
/*--------------------------------------------------------------------------------------*/

/*---------------------------------- private constants ---------------------------------*/

enum { Buffer_size   = 1024,
       End_of_line   = '\n',
       End_of_string = '\0' };

static const TXT_type Scan_text = "Scan";

typedef VOI_type (* THK_type) (VOI_type);

/*---------------------------------- private variables ---------------------------------*/

static CHR_type Current_character;
static ADR_type Current_port_address;
static UNS_type Current_position;
static UNS_type Current_size;
static TXT_type Current_text;
static UNS_type Cursor;
static THK_type Next_character;

/*---------------------------------- private prototypes --------------------------------*/

static TKN_type next_token(VOI_type);

/*---------------------------------- character category --------------------------------*/

typedef
  enum { Apo,                                      /* '                                 */
         Att,                                      /* @                                 */
         Bkq,                                      /* `                                 */
         Bks,                                      /* \                                 */
         Com,                                      /* ,                                 */
         Dgt,                                      /* 0...9                             */
         Eoi,                                      /*                                   */
         Eol,                                      /* <LF> <CR>                         */
         Exp,                                      /* e E                               */
         Fls,                                      /* f F                               */
         Hsh,                                      /* #                                 */
         Ill,                                      /*                                   */
         Lpr,                                      /* (                                 */
         Ltr,                                      /* a A...z Z w/o e E f F n N t T     */
         Nwl,                                      /* n N                               */
         Opr,                                      /* ! $ % & * / : < = > ? ^ _ ~       */
         Per,                                      /* .                                 */
         Quo,                                      /* "                                 */
         Rpr,                                      /* )                                 */
         Sgn,                                      /* + -                               */
         Smc,                                      /* ;                                 */
         Tru,                                      /* t T                               */
         Wsp } Cat_type;                           /* <HT> <FF> <SP>                    */

static const Cat_type ASCII_table[] =
                      /*NUL SOH STX ETX EOT ENQ ACK BEL BS  HT  LF  VT  FF  CR  SO  SI  */
                      { Eoi,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Wsp,Eol,Ill,Wsp,Eol,Ill,Ill,
                      /*DLE DC1 DC2 DC3 DC4 NAK SYN ETB CAN EM  SUB ESC FS  GS  RS  US  */
                        Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,
                      /*SP   !   "   #   $   %   &   '   (   )   *   +   ,   -   .   /  */
                        Wsp,Opr,Quo,Hsh,Opr,Opr,Opr,Apo,Lpr,Rpr,Opr,Sgn,Com,Sgn,Per,Opr,
                      /* 0   1   2   3   4   5   6   7   8   9   :   ;   <   =   >   ?  */
                        Dgt,Dgt,Dgt,Dgt,Dgt,Dgt,Dgt,Dgt,Dgt,Dgt,Opr,Smc,Opr,Opr,Opr,Opr,
                      /* @   A   B   C   D   E   F   G   H   I   J   K   L   M   N   O  */
                        Att,Ltr,Ltr,Ltr,Ltr,Exp,Fls,Ltr,Ltr,Ltr,Ltr,Ltr,Ltr,Ltr,Nwl,Ltr,
                      /* P   Q   R   S   T   U   V   W   X   Y   Z   [   \   ]   ^   _  */
                        Ltr,Ltr,Ltr,Ltr,Tru,Ltr,Ltr,Ltr,Ltr,Ltr,Ltr,Ill,Bks,Ill,Opr,Opr,
                      /* `   a   b   c   d   e   f   g   h   i   j   k   l   m   n   o  */
                        Bkq,Ltr,Ltr,Ltr,Ltr,Exp,Fls,Ltr,Ltr,Ltr,Ltr,Ltr,Ltr,Ltr,Nwl,Ltr,
                      /* p   q   r   s   t   u   v   w   x   y   z   {   |   }   ~  DEL */
                        Ltr,Ltr,Ltr,Ltr,Tru,Ltr,Ltr,Ltr,Ltr,Ltr,Ltr,Ill,Ill,Ill,Opr,Ill,
                      /*                                                                */
                        Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,
                      /*                                                                */
                        Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,
                      /*                                                                */
                        Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,
                      /*                                                                */
                        Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,
                      /*                                                                */
                        Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,
                      /*                                                                */
                        Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,
                      /*                                                                */
                        Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,
                      /*                                                                */
                        Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill,Ill };

static Cat_type get_category(VOI_type)
  { return ASCII_table[Current_character]; }

/*------------------------------------ character sets-----------------------------------*/

typedef
  enum { bks_set = (1<<Bks),
         del_set = (1<<Eoi)|
                   (1<<Eol)|
                   (1<<Lpr)|
                   (1<<Quo)|
                   (1<<Rpr)|
                   (1<<Smc)|
                   (1<<Wsp),
         dig_set = (1<<Dgt),
         end_set = (1<<Eoi)|
                   (1<<Eol),
         exp_set = (1<<Exp),
         per_set = (1<<Per),
         quo_set = (1<<Quo),
         sgn_set = (1<<Sgn),
         sub_set = (1<<Att)|
                   (1<<Dgt)|
                   (1<<Exp)|
                   (1<<Fls)|
                   (1<<Hsh)|
                   (1<<Ltr)|
                   (1<<Nwl)|
                   (1<<Opr)|
                   (1<<Per)|
                   (1<<Sgn)|
                   (1<<Tru),
         trm_set = (1<<Eoi)|
                   (1<<Eol)|
                   (1<<Quo),
         wsp_set = (1<<Eol)|
                   (1<<Wsp) } Set_type;

static BLN_type is_member(Set_type Set)
  { Cat_type category;
    category = get_category();
    return (BLN_type)((Set >> category) & 1); }

/*---------------------------------- private functions ---------------------------------*/

static VOI_type error(UNS_type Error)
  { Main_Error(Error,
               Scan_text); }

static VOI_type next_input_port_character(VOI_type)
  { STA_type status;
    status = Main_Read_Char(&Current_character,
                            Current_port_address);
    switch (status)
      { case Main_Aborted_Status:
          error(RDF_error);
          return;
        case Main_Terminated_Status:
          Current_character = End_of_string;
          return;
        case Main_Successful_Status:
          return; }}

static VOI_type next_string_character(VOI_type)
  { if (Current_position == Current_size)
      Current_character = End_of_string;
    else
      Current_character = Current_text[Current_position++]; }

/*--------------------------------------------------------------------------------------*/

static VOI_type get_until(Set_type Set)
  { do
      Next_character();
    while (!is_member(Set)); }

static VOI_type get_while(Set_type Set)
  { do
      Next_character();
    while (is_member(Set)); }

/*--------------------------------------------------------------------------------------*/

static VOI_type copy_char(CHR_type character)
  { if (Cursor == Buffer_size)
      error(BFO_error);
    Scan_Text[Cursor++] = character; }

static VOI_type copy_and_get_char(VOI_type)
  { copy_char(Current_character);
    Next_character(); }

static VOI_type copy_and_get_until(Set_type Set)
  { do
      copy_and_get_char();
    while (!is_member(Set)); }

static VOI_type copy_and_get_while(Set_type Set)
  { do
      copy_and_get_char();
    while (is_member(Set)); }

static VOI_type stop_copy_text(VOI_type)
  { copy_char(End_of_string); }

/*--------------------------------------------------------------------------------------*/

static TKN_type error_return(UNS_type Error)
  { error(Error);
    return EOI_token; }

static TKN_type next_character_return(TKN_type Token)
  { Next_character();
    return Token; }

static TKN_type stop_copy_text_return(TKN_type Token)
  { stop_copy_text();
    return Token; }

/*--------------------------------------------------------------------------------------*/

static TKN_type character(VOI_type)
  { Next_character();
    copy_and_get_until(del_set);
    stop_copy_text();
    return CHA_token; }

static VOI_type integer(VOI_type)
  { if (!is_member(dig_set))
      error(DRQ_error);
    copy_and_get_while(dig_set); }

static TKN_type number(VOI_type)
  { TKN_type token;
    token = INT_token;
    copy_and_get_while(dig_set);
    if (is_member(per_set))
      { token = REA_token;
        copy_and_get_char();
        integer(); }
    if (is_member(exp_set))
      { token = REA_token;
        copy_and_get_char();
        if (is_member(sgn_set))
          copy_and_get_char();
        integer(); }
    return stop_copy_text_return(token); }

static TKN_type string(VOI_type)
  { Cat_type category;
    Next_character();
    while (!is_member(trm_set))
      { if (is_member(bks_set))
          { Next_character();
            category = get_category();
            switch (category)
              { case Bks:
                  break;
                case Nwl:
                  Current_character = End_of_line;
                case Quo:
                  break;
                default:
                  return error_return(ILB_error); }}
        copy_and_get_char(); }
    stop_copy_text();
    if (!is_member(quo_set))
      return error_return(QRQ_error);
    return next_character_return(STR_token); }

static TKN_type symbol(VOI_type)
  { copy_and_get_while(sub_set);
    return stop_copy_text_return(SYM_token); }

/*--------------------------------------------------------------------------------------*/

static TKN_type apostrophe(VOI_type)
  { return next_character_return(QUO_token); }

static TKN_type commercial_at(VOI_type)
  { return error_return(ILA_error); }

static TKN_type backquote(VOI_type)
  { Cat_type category;
    Next_character();
    category = get_category();
    switch (category)
      { case Att:
          Next_character();
          return QSY_token;
        default:
          return QQU_token; }}

static TKN_type backslash(VOI_type)
  { return error_return(ILB_error); }

static TKN_type comma(VOI_type)
  { Cat_type category;
    Next_character();
    category = get_category();
    switch (category)
      { case Att:
         Next_character();
         return UQS_token;
        default:
          return UNQ_token; }}

static TKN_type digit(VOI_type)
  { return number(); }

static TKN_type end_of_information(VOI_type)
  { return EOI_token; }

static TKN_type false(VOI_type)
  { return symbol(); }

static TKN_type hash(VOI_type)
  { Cat_type category;
    Next_character();
    category = get_category();
    switch (category)
      { case Bks:
          return character();
        case Fls:
          return next_character_return(FLS_token);
        case Lpr:
          return next_character_return(VEC_token);
        case Tru:
          return next_character_return(TRU_token);
        default:
          return error_return(ILH_error); }}

static TKN_type illegal(VOI_type)
  { return error_return(ILC_error); }

static TKN_type left_parenthesis(VOI_type)
  { return next_character_return(LPR_token); }

static TKN_type letter(VOI_type)
  { return symbol(); }

static TKN_type operator(VOI_type)
  { copy_and_get_while(sub_set);
    return stop_copy_text_return(SYM_token); }

static TKN_type period(VOI_type)
  { return next_character_return(PER_token); }

static TKN_type quote(VOI_type)
  { return string(); }

static TKN_type right_parenthesis(VOI_type)
  { return next_character_return(RPR_token); }

static TKN_type sign(VOI_type)
  { copy_and_get_char();
    if (is_member(dig_set))
      return number();
    if (is_member(sub_set))
      return symbol();
    return stop_copy_text_return(SYM_token); }

static TKN_type semicolon(VOI_type)
  { get_until(end_set);
    Next_character();
    return next_token(); }

static TKN_type true(VOI_type)
  { return symbol(); }

static TKN_type whitespace(VOI_type)
  { get_while(wsp_set);
    return next_token(); }

/*--------------------------------------------------------------------------------------*/

static TKN_type next_token(VOI_type)
  { Cat_type category;
    Cursor = 0;
    Scan_Text[0] = End_of_string;
    category = get_category();
    switch (category)
      { case Apo: return apostrophe();
        case Att: return commercial_at();
        case Bkq: return backquote();
        case Bks: return backslash();
        case Com: return comma();
        case Dgt: return digit();
        case Eoi: return end_of_information();
        case Exp: return letter();
        case Fls: return false();
        case Hsh: return hash();
        case Ill: return illegal();
        case Lpr: return left_parenthesis();
        case Ltr: return letter();
        case Eol: return whitespace();
        case Nwl: return letter();
        case Opr: return operator();
        case Per: return period();
        case Quo: return quote();
        case Rpr: return right_parenthesis();
        case Sgn: return sign();
        case Smc: return semicolon();
        case Tru: return true();
        case Wsp: return whitespace(); }}

/*---------------------------------- public variables ----------------------------------*/

CHR_type Scan_Text[Buffer_size];

/*---------------------------------- public functions ----------------------------------*/

TKN_type Scan_Input_Port(ADR_type Port_address)
  { Current_port_address = Port_address;
    Next_character = next_input_port_character;
    Next_character();
    return next_token(); }

TKN_type Scan_Token(VOI_type)
  { return next_token(); }

TKN_type Scan_String(TXT_type Scan_input_text)
  { Current_text = Scan_input_text;
    Current_position = 0;
    Current_size = C_STRLEN(Scan_input_text);
    Next_character = next_string_character;
    Next_character();
    return next_token(); }

/*------------------------------------- initialize -------------------------------------*/

VOI_type Scan_Initialize(VOI_type)
  { return; }

