                     /*---------------------------------------------*/
                     /*                  >>>Slip<<<                 */
                     /*                 Theo D'Hondt                */
                     /*          VUB Software Languages Lab         */
                     /*                  (c) 2018                   */
                     /*---------------------------------------------*/
                     /*                final version                */
                     /*---------------------------------------------*/
                     /*                    thread                   */
                     /*---------------------------------------------*/
                     
#include "SlipPlatform.h"

#include "SlipGrammar.h"

#include "SlipCache.h"

#include "SlipThread.h"

/*--------------------------------------------------------------------------------------*/
/*---------------------------------- thread manipulation -------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*----------------------------------- private macros -----------------------------------*/

#define NULL_THREAD (THR_type)Unique_Null

/*------------------------------------ private types -----------------------------------*/

BEGIN_CHUNK(THR);
  DEFINE_SLOT(SMI, tag);
  DEFINE_SLOT(THR, thk);
  DEFINE_SLOT(EXP, res);
  DEFINE_SLOT(EXP, exp[]);
END_CHUNK(THR);

/*---------------------------- private protected variables -----------------------------*/

static THR_type Current_Thread_P;

/*---------------------------------- private prototypes --------------------------------*/

static THR_type  make_thread_M(UNS_type);
static VOI_type release_thread(THR_type);
static UNS_type  thread_extent(THR_type);

/*---------------------------------- private functions ---------------------------------*/

static VOI_type clone_thread_M(VOI_type)
  { CHK_type thread_chunk;
    EXP_type expression,
             value_expression;
    SMI_type thread_tag_smallint;
    THR_type new_thread,
             next_thread;
    UNS_type chunk_size,
             extent,
             index;
    if (Current_Thread_P == NULL_THREAD)
      return;
    thread_chunk = (CHK_type)Current_Thread_P;
    if (marked_CHK(thread_chunk))
      { chunk_size = size_CHK(thread_chunk);
        extent = thread_extent(Current_Thread_P);
        new_thread = make_thread_M(chunk_size);
        thread_tag_smallint = Current_Thread_P->tag;
        value_expression    = Current_Thread_P->res;
        next_thread         = Current_Thread_P->thk;
        new_thread->tag = thread_tag_smallint;
        new_thread->res = value_expression;
        new_thread->thk = next_thread;
        for (index = 0;
             index < extent;
             index += 1)
          { expression = Current_Thread_P->exp[index];
            new_thread->exp[index] = expression; }
        Current_Thread_P = new_thread; }}

static VOI_type flush(VOI_type)
  { THR_type next_thread;
    for (;
         !is_NUL(Current_Thread_P);
         Current_Thread_P = next_thread)
      { next_thread = Current_Thread_P->thk;
        release_thread(Current_Thread_P); }}

static THR_type make_thread_M(UNS_type Thread_size)
  { CHK_type thread_chunk;
    THR_type thread;
    thread_chunk = Cache_Make_Chunk_M(Thread_size);
    thread = (THR_type)thread_chunk;
    return thread; }

static VOI_type mark_thread(THR_type Thread)
  { EXP_type expression;
    UNS_type extent,
             index;
    mark_CHK((CHK_type)Thread);
    extent = thread_extent(Thread);
    for (index = 0;
         index < extent;
         index += 1)
      { expression = Thread->exp[index];
        if (is_CHK(expression))
          mark_CHK((CHK_type)expression); }}

static VOI_type release_thread(THR_type Thread)
  { CHK_type thread_chunk;
    thread_chunk = (CHK_type)Thread;
    Cache_Release_Chunk(thread_chunk); }

static UNS_type thread_extent(THR_type Thread)
  { CHK_type thread_chunk;
    UNS_type chunk_size,
             extent;
    thread_chunk = (CHK_type)Thread;
    chunk_size = size_CHK(thread_chunk);
    extent = chunk_size - THR_size;
    return extent; }

/*---------------------------------- hidden functions ----------------------------------*/

EXP_type _Thread_Get_Result(VOI_type)
  { EXP_type result_expression;
    result_expression = Current_Thread_P->res;
    return result_expression; }

THR_type _Thread_Keep_M(VOI_type)
  { clone_thread_M();
    return Current_Thread_P; }

THR_type _Thread_Patch_M(SMI_type Thread_id_smallint)
  { clone_thread_M();
    Current_Thread_P->tag = Thread_id_smallint;
    return Current_Thread_P; }

THR_type _Thread_Peek(VOI_type)
  { return Current_Thread_P; }

THR_type _Thread_Poke_M(SMI_type Thread_id_smallint,
                        UNS_type Thread_size)
  { THR_type new_thread,
             next_thread;
    new_thread = make_thread_M(Thread_size);
    next_thread = Current_Thread_P->thk;
    release_thread(Current_Thread_P);
    Current_Thread_P = new_thread;
    Current_Thread_P->tag = Thread_id_smallint;
    Current_Thread_P->thk = next_thread;
    Current_Thread_P->res = Unique_Unspecified;
    return Current_Thread_P; }

THR_type _Thread_Push_M(SMI_type Thread_id_smallint,
                        UNS_type Thread_size)
  { THR_type new_thread,
             next_thread;
    new_thread = make_thread_M(Thread_size);
    next_thread = Current_Thread_P;
    Current_Thread_P = new_thread;
    Current_Thread_P->tag = Thread_id_smallint;
    Current_Thread_P->thk = next_thread;
    Current_Thread_P->res = Unique_Unspecified;
    return Current_Thread_P; }

VOI_type _Thread_Set_Result(EXP_type Result_expression)
  { Current_Thread_P->res = Result_expression; }

VOI_type _Thread_Zap(VOI_type)
  { THR_type next_thread;
    next_thread = Current_Thread_P->thk;
    release_thread(Current_Thread_P);
    Current_Thread_P = next_thread; }

/*---------------------------------- public functions ----------------------------------*/

THR_type Thread_Mark(VOI_type)
  { THR_type thread;
    for (thread = Current_Thread_P;
         thread != NULL_THREAD;
         thread = thread->thk)
      mark_thread(thread);
    return Current_Thread_P; }

VOI_type Thread_Replace(THR_type Thread)
  { flush();
    Current_Thread_P = Thread; }

VOI_type Thread_Reset(VOI_type)
  { flush();
    Current_Thread_P = NULL_THREAD; }

/*--------------------------------------------------------------------------------------*/
/*----------------------------------- thread procedures --------------------------------*/
/*--------------------------------------------------------------------------------------*/

/*---------------------------------- private constants ---------------------------------*/

enum { Thunk_table_size   = 89 };

/*---------------------------------- private variables ---------------------------------*/

static TFN_type Thunk_table[Thunk_table_size];
static UNS_type Thunk_table_counter;

/*---------------------------------- public functions ----------------------------------*/

VOI_type Thread_Loop_M(VOI_type)
  { TFN_type native_thunk_M;
    SMI_type thread_tag_smallint;
    UNS_type thread_tag;
    for (;;)
      { thread_tag_smallint = Current_Thread_P->tag;
        thread_tag = get_SMU(thread_tag_smallint);
        native_thunk_M = Thunk_table[thread_tag];
        native_thunk_M(); }}

SMI_type Thread_Register(TFN_type Thread_thunk_C)
  { SMI_type thread_tag_smallint;
    C_ASSERT(Thunk_table_counter < Thunk_table_size);
    Thunk_table[Thunk_table_counter] = Thread_thunk_C;
    thread_tag_smallint = make_SMU(Thunk_table_counter);
    Thunk_table_counter += 1;
    return thread_tag_smallint; }

/*------------------------------------- initialize -------------------------------------*/

VOI_type Thread_Initialize_M(VOI_type)
  { Cache_Initialize_M();
    Current_Thread_P = NULL_THREAD;
    Thunk_table_counter = 0;
    REGISTER(Current_Thread_P); }

