                     /*---------------------------------------------*/
                     /*                  >>>Slip<<<                 */
                     /*                 Theo D'Hondt                */
                     /*          VUB Software Languages Lab         */
                     /*                  (c) 2016                   */
                     /*---------------------------------------------*/
                     /*                final version                */
                     /*---------------------------------------------*/
                     /*                    Test                     */
                     /*---------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>

#include "Slip.h"

enum { Buffer_size = 1024 };

static const char* Error_Messages[] =
  /*AL1_error*/ { "at least one argument required",
  /*AM1_error*/   "at most one argument allowed",
  /*AM2_error*/   "at most two arguments allowed",
  /*ARI_error*/   "arguments restricted to integers",
  /*ARN_error*/   "arguments restricted to numbers",
  /*BFO_error*/   "buffer overflow",
  /*BND_error*/   "binding expected",
  /*BVR_error*/   "byte value required",
  /*CIP_error*/   "close input port error",
  /*COP_error*/   "close output port error",
  /*COR_error*/   "character out of range",
  /*CRQ_error*/   "continuation requires single value",
  /*DBZ_error*/   "division by zero",
  /*DLV_error*/   "duplicate let variable",
  /*DPN_error*/   "duplicate parameter name",
  /*DRQ_error*/   "digit required",
  /*E1X_error*/   "exactly one expression required",
  /*EX0_error*/   "no arguments allowed",
  /*EX1_error*/   "exactly one argument required",
  /*EX2_error*/   "exactly two arguments required",
  /*EX3_error*/   "exactly three arguments required",
  /*ICL_error*/   "invalid character literal",
  /*IDF_error*/   "illegal decimal fraction",
  /*IDS_error*/   "illegal decimal scale",
  /*ILC_error*/   "illegal character",
  /*ILA_error*/   "illegal use of @",
  /*ILB_error*/   "illegal use of \\",
  /*ILH_error*/   "illegal use of #",
  /*ILS_error*/   "illegal session",
  /*IMB_error*/   "illegal macro body",
  /*IMS_error*/   "illegal macro shadowing",
  /*IMV_error*/   "illegal macro value",
  /*INM_error*/   "Insufficient memory",
  /*IOR_error*/   "index out of range",
  /*IPA_error*/   "invalid parameter",
  /*IRP_error*/   "illegal right parenthesis",
  /*IRS_error*/   "invalid response",
  /*ISF_error*/   "input stream failure",
  /*ITA_error*/   "improperly terminated argument list",
  /*ITF_error*/   "improperly terminated form",
  /*ITK_error*/   "improper token",
  /*IUK_error*/   "illegal use of keyword",
  /*IUP_error*/   "illegal use of period",
  /*IVP_error*/   "invalid pattern in form",
  /*IVS_error*/   "invalid variable shadowing",
  /*IVV_error*/   "invalid variable in form",
  /*IXT_error*/   "invalid expression type",
  /*MMO_error*/   "memory overflow",
  /*MNS_error*/   "memory not synchronized",
  /*MRP_error*/   "missing right parenthesis",
  /*MSP_error*/   "missing pattern in form",
  /*MSY_error*/   "macro symbol",
  /*NAB_error*/   "not a boolean",
  /*NAC_error*/   "not a character",
  /*NAL_error*/   "not a list",
  /*NAN_error*/   "not a number",
  /*NAP_error*/   "not a pair",
  /*NAS_error*/   "not a string",
  /*NAV_error*/   "not a vector",
  /*NIP_error*/   "not an input port",
  /*NNN_error*/   "non-negative number required",
  /*NOP_error*/   "not an output port",
  /*NPR_error*/   "not a promise",
  /*NSY_error*/   "not a symbol",
  /*NTL_error*/   "number too large",
  /*NVS_error*/   "not a valid symbol",
  /*NYI_error*/   "not yet implemented",
  /*OIF_error*/   "open input file error",
  /*OOF_error*/   "open output file error",
  /*OOR_error*/   "out of range error",
  /*OPR_error*/   "operator required",
  /*OSF_error*/   "output stream failure",
  /*PAC_error*/   "port already closed",
  /*PEP_error*/   "premature end of program",
  /*PNR_error*/   "positive number required",
  /*PRE_error*/   "protocol error",
  /*QBR_error*/   "quote or backslash required",
  /*QRQ_error*/   "quote required",
  /*RDF_error*/   "read failure",
  /*SOF_error*/   "session output failure",
  /*STL_error*/   "string too long",
  /*TFO_error*/   "too few operands in application",
  /*TFX_error*/   "too few expressions in form",
  /*TMO_error*/   "too many operands in application",
  /*TMS_error*/   "too many sessions",
  /*TMX_error*/   "too many expressions in form",
  /*UDF_error*/   "undefined variable",
  /*UQQ_error*/   "unquote(-splicing) not in quasiquote",
  /*ULS_error*/   "unquote-splicing not from within list",
  /*UXT_error*/   "unexpected expression type ",
  /*VNF_error*/   "variable not found",
  /*VTL_error*/   "vector too large",
  /*WRF_error*/   "write failure",
  /*XIT_error*/   "excess input tokens" };

static Slip_char Buffer[Buffer_size];

static void read(void)
  { int cnt;
    int ch;
    for (cnt = 0; cnt < Buffer_size; cnt += 1)
      { ch = getc(stdin);
        if (ch == EOF)
          break;
        if (ch == '\n')
          break;
        Buffer[cnt] = (Slip_char)ch; }
    Buffer[cnt] = (Slip_char)'\0'; }

Slip_response Slip_Close(Slip_mode Mode,
                         Slip_address Address)
  { FILE * file_ref = (FILE *)Address;
    int integer;
    if (file_ref)
      { integer = fclose(file_ref);
        if (integer != EOF)
          return Slip_success_response; }
    return Slip_invalid_stream_response; }

Slip_void Slip_Error(Slip_error Error,
                     Slip_text Text)
  { char * text = (char *)Text;
    unsigned error = (unsigned)Error;
    printf("error: %s (%s)", Error_Messages[error], text); }

Slip_response Slip_Open(Slip_text Path,
                        Slip_mode Mode,
                        Slip_address * Address_ref)
  { FILE * file_ref;
    char * path = (char *)Path;
    *Address_ref = (Slip_address)0;
    if (path)
      switch (Mode)
        { case Slip_read_mode:
            file_ref = fopen(Path, "r");
            break;
          case Slip_write_mode:
            file_ref = fopen(Path, "w");
            break; }
    else
      switch (Mode)
        { case Slip_read_mode:
            file_ref = stdin;
            break;
          case Slip_write_mode:
            file_ref = stdout;
            break; }
    if (file_ref)
      *Address_ref = (Slip_address)file_ref;
    else
      return Slip_invalid_stream_response;
    return Slip_success_response; }

Slip_response Slip_Read(Slip_char * Char_ref,
                        Slip_address Address)
  { FILE * file_ref = (FILE *)Address;
    int integer;
    *Char_ref = (Slip_char)0;
    if (file_ref)
      { integer = getc(file_ref);
        if (integer == EOF)
          return Slip_end_of_stream_response; }
    else
      { integer = getchar();
        if (integer == '\n')
          return Slip_end_of_stream_response; }
    *Char_ref = (Slip_char)integer;
    return Slip_success_response; }

Slip_response Slip_Write(Slip_char Char,
                         Slip_address Address)
  { FILE * file_ref = (FILE *)Address;
    char character = (char)Char;
    int integer;
    if (file_ref)
      integer = putc(character, file_ref);
    else
      integer = putchar(character);
    if (integer == EOF)
      return Slip_invalid_stream_response;
    return Slip_success_response; }

int main (int argc, const char * argv[])
  { for(;;)
      { printf(">");
        read();
        Slip_Eval(Buffer);
        printf("\n"); }
    return 0; }

